freeze;
/*****************************************************
  Construction of the Clifford algebra associated to
  a quadratic form defined on a vector space
  
  This package depends on quadform.m for utilities
  related to quadratic forms.
  
  July 2010 Don Taylor
  

  $Id: clifford.m 41773 2012-12-18 06:11:09Z don $
  
******************************************************/


/* Suppose that Q is a quadratic form defined on a
   vector space V of dimension n over a field F.
   Represent the elements of the Clifford algebra C(V)
   by sequences of "monomials" <S,a>, where S is a subset
   of {1..n}, represented by an ordered sequence and
   a is a field element.
   
   Ensure that the monomials in every element of the 
   Clifford algebra are ordered by their first component.
   
   A quadratic form is represented by an n x n upper 
   triangular matrix Q.  The polar form is represented
   by the symmmetric matrix Q + Transpose(Q).
   
   The zero element is the empty sequence [].
   The identity element is <[Integers()|],1>.
*/

declare attributes AlgClff :
  space,
  embedding,
  mainInvolutionMatrix;
  
/* space:      the quadratic space from which the Clifford is derived

   embedding:  the standard embedding of the quadratic space into the
               Clifford algebra
               
   mainInvolutionMatrix: the matrix of the antiautomorphism which 
               reverses the multiplication.
*/

cadd := function(s1,s2)
// Add the values of terms with the same first component
  if IsEmpty(s1) then
    vec := s2;
  elif IsEmpty(s2) then
    vec := s1;
  else
    U := Universe(s1);
    ndx1 := {@ p[1] : p in s1 @};
    ndx2 := {@ p[1] : p in s2 @};
    ndx := Sort(ndx1 join ndx2);
    vec := [U|];
    for i in ndx do
      i1 := Position(ndx1,i);
      i2 := Position(ndx2,i);
      if i1 gt 0 then
        if i2 gt 0 then
          val := s1[i1][2]+s2[i2][2];
          if val ne 0 then
            Append(~vec,<i,val>);
          end if;
        else
          Append(~vec,s1[i1]);
        end if;
      else
        Append(~vec,s2[i2]);
      end if;
    end for;
  end if;
  return vec;
end function;

forward termbyterm;
seqbyterm := function(B,s1,t2)
  vec := [Universe(s1)|];
  for t1 in s1 do
    vec := cadd(vec,termbyterm(B,t1,t2));
  end for;
  
  return vec;
end function;

mprod := function(B,L1,j)
// B is the bilinear form - not assumed symmetric
// L1 represents a basis element of the Clifford algebra
// j is the index of a basis vector of the vector space
  K := BaseRing(B);
  if IsEmpty(L1) then
    return [<[j],K!1>];
  end if;
  last := L1[#L1];
  stub := Prune(L1);
  if last lt j then
    vec := [<Append(L1,j),K!1>];
  elif last eq j then
    vec := [<stub,B[j,j]>];
  else // last gt j
    vec := seqbyterm(B,termbyterm(B,<stub,K!1>,<[j],-K!1>),<[last],K!1>);
    vec := cadd(vec,termbyterm(B,<stub,K!1>,<[Integers()|],B[last,j]+B[j,last]>));
  end if;

  return vec;
end function;

termbyterm := function(B,t1,t2)
  // Return the product of two terms <i1,a1> and <i2,a2>
  L1 := t1[1];
  L2 := t2[1];
  val := t1[2]*t2[2];
  if val eq 0 then
    vec := [car<Parent(L1),Parent(val)>|]; // This should not occur unless called directly
  elif IsEmpty(L1) then
    vec := [<L2,val>];
  elif IsEmpty(L2) then
    vec := [<L1,val>];
  else
    vec := seqbyterm(B,mprod(B,L1,L2[1]),<Remove(L2,1),val>);
  end if;
  return vec;
end function;


aprod := function(B,s1,s2)
// s1 and s2 are sequences of pairs <i,a>
  vec := [Universe(s1)|];
  for q in s2 do
    vec := cadd(vec,seqbyterm(B,s1,q));
  end for;
  return vec;
end function;

/* Construction of a Clifford algebra of dimension n as a structure constant 
   algebra.  The basis elements are indexed by subset of {1..n}. In order
   to construct the structure constant algebra, the subsets are represented
   by integers in the range [1..2^n]
*/

decode := func< k | [Integers()| x : x in [1..#S] | S[x] eq 1] where S is Intseq(k-1,2) >;
encode := func< L | IsEmpty(L) select 1 else 1 + &+[ 2^(k-1) : k in L ] >;

intrinsic CliffordAlgebra(Q::AlgMatElt) -> AlgClff, ModTupFld, Map
{The triple C, V, f, where C is Clifford algebra C of the quadratic form Q,
  V is the quadratic space of Q, and f is the standard embedding of V into C}
  V := QuadraticSpace(Q);
  n := Dimension(V);
  K := BaseRing(V);
  // T will hold the multiplication table as a sequence of 4-tuples
  T := [];
  for i := 1 to 2^n do
    e := [<decode(i),K!1>];
    for j := 1 to 2^n do
      for p in aprod(Q,e,[<decode(j),K!1>]) do
        Append(~T,<i,j,encode(p[1]),K ! p[2]>);
      end for;
    end for;
  end for;
  A := CliffordAlgebra<K, 2^n | T >;
  f := hom< V -> A | [A.encode([i]) : i in [1..n]] >;
//  f := hom< V -> Module(A) | [Vector(A.encode([i])) : i in [1..n]] >;
  A`space := V;
  A`embedding := f;
  return A, V, f;
end intrinsic;

intrinsic CliffordAlgebra(V::ModTupFld) -> AlgClff, Map
{The pair C, f, where C is the Clifford algebra of the quadratic
 space V and f is the standard embedding of V into C}
  Q := QuadraticFormMatrix(V);
  C, _, f := CliffordAlgebra(Q);
  return C, f;
  
end intrinsic;

/*
  K is a field, B is an n x n matrix (not necessarily symmetric) over K 
  The Clifford algebra is an associative K-algebra of dimension 2^n

  Deprecated: this is only here for consistency with the previous release
*/
intrinsic CliffordAlgebra(K::Fld,Q::AlgMatElt) -> AlgClff, ModTupFld, Map
{The Clifford algebra C of the n x n quadratic form Q over a field K,
 the quadratic space K^n, and the standard embedding of K^n into C}

  return CliffordAlgebra(Q);

end intrinsic;


intrinsic MainInvolution(C::AlgClff) -> Map
{The main involution of the Clifford algebra C}
  nu := Dimension(C);
  return hom< C -> C | [ IsEven(#decode(i)) select C.i else -C.i : i in [1..nu]]>;
end intrinsic;

intrinsic BasisProduct(C::AlgClff, L::SeqEnum) -> AlgClffElt
{ The product of the basis elements of the underlying vector space of C,
 indexed by L }
  return IsEmpty(L) select One(C) else &*[ C.encode([i]) : i in L];
end intrinsic;

intrinsic BasisElement(C::AlgClff, L::SetEnum) -> AlgClffElt
{ The basis element of C indexed by L }
  return C.encode(L);
end intrinsic;

intrinsic mainInvolution(C::AlgClff) -> Map
{The main anti-automorphism of the Clifford Algebra C}
  if not assigned C`mainInvolutionMatrix then
    m := Dimension(C);
    C`mainInvolutionMatrix := 
      Matrix(m,m,[ ElementToSequence(BasisProduct(C,Reverse(decode(i))))
           : i in [1..m]]);
  end if;
  return map< C -> C | x :-> x*C`mainInvolutionMatrix >;
end intrinsic;



