//findperfectform.m
//Imaginary quadratic fields

freeze;

import "defs.m": PerfectFormRec;
import "defs.m": ModFrmDataRec;

/* replaced by version below
function CheckPositive(F,a,b)
    //Checks positive definite for form which evaluates to
    //1 on e1,e2,e1+e2, and [a,b]
    den:=F.1*(F`Conjugate(a)*b-a*F`Conjugate(b));
    
    if den ne 0 then
	num:=1-a*F`Conjugate(a)-b*F`Conjugate(b)+
	    1/2*(F`Conjugate(a)*b+a*F`Conjugate(b));
	ans:=RationalField()!(num/den);
	tf:=RationalField()!(-ans^2*F.1^2) lt 3/4;
    else
	ans:=Infinity();
	tf:=false;
    end if;
    
    return ans,tf;
end function;
*/
function CheckPositive(F,a,b)
   //Checks positive definite for form which evaluates to
   //1 on e1,e2,e1+e2, and [a,b]
   den:=(F`Conjugate(a)*b-a*F`Conjugate(b));
   if den ne 0 then
      num := 1 - a*F`Conjugate(a) + F`Conjugate(b)*a - b*F`Conjugate(b);
      beta := num/den;
      det := RationalField()!(1 - beta*F`Conjugate(beta));
      tf := det gt 0;
      ans := det;
   else;
      ans := Infinity();
      tf := false;
   end if;
   return ans,tf;
end function;

function FindCandidates(F,r)
    vprintf Bianchi,3: "Finding candidate vectors for %o.\n",r;
    L:=StandardLattice(4);
    O:=MaximalOrder(F);
    cand:=[];
    P:=ShortVectorsProcess(L,r,r);
    v,n:=NextVector(P);
    while n ne -1 do
	a:=v[1]+O.2*v[2];
	b:=v[3]+O.2*v[4];
	c,tf:=CheckPositive(F,a,b);
	if tf then
	    Append(~cand,[a,b]);
	end if;
	
	v,n:=NextVector(P);
    end while;
    Sort(~cand,func<x,y|
	CheckPositive(F,x[1],x[2])-CheckPositive(F,y[1],y[2])>);
    return cand;
end function;

function IsUniqueForm(M,S)
    //Given F and set of cone vectors representing possible minimal vectors,
    //finds a form that evluates to 1 on the set.  Returns a solution form
    //and a boolean for if the solution is unique.
    V:=Transpose(Matrix(S));
    Q:=InnerProductMatrix(M`ModFrmData`cone_space);
    ones:=Vector([RationalField()|1:i in [1..#S]]);
    tf,form,space:=IsConsistent(Q*V,ones);
    if tf then 
	if
	    Dimension(space) eq 0 then 
	    return true,form;
	else
	    return Basis(space),form;
	end if;
    else
	return false,_;
    end if;
end function;

function ipm(M,phi)
    //Given F a number field and form in cone coordinates, return the
    //inner product matrix that it corresponds to.
    //Does not check for positive definiteness
    V:=M`ModFrmData`cone_space;
    ipMat:=ZeroMatrix(RationalField(),4,4);
    phi:=V!phi;
    E:=M`ModFrmData`O2_in_cone_space;
    Eij:=M`ModFrmData`O2_in_cone_space_ij;
    for i in [1..3] do
	for j in [i+1..4] do 
	    //ipMat[i,j]:=((phi,V!ToBianchiCone(F,e[i]+e[j])) -
	    ipMat[i,j]:=((phi,V!Eij[i,j]) -
		(phi,V!E[i]) - (phi,V!E[j]))/2 ;
	    ipMat[j,i]:=ipMat[i,j];
	end for;
    end for;
    for i in [1..4] do
	ipMat[i,i]:=(phi,V!E[i]);
    end for;    
    return ipMat;
end function;


function ToLat(M,form)
    //Given a field and a form in cone coordinates, returns the associated
    //lattice.  If the form does not correspond to a positive definite
    //form, an error is returned.
    ipMat:=ipm(M,form);
    
    if not IsPositiveDefinite(ipMat) then
	error "Not positive definite form.";
    end if;
    
    L:=LatticeWithBasis(IdentityMatrix(RationalField(),4),ipMat);
    return L;
end function;


function MinimalVectors(M,form)
    //Creates a rank 4 lattice with inner product given by the Hermitian
    //form.  Returns a list of the shortest vectors in the lattice and
    //the lattice;
    F:=BaseField(M);
    Lat:=ToLat(M,form);
    O:=MaximalOrder(F);
    //B:=Basis(O);
    B:=Basis(O);
    O2Zminvecs:=ShortestVectors(Lat);
    O2minvecs:=[Vector([O|a[1]*B[1]+a[2]*B[2],a[3]*B[1]+a[4]*B[2]]):
	a in O2Zminvecs];
    O2mvs:=[];
    //Need to get only 1 cone vector for each (up to torsion) minimal
    //vector in O2minvecs to deal with fields where torsion unit group
    //has order greater than 2
    mvs:=[];
    
    for v in O2minvecs do
	cv:=ToBianchiCone(F,v);
	if not cv in mvs then 
	    Append(~O2mvs,v);
	    Append(~mvs,cv);
	end if;
    end for;
    return mvs,O2mvs;
end function;

function IsPerfectForm(M,L)
    //Given L, a list of cone vectors representing possible minimal vectors,
    //returns Boolean is the form defined perfect.  If it is perfect,
    //also returns the minimal vectors (as cone vectors) and the form as
    //a cone vector
    isunique,form:=IsUniqueForm(M,L);
    F:=BaseField(M);
    if isunique then
	vprint Bianchi, 3: "Uniquely determined.  Is it perfect?";
	minvecs,O2minvecs:=MinimalVectors(M,form);
	tf:=Seqset(L) subset Seqset(minvecs);
	vprint Bianchi, 3:tf;
	if tf then 
	    return true, form, minvecs,O2minvecs;
	else
	    return false,form,_,_;
	end if;
    else
	return false, _,_,_;
    end if;
end function;



function FindPerfectForm(M)
    //Given F, finds initial perfect form and attached PerfectFormRec to F.
    //Assumes e1,e2,e1+e2 and v are minimal vectors
    if assigned M`ModFrmData`initial_form then
	return M`ModFrmData`initial_form;
    else
	F:=BaseField(M);
	preS:=[ToBianchiCone(F,[1,0]),ToBianchiCone(F,[0,1]),ToBianchiCone(F,[1,1])];
	r:=2;
	while true do 
	    candidates:=FindCandidates(F,r);
	    Ccandidates:=[ToBianchiCone(F,a):a in candidates];
	    vprintf Bianchi,3:  "Checking %o.\n",r;
	    for u in Ccandidates do
		S:=Include(preS,u);
		if #S gt 3 then 
		    tf,form,minvecs,O2minvecs:=IsPerfectForm(M,S);
		    if tf then 
			form:=rec<PerfectFormRec|form:=form>;
			form`minimal_vectors:=minvecs;
			form`O2minimal_vectors:=O2minvecs;
			vprint Bianchi,2: "Found perfect form.";
			M`ModFrmData`initial_form:=form;
			return form;
		    end if;    
		end if;
	    end for;
	    r:=r+1;
	end while;
    end if;
    
end function;


