freeze;

/********************************************************************

  Saturation for a given subgroup of the Mordell-Weil group of E/Q
 
  Steve Donnelly, November 2005

  Last modified: November 2010

**********************************************************************

OUTLINE AND REMARKS

   The basic idea is explained in Samir Siksek's paper "Infinite Descent
   on Elliptic Curves", and runs as follows: for P in E(Q), if the
   equation p*R = P admits no solution R in E(Q), then it admits
   no solution over F_q for some suitable prime q. Thus, if we are given
   a subgroup G < E(Q), we hope to efficiently identify the "p-divisible part of G",
   namely { P in G : p*R = P for some R in E(Q) }, by reducing at enough primes q.
   
   For each prime p,
      1) Except in the "special case", find a prime q such that #E(Fq)[p] = p. 
      2) Construct a homomorphism G -> E(Fq)[p] = Z/p by
      				  P :-> m*( Pbar ), where m = #E(Fq)/p.
         The "p-divisible part of G" must be contained in the kernel. 
      3) Repeat steps 1 and 2, intersecting the kernels, until empty, 
         or until we've tried at least 30+Rank(E) different primes q.
      4) If we're not done, or if we're in the "special case", we go through
         an analogous procedure, this time using primes q such that #E(Fq)[p] = p^2.
	 This time for each q we get two homomorphisms are, namely 
	 for T = T1 or T2 where E(Fq)[p] = <T1,T2>,
	 	G  -> E(Fq)[p]    ->   mu_p(Fq) = Z/p	
		P :-> m*( Pbar ) :-> WeilPairing( T, . )
		  where m = 1/p( exponent of E(Fq) ).
      5) If we haven't proved p-saturation, test each P in the
         remaining subgroup by actually trying to solve p*R = P.
	 (well actually, test one P in each cyclic subgroup).

   Fact: It is always possible to find q as in Step 1, except in the 
         "SPECIAL CASE": when p=2 and the discriminant of E is a square.
   Proof sketch: For odd p, the extension K_1/Q obtained by adjoining one
         element of E[p] is smaller than K_all/Q obtained by adjoining all of them
	 (because over the reals, you get one cyclic subgroup but not all).
	 So, by Cebotarev, there exist primes q with a degree one factor
	 in K_1 but not in K_all.
	 For p=2, one has K_1 = K_all if and only if K_all is cyclic of degree 3
	 (or K_all = Q), in other words when the discriminant of the cubic is a square.
   Obviously, it is always possible to find q as in Step 4.
	
   Assuming the group IS saturated, steps 4 and 5 will only arise for small primes (in practice), 
   so needn't be coded efficiently.
   Steps 1 and 2 should be coded as efficiently as possible.
   Step 2 involves taking discrete logs in E(Fq)[p], where heuristically q = O(p*log(p)).
 
   The THEORETICAL  QUESTION arises: can it happen that P is not in p*E(Q), 
       but is in p*E(Fq) for every q that we are allowed to use (in step 1, or step 4, or both)?
   ANSWER: 
       1) If P is not in p*E(Q), then P is not in p*E(Fq) for all primes q.
          This is proved in Martin Prickett's thesis (available online). 
          In other words, we would be unlucky to reach step 5 if P is not in p*E(Q).
       2) In some cases Step 4 will certainly be needed to prove saturation.
          One such situation arises from the fact that decomposition groups 
          (at unramified primes) must be cyclic. Let K_1 := Q( one p-torsion point ) 
          and let L := Q( all p-torsion points, and all R satisfying p*R=P ).
	  L/K_1 is Galois, and it can happen that every cyclic subgroup of Gal(L/K_1)
	  either fixes all the p-torsion, or fixes one of the R's. Therefore, for
	  every q such that E(Fq) has one p-torsion point, E(Fq) must either contain
	  all the p-torsion, or contain one of the R's. In particular, for every
	  prime allowed in Step 1, P will be in p*E(Fq). (This is why we try Step 4
	  if we're not done after Step 1.)

TO DO:

   + program the calculation that determines N (ie the bound on the index of the
     given group in the whole Mordell-Weil group).
   + do step 5 (the division step) efficiently. At the moment the assumption is 
     that the group is almost saturated, ie this routine is not designed to actually
     "saturate" a non-p-saturated group for large p.
   + extend to number fields

***********************************************************************************/

declare verbose ECSaturation, 3;

// Precomputation: for each prime p in [1..N], list primes q with p|#EFq
// So pqDB[p] = < sorted sequence of q's, associative array with keys q and values #EFq >

function pqDatabase(E, N, B)
    Z := Integers();
    ps := PrimesUpTo(N);
    qs := PrimesUpTo(B);
    pqDB := [];
    for p in ps do 
        pqDB[p] := < [Z| ], AssociativeArray(Z) >;
    end for;
    bad := LCM([Denominator(c) : c in Coefficients(E)] cat [Numerator(Discriminant(E))]);
    for q in qs do
        if bad mod q ne 0 then
	    orderq := q+1-FrobeniusTraceDirect(E,q);
	    for tup in Factorization(orderq) do
	    	p := tup[1];
		if p le N then 
                   pqDB[p][2][q] := orderq;
		end if;
	    end for;
	end if;
    end for;
    // For each p, we need the qs in increasing order
    // (and there's no point keeping too many of them)
    for p in ps do 
        pqs := Sort(Setseq(Keys(pqDB[p][2])));
        if #pqs gt 200 then
          pqs := pqs[1..200];
        end if;
        pqDB[p][1] := pqs;
    end for;
    return pqDB;
end function;


// Find a good prime q>q0 such that #E(F_q)[p] = p. 
// Return q, #EFq/p, and (for efficiency), EFq.

function findq( E, p, q0, pqDB : DBbound:=0)

     vprintf ECSaturation,3: "looking for q larger than %o\n", q0;
     Ecoeffs := aInvariants(E);

     // First look in pqDB
     for q in pqDB[p,1] do 
	 if q gt q0 and (p eq 2 or (q mod p) ne 1) then
            order := pqDB[p,2][q];
	    Fq := GF(q);
 	    EFq := EllipticCurve(ChangeUniverse(Ecoeffs, Fq));
	    if p ne 2 or #TwoTorsionSubgroup(EFq) eq 2 then 	
	    	return q, order div p, EFq;
	    end if;
	 end if;
     end for;

     vprintf ECSaturation, 2: "Searching for q (not precomputed): ";
     timeq := Cputime();

     // Start from where the database stopped.
     q0 := Max(q0, DBbound);

     bad := LCM([Denominator(c) : c in Coefficients(E)] cat [Numerator(Discriminant(E))]);

     // If #EFq = A*p , then A*p must be in a Hasse interval around q,
     // and so roughly speaking, q must be in a "Hasse interval" around A*p.
     // In view of Sato-Tate, we use only the middle part of each Hasse interval
     ST := 0.7;
     A := Ceiling(q0/p);
     delta := Ceiling(2*ST*Sqrt(A*p));
     if 2*delta lt p then
        interval := [A*p - delta, A*p + delta];
     else
        // intervals would overlap, so don't use intervals
        interval := [q0, Infinity()];
     end if;
     q := Max(q0, interval[1]);
     while true do 
	 q := NextPrime(q);
	 if q gt interval[2] then
	    A +:= 1;
            delta := Ceiling(2*ST*Sqrt(A*p));
            if 2*delta lt p then
               interval := [A*p - delta, A*p + delta];
            else
               // intervals now overlap, so stop using intervals
               interval := [interval[2], Infinity()];
            end if;
	    q := NextPrime(interval[1]);
	 end if;

	 // Ensure #E(F_q)[p] ne p^2
	 if p ne 2 then
	    while bad mod q eq 0 or q mod p eq 1 do
	       q := NextPrime(q);
	    end while;
	 else 
	    while bad mod q eq 0 or
               #TwoTorsionSubgroup(EllipticCurve(ChangeUniverse(Ecoeffs,GF(q)))) eq 4
            do
	       q := NextPrime(q);
	    end while;
	 end if;
         
	 // Test whether #EFq is divisible by p
	 order := q+1-FrobeniusTraceDirect(E,q);
	 bool, AA := IsDivisibleBy(order, p);
	 if bool then
            break;
	 end if;
         /* 
         // TO DO: is this trick ever better?
         if (q is large) and p le 4*Sqrt(q) then
            // A*p is the only multiple of p in the Hasse interval around q.
            // So p|#EFq <==> #EFq = A*p, and if this holds,
            // A*pt has order p with high probability for a random pt.
            // (Assume p and q are large here)
            EFq := EllipticCurve(ChangeUniverse(Ecoeffs, GF(q)));
	    pt := A*Random(EFq);
	    if p*pt eq EFq!0 and pt ne EFq!0 then 
               AA := A;
               break;
            end if;
	 end if;
         */
     end while;
     
     vprintf ECSaturation, 2: "found q = %o, %os\n", q, Cputime(timeq);

     EFq := EllipticCurve(ChangeUniverse(Ecoeffs, GF(q)));
     return q, AA, EFq;

end function;


// find q>q0 such that #E(Fq)[p] = p^2.
// (This is not intended to be efficient, and is only called when p is small).
// Return q, m := the prime-to-p part of #EFq, e := exponent of the p-part,
// and (for efficiency), EFq.

function findqSpecialCase(E, p, q0, pqDB )

   vprintf ECSaturation,3: "looking for q larger than %o\n", q0;
   foundq := false;
   Ecoeffs := aInvariants(E);

   // First look in pqDB
   for qq in pqDB[p,1] do
        q := qq;
        order := pqDB[p,2][q];
	if q gt q0 and (q mod p) eq 1 then 	// we want q = 1 mod p.
    	    EFq := EllipticCurve(ChangeUniverse(Ecoeffs, GF(q)));
            EFq_grp := AbelianGroup(EFq);
            if #[ invt : invt in Invariants(EFq_grp) | IsDivisibleBy(invt,p) ] eq 2 then
	       foundq := true;
	       break;
	    end if;
	end if;
   end for;

   if not foundq then	  // search stupidly
      vprintf ECSaturation,3: "No suitable q in the precomputed database, so search naively ... ";
      bad := LCM([Denominator(c) : c in Coefficients(E)] cat [Numerator(Discriminant(E))]);
      q := 1 + p*Floor(q0/p);     	
      while true do
	  q := p + q;
          if not IsPrime(q) or IsDivisibleBy(bad,q) then 
             continue; 
          end if;
          order := q+1-FrobeniusTraceDirect(E,q);
          if not IsDivisibleBy(order,p) then continue; end if;
    	  EFq := EllipticCurve(ChangeUniverse(Ecoeffs, GF(q)));
          EFq_grp := AbelianGroup(EFq);
          if #[ invt : invt in Invariants(EFq_grp) | IsDivisibleBy(invt,p) ] eq 2 then
	     break; 
	  end if;
      end while;
   end if;
   // figure out the p-exponent e
   e := 1;
   gensEFq := Generators(EFq_grp);
   while &or[ IsDivisibleBy( Order(p^e*gen), p) : gen in gensEFq ] do
	e +:= 1;
   end while;
   m := order div GCD(order, p^(2*e));
   return q, m, e, EFq;
end function;


intrinsic Saturation(points::SeqEnum[PtEll], N::RngIntElt :
                     OmitPrimes := [],
                     TorsionFree := false,
                     Check := true ) 
       -> SeqEnum[PtEll]
{Given a list of points [P1,...,Pn] on an elliptic curve E defined over Q, 
returns a list of generators for the smallest subgroup of E(Q)
that contains the points, and is p-saturated in E(Q) for all primes p up to N.}	
   
   if IsEmpty(points) then 
      return points; 
   end if;

   E := Curve(Universe(points));

   if Check then 
      points := ReducedBasis(points); 
   end if;
   T, Tmap := TorsionSubgroup(E);
   t := #Invariants(T);
   r := #points;

   if r eq 0 then 
       if TorsionFree then
           return points;
       else
           return [ Tmap(g) : g in Generators(T) ] cat points;
       end if;
   end if;

   // need to include torsion when checking p-divisibility (at least when p | #T)
   points := [ Tmap(g) : g in Generators(T) ] cat points;

   // For each p, we will usually need r primes q which satisfy p|#E(Fq).  
   // This occurs with probability 1/p, so we will typically need to try 
   // the first O(r*N) primes.  Therefore taking C = r+1 below is natural. 
   // The "fudge constant" Cf then balances implemenational trade-offs
   // (the cost of precomputing primes that don't get used, vs the slower 
   // search for extra primes when not enough were precomputed).
   // Cf = 2 or 3 seems about right, currently; Cf = 3 means almost all 
   // the q's are precomputed, but this doesn't cost too much time/memory.
   C := r + 1;
   Cf := 2.5;
   B := Ceiling( Cf * C*N * Log(C*N) );

   vprintf ECSaturation: "Precomputing table of auxiliary primes q up to %o: ", B;
   vtime ECSaturation:
   pqDB := pqDatabase(E,N,B);

   p := 1; 
   while NextPrime(p) le N do 
       p := NextPrime(p);
       while p in OmitPrimes do 
           p := NextPrime(p); 
       end while;

       vprintf ECSaturation,2: "Saturating at p = %o\n",p;
	
       // Determine the subgroup of G consisting of those elements
       // whose image in E(F_q) is p-divisible for suitable primes q
       //   (primes such that p divides #E(F_q) exactly once). 
       GFp := GF(p);
       Gmodp := RSpace( GFp, #points ); // abstract copy of G/p*G as a Z/p module
       Torsmodp := sub< Gmodp | [Gmodp.i : i in [1..t]] >;
       currentkernel := Gmodp;
       // The subgroup currentkernel will always (throughout the loop) 
       // contain the image of the subgroup of G that is p-divisible in E(Q).
       // With high probability, we will have detected this "non-p-saturated" 
       // subgroup after using Rank(E)+30 primes q. 

       // GENERAL CASE
       if p ne 2 or not IsSquare(Discriminant(E)) then
	    q := 2;
	    for i := 1 to #points+30 do 
		vprintf ECSaturation,3: "currentkernel is %o\n", currentkernel;
		if currentkernel subset Torsmodp then 
                    break; 
                end if;
		q, m, EFq := findq(E, p, q, pqDB : DBbound:=B);
                Fq := BaseRing(EFq);
		vprintf ECSaturation,3: "q = %o\n",q;
		// Apply the composite map
		//	   phi: pdivG -> E(F_q) -> E(F_q)[p]	
		//   to the generators of pdivG.
		//   (where the first map is reduction mod q, and 
		//   the second map is multiplication by m.)
                phiofpoints := [EFq(Fq) | m*EFq!P : P in points ];
		vprint ECSaturation,3:  "images of generators under phi are ", phiofpoints;
		S := [ x : x in phiofpoints | not IsZero(x) ];
		if #S eq 0 then 
                    continue; 
                end if;
		P := S[1];
                logs := Matrix(GFp, 1, [GFp| Log(P, Pi) : Pi in phiofpoints]);
		currentkernel := currentkernel meet Kernel(logs);
	    end for;
	    vprintf ECSaturation,3: "currentkernel is %o\n", currentkernel;
	end if;

	// SPECIAL CASE: 
	if p lt 10 and not currentkernel subset Torsmodp then 	// STEP 4 IS REQUIRED
	    vprintf ECSaturation,2: 
	          "Special case: Now trying primes q such that #E(Fq)[%o] = %o^2\n", p, p; 

	    // With pretty high probability, we will have detected the saturated subgroup
	    // after using Rank(E)+10 primes q. 
	    q := 2;
	    for i := 1 to #points+10 do
		if currentkernel subset Torsmodp then break; end if;
                q, m, e, EFq := findqSpecialCase(E, p, q, pqDB);
			// m is the order of the prime-to-p part of EFq, and
			// p^e is the exponent of the p-part (which equals that of E(Q)_tors).
		vprintf ECSaturation,3: "q = %o\n",q;
                // find ppow := the power of p dividing q-1
		a := q-1; ppow := 0;
		while IsDivisibleBy(a,p) do a := a div p; ppow +:= 1; end while;

		// Find a random basis [T1,T2] of EFq[p]
  		ptors := Points( TorsionSubgroupScheme(EFq,p) );
		T1 := EFq!Random(ptors);
		while T1 eq EFq!0 do T1 := EFq!Random(ptors); end while;
		T2 := EFq!Random(ptors);
		while WeilPairing(T1,T2,p) eq 1 do T2 := EFq!Random(ptors); end while;

		// Consider the composite map
		//   phi:  Gmodp -> E(Q) -> E(F_q) -> E(F_q)[p] ->    mu_p     -> Z/p
		// defined as 
		//     P mod pG :->  P  :->   . . . .    :-> Weil( m*p^(e-1)*P, T) :-> Z/p
		//					          where T = T1 or T2
		for T in [T1, T2] do 
		   // define a linear functional phi : Gmodp -> GF(p)
		   Fp := RSpace( GFp, 1);
		   phi := hom< Gmodp -> Fp | 
			     [ (Log(WeilPairing(m*p^(e-1)*EFq!P , T, p)) div p^(ppow-1))*Fp.1 
							    : P in points ] >;
		   currentkernel := currentkernel meet Kernel(phi);
		end for;
		vprintf ECSaturation,3 : "currentkernel = %o\n", currentkernel;
	    end for;
	end if;

	// STEP 5: Naive division by p in E(Q).
	if not currentkernel subset Torsmodp then
          vprintf ECSaturation,1: "Looks like G is divisible by %o\n",p;
          // If there exists any p-divisible element, we can assume it is
          // non-torsion, and that the first nonzero entry in the 
          // corresponding vector in Gmodp/Torsmodp is a 1.
          // So, search over cosets of the form
          //    currentkernel/Torsmodp intersect {[0,...,0,1,*, ...,*]}
          // with 1 in the kth spot, until one p-divisible element is found, 
          // then increase k.
          // The coefficients in the torsion spots can also be assumed 0
          // when the order of the corresponding generator is prime to p.
          // Recall dim Gmodp = #points, and dim Torsmodp = t.
          newpts := [E(Rationals()) | ];
          pivots := [Integers()| ];
          for k := t+1 to #points do
            free_spots := { i : i in [1..t] | IsDivisibleBy(Invariants(T)[i],p) }
                          join { k+1 .. #points }; 
            free_subgrp  := currentkernel meet sub< Gmodp | [Gmodp.i : i in free_spots] >;
            subgrp_with_k:= currentkernel meet sub< Gmodp | [Gmodp.i : i in free_spots join {k}] >;
            if Dimension(subgrp_with_k) eq 0 then break; end if;
            bool := exists(v0){ v/v[k] : v in Basis(subgrp_with_k) | v[k] ne 0 };
            if not bool then continue; end if;  // this coset is empty, so increase k
            for w in free_subgrp do 
               v := v0 + w;
               P := &+[ Integers()!v[i]*points[i] : i in [1..#points] ];
               bool, newpt := IsDivisibleBy(P, p);
               if bool then 
                 vprintf ECSaturation,1: "Found a new point: %o\n",newpt;
                 Append( ~newpts, newpt);
                 Append( ~pivots, k);
                 break;
               end if;
            end for;
          end for;
          points := [ points[k] : k in [1..#points] | k notin pivots ] cat newpts;
          // this keeps the torsion part the same.
          // Nothing is lost here because {p*newpt : newpt in newpts},
          // together with the points we keep, generate the same group
          // as the old points.

          // Now continue saturation testing, starting with p again.
          if #newpts gt 0 then p -:= 1; end if;
        end if;   // end of Step 5.

   end while; 
  	
   if TorsionFree then
       return [E(Rationals()) | P : P in points | Order(P) eq 0];
   else
       return points;
   end if;

end intrinsic;
