freeze;

///////////////////////////////////////////////////////////////////////
//                                                                   //
//  Reduction of binary quartics over number fields                  //
//                                                                   //
//  Steve Donnelly (with some input from Tom Fisher)                 //
//  October 2011                                                     //
//                                                                   //
///////////////////////////////////////////////////////////////////////

function ConjMat(M)
  R := BaseRing(M);
  if ISA(Type(R), FldRe) then
    return M;
  else
    return Matrix(Ncols(M), [Conjugate(m): m in Eltseq(M)]);
  end if;
end function;

function TwoTorsionOrbits(E)
  a1,_,a3 := Explode(aInvariants(E));
  b2 := Explode(bInvariants(E));
  c4,c6 := Explode(cInvariants(E));
  K := BaseRing(E);
  P<X> := PolynomialRing(K);
  F := X^3 - 27*c4*X - 54*c6;
  FF := [f[1] : f in Factorization(F)];
  function torspt(L,rt)
    x := (rt - 3*b2)/36;
    return E(L)! [x, -(a1*x + a3)/2];
  end function;
  TT := <torspt(K,Roots(f)[1][1]) : f in FF | Degree(f) eq 1>;
  if exists(f){f : f in FF | Degree(f) gt 1} then
    L := NumberField(f : Check:=false);
    Append(~TT, torspt(L,L.1));
  end if;
  return TT;
end function;

function TwoTorsionOrbitsFromPrevious(E, TTPrev);
  E0, TT0 := Explode(TTPrev);
  bool, phi := IsIsomorphic(E0, E); assert bool;
  TT := < pt@phi : pt in TT0 >;
  // assert forall{T: T in TT | 2*T eq E!0};
  return TT;
end function;

function torsmat2(quartic,xT)
  a,b,c,d,e := Explode(Eltseq(quartic));
  phi := (-1/12)*xT;
  alpha1 := 4*a*phi - 8*a*c + 3*b^2;
  alpha2 := b*phi - 6*a*d + b*c;
  alpha3 := -2/3*phi^2 + 2/3*c*phi - 3*b*d + 4/3*c^2;
  M := Matrix(2,[-alpha2,-alpha3,alpha1,alpha2]);
  if alpha1 eq 0 then 
    beta1 := alpha2;
    beta2 := 1/3*phi^2 + 2/3*c*phi - 12*a*e + 1/3*c^2;
    beta3 := d*phi - 6*b*e + c*d;
    M := Matrix(2,[-beta2,-beta3,beta1,beta2]);
  end if;
  return M;
end function;

function TwoTorsionMatrices(E,quartic,torspts)
  assert cInvariants(E) eq cInvariants(quartic); // this is essential
  a1,a2,a3 := Explode(aInvariants(E));
  mats := < torsmat2(quartic, 36*T[1] + 3*(a1^2 + 4*a2)) : T in torspts >;
  return mats;
end function;

function Emb(M, indx, prec)
  return Matrix(2, [Conjugate(M[i,j],indx:Prec:=prec) : i,j in [1..2]]);
end function;

function Embs(M, K, u, prec)
  L := BaseRing(M);
  inds := L eq K select [[u]] else [[u,v] : v in [1..Degree(L,K)]];
  return [Emb(M,i,prec): i in inds];
end function;

// Reduction covariants associated to the given binary quartic 
// (over a number field K) given as list of (symmetric or Hermitian) 
// matrices corresponding to the real and complex embeddings of K

// The standard action on the ReductionCovariants corresponding to 
// the usual left action model :-> T*model 
// is M :-> Conjugate(T)*M*Transpose(T)

// TO DO: is there much value in TwoTorsionOrbitsFromPrevious?

function ReductionCovariants(model, prec : TTPrev:=0)
  epsilon := 10^(-prec div 2);
  K := NumberField(BaseRing(model));
  E := Jacobian(model);
if true or TTPrev cmpeq 0 then
    TT := TwoTorsionOrbits(E);
  else
    TT := TwoTorsionOrbitsFromPrevious(E, TTPrev);
  end if;
  mats := TwoTorsionMatrices(E, model, TT);
  AA := [**];
  for u in [1..Degree(K)] do
    MM := &cat[Embs(mat,K,u,prec): mat in mats];
    assert #MM eq 3;
    MM := [IdentityMatrix(BaseRing(MM[1]),2)] cat MM;
    dets := [Abs(Determinant(M)): M in MM];
    if exists{D: D in dets | D lt epsilon} then
      return false, _, _;
    end if;
    A := &+[Transpose(ConjMat(M))*M/Abs(Determinant(M)): M in MM];
    Append(~AA,A);
  end for;
  return true, AA, <E,TT>;
end function;

// covariants in upper half planes (or upper half spaces [TO DO])
// standard action is z :-> Inverse(Transpose(T)) * z

function covariants(mats, prec)
  epsilon := 10^(-prec div 2);
  R := RealField(prec);
  C := ComplexField(prec);
  covs := [C| ];
  for M in mats do 
    a,b,c,d := Explode(Eltseq(M));
    okay := Abs(b-c) lt epsilon;
    bb := b+c; 
    // solve quadratic by hand (assume real)
    s2 := R!(4*a*d - bb^2); 
    okay and:= s2 gt epsilon;
    s := Sqrt(s2);
    okay and:= Abs(s^2-s2) lt epsilon;
    r := (-bb + s*C.1)/(2*a);
    okay and:= Abs(a*r^2+bb*r+d) lt epsilon;
    if not okay then 
      return false, _; 
    end if;
    Append(~covs, r);
  end for;
  return true, covs;
end function;

// inverse of map to MinkowskiLattice

function inverse_mink(m : eps:=10^-10)
  B := Basis(Domain(m));
  BMI := BasisMatrix(Codomain(m))^-1;
  function mi(v)
    w := Vector(v)*BMI;
    ww := [Round(c) : c in Eltseq(w)];
    assert forall{i: i in [1..#ww] | Abs(w[i]-ww[i]) le eps};
    return &+ [ww[i]*B[i] : i in [1..#B]];
  end function;
  return mi;
end function;

// lattice with basis corresponding to the given units

function UnitLattice(units : prec:=GetKantPrecision() )
  K := NumberField(Universe(units));
  R := RealField(prec);
  conjs := [ChangeUniverse(Conjugates(u:Precision:=prec), R) : u in units];
  logs := [R| Log(Abs(c)) : c in &cat conjs];
  B := Matrix(Degree(K), logs);
  L := LatticeWithBasis(B);
  return L;
end function;

// Solve s*B = v for s, and return s rounded to integers.
// Includes work-around for the case of a unit lattice
// because Solution only works when B has full rank.

function iSolution(B, v)
  n := Nrows(B);
  d := Ncols(B);
  assert n in {d,d-1};
  if n eq d-1 then
    epsilon := 10^(20 - Precision(BaseRing(B)));
    // Assumption: rows must sum to zero
    assert forall{i: i in [1..n] | Abs(&+Eltseq(B[i])) lt epsilon};
    // Add independent row with length M much larger than all coords
    M := Round(10^100 * Max([Abs(x) : x in Eltseq(B) cat Eltseq(v)]) ); 
    B := VerticalJoin(B, Vector(BaseRing(B), [M : i in [1..d]]) );
  end if;
  s := v * B^-1;
  if n eq d-1 then
    assert Abs(s[d]) lt epsilon;
  end if;
  return [Round(s[i]) : i in [1..n]];
end function;

// TO DO: ClosestVectors(L, v : Max:=1) ought to work

function closest_vector(L, v : coordinates:=false)
  n := Rank(L);
  // Get some w in L, with w close to v
  // TO DO: also try w + s for some short vectors s
  B := BasisMatrix(LLL(L));
  s := iSolution(B, v);
  w := Vector(BaseRing(B),s) * B;

  if coordinates then
    // Get coords of w as a combination of the (fixed) basis of L
    B := BasisMatrix(L);
    c := iSolution(B, w);
    // Check it's a good approximation
    delta := w - Vector(BaseRing(B),c) * B;
    assert Norm(delta) lt 10^-20;
    return true, w, c;
  else
    return true, w;
  end if;
end function;


// Find the best pair (c,d) for the given covariants.
// B is the basis matrix for the image of ZK in Minkowski space.
// Test 'Num' short vectors.

function get_cd(covs, B, ZK : Num := 100*2^Degree(ZK) );
  n := Degree(ZK);

  xx := [Real(z) : z in covs];
  yy := [Im(z) : z in covs];

  // M = matrix that maps (c,d) in ZK^2 with coords wrt the basis B
  //                 to (c,c*x+d) in two copies of Minkowski space
  X := DiagonalMatrix(xx);
  M := BlockMatrix(2, 2, [Parent(B)| B, B*X, 0, B]);

  // apply weightings
  assert forall{y : y in yy | y gt 0};
  W := DiagonalMatrix(yy cat [1/y : y in yy]);
  G := M * W * Transpose(M);

  // get some short vectors 
  try
    L := LatticeWithGram(G + Transpose(G));
  catch ERR
    assert "positive" in ERR`Object; // "argument is not positive definite"
"WARNING: Lattice positive definite test failed with precision", Precision(BaseRing(G));
    return false, _, _, _;
  end try;
  b := Ceiling(Minimum(L) * Num^(1/n)); // L has rank 2*n
  repeat
    SV := ShortVectors(L, b);
    b *:= 2;
  until #SV ge Num;
  
  // choose whichever SV will give the largest average imaginary part:
  // for each SV, calculate product of |c*z+d|^2 over the real places
  cds := [];
  denoms := [];
  for i := 1 to Num do
    vv := SV[i,1];
    v := [Round(x) : x in Eltseq(vv)];
    if GCD(v) ne 1 then
      continue;
    end if;
    c := ZK! v[1..n];
    d := ZK! v[n+1..2*n];
    I := ideal<ZK|c,d>;
    if Minimum(I) ne 1 then 
      // If I is principal, we'll probably also find the primitive (c,d).
      // Otherwise to use (c,d) we'd need to divide out the common factor,
      // using IsPrincipal, which is not fast enough
      continue;
    end if;
    Append(~cds, <c,d>);
    vM := Vector(vv)*M;
    Append(~denoms, [Abs(vM[n+i])^2 + Abs(vM[i]*yy[i])^2 : i in [1..n]]);
                     // = |c*x+d|^2 + |c*y|^2
  end for;
//Num, #denoms, "Candidates:"; [RealField(4)| &*d : d in denoms];
  _, m := Minimum([&*d : d in denoms]);
  c, d := Explode(cds[m]);
  im1 := [yy[i]/denoms[m][i] : i in [1..n]];
  return true, c, d, im1;
end function;


function get_ab(c, d)
  ZK := Parent(c);
  if d eq 0 then
    return ZK!0, ZK!(-1/c);
  elif c eq 0 then
    return ZK!(1/d), ZK!0;
  end if;
  bool, x, y := Idempotents(c*ZK, d*ZK); assert bool;
  b := -x div c;
  a := y div d;
  assert a*d - b*c eq 1;
  return a, b;
end function;


intrinsic Reduce(model::ModelG1[FldAlg] : Effort:=10) -> ModelG1, TransG1
{A reduced genus one model integrally equivalent to the given one}
  
  require #Eltseq(model) eq 5 : "Implemented only for models without cross terms";

  K := BaseRing(model);
  require IsAbsoluteField(K): "The number field must be an absolute extension of Q";
  require IsTotallyReal(K): "Currently implemented only for totally real fields";

  n := Degree(K);

  if n eq 1 then
     modelQ, trQ := Reduce(ChangeRing(model, Rationals()));
     return ChangeRing(modelQ, K), ChangeRing(trQ, K);
  end if;

  ZK := LLL(Integers(K));
  U, mU := IndependentUnits(K : Al:="ClassGroup");

  // For verbose:
  RR := RealField(4);              
  CC<i> := ComplexField(4);

  // Translate by upper triangular matrix to shove everything near the middle
  procedure translate(~covs, ~T, LZK, fromLZK)
    xx := Vector([Re(z) : z in covs]); 
    ok, w := closest_vector(LZK, xx);
    assert ok;
    if ok and not IsZero(w) then
      a := ZK!(w @ fromLZK);
      vprintf Reduction, 2: "Translating by %o\n", [RR!c : c in Eltseq(w)];
      T1 := Matrix(2, [ZK| 1, -a, 0, 1]);
      T := Transpose(T1)^-1 * T;
      covs := [covs[i] - w[i] : i in [1..n]];
      vprint Reduction, 2: "Translated covariants =", ChangeUniverse(covs, CC);
    end if;
  end procedure;

  // The action by T sending model :-> T * model corresponds to 
  // the action on covariants in upper half planes sending
  //  z :-> Inverse(Transpose(T)) * z

  T := IdentityMatrix(ZK, 2); // T*model = reduced model
  Tprev := {};
  Swap := Matrix(2, [ZK| 0, -1, 1, 0]);

  // Stopping condition -- this gets gradually weakened below
  // TO DO: what should we hope for? 
  ybound := Round(Sqrt(Abs(Discriminant(ZK)))); 
  ybound := 1;

  // TO DO: what precision to start with? has to be the size of the coeffs?
  // but should only use large precision for the calculations where its needed
  model_size := Max([Ceiling(AbsoluteLogarithmicHeight(c)/Log(10)) : c in Eltseq(model)]);
  prec0 := Ceiling(3*model_size/100)*100;
  prec0 := Max(prec0, 100*n);

  inc := 0;
  iterations := 0;
  small_enough := false;
  best := <T, 0, 0>;     // will be < T, covs, ymin >

  while not small_enough do 

    inc +:= 1;
    prec := inc*prec0;
    increased_prec := true;
    vprint Reduction: "Initializing with precision", prec;
if inc ge 20 then
 "WARNING (in Reduce) : seems to need precision", prec;
end if;
    R := RealField(prec);
    epsilon := 10^(-prec div 2); // used in assertions

    LZK, toLZK := MinkowskiLattice(ZK : Precision:=prec);
    fromLZK := inverse_mink(toLZK : eps:=epsilon);

    enough_effort := false;
    small_enough := false;
    c0flag := false;

    while true do 
      iterations +:= 1;
      Tprev join:= {T,-T};
 
      // Recompute the covariants regularly (TO DO: how often?)
      MM := 2;
      if increased_prec or (iterations-1) mod MM eq 0 then
        m := tr*model where _,tr is IsTransformation(2, <K!1, T>);
        bool, mats, TT := ReductionCovariants(m, prec :
                          TTPrev := assigned TT select TT else 0);
//if not bool then printf " BAD RC "; end if;
        if bool then
          bool, covs1 := covariants(mats, prec);
//if not bool then printf " BAD cov "; end if;
        end if;
        if not bool then
          break; // increase precision and re-initialize
        end if;
        if increased_prec then
          increased_prec := false;
          covs := covs1;
          // also need the best covariants to the same precision
          mbest := tr*model where _,tr is IsTransformation(2, <K!1, best[1]>);
          bool, mats := ReductionCovariants(mbest, prec : TTPrev:=TT);
//if not bool then printf " BAD RC (best) "; end if;
          if bool then
            bool, covs_best := covariants(mats, prec);
//if not bool then printf " BAD cov (best) "; end if;
          end if;
          if not bool then
            break; // increase precision and re-initialize
          end if;
          best := < best[1], covs_best, Min([Im(z) : z in covs_best]) >;
        else
          // check we didn't lose too much precision
          // (transformed covs may be lower half plane, so flip them first)
          covs0 := [Im(c) gt 0 select c else Conjugate(c) : c in covs];
          if not forall{i: i in [1..n] | Abs((covs1[i]-covs0[i])/covs1[i]) lt epsilon} then
             // assert false;
             rp := Max([ Abs((covs1[i]-covs0[i])/covs1[i]) : i in [1..n] ]);
lrp := Round(-Log(10,rp));
"WARNING!!! Covariants were only correct to relative precision", lrp;
assert lrp ge 50;
             break; // increase precision and re-initialize
          end if;
          covs := covs1;
        end if;
      end if;

      vprint Reduction, 3: "Current model:\n", Equation(tr*model) where _,tr is IsTransformation(2, <K!1, T>);
      if IsVerbose("Reduction", 2) then
        vprint Reduction, 2: "Current covariants =", ChangeUniverse(covs, CC);
      elif IsVerbose("Reduction") then
        vprintf Reduction: "Current heights: %o\n", [RR| Abs(Im(z)) : z in covs];
      end if;

      bool, c, d, im1 := get_cd(covs, BasisMatrix(LZK), ZK);
      if not bool then
        break; // increase precision and start again
      end if;

      a, b := get_ab(c, d);
      T1 := Matrix(2, [a, b, c, d]);
      T := Transpose(T1)^-1 * T;
      covs := [ (Conjugate(a,i:Precision:=prec)*covs[i] + Conjugate(b,i:Precision:=prec)) / 
                (Conjugate(c,i:Precision:=prec)*covs[i] + Conjugate(d,i:Precision:=prec)) 
              : i in [1..n]];
      vprint Reduction, 2: "Improved covariants =", ChangeUniverse(covs, CC);

      // In principle, not necessary to translate at each step, 
      // but it will be much less stable if we don't
      translate(~covs, ~T, LZK, fromLZK);

      // Keep track of the best so far (measured by inf of imaginary parts)
      ymin := Min([RR| Abs(Im(z)) : z in covs]);
      if ymin gt best[3] then
        vprint Reduction: "Best height now:  ", ymin;
        best := < T, covs, ymin >;
      end if;

      // If we get into a loop, escape!
      if T notin Tprev then
        Tprev join:= {T,-T};
        continue;
      else
        // TO DO: exit strategy
        enough_effort := true;
        small_enough := true;
      end if;

      // The current set of iterations has got into a cycle
    //small_enough := best[3] gt 1/ybound;
    //enough_effort := iterations ge 100*Effort;
      small_enough or:= best[3] gt 1/ybound;
      enough_effort or:= iterations ge 100*Effort;
      if small_enough and enough_effort then
        break; 
      else
        // Stopping criterion needs to gradually weaken 
        // (does this weaken too rapidly?)
        if enough_effort then
          ybound +:= (ybound lt 2*w) select 1 else ybound div w where w is 10*Effort;
        end if;
      end if;

    end while; // true

  end while; // not small_enough

  T, covs := Explode(best);
  // 'best' has the best heights, but has not necessarily been translated

  translate(~covs, ~T, LZK, fromLZK); // TO DO: not if translation was the previous move

  scal := K!1;
  _,tr := IsTransformation(2, <scal, T>);
  redmodel := tr * model;

if true then
  // Scale quartic by a square unit
  // (don't need high precision here)
  // TO DO: this step should be redundant
  lowprec := 100;
  coeffs := [c : c in Eltseq(redmodel) | c ne 0];
  conjs := [Conjugates(c:Precision:=lowprec) : c in coeffs];
  logs := [[Log(Abs(conjs[i][j])) : i in [1..#conjs]] : j in [1..n]];
  ll := [&+l/#coeffs : l in logs]; // average log sizes
  ll_avg := (&+ll)/n;
  ll1 := [(l-ll_avg)/2 : l in ll];
  assert Abs(&+ll1) lt 10^(20-lowprec);
  assert Order(U.1) eq 2;
  units := [Codomain(mU) | U.i@mU : i in [2..n]];
  ok, _, c := closest_vector(UnitLattice(units:prec:=lowprec), Vector(ll1) : coordinates);
  assert ok; // otherwise closest_vector failed
  if ok and exists{x: x in c | x ne 0} then
    scal *:= K! &* [ZK| units[i]^-c[i] : i in [1..#units]];
    vprintf Reduction, 2: "Scaling quartic by square of unit %o\n", scal;
    _,tr := IsTransformation(2, <scal, T>);
    redmodel := tr * model;
  end if;
end if;

if true then
  // Scale quartic by a square: result has integral coeffs,
  // we want coeffs to have small height, and small content
  // TO DO: select appropriate prec here (then redefine LZK)
  coeffs := [c : c in Eltseq(redmodel) | c ne 0];
  conjs := [Conjugates(c:Precision:=prec) : c in coeffs];
  W := [Max([Abs(Real(conjs[i,j])) : i in [1..#coeffs]]) : j in [1..n]];
  content := Factorization(ideal< ZK | coeffs >);
  assert forall{t: t in content | t[2] ge 0};
  content2 := &* [PowerIdeal(ZK)| t[1] ^ (t[2] div 2) : t in content | t[2] ge 2];
  content2i := content2 ^ -1;
  bas2i := Basis(content2i);
  B := Matrix(RealField(prec), [Conjugates(b:Precision:=prec) : b in bas2i]);
  G := B * DiagonalMatrix(W) * Transpose(B);
  L := LatticeWithGram(G + Transpose(G));
  v := ShortestVectors(L)[1];
  try
    c := Coordinates(v);
    ok := true;
  catch ERR
    ok := false;
  end try;
error if not ok, "Coordinates failed (as usual)";
  if ok then
    scal *:= &+ [K| c[i] * bas2i[i] : i in [1..n]]; 
    vprintf Reduction, 2: "Scaling quartic by square of non-unit %o\n", scal;
    _, tr := IsTransformation(2, <scal, T>);
//Norm(ideal<ZK|Eltseq(redmodel)>); [Conjugates(c:Precision:=6) : c in Eltseq(redmodel)];
    redmodel := tr * model;
    assert forall{c : c in Eltseq(redmodel) | IsIntegral(c)};
//Norm(ideal<ZK|Eltseq(redmodel)>); [Conjugates(c:Precision:=6) : c in Eltseq(redmodel)]; "";
  end if;

end if;

  return redmodel, tr;

end intrinsic;

/**********************************************************************************
//////////  Examples  ///////////

SetVerbose("Reduction", 2);

F, ZF, w := nf(x^2 - 5);
E := EllipticCurve([F| 0, 2767587264 ]);
td := TwoDescent(E); td; 
// one of the three was badly reduced, even though can be written over Q

F, ZF, w := nf (x^2 - 2);
E := EllipticCurve([0, (220259326464*w + 292251234816) ]);
TD := TwoDescent(E); TD;

F, ZF, w := nf(x^2 - x - 52);
E := EllipticCurve([F| 0, (-29259832766899423680*w - 196872331259043751680) ]);
TD := TwoDescent(E); TD;
// one of the 7 is sometimes less well reduced

F, ZF, w := nf([1, -1, -4, 0, 1]);
E := EllipticCurve([ 0, -1728 * (3*w^3 - 18*w^2 - 32*w + 16) ]);
TD := TwoDescent(E : RemoveTorsion); TD;

//////////  Testing  ///////////

SetVerbose("Selmer", 0);
SetVerbose("Conic", 0);
SetVerbose("Reduction", 2);

F, ZF, w := nf(x^3 - x^2 - 3*x + 1);

for n := 1 to 100 do 
    A := EllipticCurve([1, n*w]); 

for E in CremonaDatabase() do
    A := BaseChange(E,F);
    Norm(Conductor(A)), A;
    td := TwoDescent(A : MinRed:=false); td;
    if #td gt 0 then
        ProStart();
        td := TwoDescent(A : MinRed:=true); td;
        ProShow(20);
    end if;
end for;

**********************************************************************************/

