freeze;

////////////////////////////////////////////////////////////
// Chabauty + Mordell-Weil Sieve                          //
//                                                        //
// Michael Stoll, 2007-2008                               //
////////////////////////////////////////////////////////////

// Main function:
//   Chabauty(ptJ, ptC)
// where 
// * ptC is a rational point on C, a genus 2 curve over Q
// * ptJ is a rational point on J, the Jacobian of C,
//   such that ptJ generates the Mordell-Weil group J(Q) modulo torsion.
// The function returns three values:
// 1. the set of rational points on C,
// 2. the set of primes p such that information mod p was used,
// 3. the sequence of primes q that were used to build B,
//    where B is a number such that
//    + C(Q) --> J(Q) --> J(Q)/B J(Q) is injective  and
//    + for the cosets in J(Q)/ B J(Q) not in the image of the map above,
//      the information from the set of primes p (second return value)
//      proves that they are not in the image.

/* Example:
 P<x> := PolynomialRing(Rationals());
 C := HyperellipticCurve(P![1,8,22,22,5,6,1]);
 ptC := C![0,1];
 ptJ := C![0,1] - C![0,-1]; // known to generate J(Q)
 SetVerbose("Chabauty", 1); // to see something...
 Chabauty(ptJ);
 Chabauty(ptJ : ptC:=ptC);
*/

/***********************************************************************
  Changes log
  -----------

 Early 2008, Steve: 
   -- Define Kummer surface in Chabauty to avoid repeated construction

 November 2008 (for release in V2.15), Steve:
   -- ptC is now an optional argument (if not given, search for one)

   -- start with Epsilon:=0.1 instead of 0.01, often a bit quicker 
      for simple curves (I'll experiment more with parameters though)
  
 ***********************************************************************/

/***********************************************************
 disclog.m
 
 General purpose discrete log for groups of smooth order
 
 M. Stoll, started 2005-06-07
 converted to package file 2006-03-05

 **********************************************************/

// given: Abelian group G, bijective map G -> X, X some structure
// #G smooth (so that for groups of order p^f|#G, lookup is feasible)

// Exported intrinsics: 
// DiscreteLogMapSmooth(G, m)

function MakeLookup1(G, m)
  return pmap<Codomain(m) -> G| [<m(g), g> : g in G]>;
end function;

function MakeDLp1(G, m, p)
  // G a p-group
  if #G le 25 then
    return MakeLookup1(G, m);
  end if;
  invs := Invariants(G);
  // printf "MakeDLp: Invariants(G) = %o\n", invs;
  pp := ExactQuotient(invs[#invs], p);
  if pp eq 1 then
    return MakeLookup1(G, m);
  end if;
  // printf "MakeDLp: pp = %o\n", pp;
  h := hom<G -> G | [pp*G.i : i in [1..#invs]]>;
  G1 := Image(h);
  // printf "MakeDLp: Invariants(Image(h)) = %o\n", Invariants(G1);
  m1 := map<G1 -> Codomain(m) | x:->m(x)>;
  f1 := MakeLookup1(G1, m1);
  G2 := Kernel(h);
  // printf "MakeDLp: Invariants(Kernel(h)) = %o\n", Invariants(G2);
  m2 := map<G2 -> Codomain(m) | x:->m(x)>;
  f2 := MakeDLp1(G2, m2, p);
  return pmap<Codomain(m) -> G |
               x :-> f2(x - m(a)) + a where a := f1(pp*x) @@ h>;
end function;

intrinsic DiscreteLogMapSmooth(G::GrpAb, m::Map) -> Map
{Given a bijection  m : G -> X, this returns an inverse map. 
 X must have group structure (compatible with G via m) and must have 
 operations 'x+y' and 'n*x' defined for x,y in X and integers n. 
 The returned inverse map X -> G may be more efficient to evaluate than 
 the standard inverse given by Inverse(m), in the case where #G is smooth.
 (This intrinsic is temporary and will be removed in a later release)}

  n := #Invariants(G);
  f := Factorization(#G);
  cofs := [&*[Integers()|f[i,1]^f[i,2] : i in [1..#f] | i ne j] : j in [1..#f]];
  _, refs := XGCD(cofs);
  assert &+[Integers()|refs[i]*cofs[i] : i in [1..#f]] eq 1;
  DLs := [**];
  for i := 1 to #f do
    p := f[i,1];
    hp := hom<G -> G | [cofs[i]*G.j : j in [1..n]]>;
    Gp := Image(hp);
    mp := map<Gp -> Codomain(m) | x:->m(x)>;
    DLp := MakeDLp1(Gp, mp, p);
    Append(~DLs, DLp);
  end for;
  return pmap<Codomain(m) -> G
               | x :-> &+[G|refs[i]*G!(DLs[i](cofs[i]*x)) : i in [1..#f]]>;
end intrinsic;

/************************************************************
 * chabauty.m                                               *
 *                                                          *
 * Michael Stoll, started 2006-07-19                        *
 *                                                          *
 * Routines for Chabauty's method on genus 2 curves         *
 ************************************************************/

declare verbose Chabauty, 3;

// Find the reduction mod p of the differential over Q_p that kills
// the subgroup generated by a point on the Jacobian
intrinsic KillingDifferentialModp(pt::JacHypPt, p::RngIntElt) -> Pt
{ Given a point pt of infinite order on the Jacobian of a genus 2 curve
  over the rationals of the form y^2 = f(x) and a prime p of good reduction,
  this returns a point in P^1(F_p) which is the image of the two zeros
  of omega mod p, where omega is a Q_p-differential on the curve that kills
  the group generated by pt. }
  require Order(pt) eq 0: "Point pt must have infinite order.";
  J := Parent(pt);
  require Dimension(J) eq 2: "Point pt must be on a genus 2 Jacobian.";
  C := Curve(J);
  Jp := BaseChange(J, GF(p));
  n := Order(Jp!pt);
  prec := 1;
  K := KummerSurface(J);
  repeat
    prec +:= 2;
    // TO DO: this BaseChange involves BaseChange of J, which takes sig time
    K1 := BaseChange(K, pAdicField(p, prec)); 
    ptK := n*K1!K!pt;
  until Min([Valuation(ptK[i]) : i in [1..3]]) eq prec - 1;
  coords := [GF(p) | ExactQuotient(Integers()!ptK[i], p^(prec-1))
                   : i in [1..3]];
  assert coords[2]^2 eq 4*coords[1]*coords[3];
  Pr1 := ProjectiveSpace(GF(p), 1);
  return coords[1] eq 0 select Pr1![1,0] else Pr1![coords[2]/2,coords[1]];
end intrinsic;

function GroupInfo(J, bound, bas, ptC, GIlb, PrimeBound, SmoothBound)
// Construct a list of "interesting" primes and information on
// C(F_p) and J(F_p)
// An entry is [* p, imbas, imC, flag *],
// where p is the prime,
//       imbas is the image of the given generators of J(Q) in G ~ J(F_p),
//       imC is the image of C(F_p) in G,
//       flag is true iff points in C(F_p) are separated in the image group.
// GIlb is a lower bound on the primes.
  bp := Seqset(BadPrimes(J));
  res := [Parent([**]) | ];
  lb := Max(3, IsEven(GIlb) select GIlb+1 else GIlb);
  for p in Type(bound) eq RngIntElt select [lb..bound by 2] else bound do
    if IsPrime(p) and p notin bp then
      Jp := BaseChange(J, GF(p));
      oG := #Jp;
      if Max(PrimeDivisors(oG)) lt SmoothBound then
        G, m := AbelianGroup(Jp);
        vprintf Chabauty, 2: " GroupInfo: p = %o...\n", p;
        I := Invariants(G);
        Cp := BaseChange(Curve(J), Bang(Rationals(), GF(p)));
        pts := Points(Cp);
        vprintf Chabauty, 2:"   #C(F_p) = %o, Invariants(G) = %o\n", #pts, I;
        fI := Factorization(I[#I]);
        vprintf Chabauty, 2: "   Exponent = %o\n", fI;
        DL := DiscreteLogMapSmooth(G, m);
        imbas := [DL(Jp!b) : b in bas];
        orders := [Order(g) : g in imbas];
        fL := Factorization(LCM(orders));
        vprintf Chabauty, 3: "   Exponent of image = %o\n", fL;
        inj := func<pt | Jp![Cp!pt, bpt]> where bpt := Cp!ptC;
        imC := {DL(inj(pt)) : pt in pts};
        Gsub := sub<G | imbas>;
        imC := {Gsub | pt : pt in imC | pt in Gsub};
        imbas := ChangeUniverse(imbas, Gsub);
        e1 := &*[Integers()| a[1]^a[2] : a in fL | a[1] lt PrimeBound];
        if e1 lt LCM(orders) then
          // kick out primes larger than PrimeBound
          Gq, qmap := quo<Gsub | [e1*g : g in Generators(Gsub)]>;
          imbas := [qmap(b) : b in imbas];
          imCold := imC;
          imC := {qmap(c) : c in imC};
          if #imC eq #Gq then
            vprintf Chabauty, 2:
                    "   Relevant part does not give information ==> discard.\n";
          else
            Append(~res, [* p, imbas, imC, #imC eq #imCold *]);
          end if;
        else // e1 = LCM(orders)
          Append(~res, [* p, imbas, imC, true *]);
        end if; // e1 lt LCM(orders)
      end if; // Max(PrimeDivisors(oG)) lt SmoothBound
    end if; // IsPrime(p) and p notin bp
  end for; // p
  return res;
end function;

function CheckKillingDifferential(f, ptJ, entry)
// Select the entries in GI (setting entry[4] = true), for which
// the Q_p-differential killing the group generated by ptJ does not vanish
// on any point of the curve mod p
  if not entry[4] then return entry; end if;
  p := entry[1];
  cl := KillingDifferentialModp(ptJ, p);
  fp := PolynomialRing(GF(p))!f;
  if cl[2] eq 0 select not IsSquare(Coefficient(fp, 6))
                else not IsSquare(Evaluate(fp, cl[1]/cl[2]))
  then
    vprintf Chabauty, 2:
            "Found prime p = %o with killing differential not vanishing on C(F_p).\n", p;
    return entry;
  else
    return [* p, entry[2], entry[3], false *];
  end if;
end function;

function GetMinimalDivisors(set)
  res := {};
  while not IsEmpty(set) do
    a := Min(set);
    set := {s : s in set | not IsDivisibleBy(s, a)};
    Include(~res, a);
  end while;
  return res;
end function;

function FindMinimalBs(GI)
  function bestB(entry)
    G := Universe(entry[2]);
    for d in Divisors(Exponent(G)) do
      qG, qmap := quo<G | [d*g : g in OrderedGenerators(G)]>;
      if #{qmap(c) : c in entry[3]} eq #entry[3] then
        vprintf Chabauty, 2: "Best B for p = %o is %o\n", entry[1], d;
        return d;
      end if;
    end for;
    // never reached
  end function;
  return GetMinimalDivisors({bestB(entry) : entry in GI | entry[4]});
end function;

function dryRun1(GI, Bound, eps, tors, Bs)
  vprintf Chabauty, 1: "dryRun: Bound = %o, eps = %o\n", Bound, eps;
  function frac(B, entry)
    I := Invariants(Universe(entry[2]));
    Is := [Integers() | GCD(i,B) : i in I];
    n := &*Is;
    if Is eq I then return 1.0*#entry[3]/n; end if;
    cs := {};
    for cc in entry[3] do
      Include(~cs, [c[i] mod Is[i] : i in [1..#Is]] where c := Eltseq(cc));
      if #cs eq n then return 1.0; end if;
    end for;
    return 1.0*#cs/n;
  end function;
  red := func<B | B*tors*&*[frac(B, e) : e in GI]>;
  prl := [2] cat [p : p in [3..Bound by 2] | IsPrime(p)];
  max := &*[p^Max(&cat[[Valuation(i,p) : i in Invariants(Universe(e[2]))]
                         : e in GI])
             : p in prl];
  vprintf Chabauty, 2: "        max = %o\n", max;
  pl := [p : p in prl | IsDivisibleBy(max, p)];
  // optimze according to max red(B_i)
  // if B_0 = 1, B_{i+1} = B_i*p_i
  agendas := [[<1, 1, 1.0>]]; // entries <p, B, s>
  agendav :=[1.0]; 
  reached := {1}; // lists the values of B that have been seen so far
  while true do
    v, pos := Min(agendav); // try to extend cheapest path so far
    l := agendas[pos];
    vprintf Chabauty, 2: "  dryRun: l = %o,\n   v = %o\n",
                   [l[i,1] : i in [2..#l]], v;
    Remove(~agendas, pos);
    Remove(~agendav, pos);
    B := l[#l,2];
    s := l[#l,3];
    if B eq max or (s lt eps and exists{b : b in Bs | IsDivisibleBy(B,b)}) then 
      break;  // reached success criterion
    end if;
    // list possible extensions
    new := [<p, Bp, red(Bp)> : p in pl 
                             | Bp notin reached and IsDivisibleBy(max, Bp)
                               where Bp := B*p];
    // and update agenda
    agendas cat:= [Append(l, n) : n in new];
    agendav cat:= [n[3] : n in new];
    reached join:= {n[2] : n in new};
  end while;
  vprintf Chabauty, 1: "dryRun:\n Multipliers: %o\n",
                [l[i,1] : i in [2..#l]];
  vprintf Chabauty, 1: " Predicted sizes: %o\n", [Round(l[i,3]) : i in [2..#l]];
  return [<l[i,1], l[i,3]> : i in [2..#l]], s lt eps;
end function;

// Given a set of candidates <a, t> mod B, return all pairs mod B*p lifting
// them and compatible with GI. Return "primes" used in the process as second
// value.
function LiftInformation(S, GI, B, p)
  // S = {<a, t>, ...} , a : integer, t : element of abstract torsion subgroup
  // preparation: find relevant entries in GI
  vpB := Valuation(B, p);
  Bp := B*p;
  tests := [e : e in GI | Valuation(Exponent(Universe(e[2])), p) gt vpB];
  tests := [[* e[1], [q(b) : b in e[2]], {q(c) : c in e[3]} *]
              where _, q := quo< G | [Bp*g : g in Generators(G)] >
              where G := Universe(e[2])
             : e in tests];
  tests := [e : e in tests | #e[3] lt #Universe(e[3])];
  Sort(~tests, func<a,b | #a[3]/#Universe(a[3]) - #b[3]/#Universe(b[3])>);
  if IsVerbose("Chabauty", 2) then
    printf "LiftInformation: B = %o, p = %o, #S = %o\n", B, p, #S;
    if IsEmpty(tests) then
      printf "  no relevant entries.\n";
    else
      printf "  relevant entries:\n";
      for e in tests do
        printf "    %o : #G = %o, #imC = %o\n", e[1], #Universe(e[3]), #e[3];
      end for;
    end if;
  end if;
  res := {};
  primes := {};
  for tup in S do
    cands := {<a + i*B, t> : i in [0..p-1]}
               where a := tup[1] where t := tup[2];
    size := #cands;
    for e in tests do
      cands := {c : c in cands
                  | c[1]*e[2][1]
                     + &+[Universe(e[3])|ts[i]*e[2][i+1] : i in [1..#ts]]
                     in e[3]
                     where ts := Eltseq(c[2])};
      if #cands lt size then
        size := #cands; Include(~primes, e[1]);
      end if;
    end for;
    res join:= cands;
  end for;
  if IsVerbose("Chabauty", 2) then
    if not IsEmpty(tests) then
      printf "  entries used: %o\n", primes;
    end if;
    printf "  size of conditions mod %o is %o\n", Bp, #res;
  end if;
  return res, primes;
end function;

// Compute a heuristic value indicating whether we have collected
// enough information.
function CheckInvariant(GI, tors)
  rep := 10;
  function frac(B, entry)
    I := Invariants(Universe(entry[2]));
    Is := [Integers() | GCD(i,B) : i in I];
    n := &*Is;
    if Is eq I then return 1.0*#entry[3]/n; end if;
    cs := {};
    for cc in entry[3] do
      Include(~cs, [c[i] mod Is[i] : i in [1..#Is]] where c := Eltseq(cc));
      if #cs eq n then return 1.0; end if;
    end for;
    return 1.0*#cs/n;
  end function;
  //
  Gs := [Universe(e[2]) : e in GI];
  primes := &join[Seqset(PrimeDivisors(#G)) : G in Gs];
  // for each prime, find largest exponents in invariants of product
  Bs := [1 : j in [1..rep]];
  for p in primes do
    vals := &join[{* Valuation(i, p) : i in Invariants(G) *} : G in Gs];
    m := Max(vals);
    Exclude(~vals, m);
    for j := 1 to rep do
      Bs[j] *:= p^m;
      if m ne 0 then
        if IsEmpty(vals) then
          m := 0;
        else
          m := Max(vals);
          Exclude(~vals, m);
        end if;
      end if;
    end for;
  end for;
  // compute estimated number of candidates for B
  rs := [1.0*B*tors : B in Bs];
  for e in GI do 
    rs := [rs[j]*frac(Bs[j], e) : j in [1..rep]]; 
  end for;
  return Min(rs);
end function;

function CheckPoint(gen, l, t, ptC, ht)
// Checks if l*gen + t (t a torsion point) gives
// a point on J that is of the form pt - ptC.
// Returns a flag, and (if successful) the point.
// ht is a bound for the logarithmic naive height of the point on J
  C := Parent(ptC);
  prec := 0.0;
  p := 101;
  disc := Integers()!Discriminant(Curve(Parent(ptC)));
  ht1 := ht + 2*Log(Max([Abs(ptC[1]), Abs(ptC[3])])) + Log(3);
  J := Parent(gen);
  while prec lt ht1 do
    if not IsDivisibleBy(disc, p) then
      Jp := BaseChange(J, GF(p));
      ptCp := Curve(Jp)!ptC;
      ptJp := l*(Jp!gen) + Jp!t;
      if not ( ptJp eq Jp!0 or
             ( ptCp[3] eq 0
               select (Degree(ptJp[1]) lt 2 
                        and Coefficient(ptJp[2], 3) eq -ptCp[2])
               else (Evaluate(ptJp[1], x) eq 0
                      and Evaluate(ptJp[2], x) eq -ptCp[2]/ptCp[3]^3
                     where x := ptCp[1]/ptCp[3]) ) )
      then
        vprintf Chabauty, 3: "  CheckPoint: failed at p = %o\n", p;
        return false, _;
      end if;
    end if;
    prec +:= Log(p);
    p := NextPrime(p);
  end while;
  // Now there should be a point; find it!
  ptJ := l*gen + t;
  assert ptJ eq J! 0 or
         ( ptC[3] eq 0
               select (Degree(ptJ[1]) lt 2 
                        and Coefficient(ptJ[2], 3) eq -ptC[2])
               else (Evaluate(ptJ[1], x) eq 0
                      and Evaluate(ptJ[2], x) eq -ptC[2]/ptC[3]^3
                     where x := ptC[1]/ptC[3]) );
  if ptJ eq J!0 then return true, ptC; end if;
  if ptC[3] eq 0 then
    if Degree(ptJ[1]) lt 1 then
      // other point also at infinity
      return true, C![1, Coefficient(ptJ[2], 3), 0];
    else
      return true, C![x, Evaluate(ptJ[2], x), 1]
                   where x := -Coefficient(ptJ[1], 0);
    end if;
  else
    if Degree(ptJ[1]) lt 2 then
      // other point at infinity
      return true, C![1, Coefficient(ptJ[2], 3), 0];
    else
      return true, C![x, Evaluate(ptJ[2], x), 1]
                   where x := -Coefficient(ptJ[1], 1) - ptC[1]/ptC[3];
    end if;
  end if;
end function;

//=========================================================================
// Find degree 2 and degree 3 genus 1 subcovers of a genus 2 curve 
// when they exist.

// Degree 2
// ========

// We need an "extra involution" in the automorphism group of C.

function ExtraInvolutions(C)
  A, mA := AutomorphismGroup(C);
  moebius := func<aut | Matrix([[a,b], [c,d]])
                        where a := MonomialCoefficient(ax, P.1)
                        where b := MonomialCoefficient(az, P.1)
                        where c := MonomialCoefficient(ax, P.3)
                        where d := MonomialCoefficient(az, P.3)
                        where P := Parent(ax)
                        where ax,_,az := Explode(DefiningPolynomials(aut))>;
  I := moebius(mA(A!1));
  return {moe : a in A
              | Order(a) eq 2 and moe ne I where moe := moebius(mA(a))};
end function;

// Given an "extra involution" (as a matrix), compute the subcover and map.

function Degree2Subcover(C, inv)
  trmat := Matrix([Eltseq(Basis(Eigenspace(inv, 1))[1]),
                   Eltseq(Basis(Eigenspace(inv, -1))[1])]);
  f := HyperellipticPolynomials(C);
  P := Parent(f);
  l1 := P!Eltseq(Transpose(trmat)[1]);
  l2 := P!Eltseq(Transpose(trmat)[2]);
  pol := &+[Coefficient(f, j)*l1^j*l2^(6-j) : j in [0..6]];
  assert Evaluate(pol, -P.1) eq pol;
  C1 := HyperellipticCurve(pol);
  tri := trmat^-1;
  mz := tri[1,1]*C.1 + tri[2,1]*C.3;
  mx := tri[1,2]*C.1 + tri[2,2]*C.3;
  mCC1 := map<C -> C1 | [mx, C.2, mz]>;
  pol3 := P![Coefficient(pol, 2*j) : j in [0..3]];
  E1 := HyperellipticCurveOfGenus(1, pol3);
  mC1E1 := map<C1 -> E1 | [C1.1^2, C1.2*C1.3, C1.3^2]>;
  EE1, mE1E := EllipticCurve(E1);
  E2 := HyperellipticCurveOfGenus(1, P![Coefficient(pol3, 3-j) : j in [0..3]]);
  mC1E2 := map<C1 -> E2 | [C1.3^2, C1.2*C1.1, C1.1^2]>;
  EE2, mE2E := EllipticCurve(E2);
  return <EE1, Expand(mCC1*mC1E1*mE1E)>, <EE2, Expand(mCC1*mC1E2*mE2E)>;
end function;

// Main function, returns all degree 2 genus 1 subcovers as a list
// of pairs <E, mCE>.

intrinsic Degree2Subcovers(C::CrvHyp) -> List
{Returns a list of pairs <E, mCE>, where E is an elliptic curve and mCE
 is a map of degree 2 from C to E. C must be of genus 2 and of the form
 y^2 = f(x).}
  f, h := HyperellipticPolynomials(C);
  require h eq 0: "Curve must be of the form y^2 = f(x)";
  require Genus(C) eq 2: "Curve must be of genus 2";
  result := [**];
  for inv in ExtraInvolutions(C) do
    cov1, cov2 := Degree2Subcover(C, inv);
    Append(~result, cov1);
    Append(~result, cov2);
  end for;
  return result;
end intrinsic;

// Degree 3
// ========

// Here, the right hand side in  y^2 = f(x)  has to factor into two cubics
//  f(x) = g(x) h(x) .
// The covering map  C --> E  induces a map of degree 3,  phi: P^1 --> P^1
// that sends the roots of g to infinity and is ramified at some rational point.
// So we can write  phi(x) = (x - u1)^2 (x - u2) / g(x)
// (with the understanding that u1 or/and u2 might be infinity).
// The other three ramification points must be the roots of h.
// 
// We compute
//  d(x) := disc_t(x g(t) - (t-u1)^2 (t-u2))/x,  a cubic in x,
// and then
//  d(phi(x)) = -g(x) H(x,u1,u2) (k(x,u1,u2)/g(x)^2)^2 .
// If  H(x,u1,u2) = const h(x) , then 
//  E : y^2 = -const d(x)  is the subcover,
// and
//  C --> E,  (x, y) |--> (phi(x), const y k(x,u1,u2)/g(x)^2)
// is the map.
//
// The polynomials H and k can be computed generically.

function Hpoly(g, u1, u2)
  // Gives the coefficients of H
  a0, a1, a2, a3 := Explode([Coefficient(g, j) : j in [0..3]]);
  return
    [- 4*a0^2*u1 + a0^2*u2 - 4*a0*a1*u1^2 - 4*a0*a2*u1^3 - 2*a0*a2*u1^2*u2 
        - 4*a0*a3*u1^4 - 8*a0*a3*u1^3*u2 - 4*a1*a3*u1^4*u2 + a2^2*u1^4*u2,
     3*a0^2 + 2*a0*a1*u2 + 6*a0*a2*u1^2 + 4*a0*a2*u1*u2 + 12*a0*a3*u1^3 
          + 18*a0*a3*u1^2*u2 - 4*a1^2*u1^2 - 4*a1*a2*u1^3 - 2*a1*a2*u1^2*u2 
          + 8*a1*a3*u1^3*u2 - a2^2*u1^4 - 4*a2^2*u1^3*u2 - 2*a2*a3*u1^4*u2,
     2*a0*a1 - 8*a0*a2*u1 - 18*a0*a3*u1^2 - 12*a0*a3*u1*u2 + 4*a1^2*u1 
          + a1^2*u2 + 2*a1*a2*u1^2 + 4*a1*a2*u1*u2 - 4*a1*a3*u1^3 
          - 6*a1*a3*u1^2*u2 + 4*a2^2*u1^2*u2 - 2*a2*a3*u1^4 - 3*a3^2*u1^4*u2,
     4*a0*a2 + 8*a0*a3*u1 + 4*a0*a3*u2 - a1^2 + 2*a1*a3*u1^2 + 4*a1*a3*u1*u2 
       + 4*a2*a3*u1^2*u2 - a3^2*u1^4 + 4*a3^2*u1^3*u2];
end function;

function kpoly(g, u1, u2)
  a0, a1, a2, a3 := Explode([Coefficient(g, j) : j in [0..3]]);
  return Parent(g)![-a0*u1 - 2*a0*u2 - a1*u1*u2,
                     3*a0 - a1*u2 - 2*a2*u1*u2,
                     2*a1 + a2*u1 - 3*a3*u1*u2,
                     a2 + 2*a3*u1 + a3*u2];
end function;

// Now the function that checks whether the factors g and h work.
// It returns a flag; if true, the second and third return values
// are the elliptic curve E and the map C --> E.

function TestDeg3Subcover(g, h, C)
  P := Parent(g);
  P2<u1,u2> := PolynomialRing(CoefficientRing(P), 2);
  H := Hpoly(g, u1, u2);
  hs := [Coefficient(h, j) : j in [0..3]];
  I := ideal<P2 | [hs[j]*H[k] - hs[k]*H[j] : k in [j+1..4], j in [1..4]]>;
  vI := Variety(I);
  if not IsEmpty(vI) then
    assert #vI eq 1;
    sol := [vI[1,1], vI[1,2]];
    Hi := P![Evaluate(c, sol) : c in H];
    assert IsDivisibleBy(Hi, h);
    const := Evaluate(ExactQuotient(Hi, h), 0);
    PP := PolynomialRing(P);
    t := P.1*Evaluate(g, PP.1) - (PP.1-sol[1])^2*(PP.1-sol[2]);
    d := ExactQuotient(Discriminant(t), P.1);
    assert IsSquarefree(d);
    d *:= -const;
    E1 := HyperellipticCurveOfGenus(1, d);
    E, mE1E := EllipticCurve(E1);
    k := kpoly(g, sol[1], sol[2]);
    mx := (C.1 - sol[1]*C.3)^2*(C.1 - sol[2]*C.3);
    my := const*C.2*&+[Coefficient(k, j)*C.1^j*C.3^(3-j) : j in [0..3]];
    mz := &+[Coefficient(g, j)*C.1^j*C.3^(3-j) : j in [0..3]];
    mCE1 := map<C -> E1 | [mx, my, mz]>;
    return true, E, mCE1*mE1E;
  end if;
  // check special case  u1 = infinity
  H1 := [Evaluate(Coefficient(c, u1, 4), [0, P.1]) : c in H];
  seq := [hs[j]*H1[k] - hs[k]*H1[j] : k in [j+1..4], j in [1..4]];
  mat := Transpose(Matrix([[Coefficient(c, k) : k in [0,1]] : c in seq]));
  ker := Kernel(mat);
  if Dimension(ker) gt 0 then
    assert Dimension(ker) eq 1;
    b := Basis(ker)[1];
    if b[2] eq 0 then
      // u2 = infty as well
      Hi := P![Coefficient(c, 1) : c in H1];
      assert IsDivisibleBy(Hi, h);
      const := Evaluate(ExactQuotient(Hi, h), 0);
      PP := PolynomialRing(P);
      t := P.1*Evaluate(g, PP.1) - 1;
      d := ExactQuotient(Discriminant(t), P.1);
      assert IsSquarefree(d);
      d *:= -const;
      E1 := HyperellipticCurveOfGenus(1, d);
      E, mE1E := EllipticCurve(E1);
      k := P![-a0-a1, -2*a2, a2-3*a3, 0]
             where a0,a1,a2,a3 := Explode([Coefficient(g, j) : j in [0..3]]);
      mx := C.3^3;
      my := const*C.2*&+[Coefficient(k, j)*C.1^j*C.3^(3-j) : j in [0..3]];
      mz := &+[Coefficient(g, j)*C.1^j*C.3^(3-j) : j in [0..3]];
      mCE1 := map<C -> E1 | [mx, my, mz]>;
      return true, E, mCE1*mE1E;
    else
      sol2 := -b[1]/b[2];
      Hi := P![Evaluate(c, sol2) : c in H1];
      assert IsDivisibleBy(Hi, h);
      const := Evaluate(ExactQuotient(Hi, h), 0);
      PP := PolynomialRing(P);
      t := P.1*Evaluate(g, PP.1) - (PP.1-sol2);
      d := ExactQuotient(Discriminant(t), P.1);
      assert IsSquarefree(d);
      d *:= -const;
      E1 := HyperellipticCurveOfGenus(1, d);
      E, mE1E := EllipticCurve(E1);
      k := P![-a0-a1*sol2, -2*a2*sol2, a2-3*a3*sol2, 2*a3]
             where a0,a1,a2,a3 := Explode([Coefficient(g, j) : j in [0..3]]);
      mx := C.3^2*(C.1 - sol2*C.3);
      my := const*C.2*&+[Coefficient(k, j)*C.1^j*C.3^(3-j) : j in [0..3]];
      mz := &+[Coefficient(g, j)*C.1^j*C.3^(3-j) : j in [0..3]];
      mCE1 := map<C -> E1 | [mx, my, mz]>;
      return true, E, mCE1*mE1E;
    end if;
  end if;
  // check special case  u2 = infinity
  H1 := [Evaluate(Coefficient(c, u2, 1), [P.1, 0]) : c in H];
  pol := GCD([hs[j]*H1[k] - hs[k]*H1[j] : k in [j+1..4], j in [1..4]]);
  if Degree(pol) gt 0 then
    assert Degree(pol) eq 1;
    sol1 := -Coefficient(pol, 0)/Coefficient(pol, 1);
    Hi := P![Evaluate(c, sol1) : c in H1];
    assert IsDivisibleBy(Hi, h);
    const := Evaluate(ExactQuotient(Hi, h), 0);
    PP := PolynomialRing(P);
    t := P.1*Evaluate(g, PP.1) - (PP.1-sol1)^2;
    d := ExactQuotient(Discriminant(t), P.1);
    assert IsSquarefree(d);
    d *:= -const;
    E1 := HyperellipticCurveOfGenus(1, d);
    E, mE1E := EllipticCurve(E1);
    k := P![-2*a0-a1*sol1, -a1-2*a2*sol1, -3*a3*sol1, a3]
           where a0,a1,a2,a3 := Explode([Coefficient(g, j) : j in [0..3]]);
    mx := (C.1 - sol1*C.3)^2*C.3;
    my := const*C.2*&+[Coefficient(k, j)*C.1^j*C.3^(3-j) : j in [0..3]];
    mz := &+[Coefficient(g, j)*C.1^j*C.3^(3-j) : j in [0..3]];
    mCE1 := map<C -> E1 | [mx, my, mz]>;
    return true, E, mCE1*mE1E;
  end if;
  return false, _, _;
end function;

// A helper function: Given a sequence of positive integers,
// it finds all subsets of the index set such that the sum of the
// corresponding numbers is n.

function sumnsubsets(seq, n)
  if IsEmpty(seq) then
    return n eq 0 select {{Integers()|}} else {};
  elif seq[#seq] gt n then
    return sumnsubsets(Prune(seq), n);
  else
    sss := sumnsubsets(Prune(seq), n - seq[#seq]);
    return {Include(s, #seq) : s in sss} join sumnsubsets(Prune(seq), n);
  end if;
end function;


// The main function takes a curve C and returns all the subcovers
// as a list of pairs <E, mCE>.

intrinsic Degree3Subcovers(C::CrvHyp) -> List
{Returns a list of pairs <E, mCE>, where E is an elliptic curve and mCE
 is a map of degree 3 from C to E. C must be of genus 2 and of the form
 y^2 = f(x).}
  f, h := HyperellipticPolynomials(C);
  require h eq 0: "Curve must be of the form y^2 = f(x)";
  require Genus(C) eq 2: "Curve must be of genus 2";
  ff := Factorization(f);
  ff := [a[1] : a in ff];
  // find all factorizations into two (projective) cubics
  degs := [Degree(a) : a in ff];
  sss := sumnsubsets(degs, 3);
  if Degree(f) eq 5 then sss join:= sumnsubsets(degs, 2); end if;
  result := [**];
  for s in sss do
    g := &*[ff[j] : j in s];
    h := ExactQuotient(f, g);
    flag, E, mCE := TestDeg3Subcover(g, h, C);
    if flag then Append(~result, <E, Expand(mCE)>); end if;
  end for;
  return result;
end intrinsic;

//==========================================================================

intrinsic Chabauty(ptJ::JacHypPt : ptC := 0, 
                   PrimeBound := 30, SmoothBound := 50, Epsilon := 1e-1,
                   CheckBound := 300, GiveUpBound := 10000)
                    -> SetEnum, SetEnum, SeqEnum
{For a curve C of genus 2 over Q, this returns the full set of rational points,
 using Chabauty's method combined with a Mordell-Weil sieve. There are two
 conditions: (i) the Jacobian of C has rank 1, generated by the given point ptJ, 
 and (ii) the set of rational points on C is non-empty.
 The optional argument ptC, if given, should be a rational point on C.}

  J := Parent(ptJ);
  C := Curve(J);
  require Genus(C) eq 2: "The curve should have genus 2.";

  if ptC cmpne 0 then 
    require Curve(Parent(ptC)) cmpeq C and Ring(Parent(ptC)) cmpeq Rationals() :
           "The optional argument ptC should be a rational point on the curve";
  else
    vprintf Chabauty: " Searching for a rational point on the curve ... "; vtime Chabauty:
    for n := 2 to 100 by 2 do 
      pts := Points(C : Bound:=2^n );
      if #pts gt 0 then break; end if;
    end for;
    ptC := pts[1];
  end if; 

  f, h := HyperellipticPolynomials(C);
  require h eq 0: "Curve must be of the form y^2 = f(x).";
  require Order(ptJ) eq 0: "ptJ must be a point of infinite order.";
  // find torsion subgroup and hence generators of J(Q)
  T, mT := TorsionSubgroup(J);
  vprintf Chabauty, 1: " Torsion subgroup has invariants %o\n", Invariants(T);
  bas := [ptJ] cat [mT(t) : t in OrderedGenerators(T)];
  ht := Height(ptJ);
  hc := HeightConstant(J);
  // compute GroupInfo for sufficiently many primes
  oldbound := 2;
  GI := [];
  b := 10;
  est := 1.0;
  Bs := {};
  while true do
    vprintf Chabauty, 1: " Computing GroupInfo...\n";
    checked := false;
    while est ge Epsilon or IsEmpty(Bs) do 
      if b ge CheckBound and not checked and est lt Epsilon then
        // Check for elliptic subcovers
        vprintf Chabauty, 1: " No useable prime found so far.\n";
        vprintf Chabauty, 1: "   checking for elliptic subcovers...\n\n";
        sub := Degree2Subcovers(C);
        vprintf Chabauty, 1: "  %o degree 2 subcovers found.\n\n", #sub;
        if IsEmpty(sub) then
          sub := Degree3Subcovers(C);
          vprintf Chabauty, 1: "  %o degree 3 subcovers found.\n\n", #sub;
        end if;
        if not IsEmpty(sub) then
          // find an E such that ptJ maps to torsion
          ptJa := ptJ[1];
          ptJb := ptJ[2];
          ptJd := ptJ[3];
          facta := Factorization(ptJa);
          infd := ptJd - Degree(ptJa);
          // take representative points with multiplicities
          rep := [];
          for ff in facta do
            if Degree(ff[1]) eq 1 then
              xP := -Coefficient(ff[1], 0)/Coefficient(ff[1], 1);
              Append(~rep, <C![xP, Evaluate(ptJb, xP)], ff[2]>);
            else
              KP := ext<BaseField(C) | ff[1]>;
              xP1 := KP.1;
              xP2 := -Coefficient(ff[1], 1)/Coefficient(ff[1], 2) - xP1;
              Append(~rep, <C(KP)![xP1, Evaluate(ptJb, xP1)], ff[2]>);
              Append(~rep, <C(KP)![xP2, Evaluate(ptJb, xP2)], ff[2]>);
            end if;
          end for;
          if infd gt 0 then
            Append(~rep, <C![1, Coefficient(ptJb, 3), 0], infd>);
          end if;
          // Find an E where the image of ptJ is torsion, and check 
          // for preimages of torsion points on E.  
          // Note: we assume without checking that this E has rank 0,
          // which follows from the assumptions that J has rank <= 1
          // and ptJ has infinite order.
          for pair in sub do
            E := pair[1];
            vprintf Chabauty, 1: "  Checking image of ptJ on\n  %o...\n", E;
            mCE := Extend(pair[2]);
            // map to E and add
            imP := E(BaseField(E))!&+[a[2]*mCE(a[1]) : a in rep];
            if Order(imP) ne 0 then
              // point is torsion ==> OK
              vprintf Chabauty, 1: "  ... image is torsion ==> OK!\n\n";
              T, mT := TorsionSubgroup(E);
              vprintf Chabauty, 1: "  E has %o torsion points\n", #T;
              ptsE := {mT(t) : t in T};
              result := &join{Points(pt @@ mCE) : pt in ptsE};
              vprintf Chabauty, 1: "giving points on C: %o\n", result;
              return result, {}, [];
            else
              vprintf Chabauty, 1: 
                      "  ... image has infinite order ==> continue\n\n";
            end if;
          end for; 
          require false: "The given Jacobian J should have rank 1.\n"*
                         "(Instead, this J is isogenous to a product of \n"*
                         "elliptic curves that both have rank at least 1)";
        end if;
        checked := true;
      end if; 
      if b ge GiveUpBound and est lt Epsilon then
        // give up
        require false: "No useable prime found.\nMost likely, "*
                       "C covers an elliptic curve, with map of degree > 3.";
      end if;
      b *:= 2;
      vprintf Chabauty, 1: "   for bound = %o...\n", b;
      nGI := GroupInfo(J, b, bas, ptC, oldbound, PrimeBound, SmoothBound);
      nGI := [CheckKillingDifferential(f, ptJ, entry) : entry in nGI];
      Bs := GetMinimalDivisors(Bs join FindMinimalBs(nGI));
      vprintf Chabauty, 1: "  set of minimal Bs found so far:\n    %o\n", Bs;
      oldbound := b;
      GI cat:= nGI;
      est := CheckInvariant(GI, #T);
      vprintf Chabauty, 1: "   check value is %o.\n", est;
    end while;
    run, flag := dryRun1(GI, PrimeBound, 5*Epsilon, #T, Bs);
    if not flag then
      // expected size goal not reached: need more information
      vprintf Chabauty, 1:
              "No good run found  ==>  reduce Epsilon and start over.\n";
      Epsilon *:= 0.5;
      continue;
    end if;
    result := {C|};
    primes := {};
    B := 1;
    cands := {<0, t> : t in T};
    for j := 1 to #run do
      p := run[j,1];
      cands, pr := LiftInformation(cands, GI, B, p);
      primes join:= pr;
      B *:= p;
      vprintf Chabauty, 1: "    B = %o, size = %o (predicted: %o)\n",
                           B, #cands, Round(run[j,2]);
      if IsEmpty(cands) then
        vprintf Chabauty, 1: "  No more points!\n";
      //vprintf Chabauty, 1: "Final list of points: %o\n", result;
      assert Points(C:Bound:=1000) subset result;
        return result, primes, [run[i,1] : i in [1..j]];
      end if; 
      // otherwise check if rational points are separated
      if exists{b : b in Bs | IsDivisibleBy(B, b)} then
        // for each candidate, try to identify a rational point
        if IsVerbose("Chabauty", 3) then
          printf "   current list:\n   ";
          for c in cands do printf " %o", c; end for;
          printf "\n";
        end if;
        oldcands := cands;
        cands := {};
        while not IsEmpty(oldcands) do
          vprintf Chabauty, 2: "  %o candidates remaining.\n", #oldcands;
          c := Rep(oldcands);
          vprintf Chabauty, 2: "   Checking candidate %o...\n", c;
          l := 2*c[1] le B select c[1] else c[1]-B;
          flag, pt := CheckPoint(bas[1], l, mT(c[2]), ptC, l^2*ht+hc);
          if flag then
            vprintf Chabauty, 1: "  Point %o on C found!\n", pt;
            Include(~result, pt);
            // remove all suitable candidates
            dl := {d : d in Bs | IsDivisibleBy(B, d)};
            for d in dl do
              Td := sub<T | [d*t : t in OrderedGenerators(T)]>;
              oldcands := {oc : oc in oldcands
                              | not (IsDivisibleBy(oc[1]-c[1], d)
                                      and oc[2]-c[2] in Td)};
              cands :=    {oc : oc in cands
                              | not (IsDivisibleBy(oc[1]-c[1], d)
                                      and oc[2]-c[2] in Td)};
            end for;
          else
            // move to cands
            Exclude(~oldcands, c);
            Include(~cands, c);
          end if; // flag
        end while; // not IsEmpty(oldcands)
        if IsEmpty(cands) then
          vprintf Chabauty, 1: "  No more points!\n";
        //vprintf Chabauty, 1: "Final list of points: %o\n", result;
        assert Points(C:Bound:=1000) subset result;
          return result, primes, [run[i,1] : i in [1..j]];
        end if;
      end if; // exists{b : b in Bs | IsDivisibleBy(B, b)}
    end for; // j := 1 to #run
    // run used up, but some classes left undecided
    vprintf Chabauty, 1:
            "Some classes unresolved ==> reduce Epsilon and start over.\n";
    Epsilon *:= 0.1;
  end while;
  // never reached
end intrinsic;


