freeze;

/********************************************************************

                   BASIC HEIGHT FUNCTIONS 
                    
                  Steve Donnelly, Sep 2006

*********************************************************************


   So far we have C-level functions 
     Height(FldRatElt), and 
     AbsoluteLogarithmicHeight(FldAlgElt)

   Below are defined:
     Height(RngIntElt)   
     Height(Pt : Absolute:=false ) 
       for Pt on any affine or projective variety (including weighted)
       over Q, number fields, function fields ...
       (exponential height, not logarithmic)

*********************************************************************/

forward nonarch_part;

intrinsic Height(n::RngIntElt) -> RngIntElt
{The absolute height of the integer n}
  if n eq 0 then return 1; end if;
  return Abs(n); 
end intrinsic;


intrinsic HeightOnAmbient(P::Pt : Absolute:=false, Prec:=30 ) -> FldReElt
{The relative, exponential height of a point on a scheme over the rationals
or a number field. The ambient space of the scheme may be affine or projective
(including graded projective).}

debug := false;

  basering := Ring(Parent(P));
  Amb := Ambient(Scheme(P));

  // Get nonzero coords:
  if Type(Amb) eq Aff then 
    coords := Eltseq(P) cat [1];
  elif Type(Amb) eq Prj then
    grrr := Gradings(Amb);
    require #grrr eq 1 : "There is more than one grading on the ambient projective space";
    coords := Eltseq(P);
    coords := [ coords[i]^grrr[1][i] : i in [1..#coords] ];
  else require false : "The ambient space must be either affine or projective (possibly weighted).";
  end if;
  coords := [ a : a in coords | a ne 0 ];
  assert basering eq Universe(coords);

  // Easy cases:
  require #coords gt 0 : "Silly error in Height, all coords are 0";
  if #coords eq 1 then return 1; end if;
  
  K := Parent(basering.1/1);
  d := AbsoluteDegree(K);
  dd := Absolute select d else 1;

  if #coords eq 2 and ISA(Type(K), FldAlg) and not debug then
    return Exp(dd*AbsoluteLogarithmicHeight(coords[1]/coords[2])); 
  elif #coords eq 2 and Type(K) eq FldRat then
    height := Height(coords[1]/coords[2]); 
    if dd gt 1 then height := height^(1/dd); end if;
    return height;
  end if;

  if debug then time0 := Cputime(); end if;
  
  // Clear denominators.
  // We want the coords to be algebraic integers,
  // and to have integral representations in terms of K.1.
  // If K.1 is integral, the second property implies the first.
  if Type(K) eq FldRat or IsIntegral(K.1) then
    minpols := []; 
  else 
    minpols := [ Eltseq( MinimalPolynomial(c,Rationals()) ) : c in coords ]; 
  end if;
  eltseqs := [ Eltseq(c) : c in coords ];
  while exists(a){ a : a in &cat minpols cat &cat eltseqs | Denominator(a) gt 1 } do
    denom := Denominator(a);
    bool, denomroot := IsPower(denom);
    if bool then denom := denomroot; end if;
    coords := [ denom*c : c in coords ];
    minpols := [ [mpol[i+1]*denom^(degr-i) : i in [0..degr]] where degr:=#mpol-1 : mpol in minpols ];
    eltseqs := [ Eltseq(c) : c in coords ];
  end while;
  
  //if debug then Cputime(time0); print "\nnorms"; end if;
  
  // Get the norms. 
  // It's quicker to get them from the complex embeddings, 
  // given that the embeddings of K.1 have to be computed anyway.
  // Note also, extra precision comes cheap.
  if Type(K) eq FldRat then norms := [];   // don't need norms
  elif not IsEmpty(minpols) then 
    norms := [ seq[1] : seq in minpols ];
  else 
    mynorms := [ (&*[ Abs(Evaluate(c,pl : Precision:=100+5*d ))^e where e := IsReal(pl) select 1 else 2 
                     : pl in InfinitePlaces(K) ]) : c in coords ];
    norms := [ Round(mn) : mn in mynorms ];
    for i := 1 to #norms do 
      if Abs(mynorms[i]-norms[i]) gt 10^-20 then norms[i] := AbsoluteNorm(coords[i]); end if; end for;
  end if;

  //if debug then time0 := Cputime(); end if;
  
  // The infinite places:
  if Type(K) eq FldRat then height := Max([ Abs(c) : c in coords ]);
  elif ISA(Type(K), FldAlg) then
    places := InfinitePlaces(K);
    height := 1;
    for pl in places do 
      e := IsReal(pl) select 1 else 2;
      height *:= Max([ Abs(Evaluate(c,pl:Precision:=Prec)) : c in coords ])^e;
    end for;
  else require false : "Not yet implemented for points defined over this ring";
  end if;
  
  //if debug then print "infinite height is ", Log(height); print Cputime(time0); end if;
  
  if Type(K) eq FldRat then 
    return height/GCD([ Integers()!c : c in coords ]); 
  else  // FldAlg
    height /:= nonarch_part(K, coords, norms : debug:=debug);
    if dd gt 1 then height := height^(1/dd); end if;
    return height;
  end if;

end intrinsic;


// the finite part
function nonarch_part(K, coords, norms : debug:=false )  // -> RngIntElt
  error if not ISA(Type(K), FldAlg), "Not yet implemented for points defined over this ring";

  // Try to avoid using MaximalOrder ... use pMaximalOrder instead, for as few p as possible
 
  normsGCD := GCD([ Integers()!n : n in norms ]);
  if normsGCD eq 1 then return 1; end if;

  // if K.1 is not integral, then no equation order exists, so rescale ...
  if not IsIntegral(K.1) then
    if BaseField(K) ne Rationals() and not IsIntegral(BaseField(K).1) then 
      return Norm(ideal< Integers(K) | coords >);   // give up and use MaximalOrder
    else
      coeffs := Eltseq( DefiningPolynomial(K) );
      d := Degree(K); 
      lc := coeffs[d+1];
      if lc ne 1 then
        coeffs := [c/lc : c in coeffs];
      end if;
      // rescale coeffs
      scaling := 1;
      while exists(coeff){coeff : coeff in coeffs | not IsIntegral(coeff)} do
        denom := Denominator(coeff);
        bool, denomroot := IsPower(denom);
        if bool then denom := denomroot; end if;
        coeffs := [ coeffs[i+1]*denom^(d-i) : i in [0..d] ];
        scaling *:= denom;
      end while;
      KK := ext< BaseField(K) | Polynomial(coeffs) >;
      KtoKK := hom< K -> KK | KK.1/scaling >;
      coords := [ KtoKK(c) : c in coords ];
      O := EquationOrder(KK);
      // now rescale coords into O
      coords_scaling := 1;
      while exists(a){a : a in &cat [Eltseq(c) : c in coords] | Denominator(a) gt 1 } do
        denom := Denominator(a);
        bool, denomroot := IsPower(denom);
        if bool then denom := denomroot; end if;
        coords := [ denom*c : c in coords ];
        coords_scaling *:= denom;
      end while;
    end if;
  else 
    O := EquationOrder(K);  
    coords_scaling := 1;
  end if;

  disc := Discriminant(O);

  if debug then time0 := Cputime(); end if;
  
  IO := ideal<O| coords >;     if debug then print "ideal:", Cputime(time0); end if;
  normIO := Index(O,IO);
  g := GCD(disc, normIO);
  if g eq 1 then return normIO/coords_scaling; end if;

  // TEMPORARY, until bugs in pMaximalOrder get fixed:
  //  if IsEven(g) then return Norm( ideal< Integers(K) | Generators(IO) >); end if;

  if debug then print "Factoring", g; end if;
  ps := PrimeDivisors(g);
  for p in ps do O := pMaximalOrder(O,p); end for;
  
  if debug then print "index"; time ans := Index(O, ideal<O| Generators(IO) >); return ans; end if;
  return Index(O, ideal<O| Generators(IO) >);
end function;
