freeze;
///////////////////////////////////////////////////////////////////
//     Intrinsics for computing various surface  		 //
//  invariants for ord proj surfaces which are at least	 	 //
//   			Gorenstein	 			 //
//             		mch - 10/2012                            //
///////////////////////////////////////////////////////////////////

import "kod_enr.m": h0;
import "singularity.m": is_known_gor;

function get_arith_genus(S)

    if (assigned S`pg) and (assigned S`q) then
	return (S`pg)-(S`q);
    end if;
    return ArithmeticGenus(S);

end function;

intrinsic GeometricGenus(S::Srfc : CheckGor := false, UseCohom := false) 
		-> RngIntElt
{ Returns the geometric genus of ordinary projective Gorenstein surface S.
  Uses cohomology if UseCohom is true (default: false). Checks the Gorenstein
  condition on S if CheckGor is true (default: false)}

    if assigned S`pg then return S`pg; end if;
    require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
    if CheckGor then
	require IsGorenstein(S) : 
	   "Error: S is not a Gorenstein surface";
    end if;

    if UseCohom then
	OS := StructureSheaf(S);
	SaturateSheaf(~OS); //usually faster to work with full module
	g := CohomologyDimension(OS,2,0);
    else
	K := CanonicalSheaf(S);
	g := h0(K);
    end if;

    S`pg := g;
    return g;

end intrinsic;

intrinsic Plurigenus(S::Srfc, n::RngIntElt : CheckGor := false) -> RngIntElt
{ Returns the nth plurigenus of ordinary projective Gorenstein surface S.
  Checks the Gorenstein condition on S if CheckGor is true (default: false)}

    require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
    require n ge 0: "n should be a non-negative integer";
    if n eq 0 then return 1; end if;
    if n eq 1 then return GeometricGenus(S : CheckGor := CheckGor); end if;

    if CheckGor then
	require IsGorenstein(S) : 
	   "Error: S is not a Gorenstein surface";
    end if;

    // first see if we can deduce result from Kodaira type
    if assigned S`kod_dim then
      case S`kod_dim:
	when -1: return 0;
	when 0: r := S`kod_enr2;
		if r eq -3 then // Enriques
		  return IsEven(n) select 1 else 0;
		elif (r eq -2) or (r eq -1) then // K3 or torus
		  return 1;
		elif r gt 0 then // bielliptic with known order of K
		  return IsDivisibleBy(n,r) select 1 else 0;
		end if;
	when 2:
	  pa := get_arith_genus(S);
	  if not (assigned S`K2m) then
	    K := CanonicalSheaf(S);
	    SaturateSheaf(~K);
	    p2 := h0(TensorPower(K,2));
	    S`K2m := p2-1-pa;
	  end if;
	  return ((n*(n-1)) div 2)*(S`K2m)+1+pa;
      end case;
    end if;

    // just expand K^(tensor)n and take h0
    K := CanonicalSheaf(S);
    SaturateSheaf(~K);
    return h0(TensorPower(K,n));

end intrinsic;

intrinsic Irregularity(S::Srfc : CheckGor := false, UseCohom := false)
			-> RngIntElt
{ Returns the irregularity of ordinary projective surface S.
  Checks the Gorenstein condition on S if CheckGor is true (default: false).
  If S is known Gorenstein, doesn't use cohomology unless UseCohom
  is true (default: false)}

    if assigned S`q then return S`q; end if;
    require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
    if CheckGor then
	boo := IsGorenstein(S);
    end if;
    boo, boo1 := is_known_gor(S);
    if boo then boo := boo1; end if;
    if (not boo) or UseCohom then
	OS := StructureSheaf(S);
	SaturateSheaf(~OS); //usually faster to work with full module
	S`q := CohomologyDimension(OS,1,0);
    else
	pg := GeometricGenus(S);
	pa := get_arith_genus(S);
	S`q := pg-pa;
    end if;
    return S`q;

end intrinsic;

intrinsic ChernNumber(S::Srfc, n::RngIntElt : CheckADE := false)
		-> RngIntElt
{ Returns the nth (n = 1 or 2) ChernNumber of S, an ordinary projective
  surface with only ordinary (A-D-E) singularities. Checks the singularity
  condition on S if CheckADE is true (default: false)}

    require (n eq 1) or (n eq 2): "Second argument must be 1 or 2";
    if not (assigned S`K2) then
      require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
      if CheckADE then
	require HasOnlySimpleSingularities(S):
		"Error: S does not have only simple singularities";
      end if;
      K := CanonicalSheaf(S);
      S`K2 := IntersectionPairing(K,K);
    end if;
    if n eq 1 then
      return S`K2;
    else
      pa := get_arith_genus(S);
      return 12*(1+pa)-(S`K2);
    end if;

end intrinsic;

intrinsic MinimalChernNumber(S::Srfc, n::RngIntElt : CheckADE := false)
		-> RngIntElt
{ Returns the nth (n = 1 or 2) ChernNumber of a minimal model of a desingularization
  of S, an ordinary projective
  surface with only ordinary (A-D-E) singularities. Checks the singularity
  condition on S if CheckADE is true (default: false)}

    require (n eq 1) or (n eq 2): "Second argument must be 1 or 2";
    if not (assigned S`K2m) then
      require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
      if CheckADE then
	require HasOnlySimpleSingularities(S):
		"Error: S does not have only simple singularities";
      end if;
      if not (assigned S`kod_dim) then
	kd := KodairaEnriquesType(S);
      else
	kd := S`kod_dim;
      end if;
      case kd:
	when -1: assert assigned S`q;
		 S`K2m := ((S`q) eq 0) select 9 else 8;
	when 0: S`K2m := 0;
	when 1: S`K2m := 0;
	when 2:
	  if not (assigned S`K2m) then
	    p2 := Plurigenus(S,2); //this will assign S`K2m !
	  end if;
      end case;
    end if;
    if n eq 1 then
      return S`K2m;
    else
      pa := get_arith_genus(S);
      return 12*(1+pa)-(S`K2m);
    end if;

end intrinsic;

intrinsic HodgeNumber(S::Srfc, i::RngIntElt, j::RngIntElt : CheckADE := false)
		-> RngIntElt
{ Returns the h^(i,j) (0 <= i,j <= 2) HodgeNumber of S, an ordinary projective
  surface with only ordinary (A-D-E) singularities. Checks the singularity
  condition on S if CheckADE is true (default: false)}

    require (i ge 0) and (i le 2) and (j ge 0) and (j le 2) : 
		"Arguments i and j must both be 0, 1 or 2";
    require IsOrdinaryProjective(S) : 
		"S should be a surface in ordinary projective space";
    if CheckADE then
	require HasOnlySimpleSingularities(S):
		"Error: S does not have only simple singularities";
    end if;

    // use symmetry to get i, j in standard form
    if i+j ge 2 then
	i := 2-i; j := 2-j;
    end if;
    if i gt j then
	k := i; i := j; j := k;
    end if;
    
    if i+j eq 0 then
	return 1;
    elif (i+j eq 1) and (assigned S`q) then
	return S`q;
    end if;
    pg := GeometricGenus(S);
    if i+j eq 1 then
	pa := get_arith_genus(S);
	S`q := pg-pa;
	return S`q;
    else //i+j eq 2
	if i eq 0 then return pg; end if;
	// h^{1,1} case
	pa := get_arith_genus(S);
	if not (assigned S`K2) then
	  K := CanonicalSheaf(S);
	  S`K2 := IntersectionPairing(K,K);
	end if;
	return 10+8*pa+2*pg-(S`K2);
    end if;

end intrinsic;
