freeze;

/*
  Functions and intrinsics which apply to all classical groups
  
  linear
  symplectic
  unitary
  orthogonal

  $Id: generic.m 41116 2012-11-28 00:43:35Z billu $

  Don Taylor
  16 August 2011
*/


// This returns \prod_{j=1}^s(1-t^j)^{-1} up to degree m
TruncatedEulerProduct := function(t,s,m)
// t must be an indeterminate in a polynomial ring.
  P := Parent(t);
  f := P!1;
  for j := 1 to Min(m,s) do
    f *:= &+[P| t^(j*i) : i in [0..(m div j)]];
  end for;
  c := Rank(P) eq 1 select Coefficients(f) else Coefficients(f,t);
  return &+[c[i+1]*t^i : i in [0..Min(#c-1,m)]];
end function;

// The number of conjugacy classes in GL(n,q) as a polynomial in q
NclassesGL := function(n)
  P<t,qq> := PolynomialRing(Integers(),2);
  gf := P!0;
  for r := 0 to n do
    term := 0;
    for s := 0 to n do
      h := r*(s+1)+(s*(s+1) div 2);
      if h gt n then break; end if;
      term +:= (-1)^s*t^h*TruncatedEulerProduct(t,s,n-h);
    end for;
    gf +:= qq^r*term;
  end for;
  // Could use UnivariatePolynomial but I want q to be named.
  Q<q> := PolynomialRing(Integers());
  return Evaluate(Coefficient(gf,t,n),[1,q]);
end function;

// This version is several orders of magnitude slower!
/*
NclassesGLalt := function(n)
  P<t,qq> := PolynomialRing(Integers(),2);
  f := P!1;
  for r := 1 to n do
    f *:= 1+(qq-1)*&+[P| qq^(i-1)*t^(i*r) : i in [1..n div r]];
  end for;
  Q<q> := PolynomialRing(Integers());
  return Evaluate(Coefficient(f,t,n),[1,q]);
end function;
*/

// The number of conjugacy classes in U(n,q) as a polynomial in q
NclassesU := function(n)
  P<t,qq> := PolynomialRing(Integers(),2);
  gf := P!0;
  for r := 0 to n do
    term := 0;
    for s := 0 to n do
      h := s*(s+1) div 2;
      if h gt n then break; end if;
      term +:= t^h*TruncatedEulerProduct(t,s,n-h);
    end for;
    gf +:= qq^r*t^r*term*TruncatedEulerProduct(t,r,n-r);
  end for;
  Q<q> := PolynomialRing(Integers());
  return Evaluate(Coefficient(gf,t,n),[1,q]);
end function;

/*
NclassesUalt := function(n)
  P<t,qq> := PolynomialRing(Integers(),2);
  f := P!1;
  for r := 1 to n do
    f *:= 1+(qq+1)*&+[P| qq^(i-1)*t^(i*r) : i in [1..n div r]];
  end for;
  Q<q> := PolynomialRing(Integers());
  return Evaluate(Coefficient(f,t,n),[1,q]);
end function;
*/

NclassesCU := function(n)
  f := NclassesU(n);
  P<q> := Parent(f);
  return f*(q-1);
end function;

intrinsic Nclasses( type::MonStgElt, n::RngIntElt ) -> RngUPolElt
{ The number of conjugacy classes of a finite classical group of the 
given type as a polynomial in q }
  case type:
  when "linear":
    return NclassesGL(n);
  when "unitary":
    return NclassesU(n);
  when "conformalunitary":
    return NclassesCU(n);
  else
    error "available types: linear, unitary and conformalunitary";
  end case;
end intrinsic;

intrinsic CheckNclassesGL(n::RngIntElt, q::RngIntElt, u::RngIntElt) -> BoolElt
{}
    return Evaluate(Nclasses("linear",n), q) lt u;
end intrinsic;
