freeze;

intrinsic CheckCharacterTable (irreds::SeqEnum[AlgChtrElt], level::RngIntElt)
{Perform tests that the sequence irreds will pass if it is the character table
of a group. The parameter level controls how much testing is done, from 1 to 5.}

    if level le 0 then return; end if;

    /* level 1: Check degrees */
    Z := Integers();
    G := Group(Universe(irreds));
    sos := 0;
    N := Index(G, Centre(G));
    cl := ClassesData(G);
    if #cl ne #irreds then
	printf "WRONG - incorrect number of characters";
	return;
    end if;
    for x in irreds do
	flag, deg := IsCoercible(Z, Degree(x));
	if not flag then
	    printf "WRONG - degree is not an integer\n";
	    return;
	end if;
	if deg le 0  or N mod deg ne 0 then
	    printf "WRONG - degree %o is impossible\n", deg;
	    return;
	end if;
	sos +:= deg^2;
    end for;
    if sos ne #G then
	printf "WRONG - Sum of degree squares is not group order\n";
	return;
    end if;
    if level le 1 then return; end if;

    /* level 2: Check Frob-Schur indicators */
    sos := 0;
    for x in irreds do
	s := Schur(x, 2);
	flag, s_Z := IsCoercible(Z, s);
	if not flag then 
	    printf "WRONG - Schur indicator not an integer\n"; 
	    return;
	end if;
	if s_Z notin {0,1,-1} then 
	    printf "WRONG - Schur indicator %o\n", s_Z;
	    return;
	end if;
	if s_Z eq -1 and Z!Degree(x) mod 2 ne 0 then
	    printf "WRONG - Schur indicator %o, degree %o\n", s_Z, Z!Degree(x);
	    return;
	end if;
	sos +:= s_Z * Degree(x);
    end for;
    if sos ne &+[c[2]:c in cl|c[1] le 2] then
	printf "WRONG - Count of involutions is incorrect\n";
	return;
    end if;
    if level le 2 then return; end if;

    /* Level 3: Check norms - have to compute from scratch */
    pm := PowerMap(G);
    invs := [pm(i, -1):i in [1..#cl]];
    for x in irreds do
	y := Eltseq(x);
	K := Universe(y);
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := true;
	end if;
	sum := 0;
	for i := 1 to #cl do
	    sum +:= cl[i,2]*y[i]*y[invs[i]];
	end for;
	flag, sum := IsCoercible(Z, sum);
	if not flag then 
	    printf "WRONG - Norm not an integer\n"; 
	    return;
	end if;
	if sum ne #G then 
	    printf "WRONG - Norm not 1 (%o)\n", sum;
	    return;
	end if;
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := false;
	end if;
    end for;
    if level le 3 then return; end if;

    /* Level 4: Number theoretic congruences */
    G_ord := FactoredOrder(G);
    ppc := []; /* p-prime classes */
    for t in G_ord do
	p := t[1];
	ppp := [];
	for i := 1 to #cl do
	    v, r := Valuation(cl[i,1], p);
	    _,a,_:= Xgcd(p^v, r);
	    ppp[i] := pm(i, a*p^v);
	end for;
	Append(~ppc, ppp);
    end for;
    for x in irreds do
	y := [x[i]:i in [1..#cl]];
	K := Universe(y);
	R := IntegerRing(K);
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := true;
	    R`DontMarkSweep := true;
	end if;
	ChangeUniverse(~y, R);
	for i := 1 to #G_ord do
	    ppp := ppc[i];
	    p := G_ord[i,1];
	    for j := 1 to #cl do
		pppj := ppp[j];
		if not IsZero(Modexp(y[j]-y[pppj], cl[j,1] div cl[pppj,1], p)) 
		then
		    printf "WRONG - congruence with p'-part\n";
		    printf "Prime: %o, Class: %o, p' class: %o\n",p,j,pppj;
		    printf "Values: %o, %o\n", K!y[j], K!y[pppj];
		end if;
	    end for;
	end for;
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := false;
	    R`DontMarkSweep := false;
	end if;
    end for;
    if level le 4 then return; end if;

    /* Level 5: Row orthogonality */
    for i := 2 to #cl do
	for j := 1 to i-1 do
	    ip := InnerProduct(irreds[i], irreds[j]);
	    if not IsZero(ip) then
		printf "WRONG - Inner product (%o,%o) not zero (%o)\n",
		    i,j,ip;
		return;
	    end if;
	end for;
    end for;
end intrinsic;

intrinsic CheckCharacterTableHard (irreds::SeqEnum[AlgChtrElt], level::RngIntElt)
{Perform tests that the sequence irreds will pass if it is the character table
of a group. The parameter level controls how much testing is done, from 1 to 5.}

    if level le 0 then return; end if;

    /* level 1: Check degrees */
    Z := Integers();
    G := Group(Universe(irreds));
    sos := 0;
    N := Index(G, Centre(G));
    cl := ClassesData(G);
    if #cl ne #irreds then
	error "WRONG - incorrect number of characters";
	return;
    end if;
    for x in irreds do
	flag, deg := IsCoercible(Z, Degree(x));
	if not flag then
	    error "WRONG - degree is not an integer\n";
	    return;
	end if;
	if deg le 0  or N mod deg ne 0 then
	    error "WRONG - degree %o is impossible\n", deg;
	    return;
	end if;
	sos +:= deg^2;
    end for;
    if sos ne #G then
	error "WRONG - Sum of degree squares is not group order\n";
	return;
    end if;
    if level le 1 then return; end if;

    /* level 2: Check Frob-Schur indicators */
    sos := 0;
    for x in irreds do
	s := Schur(x, 2);
	flag, s_Z := IsCoercible(Z, s);
	if not flag then 
	    error "WRONG - Schur indicator not an integer\n"; 
	    return;
	end if;
	if s_Z notin {0,1,-1} then 
	    error "WRONG - Schur indicator %o\n", s_Z;
	    return;
	end if;
	if s_Z eq -1 and Z!Degree(x) mod 2 ne 0 then
	    error "WRONG - Schur indicator %o, degree %o\n", s_Z, Z!Degree(x);
	    return;
	end if;
	sos +:= s_Z * Degree(x);
    end for;
    if sos ne &+[c[2]:c in cl|c[1] le 2] then
	error "WRONG - Count of involutions is incorrect\n";
	return;
    end if;
    if level le 2 then return; end if;

    /* Level 3: Check norms - have to compute from scratch */
    pm := PowerMap(G);
    invs := [pm(i, -1):i in [1..#cl]];
    for x in irreds do
	y := Eltseq(x);
	K := Universe(y);
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := true;
	end if;
	sum := 0;
	for i := 1 to #cl do
	    sum +:= cl[i,2]*y[i]*y[invs[i]];
	end for;
	flag, sum := IsCoercible(Z, sum);
	if not flag then 
	    error "WRONG - Norm not an integer\n"; 
	    return;
	end if;
	if sum ne #G then 
	    error "WRONG - Norm not 1 (%o)\n", sum;
	    return;
	end if;
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := false;
	end if;
    end for;
    if level le 3 then return; end if;

    /* Level 4: Number theoretic congruences */
    G_ord := FactoredOrder(G);
    ppc := []; /* p-prime classes */
    for t in G_ord do
	p := t[1];
	ppp := [];
	for i := 1 to #cl do
	    v, r := Valuation(cl[i,1], p);
	    _,a,_:= Xgcd(p^v, r);
	    ppp[i] := pm(i, a*p^v);
	end for;
	Append(~ppc, ppp);
    end for;
    for x in irreds do
	y := [x[i]:i in [1..#cl]];
	K := Universe(y);
	R := IntegerRing(K);
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := true;
	    R`DontMarkSweep := true;
	end if;
	ChangeUniverse(~y, R);
	for i := 1 to #G_ord do
	    ppp := ppc[i];
	    p := G_ord[i,1];
	    for j := 1 to #cl do
		pppj := ppp[j];
		if not IsZero(Modexp(y[j]-y[pppj], cl[j,1] div cl[pppj,1], p)) 
		then
		    printf "WRONG - congruence with p'-part\n";
		    printf "Prime: %o, Class: %o, p' class: %o\n",p,j,pppj;
		    printf "Values: %o, %o\n", K!y[j], K!y[pppj];
		    error "";
		end if;
	    end for;
	end for;
	if Type(K) eq FldCyc then
	    K`DontMarkSweep := false;
	    R`DontMarkSweep := false;
	end if;
    end for;
    if level le 4 then return; end if;

    /* Level 5: Row orthogonality */
    for i := 2 to #cl do
	for j := 1 to i-1 do
	    ip := InnerProduct(irreds[i], irreds[j]);
	    if not IsZero(ip) then
		printf "WRONG - Inner product (%o,%o) not zero (%o)\n",
		    i,j,ip;
		error "";
		return;
	    end if;
	end for;
    end for;
end intrinsic;
