/******************************************************************************
 *
 *    defrepdata.m Composition Tree Defining Rep Data
 *
 *    File      : $HeadURL:: https://subversion.sfac.auckland.ac.nz/svn/prj_#$:
 *    Author    : Frank Lbeck and Henrik Brnhielm
 *    Dev start : 2008-07-18
 *
 *    Version   : $Revision:: 1938                                           $:
 *    Date      : $Date:: 2010-09-04 06:40:53 +1200 (Sat, 04 Sep 2010)       $:
 *    Last edit : $Author:: jbaa004                                          $:
 *
 *    $Id:: defrepdata.m 1938 2010-09-03 18:40:53Z jbaa004                 $:
 *
 *****************************************************************************/

freeze;

/*
##  
##  Format of the following list:
##  DEFREPDATA[n] contains information on irreducible defining characteristic 
##  representations of degree n for simple simply connected algebraic groups G.
##  
##  For n = 1..300 the entry DEFREPDATA[n]  is a list of lists of the form:
##      [[type, rank], n, [list of weights], prime_info]
##  where prime_info is either a single prime (meaning that this entry is only
##  valid in characteristic prime_info) or it is a list of list of primes
##  (meaning that the entry is valid in any characteristic which is not one of
##  the given primes).
##  
##  The 'list of weights' contains p-restricted weights. There are infinitely 
##  many highest weight representations corresponding to such an entry: If the
##  characteristic is p, then the highest weights are all linear combinations 
##  of the given list of weights with pairwise different coefficients of 
##  form p^i  (i in Z, i >= 0). 
##  
##  If we are not in the Suzuki-/Ree-group case, then the restrictions of the
##  representations of G to the finite groups G(q), q a power of p, which 
##  have  highest weight with all coefficients < q give a
##  complete set of defining characteristic irreducible representations of
##  G(q).
##  
##  The data and the notations for the highest weights are from the article:
##    Lübeck, F., Small degree representations of finite Chevalley groups in
##    defining characteristic, LMS J. Comput. Math., 4 (2001), p. 135--169
##  
##  Here are two utility functions:
##  
##  AllCases(n, q) with n <= 300 and q a prime power returns a list of lists 
##  of form  [type, rank, [list of weights]] showing which of the mentioned
##  representations occur for given n and q. The 'list of weights' is as
##  described above. The function
##  
##  AllHighestWeightsOfCase( case, q ) can be used to actually transform this
##  into an explicit list of highest weights which are relevant for the
##  representations of groups G(q). (If q is a high power of a prime p then 
##  this list can easily become quite long. For this reason these two 
##  functions are separated.)
##  
##  Example:
##  gap> cases := AllCases(56, 3^4);
##  [ [ "A", 7, [ [ 0, 0, 0, 0, 1, 0, 0 ] ] ], 
##    [ "A", 7, [ [ 0, 0, 1, 0, 0, 0, 0 ] ] ], 
##    [ "B", 2, [ [ 1, 0 ], [ 0, 2 ] ] ], 
##    [ "B", 3, [ [ 0, 0, 1 ], [ 1, 0, 0 ] ] ], 
##    [ "D", 4, [ [ 0, 1, 0, 1 ] ] ], 
##    [ "D", 4, [ [ 1, 0, 0, 1 ] ] ], 
##    [ "D", 4, [ [ 1, 1, 0, 0 ] ] ], 
##    [ "E", 7, [ [ 0, 0, 0, 0, 0, 0, 1 ] ] ], 
##    [ "C", 28, [ [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 ] ] ], 
##    [ "D", 28, 
##        [ [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 1 ] ] ], 
##    [ "A", 55, [ [ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0 ] ] ], 
##    [ "A", 55, [ [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
##                0, 0, 0, 0, 0, 0, 0, 1 ] ] ] ]
##  gap> AllHighestWeightsOfCase(cases[3], 3^4);
##  [ [ 1, 6 ], [ 1, 18 ], [ 1, 54 ], [ 3, 2 ], [ 3, 18 ], 
##    [ 3, 54 ], [ 9, 2 ], [ 9, 6 ], [ 9, 54 ], [ 27, 2 ], 
##    [ 27, 6 ], [ 27, 18 ] ]
##  
*/

DEFREPDATA :=
[*[**],[*[*[*"A",1*],2,[*[*1*]*],[*[**]*]*]*],[*[*[*"A",2*],3,[*[*0,1
*]*],[*[**]*]*],[*[*"A",2*],3,[*[*1,0*]*],[*[**]*]*],[*[*"A",1*],3,[*[*
2*]*],[*[*2*]*]*]*],[*[*[*"A",3*],4,[*[*0,0,1*]*],[*[**]*]*],[*[*"A",
3*],4,[*[*1,0,0*]*],[*[**]*]*],[*[*"B",2*],4,[*[*0,1*]*],2*],[*[*
"B",2*],4,[*[*1,0*]*],[*[**]*]*],[*[*"A",1*],4,[*[*1*],[*1*]*],[*[**]
*]*],[*[*"A",1*],4,[*[*3*]*],[*[*2,3*]*]*]*],[*[*[*"A",4*],5,[*[*0,
0,0,1*]*],[*[**]*]*],[*[*"A",4*],5,[*[*1,0,0,0*]*],[*[**]*]*],[*[*
"B",2*],5,[*[*0,1*]*],[*[*2*]*]*],[*[*"A",1*],5,[*[*4*]*],[*[*2,3*]
*]*]*],[*[*[*"A",2*],6,[*[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],6,[*[*2,0
*]*],[*[*2*]*]*],[*[*"A",3*],6,[*[*0,1,0*]*],[*[**]*]*],[*[*"A",5*],
6,[*[*0,0,0,0,1*]*],[*[**]*]*],[*[*"A",5*],6,[*[*1,0,0,0,0*]*],[*
[**]*]*],[*[*"B",3*],6,[*[*0,0,1*]*],2*],[*[*"C",3*],6,[*[*0,0,1*]
*],[*[**]*]*],[*[*"G",2*],6,[*[*0,1*]*],2*],[*[*"A",1*],6,[*[*1*],[*
2*]*],[*[*2*]*]*],[*[*"A",1*],6,[*[*5*]*],[*[*2,3,5*]*]*]*],[*[*[*
"A",2*],7,[*[*1,1*]*],3*],[*[*"A",6*],7,[*[*0,0,0,0,0,1*]*],[*[*
*]*]*],[*[*"A",6*],7,[*[*1,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",3*],7,
[*[*0,0,1*]*],[*[*2*]*]*],[*[*"G",2*],7,[*[*0,1*]*],[*[*2*]*]*],[*[*
"G",2*],7,[*[*1,0*]*],3*],[*[*"A",1*],7,[*[*6*]*],[*[*2,3,5*]*]*]
*],[*[*[*"A",2*],8,[*[*1,1*]*],[*[*3*]*]*],[*[*"A",7*],8,[*[*0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",7*],8,[*[*1,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"B",3*],8,[*[*1,0,0*]*],[*[**]*]*],[*[*"B",4*],8,[*[*0,0,
0,1*]*],2*],[*[*"C",3*],8,[*[*1,0,0*]*],2*],[*[*"C",4*],8,[*[*0,
0,0,1*]*],[*[**]*]*],[*[*"D",4*],8,[*[*0,0,0,1*]*],[*[**]*]*],[*[*
"D",4*],8,[*[*0,1,0,0*]*],[*[**]*]*],[*[*"D",4*],8,[*[*1,0,0,0*]
*],[*[**]*]*],[*[*"A",1*],8,[*[*1*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],
8,[*[*1*],[*1*],[*1*]*],[*[**]*]*],[*[*"A",1*],8,[*[*7*]*],[*[*2,3,
5,7*]*]*]*],[*[*[*"A",2*],9,[*[*0,1*],[*1,0*]*],[*[**]*]*],[*[*"A",2
*],9,[*[*0,1*],[*0,1*]*],[*[**]*]*],[*[*"A",2*],9,[*[*1,0*],[*1,0*]
*],[*[**]*]*],[*[*"A",8*],9,[*[*0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",8*],9,[*[*1,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",4*],9,[*[*
0,0,0,1*]*],[*[*2*]*]*],[*[*"A",1*],9,[*[*2*],[*2*]*],[*[*2*]*]*],[*
[*"A",1*],9,[*[*8*]*],[*[*2,3,5,7*]*]*]*],[*[*[*"A",2*],10,[*[*0,3
*]*],[*[*2,3*]*]*],[*[*"A",2*],10,[*[*3,0*]*],[*[*2,3*]*]*],[*[*"A",
3*],10,[*[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],10,[*[*2,0,0*]*],[*[*
2*]*]*],[*[*"A",4*],10,[*[*0,0,1,0*]*],[*[**]*]*],[*[*"A",4*],10,[*
[*0,1,0,0*]*],[*[**]*]*],[*[*"A",9*],10,[*[*0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",9*],10,[*[*1,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"B",2*],10,[*[*2,0*]*],[*[*2*]*]*],[*[*"B",5*],10,[*[*0,0,0,0,
1*]*],2*],[*[*"C",5*],10,[*[*0,0,0,0,1*]*],[*[**]*]*],[*[*"D",5*],
10,[*[*0,0,0,0,1*]*],[*[**]*]*],[*[*"A",1*],10,[*[*1*],[*4*]*],[*[*
2,3*]*]*],[*[*"A",1*],10,[*[*9*]*],[*[*2,3,5,7*]*]*]*],[*[*[*"A",10
*],11,[*[*0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",10*],11,[*[*
1,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",5*],11,[*[*0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",1*],11,[*[*10*]*],[*[*2,3,5,7*]*]*]*],[*[*
[*"A",11*],12,[*[*0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
11*],12,[*[*1,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],
12,[*[*1,1*]*],5*],[*[*"B",6*],12,[*[*0,0,0,0,0,1*]*],2*],[*[*
"C",6*],12,[*[*0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",6*],12,[*[*0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",1*],12,[*[*1*],[*5*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],12,[*[*1*],[*1*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],12,[*
[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],12,[*[*11*]*],[*[*2,3,5,7,
11*]*]*]*],[*[*[*"A",12*],13,[*[*0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"A",12*],13,[*[*1,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"B",2*],13,[*[*0,2*]*],5*],[*[*"B",6*],13,[*[*0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"C",3*],13,[*[*0,1,0*]*],3*],[*[*"A",1*],13,
[*[*12*]*],[*[*2,3,5,7,11*]*]*]*],[*[*[*"A",13*],14,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",13*],14,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],14,[*[*1,0,1*]*],2
*],[*[*"B",2*],14,[*[*0,2*]*],[*[*2,5*]*]*],[*[*"B",3*],14,[*[*0,1,
0*]*],2*],[*[*"B",7*],14,[*[*0,0,0,0,0,0,1*]*],2*],[*[*"C",3*],
14,[*[*0,1,0*]*],[*[*3*]*]*],[*[*"C",3*],14,[*[*1,0,0*]*],[*[*2*]*]
*],[*[*"C",7*],14,[*[*0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",7*],14,
[*[*0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"G",2*],14,[*[*1,0*]*],[*[*3
*]*]*],[*[*"A",1*],14,[*[*1*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],
14,[*[*13*]*],[*[*2,3,5,7,11,13*]*]*]*],[*[*[*"A",14*],15,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",14*],15,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],15,[*[*0,
4*]*],[*[*2,3*]*]*],[*[*"A",2*],15,[*[*1,2*]*],[*[*2*]*]*],[*[*"A",2
*],15,[*[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],15,[*[*4,0*]*],[*[*2,3*]*]
*],[*[*"A",3*],15,[*[*1,0,1*]*],[*[*2*]*]*],[*[*"A",4*],15,[*[*0,0,
0,2*]*],[*[*2*]*]*],[*[*"A",4*],15,[*[*2,0,0,0*]*],[*[*2*]*]*],[*[*
"A",5*],15,[*[*0,0,0,1,0*]*],[*[**]*]*],[*[*"A",5*],15,[*[*0,1,0,
0,0*]*],[*[**]*]*],[*[*"B",7*],15,[*[*0,0,0,0,0,0,1*]*],[*[*2*]*]
*],[*[*"A",1*],15,[*[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],15,[*[*
14*]*],[*[*2,3,5,7,11,13*]*]*]*],[*[*[*"A",15*],16,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",15*],16,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],16,[*[*0,
0,1*],[*1,0,0*]*],[*[**]*]*],[*[*"A",3*],16,[*[*0,0,1*],[*0,0,1*]
*],[*[**]*]*],[*[*"A",3*],16,[*[*1,0,0*],[*1,0,0*]*],[*[**]*]*],[*[*
"A",3*],16,[*[*0,1,1*]*],3*],[*[*"A",3*],16,[*[*1,1,0*]*],3*],[*
[*"B",2*],16,[*[*0,1*],[*1,0*]*],2*],[*[*"B",2*],16,[*[*0,1*],[*
0,1*]*],2*],[*[*"B",2*],16,[*[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"B",2
*],16,[*[*1,1*]*],[*[*5*]*]*],[*[*"B",4*],16,[*[*1,0,0,0*]*],[*[**]
*]*],[*[*"B",8*],16,[*[*0,0,0,0,0,0,0,1*]*],2*],[*[*"C",4*],16,
[*[*1,0,0,0*]*],2*],[*[*"C",8*],16,[*[*0,0,0,0,0,0,0,1*]*],[*[*
*]*]*],[*[*"D",5*],16,[*[*0,1,0,0,0*]*],[*[**]*]*],[*[*"D",5*],16,[*
[*1,0,0,0,0*]*],[*[**]*]*],[*[*"D",8*],16,[*[*0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",1*],16,[*[*1*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],16,[*[*1*],[*1*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],16,[*[*1
*],[*1*],[*1*],[*1*]*],[*[**]*]*],[*[*"A",1*],16,[*[*3*],[*3*]*],[*[*
2,3*]*]*],[*[*"A",1*],16,[*[*15*]*],[*[*2,3,5,7,11,13*]*]*]*],[*[*
[*"A",16*],17,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",16*],17,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"B",8*],17,[*[*0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",1*],17,[*[*16*]*],[*[*2,3,5,7,11,13*]*]*]*],[*[*[*"A",17*],
18,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",17*],18,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*
*]*]*],[*[*"A",2*],18,[*[*0,1*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],
18,[*[*0,1*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],18,[*[*1,0*],[*0,2
*]*],[*[*2*]*]*],[*[*"A",2*],18,[*[*1,0*],[*2,0*]*],[*[*2*]*]*],[*[*
"A",2*],18,[*[*1,3*]*],5*],[*[*"A",2*],18,[*[*3,1*]*],5*],[*[*
"B",9*],18,[*[*0,0,0,0,0,0,0,0,1*]*],2*],[*[*"C",9*],18,[*[*0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",9*],18,[*[*0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",1*],18,[*[*1*],[*2*],[*2*]*],[*[*2*]*]
*],[*[*"A",1*],18,[*[*1*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],18,
[*[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],18,[*[*17*]*],[*[*2,3,
5,7,11,13,17*]*]*]*],[*[*[*"A",18*],19,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",18*],19,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],19,[*[*
2,2*]*],5*],[*[*"A",3*],19,[*[*0,2,0*]*],3*],[*[*"B",9*],19,[*[*
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",1*],19,[*[*18*]*],[*[*
2,3,5,7,11,13,17*]*]*],[*[*"A",18*],19,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",18*],19,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",19*],
20,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"A",19*],20,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",3*],20,[*[*0,0,3*]*],[*[*2,3*]*]*],[*[*
"A",3*],20,[*[*0,1,1*]*],[*[*3*]*]*],[*[*"A",3*],20,[*[*0,2,0*]*],
[*[*2,3*]*]*],[*[*"A",3*],20,[*[*1,1,0*]*],[*[*3*]*]*],[*[*"A",3*],
20,[*[*3,0,0*]*],[*[*2,3*]*]*],[*[*"A",5*],20,[*[*0,0,1,0,0*]*],
[*[**]*]*],[*[*"B",10*],20,[*[*0,0,0,0,0,0,0,0,0,1*]*],2*],[*[*
"B",2*],20,[*[*1,0*],[*0,1*]*],[*[*2*]*]*],[*[*"B",2*],20,[*[*3,0
*]*],[*[*2,3*]*]*],[*[*"C",10*],20,[*[*0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"D",10*],20,[*[*0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"A",1*],20,[*[*1*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],20,[*
[*1*],[*1*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],20,[*[*3*],[*4*]*],[*[*
2,3*]*]*],[*[*"A",1*],20,[*[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",19*],20,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",19*],20,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",20*],21,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",20*],
21,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",2*],21,[*[*0,1*],[*1,1*]*],3*],[*[*"A",2*],21,[*[*1,0
*],[*1,1*]*],3*],[*[*"A",2*],21,[*[*0,5*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],21,[*[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",5*],21,[*[*0,0,0,
0,2*]*],[*[*2*]*]*],[*[*"A",5*],21,[*[*2,0,0,0,0*]*],[*[*2*]*]*],[*
[*"A",6*],21,[*[*0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",6*],21,[*[*0,
1,0,0,0,0*]*],[*[**]*]*],[*[*"B",10*],21,[*[*0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"B",3*],21,[*[*0,1,0*]*],[*[*2*]*]*],[*[*
"C",3*],21,[*[*0,0,2*]*],[*[*2*]*]*],[*[*"A",1*],21,[*[*2*],[*6*]
*],[*[*2,3,5*]*]*],[*[*"A",1*],21,[*[*20*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",20*],21,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",20*],21,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",11*],22,[*
[*0,0,0,0,0,0,0,0,0,0,1*]*],2*],[*[*"C",11*],22,[*[*0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",11*],22,[*[*0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",1*],22,[*[*1*],[*10*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],22,[*[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",21*],22,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",21*],22,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],23,[*[*
1,0,0,1*]*],5*],[*[*"B",11*],23,[*[*0,0,0,0,0,0,0,0,0,0,1*]
*],[*[*2*]*]*],[*[*"A",1*],23,[*[*22*]*],[*[*2,3,5,7,11,13,17,19*]
*]*],[*[*"A",22*],23,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",22*],23,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],24,
[*[*0,1*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2*],24,[*[*1,0*],[*1,1*]*],
[*[*3*]*]*],[*[*"A",2*],24,[*[*1,3*]*],[*[*3,5*]*]*],[*[*"A",2*],24,
[*[*3,1*]*],[*[*3,5*]*]*],[*[*"A",3*],24,[*[*0,0,1*],[*0,1,0*]*],
[*[**]*]*],[*[*"A",3*],24,[*[*1,0,0*],[*0,1,0*]*],[*[**]*]*],[*[*"A",
4*],24,[*[*1,0,0,1*]*],[*[*5*]*]*],[*[*"B",2*],24,[*[*1,2*]*],7*],
[*[*"A",1*],24,[*[*1*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],24,[*
[*1*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],24,[*[*1*],[*1*],[*
5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],24,[*[*1*],[*1*],[*1*],[*2*]*],[*
[*2*]*]*],[*[*"A",1*],24,[*[*2*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],24,[*[*3*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],24,[*[*23*]*],[*
[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"C",12*],24,[*[*0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",12*],24,[*[*0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",23*],24,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",23*],
24,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",4*],25,[*[*0,0,0,1*],[*1,0,0,0*]*],[*[**]
*]*],[*[*"A",4*],25,[*[*0,0,0,1*],[*0,0,0,1*]*],[*[**]*]*],[*[*"A",
4*],25,[*[*1,0,0,0*],[*1,0,0,0*]*],[*[**]*]*],[*[*"B",2*],25,[*[*
0,1*],[*0,1*]*],[*[*2*]*]*],[*[*"B",2*],25,[*[*0,3*]*],7*],[*[*"B",
2*],25,[*[*2,1*]*],3*],[*[*"F",4*],25,[*[*0,0,0,1*]*],3*],[*[*
"A",1*],25,[*[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],25,[*[*24*]*],
[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"B",12*],25,[*[*0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",24*],25,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",24*],25,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",3*],26,[*[*0,0,2*]*],7*],[*[*
"B",4*],26,[*[*0,0,1,0*]*],2*],[*[*"C",4*],26,[*[*0,0,1,0*]*],2
*],[*[*"D",4*],26,[*[*0,0,1,0*]*],2*],[*[*"F",4*],26,[*[*0,0,0,1
*]*],[*[*3*]*]*],[*[*"F",4*],26,[*[*1,0,0,0*]*],2*],[*[*"G",2*],26,
[*[*0,2*]*],7*],[*[*"A",1*],26,[*[*1*],[*12*]*],[*[*2,3,5,7,11*]*]
*],[*[*"A",1*],26,[*[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"C",13*],26,[*[*0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",13*],26,[*[*0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",25*],26,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",25*],26,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
2*],27,[*[*0,1*],[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],27,[*[*0,
1*],[*0,1*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],27,[*[*0,1*],[*0,1*],
[*0,1*]*],[*[**]*]*],[*[*"A",2*],27,[*[*1,0*],[*1,0*],[*1,0*]*],[*[*
*]*]*],[*[*"A",2*],27,[*[*2,2*]*],[*[*2,5*]*]*],[*[*"B",3*],27,[*[*
0,0,2*]*],[*[*2,7*]*]*],[*[*"C",4*],27,[*[*0,0,1,0*]*],[*[*2*]*]
*],[*[*"E",6*],27,[*[*0,0,0,0,0,1*]*],[*[**]*]*],[*[*"E",6*],27,[*
[*1,0,0,0,0,0*]*],[*[**]*]*],[*[*"G",2*],27,[*[*0,2*]*],[*[*2,7*]
*]*],[*[*"G",2*],27,[*[*2,0*]*],3*],[*[*"A",1*],27,[*[*2*],[*8*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],27,[*[*2*],[*2*],[*2*]*],[*[*2*]*]*],
[*[*"A",1*],27,[*[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"B",13*],27,[*[*0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",26*],27,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",26*],27,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",2*],28,[*[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],28,[*[*6,0*]
*],[*[*2,3,5*]*]*],[*[*"A",6*],28,[*[*0,0,0,0,0,2*]*],[*[*2*]*]*],
[*[*"A",6*],28,[*[*2,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",7*],28,[*[*
0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",7*],28,[*[*0,1,0,0,0,0,0
*]*],[*[**]*]*],[*[*"D",4*],28,[*[*0,0,1,0*]*],[*[*2*]*]*],[*[*"A",1
*],28,[*[*1*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],28,[*[*
1*],[*1*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],28,[*[*3*],[*6*]*],[*
[*2,3,5*]*]*],[*[*"A",1*],28,[*[*27*]*],[*[*2,3,5,7,11,13,17,19,
23*]*]*],[*[*"C",14*],28,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"D",14*],28,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",27*],28,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",27*],28,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",1*],29,[*[*28*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"B",14*],29,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",28*],29,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",28
*],29,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],30,[*[*0,1*],[*0,3*]*],
[*[*2,3*]*]*],[*[*"A",2*],30,[*[*0,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*
"A",2*],30,[*[*1,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],30,[*[*1,
0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",4*],30,[*[*0,0,1,1*]*],3*],[*
[*"A",4*],30,[*[*1,1,0,0*]*],3*],[*[*"B",2*],30,[*[*0,3*]*],[*[*
2,3,7*]*]*],[*[*"A",1*],30,[*[*1*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],30,[*[*1*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
30,[*[*2*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],30,[*[*4*],[*5*]
*],[*[*2,3,5*]*]*],[*[*"A",1*],30,[*[*29*]*],[*[*2,3,5,7,11,13,
17,19,23,29*]*]*],[*[*"C",15*],30,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",15*],30,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",29*],30,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",29*],30,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],31,[*[*
30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"B",15*],31,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",30*],
31,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",30*],31,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",3*],32,[*[*1,0,2*]*],5*],[*[*"A",3*],32,[*[*2,
0,1*]*],5*],[*[*"B",5*],32,[*[*1,0,0,0,0*]*],[*[**]*]*],[*[*"C",5
*],32,[*[*1,0,0,0,0*]*],2*],[*[*"D",6*],32,[*[*0,1,0,0,0,0*]*],
[*[**]*]*],[*[*"D",6*],32,[*[*1,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",1
*],32,[*[*1*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],32,[*[*1*],[*
15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],32,[*[*1*],[*1*],[*7*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],32,[*[*1*],[*1*],[*1*],[*3*]*],[*[*
2,3*]*]*],[*[*"A",1*],32,[*[*1*],[*1*],[*1*],[*1*],[*1*]*],[*[**]*]
*],[*[*"A",1*],32,[*[*3*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],32,
[*[*31*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"C",16*],
32,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",16*],32,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"A",31*],32,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",31*],32,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],33,[*[*1,5*]*],7*],[*[*
"A",2*],33,[*[*5,1*]*],7*],[*[*"A",1*],33,[*[*2*],[*10*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],33,[*[*32*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31*]*]*],[*[*"B",16*],33,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[*2*]*]*],[*[*"A",32*],33,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",32*],33,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",5*],
34,[*[*1,0,0,0,1*]*],2*],[*[*"A",5*],34,[*[*1,0,0,0,1*]*],3*],
[*[*"A",1*],34,[*[*1*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],34,[*[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"C",
17*],34,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"D",17*],34,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",33*],34,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",33*],34,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],35,[*
[*1,4*]*],[*[*2,3*]*]*],[*[*"A",2*],35,[*[*4,1*]*],[*[*2,3*]*]*],[*
[*"A",3*],35,[*[*0,0,4*]*],[*[*2,3*]*]*],[*[*"A",3*],35,[*[*4,0,0
*]*],[*[*2,3*]*]*],[*[*"A",4*],35,[*[*0,0,0,3*]*],[*[*2,3*]*]*],[*[*
"A",4*],35,[*[*3,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",5*],35,[*[*1,0,
0,0,1*]*],[*[*2,3*]*]*],[*[*"A",6*],35,[*[*0,0,0,1,0,0*]*],[*[**]
*]*],[*[*"A",6*],35,[*[*0,0,1,0,0,0*]*],[*[**]*]*],[*[*"B",2*],35,
[*[*2,1*]*],[*[*2,3*]*]*],[*[*"B",2*],35,[*[*4,0*]*],[*[*2,3*]*]*],
[*[*"B",3*],35,[*[*2,0,0*]*],[*[*2*]*]*],[*[*"D",4*],35,[*[*0,0,0,
2*]*],[*[*2*]*]*],[*[*"D",4*],35,[*[*0,2,0,0*]*],[*[*2*]*]*],[*[*
"D",4*],35,[*[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",1*],35,[*[*4*],[*6
*]*],[*[*2,3,5*]*]*],[*[*"A",1*],35,[*[*34*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"B",17*],35,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",34*],35,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",34*],35,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",2*],36,[*[*0,2*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],
36,[*[*0,2*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],36,[*[*2,0*],[*2,0
*]*],[*[*2*]*]*],[*[*"A",2*],36,[*[*0,7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],36,[*[*2,4*]*],7*],[*[*"A",2*],36,[*[*4,2*]*],7*],[*[*
"A",2*],36,[*[*7,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",3*],36,[*[*0,1,
0*],[*0,1,0*]*],[*[**]*]*],[*[*"A",3*],36,[*[*1,0,2*]*],[*[*2,5*]*]
*],[*[*"A",3*],36,[*[*2,0,1*]*],[*[*2,5*]*]*],[*[*"A",5*],36,[*[*0,
0,0,0,1*],[*1,0,0,0,0*]*],[*[**]*]*],[*[*"A",5*],36,[*[*0,0,0,
0,1*],[*0,0,0,0,1*]*],[*[**]*]*],[*[*"A",5*],36,[*[*1,0,0,0,0*],
[*1,0,0,0,0*]*],[*[**]*]*],[*[*"A",7*],36,[*[*0,0,0,0,0,0,2*]*],
[*[*2*]*]*],[*[*"A",7*],36,[*[*2,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*
"A",8*],36,[*[*0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",8*],36,[*[*
0,1,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",3*],36,[*[*0,0,1*],[*0,
0,1*]*],2*],[*[*"B",4*],36,[*[*0,0,1,0*]*],[*[*2*]*]*],[*[*"C",3
*],36,[*[*0,0,1*],[*0,0,1*]*],[*[**]*]*],[*[*"C",4*],36,[*[*0,0,0,
2*]*],[*[*2*]*]*],[*[*"G",2*],36,[*[*0,1*],[*0,1*]*],2*],[*[*"A",1
*],36,[*[*1*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],36,[*[*1*],
[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],36,[*[*1*],[*1*],
[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],36,[*[*1*],[*1*],[*8*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],36,[*[*2*],[*11*]*],[*[*2,3,5,7,11*]*]*],
[*[*"A",1*],36,[*[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],36,[*
[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],36,[*[*5*],[*5*]*],[*[*
2,3,5*]*]*],[*[*"A",1*],36,[*[*35*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31*]*]*],[*[*"C",18*],36,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",18*],36,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",35*],36,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",35*],36,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],37,[*[*3,3*]*],7*],[*[*"A",1*],
37,[*[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"B",18
*],37,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]
*],[*[*"A",36*],37,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",36*],37,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"G",2
*],38,[*[*1,1*]*],7*],[*[*"A",1*],38,[*[*1*],[*18*]*],[*[*2,3,5,7,
11,13,17*]*]*],[*[*"A",1*],38,[*[*37*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37*]*]*],[*[*"C",19*],38,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",19*],38,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",37*],38,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",37*],38,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],39,[*[*2,3*]
*],5*],[*[*"A",2*],39,[*[*3,2*]*],5*],[*[*"A",1*],39,[*[*2*],[*12*]
*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],39,[*[*38*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37*]*]*],[*[*"B",19*],39,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",38*],39,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",38*],39,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],40,[*[*0,0,1
*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],40,[*[*0,0,1*],[*2,0,0*]*],
[*[*2*]*]*],[*[*"A",3*],40,[*[*1,0,0*],[*0,0,2*]*],[*[*2*]*]*],[*[*
"A",3*],40,[*[*1,0,0*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",4*],40,[*[*
0,0,1,1*]*],[*[*3*]*]*],[*[*"A",4*],40,[*[*0,1,0,1*]*],2*],[*[*
"A",4*],40,[*[*1,0,1,0*]*],2*],[*[*"A",4*],40,[*[*1,1,0,0*]*],[*
[*3*]*]*],[*[*"B",2*],40,[*[*1,0*],[*2,0*]*],[*[*2*]*]*],[*[*"B",2
*],40,[*[*1,2*]*],[*[*2,7*]*]*],[*[*"B",3*],40,[*[*1,0,1*]*],7*],[*
[*"C",4*],40,[*[*0,1,0,0*]*],3*],[*[*"A",1*],40,[*[*1*],[*3*],[*4
*]*],[*[*2,3*]*]*],[*[*"A",1*],40,[*[*1*],[*19*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],40,[*[*1*],[*1*],[*9*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],40,[*[*1*],[*1*],[*1*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],40,[*[*3*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],40,[*[*4
*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],40,[*[*39*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37*]*]*],[*[*"C",20*],40,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",20*],40,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"A",39*],40,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",39*],40,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"C",4*],41,[*[*1,0,0,0*]*],3*],[*[*"A",1*],41,[*[*40*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"B",20*],41,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",40*],41,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",40*],41,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",2*],42,[*[*0,2*],[*1,1*]*],3*],[*[*"A",2*],42,[*[*2,
0*],[*1,1*]*],3*],[*[*"A",2*],42,[*[*2,3*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],42,[*[*3,2*]*],[*[*2,3,5*]*]*],[*[*"C",4*],42,[*[*1,0,0,
0*]*],[*[*2,3*]*]*],[*[*"A",1*],42,[*[*1*],[*2*],[*6*]*],[*[*2,3,5
*]*]*],[*[*"A",1*],42,[*[*1*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",1*],42,[*[*2*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",
1*],42,[*[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],42,[*[*41*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"C",21*],42,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"D",21*],42,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",41*],42,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",41*],42,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",4*],43,[*[*0,0,0,2*]*],
3*],[*[*"C",5*],43,[*[*0,0,0,1,0*]*],5*],[*[*"A",1*],43,[*[*42*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"B",21*],
43,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*
2*]*]*],[*[*"A",42*],43,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",42*],43,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],44,[*[*1,1,1*]*],3*],[*[*
"B",2*],44,[*[*3,1*]*],7*],[*[*"B",4*],44,[*[*0,0,0,2*]*],[*[*2,
3*]*]*],[*[*"B",5*],44,[*[*0,0,0,1,0*]*],2*],[*[*"C",5*],44,[*[*
0,0,0,1,0*]*],[*[*5*]*]*],[*[*"D",5*],44,[*[*0,0,0,1,0*]*],2*],
[*[*"A",1*],44,[*[*1*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],44,[*[*1*],[*1*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
44,[*[*3*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],44,[*[*43*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"C",22*],44,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",22*],44,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",43*],44,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",43*],44,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],45,[*
[*0,1*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],45,[*[*0,1*],[*1,2*]
*],[*[*2*]*]*],[*[*"A",2*],45,[*[*0,1*],[*2,1*]*],[*[*2*]*]*],[*[*
"A",2*],45,[*[*0,1*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],45,[*[*1,
0*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],45,[*[*1,0*],[*1,2*]*],[*[*
2*]*]*],[*[*"A",2*],45,[*[*1,0*],[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],
45,[*[*1,0*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],45,[*[*0,8*]*],[*
[*2,3,5,7*]*]*],[*[*"A",2*],45,[*[*8,0*]*],[*[*2,3,5,7*]*]*],[*[*
"A",3*],45,[*[*0,1,2*]*],[*[*2*]*]*],[*[*"A",3*],45,[*[*2,1,0*]*],
[*[*2*]*]*],[*[*"A",4*],45,[*[*0,0,2,0*]*],3*],[*[*"A",4*],45,[*[*
0,1,0,1*]*],[*[*2*]*]*],[*[*"A",4*],45,[*[*0,2,0,0*]*],3*],[*[*
"A",4*],45,[*[*1,0,1,0*]*],[*[*2*]*]*],[*[*"A",8*],45,[*[*0,0,0,
0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",8*],45,[*[*2,0,0,0,0,0,0,0*]
*],[*[*2*]*]*],[*[*"A",9*],45,[*[*0,0,0,0,0,0,0,1,0*]*],[*[**]*]
*],[*[*"A",9*],45,[*[*0,1,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"D",5
*],45,[*[*0,0,0,1,0*]*],[*[*2*]*]*],[*[*"A",1*],45,[*[*2*],[*14*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],45,[*[*2*],[*2*],[*4*]*],[*
[*2,3*]*]*],[*[*"A",1*],45,[*[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],45,[*[*44*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43*]*]*],[*[*"B",22*],45,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",44*],45,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",44*],45,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
1*],46,[*[*1*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1
*],46,[*[*45*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]
*],[*[*"C",23*],46,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",23*],46,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",45*],
46,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",45*],46,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",6*],47,[*[*1,0,0,0,0,1*]*],7*],[*[*
"A",1*],47,[*[*46*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43*]*]*],[*[*"B",23*],47,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",46*],47,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",46*],47,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",2*],48,[*[*0,2*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],
48,[*[*2,0*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],48,[*[*1,5*]*],[*
[*2,3,5,7*]*]*],[*[*"A",2*],48,[*[*5,1*]*],[*[*2,3,5,7*]*]*],[*[*
"A",6*],48,[*[*1,0,0,0,0,1*]*],[*[*7*]*]*],[*[*"B",2*],48,[*[*1,
0*],[*1,1*]*],5*],[*[*"B",3*],48,[*[*0,0,1*],[*1,0,0*]*],2*],[*[*
"B",3*],48,[*[*1,0,1*]*],[*[*7*]*]*],[*[*"B",4*],48,[*[*0,1,0,0*]
*],2*],[*[*"C",3*],48,[*[*0,0,1*],[*1,0,0*]*],2*],[*[*"C",3*],48,
[*[*1,0,1*]*],2*],[*[*"C",4*],48,[*[*0,1,0,0*]*],[*[*3*]*]*],[*[*
"D",4*],48,[*[*0,1,0,1*]*],2*],[*[*"D",4*],48,[*[*1,0,0,1*]*],2
*],[*[*"D",4*],48,[*[*1,1,0,0*]*],2*],[*[*"A",1*],48,[*[*1*],[*2
*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],48,[*[*1*],[*3*],[*5*]*],[*
[*2,3,5*]*]*],[*[*"A",1*],48,[*[*1*],[*23*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"A",1*],48,[*[*1*],[*1*],[*2*],[*3*]*],[*[*2,3*]
*]*],[*[*"A",1*],48,[*[*1*],[*1*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",1*],48,[*[*1*],[*1*],[*1*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],
48,[*[*1*],[*1*],[*1*],[*1*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],48,[*[*
2*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],48,[*[*2*],[*15*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",1*],48,[*[*3*],[*11*]*],[*[*2,3,5,7,
11*]*]*],[*[*"A",1*],48,[*[*5*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],48,[*[*47*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47
*]*]*],[*[*"C",24*],48,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",24*],48,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",47*],48,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",47*],48,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],49,[*[*1,1*],[*
1,1*]*],3*],[*[*"A",6*],49,[*[*0,0,0,0,0,1*],[*1,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",6*],49,[*[*0,0,0,0,0,1*],[*0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",6*],49,[*[*1,0,0,0,0,0*],[*1,0,0,0,0,0*]
*],[*[**]*]*],[*[*"B",3*],49,[*[*0,0,1*],[*0,0,1*]*],[*[*2*]*]*],[*[*
"G",2*],49,[*[*0,1*],[*1,0*]*],3*],[*[*"G",2*],49,[*[*0,1*],[*0,
1*]*],[*[*2*]*]*],[*[*"G",2*],49,[*[*1,0*],[*1,0*]*],3*],[*[*"G",2
*],49,[*[*1,1*]*],3*],[*[*"A",1*],49,[*[*6*],[*6*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],49,[*[*48*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47*]*]*],[*[*"B",24*],49,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",48*],49,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",48*],49,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],50,[*[*0,3,0*]*],[*[*2,3
*]*]*],[*[*"A",4*],50,[*[*0,0,0,1*],[*0,0,1,0*]*],[*[**]*]*],[*[*
"A",4*],50,[*[*0,0,0,1*],[*0,1,0,0*]*],[*[**]*]*],[*[*"A",4*],50,
[*[*1,0,0,0*],[*0,0,1,0*]*],[*[**]*]*],[*[*"A",4*],50,[*[*1,0,0,
0*],[*0,1,0,0*]*],[*[**]*]*],[*[*"A",4*],50,[*[*0,0,2,0*]*],[*[*2,
3*]*]*],[*[*"A",4*],50,[*[*0,2,0,0*]*],[*[*2,3*]*]*],[*[*"A",5*],
50,[*[*0,0,0,1,1*]*],3*],[*[*"A",5*],50,[*[*1,1,0,0,0*]*],3*],
[*[*"B",2*],50,[*[*0,1*],[*2,0*]*],[*[*2*]*]*],[*[*"C",3*],50,[*[*
0,1,1*]*],3*],[*[*"A",1*],50,[*[*1*],[*4*],[*4*]*],[*[*2,3*]*]*],[*
[*"A",1*],50,[*[*1*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],50,[*[*4*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],50,[*[*
49*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*
"C",25*],50,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"D",25*],50,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",49
*],50,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",49*],50,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],51,[*[*0,
1,1,0*]*],3*],[*[*"A",1*],51,[*[*2*],[*16*]*],[*[*2,3,5,7,11,13
*]*]*],[*[*"A",1*],51,[*[*50*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47*]*]*],[*[*"B",25*],51,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",50*],
51,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",50*],51,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],52,[*[*
0,1,3*]*],5*],[*[*"A",3*],52,[*[*3,1,0*]*],5*],[*[*"B",2*],52,[*
[*1,0*],[*0,2*]*],5*],[*[*"B",2*],52,[*[*3,1*]*],5*],[*[*"F",4*],
52,[*[*1,0,0,0*]*],[*[*2*]*]*],[*[*"A",1*],52,[*[*1*],[*25*]*],[*[*
2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],52,[*[*1*],[*1*],[*12*]
*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],52,[*[*3*],[*12*]*],[*[*2,3,5,
7,11*]*]*],[*[*"A",1*],52,[*[*51*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47*]*]*],[*[*"C",26*],52,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",
26*],52,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",51*],52,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",51
*],52,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"D",5*],53,[*[*0,0,0,0,2*]*],5*],[*[*
"A",1*],53,[*[*52*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47*]*]*],[*[*"B",26*],53,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",52*],53,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",52*],53,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],54,[*[*0,1
*],[*1,0*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],54,[*[*0,1*],[*1,0*],
[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],54,[*[*0,1*],[*1,3*]*],5*],[*[*
"A",2*],54,[*[*0,1*],[*3,1*]*],5*],[*[*"A",2*],54,[*[*0,1*],[*0,
1*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],54,[*[*0,1*],[*0,1*],[*2,0*]
*],[*[*2*]*]*],[*[*"A",2*],54,[*[*1,0*],[*1,3*]*],5*],[*[*"A",2*],
54,[*[*1,0*],[*3,1*]*],5*],[*[*"A",2*],54,[*[*1,0*],[*1,0*],[*0,
2*]*],[*[*2*]*]*],[*[*"A",2*],54,[*[*1,0*],[*1,0*],[*2,0*]*],[*[*2
*]*]*],[*[*"B",2*],54,[*[*0,4*]*],7*],[*[*"D",5*],54,[*[*0,0,0,0,
2*]*],[*[*2,5*]*]*],[*[*"A",1*],54,[*[*1*],[*2*],[*8*]*],[*[*2,3,5,
7*]*]*],[*[*"A",1*],54,[*[*1*],[*2*],[*2*],[*2*]*],[*[*2*]*]*],[*[*
"A",1*],54,[*[*1*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],54,[*[*2*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1
*],54,[*[*2*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],54,[*[*5*],
[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],54,[*[*53*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"C",27*],54,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"D",27*],54,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",53*],54,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",53*],54,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10
*],55,[*[*0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",10*],55,[*[*
0,1,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],55,[*[*0,9*]*],[*
[*2,3,5,7*]*]*],[*[*"A",2*],55,[*[*9,0*]*],[*[*2,3,5,7*]*]*],[*[*
"A",9*],55,[*[*0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",9*],
55,[*[*2,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"B",2*],55,[*[*0,4
*]*],[*[*2,3,7*]*]*],[*[*"B",5*],55,[*[*0,0,0,1,0*]*],[*[*2*]*]*],
[*[*"C",5*],55,[*[*0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",1*],55,[*[*4
*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],55,[*[*54*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"B",27*],55,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",54*],55,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",54
*],55,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],56,[*[*0,0,1*],[*1,0,
1*]*],2*],[*[*"A",3*],56,[*[*1,0,0*],[*1,0,1*]*],2*],[*[*"A",3*],
56,[*[*0,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",3*],56,[*[*5,0,0*]*],[*[*
2,3,5*]*]*],[*[*"A",5*],56,[*[*0,0,0,0,3*]*],[*[*2,3*]*]*],[*[*
"A",5*],56,[*[*3,0,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",7*],56,[*[*0,
0,0,0,1,0,0*]*],[*[**]*]*],[*[*"A",7*],56,[*[*0,0,1,0,0,0,0*]
*],[*[**]*]*],[*[*"B",2*],56,[*[*1,0*],[*0,2*]*],[*[*2,5*]*]*],[*[*
"B",2*],56,[*[*5,0*]*],[*[*2,3,5*]*]*],[*[*"B",3*],56,[*[*0,0,1
*],[*1,0,0*]*],[*[*2*]*]*],[*[*"C",3*],56,[*[*0,0,3*]*],[*[*2,3*]*]
*],[*[*"D",4*],56,[*[*0,1,0,1*]*],[*[*2*]*]*],[*[*"D",4*],56,[*[*1,
0,0,1*]*],[*[*2*]*]*],[*[*"D",4*],56,[*[*1,1,0,0*]*],[*[*2*]*]*],[*
[*"E",7*],56,[*[*0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",1*],56,[*[*
1*],[*3*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],56,[*[*1*],[*27*]*],[*
[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],56,[*[*1*],[*1*],[*13
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],56,[*[*1*],[*1*],[*1*],[*
6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],56,[*[*3*],[*13*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],56,[*[*6*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],56,[*[*55*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53*]*]*],[*[*"C",28*],56,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",28*],
56,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",55*],56,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",55*],56,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],57,[*[*0,1
*],[*2,2*]*],5*],[*[*"A",2*],57,[*[*1,0*],[*2,2*]*],5*],[*[*"C",3
*],57,[*[*1,0,1*]*],3*],[*[*"A",1*],57,[*[*2*],[*18*]*],[*[*2,3,5,
7,11,13,17*]*]*],[*[*"A",1*],57,[*[*56*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53*]*]*],[*[*"B",28*],57,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",56*],57,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",56*],57,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],58,[*[*1,1,1*]*],5*],[*[*
"C",3*],58,[*[*0,1,1*]*],7*],[*[*"A",1*],58,[*[*1*],[*28*]*],[*[*
2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],58,[*[*57*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"C",29*],58,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",29*],58,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",57*],58,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",57*],58,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],59,[*[*58*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"B",29*],59,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",58*],59,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",58*],59,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],60,[*[*
0,2*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],60,[*[*0,2*],[*3,0*]*],
[*[*2,3*]*]*],[*[*"A",2*],60,[*[*2,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*
"A",2*],60,[*[*2,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],60,[*[*2,
4*]*],[*[*2,3,7*]*]*],[*[*"A",2*],60,[*[*4,2*]*],[*[*2,3,7*]*]*],[*
[*"A",3*],60,[*[*0,0,1*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",3*],60,[*
[*1,0,0*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",3*],60,[*[*0,1,0*],[*0,
0,2*]*],[*[*2*]*]*],[*[*"A",3*],60,[*[*0,1,0*],[*2,0,0*]*],[*[*2*]
*]*],[*[*"A",3*],60,[*[*0,2,1*]*],[*[*2*]*]*],[*[*"A",3*],60,[*[*1,
2,0*]*],[*[*2*]*]*],[*[*"B",2*],60,[*[*0,1*],[*1,1*]*],5*],[*[*"B",
2*],60,[*[*1,4*]*],11*],[*[*"A",1*],60,[*[*1*],[*2*],[*9*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],60,[*[*1*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*
[*"A",1*],60,[*[*1*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]
*],[*[*"A",1*],60,[*[*1*],[*1*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",
1*],60,[*[*1*],[*1*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],60,[*[*2*],[*3*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],60,[*[*2*],[*
19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],60,[*[*3*],[*14
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],60,[*[*4*],[*11*]*],[*[*2,
3,5,7,11*]*]*],[*[*"A",1*],60,[*[*5*],[*9*]*],[*[*2,3,5,7*]*]*],[*
[*"A",1*],60,[*[*59*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59*]*]*],[*[*"C",30*],60,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",30*],60,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",59*],60,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",59*],60,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"B",2*],61,[*[*0,5*]*],11*],[*[*"A",1*],61,[*[*60*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59*]*]*],[*
[*"B",30*],61,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",60*],61,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",60*],61,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",7*],62,[*[*1,0,0,0,0,0,1*]*],2*],[*[*
"C",3*],62,[*[*1,1,0*]*],5*],[*[*"A",1*],62,[*[*1*],[*30*]*],[*[*
2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],62,[*[*61*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*
"C",31*],62,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",31*],62,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",61*],62,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",61*],62,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",2*],63,[*[*0,1*],[*1,0*],[*1,1*]*],3*],[*[*"A",2*],63,[*[*
0,1*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],63,[*[*0,1*],[*5,0*]
*],[*[*2,3,5*]*]*],[*[*"A",2*],63,[*[*0,1*],[*0,1*],[*1,1*]*],3*],
[*[*"A",2*],63,[*[*1,0*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],63,
[*[*1,0*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],63,[*[*1,0*],[*1,
0*],[*1,1*]*],3*],[*[*"A",2*],63,[*[*1,6*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],63,[*[*3,3*]*],5*],[*[*"A",2*],63,[*[*6,1*]*],[*[*2,3,5*]
*]*],[*[*"A",7*],63,[*[*1,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"B",3*],
63,[*[*0,1,1*]*],3*],[*[*"C",3*],63,[*[*2,0,0*]*],5*],[*[*"A",1
*],63,[*[*2*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],
63,[*[*2*],[*2*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],63,[*[*6*],[*8
*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],63,[*[*62*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"B",31*],63,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",62*],63,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",62*],63,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",2*],64,[*[*1,1*],[*1,1*]*],[*[*3*]*]*],[*
[*"A",2*],64,[*[*3,3*]*],[*[*3,5,7*]*]*],[*[*"A",3*],64,[*[*0,0,1
*],[*1,0,0*],[*1,0,0*]*],[*[**]*]*],[*[*"A",3*],64,[*[*0,0,1*],[*
0,1,1*]*],3*],[*[*"A",3*],64,[*[*0,0,1*],[*1,1,0*]*],3*],[*[*
"A",3*],64,[*[*0,0,1*],[*0,0,1*],[*1,0,0*]*],[*[**]*]*],[*[*"A",3
*],64,[*[*0,0,1*],[*0,0,1*],[*0,0,1*]*],[*[**]*]*],[*[*"A",3*],64,
[*[*1,0,0*],[*0,1,1*]*],3*],[*[*"A",3*],64,[*[*1,0,0*],[*1,1,0
*]*],3*],[*[*"A",3*],64,[*[*1,0,0*],[*1,0,0*],[*1,0,0*]*],[*[**]*]
*],[*[*"A",3*],64,[*[*1,1,1*]*],[*[*3,5*]*]*],[*[*"A",7*],64,[*[*0,
0,0,0,0,0,1*],[*1,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",7*],64,[*
[*0,0,0,0,0,0,1*],[*0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",7*],
64,[*[*1,0,0,0,0,0,0*],[*1,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",
2*],64,[*[*0,1*],[*1,0*],[*1,0*]*],2*],[*[*"B",2*],64,[*[*0,1*],
[*1,1*]*],2*],[*[*"B",2*],64,[*[*0,1*],[*0,1*],[*1,0*]*],2*],[*[*
"B",2*],64,[*[*0,1*],[*0,1*],[*0,1*]*],2*],[*[*"B",2*],64,[*[*1,
0*],[*1,1*]*],[*[*5*]*]*],[*[*"B",2*],64,[*[*1,0*],[*1,0*],[*1,0*]
*],[*[**]*]*],[*[*"B",2*],64,[*[*3,1*]*],[*[*3,5,7*]*]*],[*[*"B",3*],
64,[*[*1,0,0*],[*1,0,0*]*],[*[**]*]*],[*[*"B",3*],64,[*[*0,1,1*]
*],2*],[*[*"B",3*],64,[*[*1,1,0*]*],5*],[*[*"B",4*],64,[*[*0,0,0,
1*],[*0,0,0,1*]*],2*],[*[*"B",5*],64,[*[*0,0,0,0,2*]*],11*],[*[*
"B",6*],64,[*[*0,0,0,0,1,0*]*],2*],[*[*"B",6*],64,[*[*1,0,0,0,
0,0*]*],[*[**]*]*],[*[*"C",3*],64,[*[*1,0,0*],[*1,0,0*]*],2*],[*[*
"C",3*],64,[*[*0,1,1*]*],[*[*3,7*]*]*],[*[*"C",4*],64,[*[*0,0,0,
1*],[*0,0,0,1*]*],[*[**]*]*],[*[*"C",6*],64,[*[*0,0,0,0,1,0*]*],
2*],[*[*"C",6*],64,[*[*0,0,0,0,1,0*]*],3*],[*[*"C",6*],64,[*[*1,
0,0,0,0,0*]*],2*],[*[*"D",4*],64,[*[*0,0,0,1*],[*0,1,0,0*]*],
[*[**]*]*],[*[*"D",4*],64,[*[*0,0,0,1*],[*1,0,0,0*]*],[*[**]*]*],[*
[*"D",4*],64,[*[*0,0,0,1*],[*0,0,0,1*]*],[*[**]*]*],[*[*"D",4*],
64,[*[*0,1,0,0*],[*1,0,0,0*]*],[*[**]*]*],[*[*"D",4*],64,[*[*0,1,
0,0*],[*0,1,0,0*]*],[*[**]*]*],[*[*"D",4*],64,[*[*1,0,0,0*],[*1,
0,0,0*]*],[*[**]*]*],[*[*"D",6*],64,[*[*0,0,0,0,1,0*]*],2*],[*[*
"D",7*],64,[*[*0,1,0,0,0,0,0*]*],[*[**]*]*],[*[*"D",7*],64,[*[*1,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"G",2*],64,[*[*1,1*]*],[*[*3,7*]*]
*],[*[*"A",1*],64,[*[*1*],[*3*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],64,[*[*1*],[*31*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*
[*"A",1*],64,[*[*1*],[*1*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],
64,[*[*1*],[*1*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],64,
[*[*1*],[*1*],[*1*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],64,[*[*1
*],[*1*],[*1*],[*1*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],64,[*[*1*],[*
1*],[*1*],[*1*],[*1*],[*1*]*],[*[**]*]*],[*[*"A",1*],64,[*[*3*],[*15
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],64,[*[*3*],[*3*],[*3*]*],
[*[*2,3*]*]*],[*[*"A",1*],64,[*[*7*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],64,[*[*63*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61*]*]*],[*[*"C",32*],64,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"D",32*],64,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",63*],64,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",63*],64,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],65,[*[*1,0,
0,2*]*],3*],[*[*"A",4*],65,[*[*2,0,0,1*]*],3*],[*[*"B",2*],65,[*
[*0,1*],[*0,2*]*],5*],[*[*"B",5*],65,[*[*0,0,0,0,2*]*],[*[*2,11
*]*]*],[*[*"C",6*],65,[*[*0,0,0,0,1,0*]*],[*[*2,3*]*]*],[*[*"A",1
*],65,[*[*4*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],65,[*[*64*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61*]*]
*],[*[*"B",32*],65,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",64*],
65,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",64*],65,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],66,[*
[*0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",10*],66,[*[*2,0,
0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",11*],66,[*[*0,0,0,0,0,
0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",11*],66,[*[*0,1,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],66,[*[*0,10*]*],[*[*2,3,5,7*]*]
*],[*[*"A",2*],66,[*[*10,0*]*],[*[*2,3,5,7*]*]*],[*[*"D",6*],66,[*
[*0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"A",1*],66,[*[*1*],[*2*],[*10*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],66,[*[*1*],[*32*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31*]*]*],[*[*"A",1*],66,[*[*2*],[*21*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],66,[*[*5*],[*10*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],66,[*[*65*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61*]*]*],[*[*"C",33*],66,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",33*],66,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",65*],66,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",65*],66,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],67,[*[*66*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61*]*]*],[*[*"B",33*],67,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",66*],67,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",66*],67,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],68,[*[*0,2,2*]*],5*],[*[*"A",3
*],68,[*[*2,2,0*]*],5*],[*[*"B",2*],68,[*[*2,2*]*],5*],[*[*"A",1
*],68,[*[*1*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*
[*"A",1*],68,[*[*1*],[*1*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],68,[*[*3*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
68,[*[*67*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67*]*]*],[*[*"C",34*],68,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",34*],68,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",67
*],68,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",67*],68,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",3*],69,[*[*2,0,2*]*],3*],[*[*"A",1*],69,[*[*2*],[*22*]
*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],69,[*[*68*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*
"B",34*],69,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",68*],69,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",68
*],69,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",3*],70,[*[*1,0,3*]*],[*[*2,3*]*]*],[*[*"A",3*],70,[*[*3,0,
1*]*],[*[*2,3*]*]*],[*[*"A",4*],70,[*[*0,0,0,4*]*],[*[*2,3*]*]*],[*
[*"A",4*],70,[*[*1,0,0,2*]*],[*[*2,3*]*]*],[*[*"A",4*],70,[*[*2,
0,0,1*]*],[*[*2,3*]*]*],[*[*"A",4*],70,[*[*4,0,0,0*]*],[*[*2,3*]
*]*],[*[*"A",5*],70,[*[*0,0,0,1,1*]*],[*[*3*]*]*],[*[*"A",5*],70,[*
[*1,1,0,0,0*]*],[*[*3*]*]*],[*[*"A",7*],70,[*[*0,0,0,1,0,0,0*]
*],[*[**]*]*],[*[*"B",2*],70,[*[*0,1*],[*0,2*]*],[*[*2,5*]*]*],[*[*
"C",3*],70,[*[*1,0,1*]*],[*[*2,3*]*]*],[*[*"A",1*],70,[*[*1*],[*4
*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],70,[*[*1*],[*34*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],70,[*[*4*],[*13*]*],[*
[*2,3,5,7,11,13*]*]*],[*[*"A",1*],70,[*[*6*],[*9*]*],[*[*2,3,5,7
*]*]*],[*[*"A",1*],70,[*[*69*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67*]*]*],[*[*"C",35*],70,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"D",35*],70,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",69*],70,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",69*],70,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],71,[*[*2,5*]*],
7*],[*[*"A",2*],71,[*[*5,2*]*],7*],[*[*"B",2*],71,[*[*2,2*]*],7*],
[*[*"A",1*],71,[*[*70*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67*]*]*],[*[*"B",35*],71,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",70*],71,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",70*],71,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],72,[*[*
0,1*],[*1,0*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2*],72,[*[*0,1*],[*1,
3*]*],[*[*3,5*]*]*],[*[*"A",2*],72,[*[*0,1*],[*3,1*]*],[*[*3,5*]*]
*],[*[*"A",2*],72,[*[*0,1*],[*0,1*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2
*],72,[*[*1,0*],[*1,3*]*],[*[*3,5*]*]*],[*[*"A",2*],72,[*[*1,0*],[*
3,1*]*],[*[*3,5*]*]*],[*[*"A",2*],72,[*[*1,0*],[*1,0*],[*1,1*]*],
[*[*3*]*]*],[*[*"A",1*],72,[*[*1*],[*2*],[*11*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],72,[*[*1*],[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*
"A",1*],72,[*[*1*],[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],72,
[*[*1*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],72,[*[*1*],[*35*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],72,[*[*1
*],[*1*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],72,[*[*1*],[*1*],
[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],72,[*[*1*],[*1*],
[*1*],[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],72,[*[*1*],[*1*],[*1*],[*
8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],72,[*[*2*],[*3*],[*5*]*],[*[*2,
3,5*]*]*],[*[*"A",1*],72,[*[*2*],[*23*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],72,[*[*2*],[*2*],[*7*]*],[*[*2,3,5,7*]*]*],
[*[*"A",1*],72,[*[*3*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*
"A",1*],72,[*[*5*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],72,[*
[*7*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],72,[*[*71*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*
[*"C",36*],72,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",36
*],72,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",71*],72,[*
[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",71*],72,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",1*],73,[*[*72*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*[*"B",36*],73,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",72*],73,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",72
*],73,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",4*],74,[*[*0,1,1,0*]*],2*],[*[*"A",1*],74,[*[*
1*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],
74,[*[*73*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73*]*]*],[*[*"C",37*],74,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",37*],74,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",73*],74,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",73*],74,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
2*],75,[*[*1,9*]*],11*],[*[*"A",2*],75,[*[*3,4*]*],7*],[*[*"A",2
*],75,[*[*4,3*]*],7*],[*[*"A",2*],75,[*[*9,1*]*],11*],[*[*"A",4*],
75,[*[*0,0,0,1*],[*0,0,0,2*]*],[*[*2*]*]*],[*[*"A",4*],75,[*[*0,
0,0,1*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",4*],75,[*[*1,0,0,0*],
[*0,0,0,2*]*],[*[*2*]*]*],[*[*"A",4*],75,[*[*1,0,0,0*],[*2,0,0,
0*]*],[*[*2*]*]*],[*[*"A",4*],75,[*[*0,1,1,0*]*],[*[*2,3*]*]*],[*[*
"A",1*],75,[*[*2*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],75,[*[*2
*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],75,[*[*4
*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],75,[*[*74*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73
*]*]*],[*[*"B",37*],75,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]
*],[*[*"A",74*],75,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",74*],75,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],76,
[*[*0,0,1*],[*0,2,0*]*],3*],[*[*"A",3*],76,[*[*1,0,0*],[*0,2,0
*]*],3*],[*[*"B",2*],76,[*[*1,3*]*],7*],[*[*"D",6*],76,[*[*0,0,0,
0,0,2*]*],3*],[*[*"A",1*],76,[*[*1*],[*37*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37*]*]*],[*[*"A",1*],76,[*[*1*],[*1*],[*18*]*],[*[*
2,3,5,7,11,13,17*]*]*],[*[*"A",1*],76,[*[*3*],[*18*]*],[*[*2,3,
5,7,11,13,17*]*]*],[*[*"A",1*],76,[*[*75*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"C",38
*],76,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",38*],
76,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",75*],76,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",75*],76,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",6*],77,[*[*0,0,0,0,
1,1*]*],3*],[*[*"A",6*],77,[*[*1,1,0,0,0,0*]*],3*],[*[*"B",3*],
77,[*[*0,0,3*]*],[*[*2,3*]*]*],[*[*"D",6*],77,[*[*0,0,0,0,0,2*]
*],[*[*2,3*]*]*],[*[*"E",6*],77,[*[*0,1,0,0,0,0*]*],3*],[*[*"G",2
*],77,[*[*0,3*]*],[*[*2,3*]*]*],[*[*"G",2*],77,[*[*2,0*]*],[*[*2,3
*]*]*],[*[*"A",1*],77,[*[*6*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],77,[*[*76*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73*]*]*],[*[*"B",38*],77,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[*2*]*]*],[*[*"A",76*],77,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",76*],77,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",11*],78,[*[*0,0,0,0,0,0,0,0,0,0,2*]*],[*[*
2*]*]*],[*[*"A",11*],78,[*[*2,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]
*],[*[*"A",12*],78,[*[*0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*
[*"A",12*],78,[*[*0,1,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",2*],78,[*[*0,11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],78,[*[*
11,0*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",5*],78,[*[*0,1,0,0,1*]*],
5*],[*[*"A",5*],78,[*[*1,0,0,1,0*]*],5*],[*[*"B",6*],78,[*[*0,0,
0,0,1,0*]*],[*[*2*]*]*],[*[*"C",3*],78,[*[*0,0,1*],[*0,1,0*]*],3
*],[*[*"C",6*],78,[*[*0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"E",6*],78,
[*[*0,1,0,0,0,0*]*],[*[*3*]*]*],[*[*"A",1*],78,[*[*1*],[*2*],[*12
*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],78,[*[*1*],[*38*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],78,[*[*2*],[*25*]
*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],78,[*[*5*],[*12*]
*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],78,[*[*77*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"C",
39*],78,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",
39*],78,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
77*],78,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",77*],78,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",8
*],79,[*[*1,0,0,0,0,0,0,1*]*],3*],[*[*"A",1*],79,[*[*78*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73
*]*]*],[*[*"B",39*],79,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",78*],79,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",78*],79,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",2*],80,[*[*1,1*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2
*],80,[*[*1,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],80,[*[*1,7*]*],
[*[*2,3,5,7*]*]*],[*[*"A",2*],80,[*[*7,1*]*],[*[*2,3,5,7*]*]*],[*
[*"A",3*],80,[*[*0,0,1*],[*0,0,3*]*],[*[*2,3*]*]*],[*[*"A",3*],
80,[*[*0,0,1*],[*0,1,1*]*],[*[*3*]*]*],[*[*"A",3*],80,[*[*0,0,1
*],[*0,2,0*]*],[*[*2,3*]*]*],[*[*"A",3*],80,[*[*0,0,1*],[*1,1,0*]
*],[*[*3*]*]*],[*[*"A",3*],80,[*[*0,0,1*],[*3,0,0*]*],[*[*2,3*]*]
*],[*[*"A",3*],80,[*[*1,0,0*],[*0,0,3*]*],[*[*2,3*]*]*],[*[*"A",3
*],80,[*[*1,0,0*],[*0,1,1*]*],[*[*3*]*]*],[*[*"A",3*],80,[*[*1,0,
0*],[*0,2,0*]*],[*[*2,3*]*]*],[*[*"A",3*],80,[*[*1,0,0*],[*1,1,0
*]*],[*[*3*]*]*],[*[*"A",3*],80,[*[*1,0,0*],[*3,0,0*]*],[*[*2,3*]*]
*],[*[*"A",3*],80,[*[*0,3,1*]*],5*],[*[*"A",3*],80,[*[*1,3,0*]*],
5*],[*[*"A",8*],80,[*[*1,0,0,0,0,0,0,1*]*],[*[*3*]*]*],[*[*"B",2
*],80,[*[*1,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"B",2*],80,[*[*1,0*],[*
1,0*],[*0,1*]*],[*[*2*]*]*],[*[*"B",2*],80,[*[*0,1*],[*1,1*]*],[*[*
2,5*]*]*],[*[*"B",2*],80,[*[*1,3*]*],[*[*2,3,7*]*]*],[*[*"A",1*],
80,[*[*1*],[*3*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],80,[*[*1*],
[*4*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],80,[*[*1*],[*39*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],80,[*[*1*],[*
1*],[*3*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],80,[*[*1*],[*1*],[*19*]
*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],80,[*[*1*],[*1*],[*1
*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],80,[*[*1*],[*1*],[*1*],[*1
*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],80,[*[*3*],[*19*]*],[*[*2,3,5,
7,11,13,17,19*]*]*],[*[*"A",1*],80,[*[*3*],[*3*],[*4*]*],[*[*2,3*]
*]*],[*[*"A",1*],80,[*[*4*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],80,[*[*7*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],80,[*[*79
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79*]*]*],[*[*"C",40*],80,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",40*],80,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",79*],80,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
79*],80,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],81,[*[*0,1*],[*1,0*],[*
1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],81,[*[*0,1*],[*2,2*]*],[*[*
2,5*]*]*],[*[*"A",2*],81,[*[*0,1*],[*0,1*],[*1,0*],[*1,0*]*],[*[*
*]*]*],[*[*"A",2*],81,[*[*0,1*],[*0,1*],[*0,1*],[*1,0*]*],[*[**]*]
*],[*[*"A",2*],81,[*[*0,1*],[*0,1*],[*0,1*],[*0,1*]*],[*[**]*]*],
[*[*"A",2*],81,[*[*1,0*],[*2,2*]*],[*[*2,5*]*]*],[*[*"A",2*],81,[*
[*1,0*],[*1,0*],[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],81,[*[*2,
5*]*],[*[*2,5,7*]*]*],[*[*"A",2*],81,[*[*5,2*]*],[*[*2,5,7*]*]*],[*
[*"A",8*],81,[*[*0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",8*],81,[*[*0,0,0,0,0,0,0,1*],[*0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",8*],81,[*[*1,0,0,0,0,0,0,0*],[*1,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],81,[*[*2,2*]*],[*[*2,5,7*]
*]*],[*[*"B",4*],81,[*[*0,0,0,1*],[*0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",1*],81,[*[*2*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],81,[*[*2*],[*2*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],81,
[*[*2*],[*2*],[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],81,[*[*8*],[*8*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],81,[*[*80*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79*]*]*],[*[*"B",
40*],81,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",80*],81,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",80*],81,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",2*],82,[*[*2,8*]*],11*],[*[*"A",2*],82,[*[*8,2*]*],
11*],[*[*"A",1*],82,[*[*1*],[*40*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37*]*]*],[*[*"A",1*],82,[*[*81*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79*]*]*],[*[*"C",
41*],82,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",41*],82,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"A",81*],82,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",81*],82,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",3*],83,[*[*2,0,2*]*],5*],[*[*"A",1*],
83,[*[*82*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79*]*]*],[*[*"B",41*],83,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",82*],83,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",82*],83,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],
84,[*[*0,1*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],84,[*[*0,1*],[*
6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],84,[*[*1,0*],[*0,6*]*],[*[*2,
3,5*]*]*],[*[*"A",2*],84,[*[*1,0*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*
"A",3*],84,[*[*0,1,0*],[*1,0,1*]*],2*],[*[*"A",3*],84,[*[*0,0,6
*]*],[*[*2,3,5*]*]*],[*[*"A",3*],84,[*[*0,1,3*]*],[*[*2,3,5*]*]*],
[*[*"A",3*],84,[*[*2,0,2*]*],[*[*2,3,5*]*]*],[*[*"A",3*],84,[*[*3,
1,0*]*],[*[*2,3,5*]*]*],[*[*"A",3*],84,[*[*6,0,0*]*],[*[*2,3,5*]
*]*],[*[*"A",5*],84,[*[*0,1,0,0,1*]*],[*[*5*]*]*],[*[*"A",5*],84,[*
[*1,0,0,1,0*]*],[*[*5*]*]*],[*[*"A",6*],84,[*[*0,0,0,0,0,3*]*],
[*[*2,3*]*]*],[*[*"A",6*],84,[*[*3,0,0,0,0,0*]*],[*[*2,3*]*]*],[*
[*"A",8*],84,[*[*0,0,0,0,0,1,0,0*]*],[*[**]*]*],[*[*"A",8*],84,[*
[*0,0,1,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],84,[*[*1,5*]*],13*],
[*[*"B",2*],84,[*[*6,0*]*],[*[*2,3,5*]*]*],[*[*"B",3*],84,[*[*0,0,
1*],[*0,1,0*]*],2*],[*[*"B",4*],84,[*[*0,1,0,0*]*],[*[*2*]*]*],[*
[*"C",3*],84,[*[*0,0,1*],[*0,1,0*]*],[*[*3*]*]*],[*[*"C",3*],84,[*
[*0,0,1*],[*1,0,0*]*],[*[*2*]*]*],[*[*"C",3*],84,[*[*2,0,0*]*],[*
[*2,5*]*]*],[*[*"G",2*],84,[*[*0,1*],[*1,0*]*],2*],[*[*"A",1*],84,
[*[*1*],[*2*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],84,[*[*
1*],[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],84,[*[*1*],[*41*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],84,[*[*
1*],[*1*],[*2*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],84,[*[*1*],[*1
*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],84,[*[*2*],
[*3*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],84,[*[*2*],[*27*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],84,[*[*3*],[*20*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],84,[*[*5*],[*13*]*],[*[*2,3,
5,7,11,13*]*]*],[*[*"A",1*],84,[*[*6*],[*11*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],84,[*[*83*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"C",42*],84,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",42*],
84,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",83*],84,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",83*],84,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],85,[*[*0,4,0*]*],5*],[*[*"B",2
*],85,[*[*0,6*]*],13*],[*[*"A",1*],85,[*[*4*],[*16*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],85,[*[*84*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"B",42
*],85,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",84*],85,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",84*],85,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",2*],86,[*[*2,3*]*],5*],[*[*
"A",1*],86,[*[*1*],[*42*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41*]*]*],[*[*"A",1*],86,[*[*85*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"C",43
*],86,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"D",43*],86,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"A",85*],86,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
85*],86,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],87,[*[*
3,7*]*],11*],[*[*"A",2*],87,[*[*7,3*]*],11*],[*[*"A",1*],87,[*[*2
*],[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],87,[*[*
86*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83*]*]*],[*[*"B",43*],87,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",86*],87,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",86*],87,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",1*],88,[*[*1*],[*3*],[*10*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],88,[*[*1*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43*]*]*],[*[*"A",1*],88,[*[*1*],[*1*],[*21*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],88,[*[*1*],[*1*],[*1*],[*10*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],88,[*[*3*],[*21*]*],[*[*2,3,5,7,
11,13,17,19*]*]*],[*[*"A",1*],88,[*[*7*],[*10*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],88,[*[*87*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"C",44*],88,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",
44*],88,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",87*],88,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
87*],88,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",6*],89,
[*[*0,0,0,0,0,2*]*],13*],[*[*"C",3*],89,[*[*0,2,0*]*],7*],[*[*
"C",7*],89,[*[*0,0,0,0,0,1,0*]*],7*],[*[*"A",1*],89,[*[*88*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83*]*]*],[*[*"B",44*],89,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",88*],89,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",88*],89,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",2*],90,[*[*0,1*],[*1,0*],[*0,3*]*],[*[*
2,3*]*]*],[*[*"A",2*],90,[*[*0,1*],[*1,0*],[*3,0*]*],[*[*2,3*]*]
*],[*[*"A",2*],90,[*[*0,1*],[*0,1*],[*0,3*]*],[*[*2,3*]*]*],[*[*
"A",2*],90,[*[*0,1*],[*0,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],
90,[*[*1,0*],[*1,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],90,[*[*1,
0*],[*1,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],90,[*[*0,2*],[*0,
4*]*],[*[*2,3*]*]*],[*[*"A",2*],90,[*[*0,2*],[*1,2*]*],[*[*2*]*]*],
[*[*"A",2*],90,[*[*0,2*],[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],90,[*[*
0,2*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],90,[*[*2,0*],[*0,4*]*],
[*[*2,3*]*]*],[*[*"A",2*],90,[*[*2,0*],[*1,2*]*],[*[*2*]*]*],[*[*
"A",2*],90,[*[*2,0*],[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],90,[*[*2,0
*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],90,[*[*3,4*]*],[*[*2,3,7*]*]
*],[*[*"A",2*],90,[*[*4,3*]*],[*[*2,3,7*]*]*],[*[*"A",2*],90,[*[*4,
6*]*],11*],[*[*"A",2*],90,[*[*6,4*]*],11*],[*[*"A",3*],90,[*[*0,1,
0*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",5*],90,[*[*0,0,0,0,1*],[*0,
0,0,1,0*]*],[*[**]*]*],[*[*"A",5*],90,[*[*0,0,0,0,1*],[*0,1,0,
0,0*]*],[*[**]*]*],[*[*"A",5*],90,[*[*1,0,0,0,0*],[*0,0,0,1,0*]
*],[*[**]*]*],[*[*"A",5*],90,[*[*1,0,0,0,0*],[*0,1,0,0,0*]*],[*[*
*]*]*],[*[*"A",5*],90,[*[*0,0,0,2,0*]*],3*],[*[*"A",5*],90,[*[*0,
0,1,0,1*]*],2*],[*[*"A",5*],90,[*[*0,2,0,0,0*]*],3*],[*[*"A",5
*],90,[*[*1,0,1,0,0*]*],2*],[*[*"B",6*],90,[*[*0,0,0,0,0,2*]*],
[*[*2,13*]*]*],[*[*"B",7*],90,[*[*0,0,0,0,0,1,0*]*],2*],[*[*"C",
3*],90,[*[*0,2,0*]*],[*[*2,7*]*]*],[*[*"C",7*],90,[*[*0,0,0,0,0,
1,0*]*],[*[*7*]*]*],[*[*"D",7*],90,[*[*0,0,0,0,0,1,0*]*],2*],[*[*
"A",1*],90,[*[*1*],[*2*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],90,[*[*1*],[*2*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],
90,[*[*1*],[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],90,[*[*1*],
[*44*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*
"A",1*],90,[*[*2*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],90,[*
[*2*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],
90,[*[*2*],[*2*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],90,[*[*4*],
[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],90,[*[*5*],[*14*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],90,[*[*8*],[*9*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],90,[*[*89*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*"C",45
*],90,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",45*],90,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",89*],90,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",89*],90,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",12*],91,[*[*0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2
*]*]*],[*[*"A",12*],91,[*[*2,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]
*]*],[*[*"A",13*],91,[*[*0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]
*],[*[*"A",13*],91,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",2*],91,[*[*0,12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],
91,[*[*5,5*]*],11*],[*[*"A",2*],91,[*[*12,0*]*],[*[*2,3,5,7,11*]
*]*],[*[*"B",2*],91,[*[*0,5*]*],[*[*2,3,5,11*]*]*],[*[*"D",7*],91,
[*[*0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"A",1*],91,[*[*6*],[*12*]*],
[*[*2,3,5,7,11*]*]*],[*[*"A",1*],91,[*[*90*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]
*],[*[*"B",45*],91,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",90*],91,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",90*],91,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",1*],92,[*[*1*],[*45*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],92,[*[*1*],[*1*],[*22*]
*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],92,[*[*3*],[*22*]*],
[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],92,[*[*91*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89*]*]*],[*[*"C",46*],92,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",46*],92,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",91*],92,[*
[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",91*],92,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],93,
[*[*2*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],
93,[*[*92*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89*]*]*],[*[*"B",46*],93,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",92
*],93,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
92*],93,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",1*],94,[*[*1*],[*46*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43*]*]*],[*[*"A",1*],94,[*[*93*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*
"C",47*],94,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"D",47*],94,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",93*],94,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",93*],94,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],95,[*
[*4*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],95,[*[*94*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89*]*]*],[*[*"B",47*],95,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",94*],95,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",94*],95,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
3*],96,[*[*0,0,1*],[*1,0,0*],[*0,1,0*]*],[*[**]*]*],[*[*"A",3*],
96,[*[*0,0,1*],[*0,0,1*],[*0,1,0*]*],[*[**]*]*],[*[*"A",3*],96,[*
[*1,0,0*],[*1,0,0*],[*0,1,0*]*],[*[**]*]*],[*[*"A",3*],96,[*[*0,
1,0*],[*0,1,1*]*],3*],[*[*"A",3*],96,[*[*0,1,0*],[*1,1,0*]*],3
*],[*[*"B",2*],96,[*[*1,0*],[*1,2*]*],7*],[*[*"A",1*],96,[*[*1*],[*
2*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],96,[*[*1*],[*2*],[*15*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],96,[*[*1*],[*3*],[*11*]*],
[*[*2,3,5,7,11*]*]*],[*[*"A",1*],96,[*[*1*],[*5*],[*7*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],96,[*[*1*],[*47*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*
2*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*3*],[*
5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*23*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*1*],[*
2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*1*],[*11*]
*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*1*],[*1*],[*
5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],96,[*[*1*],[*1*],[*1*],[*1*],[*1
*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],96,[*[*2*],[*3*],[*7*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],96,[*[*2*],[*31*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31*]*]*],[*[*"A",1*],96,[*[*3*],[*23*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],96,[*[*3*],[*3*],[*5*]*],[*[*2,3,5*]
*]*],[*[*"A",1*],96,[*[*5*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],96,[*[*7*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],96,[*
[*95*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89*]*]*],[*[*"C",48*],96,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",48*],96,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",95*],96,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",95*],96,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"G",2*],97,[*[*2,1*]*],5*],[*[*"A",1*],
97,[*[*96*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89*]*]*],[*[*"B",48*],97,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",96*],97,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",96*],97,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",9*],98,[*[*1,0,0,0,0,0,0,0,1*]
*],2*],[*[*"A",9*],98,[*[*1,0,0,0,0,0,0,0,1*]*],5*],[*[*"G",2
*],98,[*[*0,1*],[*1,0*]*],[*[*2,3*]*]*],[*[*"A",1*],98,[*[*1*],[*6
*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],98,[*[*1*],[*48*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],98,[*[*
6*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],98,[*[*97*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97*]*]*],[*[*"C",49*],98,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",49*],98,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",97*],98,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",97*],98,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],99,[*[*0,1*],[*1,
5*]*],7*],[*[*"A",2*],99,[*[*0,1*],[*5,1*]*],7*],[*[*"A",2*],99,[*
[*1,0*],[*1,5*]*],7*],[*[*"A",2*],99,[*[*1,0*],[*5,1*]*],7*],[*[*
"A",2*],99,[*[*1,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],99,[*[*8,1*]
*],[*[*2,3,5,7*]*]*],[*[*"A",9*],99,[*[*1,0,0,0,0,0,0,0,1*]*],
[*[*2,5*]*]*],[*[*"A",1*],99,[*[*2*],[*32*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],99,[*[*2*],[*2*],[*10*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],99,[*[*8*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],99,[*[*98*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79,83,89,97*]*]*],[*[*"B",49*],99,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",98*],99,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",98*],99,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],100,[*[*0,
3*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],100,[*[*0,3*],[*0,3*]*],[*
[*2,3*]*]*],[*[*"A",2*],100,[*[*3,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*
"A",3*],100,[*[*0,0,2*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],100,[*
[*0,0,2*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],100,[*[*2,0,0*],[*
2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],100,[*[*1,0,4*]*],7*],[*[*"A",3
*],100,[*[*4,0,1*]*],7*],[*[*"A",4*],100,[*[*0,0,1,0*],[*0,1,0,
0*]*],[*[**]*]*],[*[*"A",4*],100,[*[*0,0,1,0*],[*0,0,1,0*]*],[*[**]
*]*],[*[*"A",4*],100,[*[*0,1,0,0*],[*0,1,0,0*]*],[*[**]*]*],[*[*
"A",9*],100,[*[*0,0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",9*],100,[*[*0,0,0,0,0,0,0,0,1*],[*0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",9*],100,[*[*1,0,0,0,0,0,0,
0,0*],[*1,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],100,[*[*1,
0*],[*0,1*],[*0,1*]*],[*[*2*]*]*],[*[*"B",2*],100,[*[*1,0*],[*0,3
*]*],7*],[*[*"B",2*],100,[*[*1,0*],[*2,1*]*],3*],[*[*"B",2*],100,[*
[*0,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*"B",2*],100,[*[*2,0*],[*2,0*]
*],[*[*2*]*]*],[*[*"B",5*],100,[*[*0,0,0,0,1*],[*0,0,0,0,1*]*],2
*],[*[*"B",5*],100,[*[*0,0,1,0,0*]*],2*],[*[*"C",5*],100,[*[*0,0,
0,0,1*],[*0,0,0,0,1*]*],[*[**]*]*],[*[*"C",5*],100,[*[*0,0,1,0,
0*]*],2*],[*[*"D",5*],100,[*[*0,0,0,0,1*],[*0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",5*],100,[*[*0,0,1,0,0*]*],2*],[*[*"A",1*],100,[*[*1
*],[*4*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],100,[*[*1*],[*49*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1
*],100,[*[*1*],[*1*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],100,[*[*
1*],[*1*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
100,[*[*3*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],100,[*[*3*],[*24
*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],100,[*[*4*],[*
19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],100,[*[*9*],[*9
*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],100,[*[*99*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97
*]*]*],[*[*"C",50*],100,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",50*],100,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",99*],100,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",99*],100,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],101,[*[*100*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97*]*]*],[*[*"B",50*],101,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",100
*],101,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",100*],101,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],102,[*[*1,11*]
*],13*],[*[*"A",2*],102,[*[*11,1*]*],13*],[*[*"A",1*],102,[*[*1*],[*
2*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],102,[*[*1*],[*50
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*
"A",1*],102,[*[*2*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]
*]*],[*[*"A",1*],102,[*[*5*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],102,[*[*101*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101*]*]*],[*[*"C",51*],
102,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"D",51*],102,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",101*],102,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",101*],102,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],103,[*[*0,1,0,2*]
*],5*],[*[*"A",4*],103,[*[*2,0,1,0*]*],5*],[*[*"D",7*],103,[*[*0,
0,0,0,0,0,2*]*],7*],[*[*"A",1*],103,[*[*102*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101*]*]*],[*[*"B",51*],103,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",102*],103,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",102*],103,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",3*],104,[*[*1,1,0*]
*],3*],[*[*"B",3*],104,[*[*3,0,0*]*],5*],[*[*"D",4*],104,[*[*0,0,
0,3*]*],5*],[*[*"D",4*],104,[*[*0,0,1,1*]*],3*],[*[*"D",4*],104,
[*[*0,1,1,0*]*],3*],[*[*"D",4*],104,[*[*0,3,0,0*]*],5*],[*[*"D",
4*],104,[*[*1,0,1,0*]*],3*],[*[*"D",4*],104,[*[*3,0,0,0*]*],5*],
[*[*"D",7*],104,[*[*0,0,0,0,0,0,2*]*],[*[*2,7*]*]*],[*[*"A",1*],
104,[*[*1*],[*3*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],104,[*
[*1*],[*51*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]
*]*],[*[*"A",1*],104,[*[*1*],[*1*],[*25*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],104,[*[*1*],[*1*],[*1*],[*12*]*],[*[*2,3,5,
7,11*]*]*],[*[*"A",1*],104,[*[*3*],[*25*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],104,[*[*7*],[*12*]*],[*[*2,3,5,7,11*]*]*],
[*[*"A",1*],104,[*[*103*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103*]*]*],[*[*
"C",52*],104,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",52*],104,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",103*],104,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",103*],104,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*
"A",13*],105,[*[*0,0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*
[*"A",13*],105,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],
[*[*"A",14*],105,[*[*0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]
*],[*[*"A",14*],105,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",2*],105,[*[*0,1*],[*1,4*]*],[*[*2,3*]*]*],[*[*"A",2*],
105,[*[*0,1*],[*4,1*]*],[*[*2,3*]*]*],[*[*"A",2*],105,[*[*1,0*],[*
1,4*]*],[*[*2,3*]*]*],[*[*"A",2*],105,[*[*1,0*],[*4,1*]*],[*[*2,3
*]*]*],[*[*"A",2*],105,[*[*1,1*],[*1,2*]*],3*],[*[*"A",2*],105,[*[*
1,1*],[*2,1*]*],3*],[*[*"A",2*],105,[*[*0,13*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"A",2*],105,[*[*2,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
105,[*[*6,2*]*],[*[*2,3,5*]*]*],[*[*"A",2*],105,[*[*13,0*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",3*],105,[*[*0,4,0*]*],[*[*2,3,5*]*]*],
[*[*"A",4*],105,[*[*0,0,1,2*]*],[*[*2*]*]*],[*[*"A",4*],105,[*[*2,
1,0,0*]*],[*[*2*]*]*],[*[*"A",5*],105,[*[*0,0,0,2,0*]*],[*[*2,3*]
*]*],[*[*"A",5*],105,[*[*0,0,1,0,1*]*],[*[*2*]*]*],[*[*"A",5*],105,
[*[*0,2,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",5*],105,[*[*1,0,1,0,0*]
*],[*[*2*]*]*],[*[*"B",2*],105,[*[*4,1*]*],[*[*2,3*]*]*],[*[*"B",3*],
105,[*[*0,1,1*]*],[*[*2,3*]*]*],[*[*"B",7*],105,[*[*0,0,0,0,0,1,
0*]*],[*[*2*]*]*],[*[*"C",7*],105,[*[*0,0,0,0,0,0,2*]*],[*[*2*]*]
*],[*[*"A",1*],105,[*[*2*],[*4*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1
*],105,[*[*2*],[*34*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*
[*"A",1*],105,[*[*4*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],105,[*[*6*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
105,[*[*104*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103*]*]*],[*[*"B",52*],105,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",104*],105,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",104*],
105,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],106,[*[*1*],[*52*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],106,[*[*
105*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103*]*]*],[*[*"C",53*],106,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"D",53*],106,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",105*],106,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",105*],106,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],
107,[*[*106*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103*]*]*],[*[*"B",53*],107,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",106*],107,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
106*],107,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],108,[*[*0,1*],[*0,2
*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],108,[*[*0,1*],[*0,2*],[*0,2*]
*],[*[*2*]*]*],[*[*"A",2*],108,[*[*0,1*],[*2,0*],[*2,0*]*],[*[*2*]*]
*],[*[*"A",2*],108,[*[*0,1*],[*0,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",2
*],108,[*[*0,1*],[*2,4*]*],7*],[*[*"A",2*],108,[*[*0,1*],[*4,2*]
*],7*],[*[*"A",2*],108,[*[*0,1*],[*7,0*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],108,[*[*1,0*],[*0,2*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],
108,[*[*1,0*],[*0,2*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],108,[*[*1,
0*],[*2,0*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],108,[*[*1,0*],[*0,7
*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],108,[*[*1,0*],[*2,4*]*],7*],[*[*
"A",2*],108,[*[*1,0*],[*4,2*]*],7*],[*[*"A",2*],108,[*[*1,0*],[*
7,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],108,[*[*0,2*],[*1,3*]*],5
*],[*[*"A",2*],108,[*[*0,2*],[*3,1*]*],5*],[*[*"A",2*],108,[*[*2,0
*],[*1,3*]*],5*],[*[*"A",2*],108,[*[*2,0*],[*3,1*]*],5*],[*[*"A",1
*],108,[*[*1*],[*2*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1
*],108,[*[*1*],[*2*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],108,
[*[*1*],[*5*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],108,[*[*1*],[*
53*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],
[*[*"A",1*],108,[*[*1*],[*1*],[*2*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],108,[*[*1*],[*1*],[*2*],[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1
*],108,[*[*1*],[*1*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],108,[*[*2*],[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
108,[*[*2*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],108,[*[*2*],[*
35*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],108,[*
[*2*],[*2*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],108,[*[*2*],
[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],108,[*[*3*],[*26*]*],[*
[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],108,[*[*5*],[*17*]*],
[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],108,[*[*8*],[*11*]*],[*[*2,
3,5,7,11*]*]*],[*[*"A",1*],108,[*[*107*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107*]*]*],[*[*"C",54*],108,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",54*],
108,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",107*],108,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",107*],108,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],109,[*[*108
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107*]*]*],[*[*"B",54*],109,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[*2*]*]*],[*[*"A",108*],109,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
108*],109,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"C",5*],110,[*[*0,0,1,
0,0*]*],[*[*2*]*]*],[*[*"A",1*],110,[*[*1*],[*4*],[*10*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],110,[*[*1*],[*54*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],110,[*[*4*],[*21*]
*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],110,[*[*9*],[*10*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],110,[*[*109*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109*]*]*],[*[*"C",55*],110,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",55*],110,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",109*],110,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",109*],110,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",2*],111,[*[*0,1*],[*3,3*]*],7*],[*[*"A",2*],111,[*[*1,0*],
[*3,3*]*],7*],[*[*"A",2*],111,[*[*2,10*]*],13*],[*[*"A",2*],111,[*
[*10,2*]*],13*],[*[*"A",1*],111,[*[*2*],[*36*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],111,[*[*110*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109*]*]*],[*[*"B",55*],111,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]
*]*],[*[*"A",110*],111,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",110*],
111,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",6*],112,[*[*0,0,0,
0,1,1*]*],[*[*3*]*]*],[*[*"A",6*],112,[*[*1,1,0,0,0,0*]*],[*[*3*]
*]*],[*[*"A",7*],112,[*[*0,0,0,0,0,1,1*]*],3*],[*[*"A",7*],112,[*
[*1,1,0,0,0,0,0*]*],3*],[*[*"B",3*],112,[*[*1,0,0*],[*0,1,0*]
*],2*],[*[*"B",3*],112,[*[*1,1,0*]*],[*[*3,5*]*]*],[*[*"B",3*],112,
[*[*3,0,0*]*],[*[*2,3,5*]*]*],[*[*"B",4*],112,[*[*1,0,0,1*]*],3
*],[*[*"C",3*],112,[*[*1,0,0*],[*0,1,0*]*],2*],[*[*"C",3*],112,[*
[*1,1,0*]*],2*],[*[*"C",4*],112,[*[*0,0,1,1*]*],3*],[*[*"D",4*],
112,[*[*0,0,0,3*]*],[*[*2,3,5*]*]*],[*[*"D",4*],112,[*[*0,3,0,0
*]*],[*[*2,3,5*]*]*],[*[*"D",4*],112,[*[*3,0,0,0*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],112,[*[*1*],[*3*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],
[*[*"A",1*],112,[*[*1*],[*6*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],112,[*[*1*],[*55*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53*]*]*],[*[*"A",1*],112,[*[*1*],[*1*],[*3*],[*6*]*],[*[*2,
3,5*]*]*],[*[*"A",1*],112,[*[*1*],[*1*],[*27*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],112,[*[*1*],[*1*],[*1*],[*13*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",1*],112,[*[*1*],[*1*],[*1*],[*1*],[*6
*]*],[*[*2,3,5*]*]*],[*[*"A",1*],112,[*[*3*],[*27*]*],[*[*2,3,5,7,
11,13,17,19,23*]*]*],[*[*"A",1*],112,[*[*3*],[*3*],[*6*]*],[*[*2,
3,5*]*]*],[*[*"A",1*],112,[*[*6*],[*15*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],112,[*[*7*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],112,[*[*111*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109*]*]*],
[*[*"C",56*],112,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",56*],112,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",111*],112,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",111*],112,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
1*],113,[*[*112*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109*]*]*],[*[*"B",
56*],113,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",112*],113,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",112*],113,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",2*],114,[*[*0,2*],[*2,2*]*],5*],[*[*"A",2
*],114,[*[*2,0*],[*2,2*]*],5*],[*[*"A",2*],114,[*[*3,5*]*],7*],[*[*
"A",2*],114,[*[*5,3*]*],7*],[*[*"A",3*],114,[*[*0,1,0*],[*0,2,0
*]*],3*],[*[*"A",5*],114,[*[*1,0,0,0,2*]*],7*],[*[*"A",5*],114,[*
[*2,0,0,0,1*]*],7*],[*[*"A",1*],114,[*[*1*],[*2*],[*18*]*],[*[*2,
3,5,7,11,13,17*]*]*],[*[*"A",1*],114,[*[*1*],[*56*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],114,[*
[*2*],[*37*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*
"A",1*],114,[*[*5*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1
*],114,[*[*113*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*
"C",57*],114,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",57*],114,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",113*],114,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",113*],114,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",4*],115,[*[*0,0,0,1*],[*1,0,0,1*]*],5*],[*[*"A",4
*],115,[*[*1,0,0,0*],[*1,0,0,1*]*],5*],[*[*"B",2*],115,[*[*4,2*]
*],5*],[*[*"A",1*],115,[*[*4*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]
*]*],[*[*"A",1*],115,[*[*114*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113*]*]*],[*[*"B",57*],115,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",114*],115,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",114
*],115,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],
116,[*[*0,1,5*]*],7*],[*[*"A",3*],116,[*[*1,1,2*]*],3*],[*[*"A",3
*],116,[*[*2,1,1*]*],3*],[*[*"A",3*],116,[*[*5,1,0*]*],7*],[*[*
"B",2*],116,[*[*3,3*]*],11*],[*[*"B",2*],116,[*[*5,1*]*],7*],[*[*
"A",1*],116,[*[*1*],[*57*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53*]*]*],[*[*"A",1*],116,[*[*1*],[*1*],[*28*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],116,[*[*3*],[*28*]*],[*[*
2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],116,[*[*115*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113*]*]*],[*[*"C",58*],116,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",58*],116,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",115*],116,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",115*],116,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",2*],117,[*[*0,1*],[*2,3*]*],5*],[*[*"A",2*],117,[*[*0,1*],[*
3,2*]*],5*],[*[*"A",2*],117,[*[*1,0*],[*2,3*]*],5*],[*[*"A",2*],
117,[*[*1,0*],[*3,2*]*],5*],[*[*"A",2*],117,[*[*4,4*]*],7*],[*[*
"A",1*],117,[*[*2*],[*38*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37*]*]*],[*[*"A",1*],117,[*[*2*],[*2*],[*12*]*],[*[*2,3,5,7,11*]*]
*],[*[*"A",1*],117,[*[*8*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1
*],117,[*[*116*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*
"B",58*],117,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",116*],117,[*
[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",116*],117,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],118,[*[*
3,9*]*],13*],[*[*"A",2*],118,[*[*9,3*]*],13*],[*[*"B",7*],118,[*[*
0,0,0,0,0,0,2*]*],3*],[*[*"B",7*],118,[*[*0,0,0,0,0,0,2*]*],
5*],[*[*"B",8*],118,[*[*0,0,0,0,0,0,1,0*]*],2*],[*[*"C",8*],
118,[*[*0,0,0,0,0,0,1,0*]*],2*],[*[*"D",8*],118,[*[*0,0,0,0,
0,0,1,0*]*],2*],[*[*"A",1*],118,[*[*1*],[*58*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],118,[*[*117
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"C",59*],118,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",59*],118,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",117*],118,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",117*],118,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],119,[*[*1,0,0,0,0,0,0,0,0,1
*]*],11*],[*[*"B",7*],119,[*[*0,0,0,0,0,0,2*]*],[*[*2,3,5*]*]*],
[*[*"C",8*],119,[*[*0,0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"A",1*],
119,[*[*6*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],119,[*[*
118*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"B",59*],
119,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",118*],119,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",118*],119,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],120,
[*[*1,0,0,0,0,0,0,0,0,1*]*],[*[*11*]*]*],[*[*"A",14*],120,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",14*],120,[*
[*2,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",15*],
120,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",
15*],120,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",2*],120,[*[*1,1*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],120,[*[*
1,1*],[*1,2*]*],[*[*2,3*]*]*],[*[*"A",2*],120,[*[*1,1*],[*2,1*]*],
[*[*2,3*]*]*],[*[*"A",2*],120,[*[*1,1*],[*4,0*]*],[*[*2,3*]*]*],[*[*
"A",2*],120,[*[*0,14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",2*],120,
[*[*1,9*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],120,[*[*3,5*]*],[*[*
2,3,5,7*]*]*],[*[*"A",2*],120,[*[*5,3*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],120,[*[*9,1*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],120,[*[*
14,0*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",3*],120,[*[*0,1,0*],[*
0,0,3*]*],[*[*2,3*]*]*],[*[*"A",3*],120,[*[*0,1,0*],[*0,1,1*]*],
[*[*3*]*]*],[*[*"A",3*],120,[*[*0,1,0*],[*0,2,0*]*],[*[*2,3*]*]*],
[*[*"A",3*],120,[*[*0,1,0*],[*1,1,0*]*],[*[*3*]*]*],[*[*"A",3*],
120,[*[*0,1,0*],[*3,0,0*]*],[*[*2,3*]*]*],[*[*"A",3*],120,[*[*0,
0,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",3*],120,[*[*1,0,4*]*],[*[*2,3,
7*]*]*],[*[*"A",3*],120,[*[*4,0,1*]*],[*[*2,3,7*]*]*],[*[*"A",3*],
120,[*[*7,0,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",4*],120,[*[*0,0,0,1
*],[*1,0,0,1*]*],[*[*5*]*]*],[*[*"A",4*],120,[*[*1,0,0,0*],[*1,0,
0,1*]*],[*[*5*]*]*],[*[*"A",5*],120,[*[*0,0,0,0,1*],[*0,0,1,0,0
*]*],[*[**]*]*],[*[*"A",5*],120,[*[*1,0,0,0,0*],[*0,0,1,0,0*]*],[*
[**]*]*],[*[*"A",5*],120,[*[*1,0,0,0,2*]*],[*[*2,7*]*]*],[*[*"A",5
*],120,[*[*2,0,0,0,1*]*],[*[*2,7*]*]*],[*[*"A",7*],120,[*[*0,0,0,
0,0,0,3*]*],[*[*2,3*]*]*],[*[*"A",7*],120,[*[*3,0,0,0,0,0,0*]
*],[*[*2,3*]*]*],[*[*"A",9*],120,[*[*0,0,0,0,0,0,1,0,0*]*],[*[**]
*]*],[*[*"A",9*],120,[*[*0,0,1,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",
2*],120,[*[*1,0*],[*0,3*]*],[*[*2,3,7*]*]*],[*[*"B",2*],120,[*[*0,
1*],[*1,2*]*],7*],[*[*"B",2*],120,[*[*2,0*],[*1,1*]*],5*],[*[*"B",
2*],120,[*[*7,0*]*],[*[*2,3,5,7*]*]*],[*[*"B",3*],120,[*[*1,0,2*]
*],3*],[*[*"C",4*],120,[*[*0,0,0,3*]*],[*[*2,3*]*]*],[*[*"D",5*],
120,[*[*0,0,1,0,0*]*],[*[*2*]*]*],[*[*"D",8*],120,[*[*0,0,0,0,0,
0,1,0*]*],[*[*2*]*]*],[*[*"A",1*],120,[*[*1*],[*2*],[*3*],[*4*]*],[*
[*2,3*]*]*],[*[*"A",1*],120,[*[*1*],[*2*],[*19*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],120,[*[*1*],[*3*],[*14*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],120,[*[*1*],[*4*],[*11*]*],[*[*2,3,5,7,11
*]*]*],[*[*"A",1*],120,[*[*1*],[*5*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],120,[*[*1*],[*59*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59*]*]*],[*[*"A",1*],120,[*[*1*],[*1*],[*2*],[*9*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],120,[*[*1*],[*1*],[*4*],[*5*]*],[*
[*2,3,5*]*]*],[*[*"A",1*],120,[*[*1*],[*1*],[*29*]*],[*[*2,3,5,7,
11,13,17,19,23,29*]*]*],[*[*"A",1*],120,[*[*1*],[*1*],[*1*],[*2*],
[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],120,[*[*1*],[*1*],[*1*],[*14*]*],
[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],120,[*[*2*],[*3*],[*9*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],120,[*[*2*],[*4*],[*7*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],120,[*[*2*],[*39*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37*]*]*],[*[*"A",1*],120,[*[*3*],[*4*],[*5*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],120,[*[*3*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,
29*]*]*],[*[*"A",1*],120,[*[*4*],[*23*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],120,[*[*5*],[*19*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",1*],120,[*[*7*],[*14*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],120,[*[*9*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1
*],120,[*[*119*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*
"C",60*],120,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",60*],
120,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",119*],120,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",119*],120,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],
121,[*[*0,0,0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",10*],121,[*[*0,0,0,0,0,0,0,0,0,1*],[*0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",10*],121,[*[*1,0,0,0,0,0,0,
0,0,0*],[*1,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",4*],121,[*
[*0,0,1,3*]*],5*],[*[*"A",4*],121,[*[*3,1,0,0*]*],5*],[*[*"B",5
*],121,[*[*0,0,0,0,1*],[*0,0,0,0,1*]*],[*[*2*]*]*],[*[*"C",5*],
121,[*[*0,1,0,0,0*]*],3*],[*[*"A",1*],121,[*[*10*],[*10*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],121,[*[*120*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113*]*]*],[*[*"B",60*],121,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",120*],121,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",120*],121,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"C",5*],122,[*[*1,0,0,0,0*]*],3*],
[*[*"A",1*],122,[*[*1*],[*60*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59*]*]*],[*[*"A",1*],122,[*[*121*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113*]*]*],[*[*"C",61*],122,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",61*],122,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",121*],122,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",121*],122,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],123,[*[*4,8*]*],13
*],[*[*"A",2*],123,[*[*8,4*]*],13*],[*[*"A",1*],123,[*[*2*],[*40*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],123,[*
[*122*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"B",61*],
123,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",122*],123,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
122*],123,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",3*],124,[*[*2,0,3*]*],7*],[*[*"A",3*],124,[*[*3,0,
2*]*],7*],[*[*"A",1*],124,[*[*1*],[*61*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],124,[*[*1*],
[*1*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],
124,[*[*3*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",
1*],124,[*[*123*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*
"C",62*],124,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",62
*],124,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",123*],124,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",123*],124,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",2*],125,[*[*4,4*]*],[*[*2,3,7*]*]*],[*[*"A",
4*],125,[*[*0,0,0,1*],[*1,0,0,0*],[*1,0,0,0*]*],[*[**]*]*],[*[*
"A",4*],125,[*[*0,0,0,1*],[*0,0,0,1*],[*1,0,0,0*]*],[*[**]*]*],
[*[*"A",4*],125,[*[*0,0,0,1*],[*0,0,0,1*],[*0,0,0,1*]*],[*[**]*]
*],[*[*"A",4*],125,[*[*1,0,0,0*],[*1,0,0,0*],[*1,0,0,0*]*],[*[*
*]*]*],[*[*"B",2*],125,[*[*0,1*],[*0,3*]*],7*],[*[*"B",2*],125,[*[*
0,1*],[*2,1*]*],3*],[*[*"B",2*],125,[*[*0,1*],[*0,1*],[*0,1*]*],
[*[*2*]*]*],[*[*"G",2*],125,[*[*1,2*]*],11*],[*[*"A",1*],125,[*[*4
*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],125,[*[*
4*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],125,[*[*124*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113*]*]*],[*[*"B",62*],125,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",124*],125,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",124*],125,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",2*],126,[*[*0,1*],[*0,2*],[*1,1*]*],3*],[*[*"A",2*],
126,[*[*0,1*],[*2,0*],[*1,1*]*],3*],[*[*"A",2*],126,[*[*0,1*],[*
2,3*]*],[*[*2,3,5*]*]*],[*[*"A",2*],126,[*[*0,1*],[*3,2*]*],[*[*2,
3,5*]*]*],[*[*"A",2*],126,[*[*1,0*],[*0,2*],[*1,1*]*],3*],[*[*"A",
2*],126,[*[*1,0*],[*2,0*],[*1,1*]*],3*],[*[*"A",2*],126,[*[*1,0
*],[*2,3*]*],[*[*2,3,5*]*]*],[*[*"A",2*],126,[*[*1,0*],[*3,2*]*],[*
[*2,3,5*]*]*],[*[*"A",2*],126,[*[*0,2*],[*0,5*]*],[*[*2,3,5*]*]*],
[*[*"A",2*],126,[*[*0,2*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
126,[*[*2,0*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],126,[*[*2,0*],
[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],126,[*[*5,7*]*],13*],[*[*"A",
2*],126,[*[*7,5*]*],13*],[*[*"A",3*],126,[*[*0,2,2*]*],[*[*2,5*]*]
*],[*[*"A",3*],126,[*[*2,2,0*]*],[*[*2,5*]*]*],[*[*"A",4*],126,[*[*
0,0,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",4*],126,[*[*0,1,0,2*]*],[*[*
2,5*]*]*],[*[*"A",4*],126,[*[*2,0,1,0*]*],[*[*2,5*]*]*],[*[*"A",4
*],126,[*[*5,0,0,0*]*],[*[*2,3,5*]*]*],[*[*"A",5*],126,[*[*0,0,0,
0,1*],[*0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",5*],126,[*[*0,0,0,0,
1*],[*2,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",5*],126,[*[*1,0,0,0,0
*],[*0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",5*],126,[*[*1,0,0,0,0*],
[*2,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",5*],126,[*[*0,0,0,0,4*]*],[*
[*2,3*]*]*],[*[*"A",5*],126,[*[*0,0,1,1,0*]*],3*],[*[*"A",5*],
126,[*[*0,1,1,0,0*]*],3*],[*[*"A",5*],126,[*[*4,0,0,0,0*]*],[*[*
2,3*]*]*],[*[*"A",8*],126,[*[*0,0,0,0,1,0,0,0*]*],[*[**]*]*],[*[*
"A",8*],126,[*[*0,0,0,1,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],126,[*
[*0,6*]*],11*],[*[*"B",4*],126,[*[*2,0,0,0*]*],[*[*2*]*]*],[*[*"C",
3*],126,[*[*0,0,1*],[*0,0,2*]*],[*[*2*]*]*],[*[*"C",3*],126,[*[*0,
0,4*]*],[*[*2,3*]*]*],[*[*"C",3*],126,[*[*1,1,0*]*],[*[*2,5*]*]*],
[*[*"D",5*],126,[*[*0,2,0,0,0*]*],[*[*2*]*]*],[*[*"D",5*],126,[*[*
2,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",1*],126,[*[*1*],[*2*],[*20*]*],[*
[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],126,[*[*1*],[*2*],[*2*],
[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],126,[*[*1*],[*6*],[*8*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],126,[*[*1*],[*62*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],126,[*[*2
*],[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],126,[*[*2*],[*41*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],126,[*[*
2*],[*2*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],126,[*[*5
*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],126,[*[*6*],
[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],126,[*[*8*],[*13*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],126,[*[*125*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113*]*]*],[*[*"C",63*],126,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",63*],126,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",125*],126,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",125*],126,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",2*],127,[*[*6,6*]*],13*],[*[*"A",1*],127,[*[*126*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113*]*]*],[*[*"B",63*],127,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",126*],127,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",126*],127,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],128,[*[*0,0,1*],[*1,0,2*]*],5
*],[*[*"A",3*],128,[*[*0,0,1*],[*2,0,1*]*],5*],[*[*"A",3*],128,[*
[*1,0,0*],[*1,0,2*]*],5*],[*[*"A",3*],128,[*[*1,0,0*],[*2,0,1*]
*],5*],[*[*"B",4*],128,[*[*0,0,0,1*],[*1,0,0,0*]*],2*],[*[*"B",4
*],128,[*[*1,0,0,1*]*],[*[*3*]*]*],[*[*"B",7*],128,[*[*1,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"C",4*],128,[*[*0,0,0,1*],[*1,0,0,0*]*],2
*],[*[*"C",4*],128,[*[*1,0,0,1*]*],2*],[*[*"C",7*],128,[*[*1,0,0,
0,0,0,0*]*],2*],[*[*"D",5*],128,[*[*0,1,0,0,1*]*],5*],[*[*"D",5
*],128,[*[*1,0,0,0,1*]*],5*],[*[*"D",8*],128,[*[*0,1,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"D",8*],128,[*[*1,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",1*],128,[*[*1*],[*3*],[*15*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],128,[*[*1*],[*3*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",
1*],128,[*[*1*],[*7*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],128,[*
[*1*],[*63*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*3*],[*7*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*31*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*1*],[*3*],
[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*1*],[*15*]*],
[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*1*],[*1*],
[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*1*],[*1
*],[*1*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],128,[*[*1*],[*1*],[*1*],[*
1*],[*1*],[*1*],[*1*]*],[*[**]*]*],[*[*"A",1*],128,[*[*3*],[*31*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],128,[*[*3*],[*
3*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],128,[*[*7*],[*15*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",1*],128,[*[*127*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127*]*]*],[*[*"C",64*],128,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",64*],128,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",127*],128,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",127*],
128,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",1*],129,[*[*2*],[*42*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],129,[*[*128*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127*]*]*],[*[*"B",64*],129,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",128*],129,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",128*],129,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",2*],130,[*[*2,0*],[*0,
2*]*],5*],[*[*"A",1*],130,[*[*1*],[*4*],[*12*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],130,[*[*1*],[*64*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],130,[*[*4*],[*25*]
*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],130,[*[*9*],[*12
*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],130,[*[*129*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127*]*]*],[*[*"C",65*],130,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",65*],130,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",129*],130,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",129*],130,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],131,[*[*130*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127*]*]*],[*[*"B",65*],131,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",130*],
131,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",130*],131,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2
*],132,[*[*2,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],132,[*[*7,2*]*],[*
[*2,3,5,7*]*]*],[*[*"B",3*],132,[*[*0,2,0*]*],3*],[*[*"C",5*],
132,[*[*1,0,0,0,0*]*],[*[*2,3*]*]*],[*[*"E",7*],132,[*[*1,0,0,0,
0,0,0*]*],2*],[*[*"A",1*],132,[*[*1*],[*2*],[*21*]*],[*[*2,3,5,7,
11,13,17,19*]*]*],[*[*"A",1*],132,[*[*1*],[*5*],[*10*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],132,[*[*1*],[*65*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],132,[*[*1*],
[*1*],[*2*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],132,[*[*1*],[*1
*],[*32*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],
132,[*[*2*],[*3*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],132,[*[*2
*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*
"A",1*],132,[*[*3*],[*32*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]
*]*],[*[*"A",1*],132,[*[*5*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",1*],132,[*[*10*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1
*],132,[*[*131*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131
*]*]*],[*[*"C",66*],132,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",66*],132,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"A",131*],132,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",131*],132,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",6*],133,[*[*0,1,0,0,0,1*]*],2*],[*[*
"A",6*],133,[*[*0,1,0,0,0,1*]*],3*],[*[*"A",6*],133,[*[*1,0,0,
0,1,0*]*],2*],[*[*"A",6*],133,[*[*1,0,0,0,1,0*]*],3*],[*[*"E",7
*],133,[*[*1,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",1*],133,[*[*6*],
[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],133,[*[*132*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131*]*]*],[*[*"B",66*],
133,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",
132*],133,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",132*],133,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],134,[*[*1*],[*66*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],134,[*[*133*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131*]*]*],[*[*"C",67
*],134,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",67*],134,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"A",133*],134,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",133*],134,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],135,[*[*0,1*],[*1,0*],[*0,4*]*],
[*[*2,3*]*]*],[*[*"A",2*],135,[*[*0,1*],[*1,0*],[*1,2*]*],[*[*2*]*]
*],[*[*"A",2*],135,[*[*0,1*],[*1,0*],[*2,1*]*],[*[*2*]*]*],[*[*"A",
2*],135,[*[*0,1*],[*1,0*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],135,
[*[*0,1*],[*0,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],135,[*[*0,1*],[*
8,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],135,[*[*0,1*],[*0,1*],[*0,
4*]*],[*[*2,3*]*]*],[*[*"A",2*],135,[*[*0,1*],[*0,1*],[*1,2*]*],[*
[*2*]*]*],[*[*"A",2*],135,[*[*0,1*],[*0,1*],[*2,1*]*],[*[*2*]*]*],[*
[*"A",2*],135,[*[*0,1*],[*0,1*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2
*],135,[*[*1,0*],[*0,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],135,[*[*
1,0*],[*8,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],135,[*[*1,0*],[*1,
0*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],135,[*[*1,0*],[*1,0*],[*1,
2*]*],[*[*2*]*]*],[*[*"A",2*],135,[*[*1,0*],[*1,0*],[*2,1*]*],[*[*2
*]*]*],[*[*"A",2*],135,[*[*1,0*],[*1,0*],[*4,0*]*],[*[*2,3*]*]*],[*
[*"A",4*],135,[*[*1,0,1,1*]*],3*],[*[*"A",4*],135,[*[*1,1,0,1*]
*],3*],[*[*"D",8*],135,[*[*0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*
"A",1*],135,[*[*2*],[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
135,[*[*2*],[*44*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43
*]*]*],[*[*"A",1*],135,[*[*2*],[*2*],[*14*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],135,[*[*2*],[*2*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",
1*],135,[*[*4*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],135,[*[*8*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
135,[*[*134*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131*]*]
*],[*[*"B",67*],135,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",134*],135,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",134*],
135,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",15*],136,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",15*],136,[*[*2,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",16*],136,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",16
*],136,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",2*],136,[*[*0,15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",2*],136,
[*[*15,0*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"B",8*],136,[*[*0,0,0,
0,0,0,1,0*]*],[*[*2*]*]*],[*[*"C",8*],136,[*[*0,0,0,0,0,0,0,2
*]*],[*[*2*]*]*],[*[*"A",1*],136,[*[*1*],[*3*],[*16*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],136,[*[*1*],[*67*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],136,[*
[*1*],[*1*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],136,[*[*1*],[*1*],[*1*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],
[*[*"A",1*],136,[*[*3*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31*]*]*],[*[*"A",1*],136,[*[*7*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],
[*[*"A",1*],136,[*[*135*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131*]*]*],[*[*"C",68*],136,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"D",68*],136,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",135*],136,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",135*],136,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*
"A",1*],137,[*[*136*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131*]*]*],[*[*"B",68*],137,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",136*],137,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",136*],137,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1
*],138,[*[*1*],[*2*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],138,[*[*1*],[*68*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],138,[*[*2*],[*45*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],138,
[*[*5*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],138,[*
[*137*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137*]*]*],
[*[*"C",69*],138,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"D",69*],138,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",137*],138,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",137*],138,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
1*],139,[*[*138*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137*]*]*],[*[*"B",69*],139,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",138*],139,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",138*],139,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",3*],140,[*[*0,0,1*],[*0,0,4*]*],[*[*2,3*]*]*],[*[*"A",3*],
140,[*[*0,0,1*],[*4,0,0*]*],[*[*2,3*]*]*],[*[*"A",3*],140,[*[*1,
0,0*],[*0,0,4*]*],[*[*2,3*]*]*],[*[*"A",3*],140,[*[*1,0,0*],[*4,
0,0*]*],[*[*2,3*]*]*],[*[*"A",3*],140,[*[*0,1,4*]*],[*[*2,3*]*]*],
[*[*"A",3*],140,[*[*0,3,1*]*],[*[*2,3,5*]*]*],[*[*"A",3*],140,[*[*
1,1,2*]*],[*[*2,3*]*]*],[*[*"A",3*],140,[*[*1,3,0*]*],[*[*2,3,5*]
*]*],[*[*"A",3*],140,[*[*2,1,1*]*],[*[*2,3*]*]*],[*[*"A",3*],140,[*
[*4,1,0*]*],[*[*2,3*]*]*],[*[*"A",6*],140,[*[*0,1,0,0,0,1*]*],[*
[*2,3*]*]*],[*[*"A",6*],140,[*[*1,0,0,0,1,0*]*],[*[*2,3*]*]*],[*[*
"B",2*],140,[*[*1,0*],[*2,1*]*],[*[*2,3*]*]*],[*[*"B",2*],140,[*[*
1,0*],[*4,0*]*],[*[*2,3*]*]*],[*[*"B",2*],140,[*[*2,0*],[*0,2*]*],
[*[*2,5*]*]*],[*[*"B",2*],140,[*[*0,6*]*],[*[*2,3,5,11,13*]*]*],[*
[*"B",2*],140,[*[*1,4*]*],[*[*2,3,11*]*]*],[*[*"B",2*],140,[*[*3,2
*]*],[*[*2,3*]*]*],[*[*"A",1*],140,[*[*1*],[*4*],[*13*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],140,[*[*1*],[*6*],[*9*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],140,[*[*1*],[*69*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],140,[*[*1*],[*1
*],[*4*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],140,[*[*1*],[*1*],[*34
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],140,[*[*
3*],[*4*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],140,[*[*3*],[*34*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],140,[*[*4*],
[*27*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],140,[*[*6
*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],140,[*[*9*],
[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],140,[*[*139*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*"C",70
*],140,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"D",70*],140,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",139*],140,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",139*],140,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",5*],141,[*[*0,0,2,0,0*]*],3*],[*[*"B",3*],141,[*[*0,2,0*]
*],5*],[*[*"A",1*],141,[*[*2*],[*46*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43*]*]*],[*[*"A",1*],141,[*[*140*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*"B",70*],141,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",140*],141,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",140
*],141,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",11*],
142,[*[*1,0,0,0,0,0,0,0,0,0,1*]*],2*],[*[*"A",11*],142,[*[*1,
0,0,0,0,0,0,0,0,0,1*]*],3*],[*[*"A",1*],142,[*[*1*],[*70*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67*]*]
*],[*[*"A",1*],142,[*[*141*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139*]*]*],[*[*"C",71*],142,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",71*],142,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",141
*],142,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",141*],
142,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",11*],
143,[*[*1,0,0,0,0,0,0,0,0,0,1*]*],[*[*2,3*]*]*],[*[*"A",2*],
143,[*[*1,10*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],143,[*[*10,1*]*],[*
[*2,3,5,7*]*]*],[*[*"A",1*],143,[*[*10*],[*12*]*],[*[*2,3,5,7,11
*]*]*],[*[*"A",1*],143,[*[*142*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139*]*]*],[*[*"B",71*],143,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",142*],143,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",142*],143,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",11*],144,[*[*
0,0,0,0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0,0,0,0*]*],[*[*
*]*]*],[*[*"A",11*],144,[*[*0,0,0,0,0,0,0,0,0,0,1*],[*0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",11*],144,[*[*1,0,0,0,0,
0,0,0,0,0,0*],[*1,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
2*],144,[*[*0,1*],[*0,2*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],144,
[*[*0,1*],[*2,0*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],144,[*[*0,1
*],[*1,5*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],144,[*[*0,1*],[*5,1*]
*],[*[*2,3,5,7*]*]*],[*[*"A",2*],144,[*[*1,0*],[*0,2*],[*1,1*]*],
[*[*2,3*]*]*],[*[*"A",2*],144,[*[*1,0*],[*2,0*],[*1,1*]*],[*[*2,3
*]*]*],[*[*"A",2*],144,[*[*1,0*],[*1,5*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],144,[*[*1,0*],[*5,1*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],
144,[*[*0,2*],[*1,3*]*],[*[*2,3,5*]*]*],[*[*"A",2*],144,[*[*0,2*],
[*3,1*]*],[*[*2,3,5*]*]*],[*[*"A",2*],144,[*[*2,0*],[*1,3*]*],[*[*
2,3,5*]*]*],[*[*"A",2*],144,[*[*2,0*],[*3,1*]*],[*[*2,3,5*]*]*],[*
[*"A",2*],144,[*[*1,1*],[*1,3*]*],5*],[*[*"A",2*],144,[*[*1,1*],[*
3,1*]*],5*],[*[*"A",3*],144,[*[*0,0,1*],[*0,1,0*],[*0,1,0*]*],[*
[**]*]*],[*[*"A",3*],144,[*[*0,0,1*],[*1,0,2*]*],[*[*2,5*]*]*],[*[*
"A",3*],144,[*[*0,0,1*],[*2,0,1*]*],[*[*2,5*]*]*],[*[*"A",3*],
144,[*[*1,0,0*],[*0,1,0*],[*0,1,0*]*],[*[**]*]*],[*[*"A",3*],144,
[*[*1,0,0*],[*1,0,2*]*],[*[*2,5*]*]*],[*[*"A",3*],144,[*[*1,0,0
*],[*2,0,1*]*],[*[*2,5*]*]*],[*[*"B",2*],144,[*[*1,1*],[*1,1*]*],5
*],[*[*"B",2*],144,[*[*1,7*]*],17*],[*[*"B",4*],144,[*[*0,0,0,1*],
[*1,0,0,0*]*],[*[*2*]*]*],[*[*"B",6*],144,[*[*0,0,0,0,0,1*],[*0,
0,0,0,0,1*]*],2*],[*[*"C",6*],144,[*[*0,0,0,0,0,1*],[*0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",5*],144,[*[*0,1,0,0,1*]*],[*[*5*]*]*],
[*[*"D",5*],144,[*[*1,0,0,0,1*]*],[*[*5*]*]*],[*[*"D",6*],144,[*[*
0,0,0,0,0,1*],[*0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",1*],144,[*[*
1*],[*2*],[*3*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],144,[*[*1*],[*2
*],[*23*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],144,[*[*
1*],[*2*],[*2*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],144,[*[*1*],
[*3*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],144,[*[*1*],
[*5*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],144,[*[*1*],[*7*],
[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],144,[*[*1*],[*71*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*
[*"A",1*],144,[*[*1*],[*1*],[*2*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*
[*"A",1*],144,[*[*1*],[*1*],[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*
"A",1*],144,[*[*1*],[*1*],[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],144,[*[*1*],[*1*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],
144,[*[*1*],[*1*],[*35*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]
*],[*[*"A",1*],144,[*[*1*],[*1*],[*1*],[*2*],[*5*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],144,[*[*1*],[*1*],[*1*],[*17*]*],[*[*2,3,5,7,11,13,
17*]*]*],[*[*"A",1*],144,[*[*1*],[*1*],[*1*],[*1*],[*2*],[*2*]*],[*[*
2*]*]*],[*[*"A",1*],144,[*[*1*],[*1*],[*1*],[*1*],[*8*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],144,[*[*2*],[*3*],[*11*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],144,[*[*2*],[*5*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],144,[*[*2*],[*47*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47*]*]*],[*[*"A",1*],144,[*[*2*],[*2*],[*3*],[*3*]*],[*[*
2,3*]*]*],[*[*"A",1*],144,[*[*2*],[*2*],[*15*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"A",1*],144,[*[*3*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],144,[*[*3*],[*35*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]
*]*],[*[*"A",1*],144,[*[*3*],[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],144,[*[*5*],[*23*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],144,[*[*7*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",
1*],144,[*[*8*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],144,
[*[*11*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],144,[*[*143*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*
"C",72*],144,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"D",72*],144,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",143*],144,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",143*],144,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],145,[*[*1,0,0,3*]
*],7*],[*[*"A",4*],145,[*[*3,0,0,1*]*],7*],[*[*"B",2*],145,[*[*0,
8*]*],17*],[*[*"A",1*],145,[*[*4*],[*28*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],145,[*[*144*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139*]*]*],[*[*"B",72*],145,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",144
*],145,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
144*],145,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",1*],146,[*[*1*],[*72*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71*]*]*],[*[*"A",1*],146,[*[*145*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*
"C",73*],146,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"D",73*],146,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",145*],146,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",145*],146,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],
147,[*[*0,1*],[*1,1*],[*1,1*]*],3*],[*[*"A",2*],147,[*[*1,0*],[*
1,1*],[*1,1*]*],3*],[*[*"A",6*],147,[*[*0,0,0,0,0,1*],[*0,0,0,
0,1,0*]*],[*[**]*]*],[*[*"A",6*],147,[*[*0,0,0,0,0,1*],[*0,1,0,
0,0,0*]*],[*[**]*]*],[*[*"A",6*],147,[*[*1,0,0,0,0,0*],[*0,0,0,
0,1,0*]*],[*[**]*]*],[*[*"A",6*],147,[*[*1,0,0,0,0,0*],[*0,1,0,
0,0,0*]*],[*[**]*]*],[*[*"B",3*],147,[*[*0,0,1*],[*0,1,0*]*],[*[*2
*]*]*],[*[*"B",4*],147,[*[*0,0,0,3*]*],11*],[*[*"B",4*],147,[*[*0,
0,1,1*]*],3*],[*[*"A",1*],147,[*[*2*],[*6*],[*6*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],147,[*[*2*],[*48*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47*]*]*],[*[*"A",1*],147,[*[*6*],[*20*]*],[*[*2,3,
5,7,11,13,17,19*]*]*],[*[*"A",1*],147,[*[*146*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*"B",73*],147,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]
*]*],[*[*"A",146*],147,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",146*],147,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"G",2*],148,[*[*3,0*]*],11*],[*[*"A",1
*],148,[*[*1*],[*73*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],148,[*[*1*],[*1*],[*36
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],148,[*[*
3*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],
148,[*[*147*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139*]*]*],[*[*"C",74*],148,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",74*],148,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
147*],148,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",147*],148,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",3*],149,[*[*0,2,4*]*],7*],[*[*"A",3*],149,
[*[*4,2,0*]*],7*],[*[*"B",2*],149,[*[*4,2*]*],7*],[*[*"A",1*],149,
[*[*148*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139*]*]*],[*[*"B",74*],149,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",148*],149,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",148*],149,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",2*],150,[*[*0,3*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],150,[*
[*0,3*],[*1,2*]*],[*[*2,3*]*]*],[*[*"A",2*],150,[*[*0,3*],[*2,1*]
*],[*[*2,3*]*]*],[*[*"A",2*],150,[*[*0,3*],[*4,0*]*],[*[*2,3*]*]*],
[*[*"A",2*],150,[*[*3,0*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],150,
[*[*3,0*],[*1,2*]*],[*[*2,3*]*]*],[*[*"A",2*],150,[*[*3,0*],[*2,1
*]*],[*[*2,3*]*]*],[*[*"A",2*],150,[*[*3,0*],[*4,0*]*],[*[*2,3*]*]
*],[*[*"A",3*],150,[*[*0,0,2*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",3*],
150,[*[*2,0,0*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",4*],150,[*[*0,0,
0,1*],[*0,0,1,1*]*],3*],[*[*"A",4*],150,[*[*0,0,0,1*],[*1,1,0,
0*]*],3*],[*[*"A",4*],150,[*[*1,0,0,0*],[*0,0,1,1*]*],3*],[*[*
"A",4*],150,[*[*1,0,0,0*],[*1,1,0,0*]*],3*],[*[*"A",4*],150,[*[*
0,0,1,0*],[*0,0,0,2*]*],[*[*2*]*]*],[*[*"A",4*],150,[*[*0,0,1,0
*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",4*],150,[*[*0,1,0,0*],[*0,0,
0,2*]*],[*[*2*]*]*],[*[*"A",4*],150,[*[*0,1,0,0*],[*2,0,0,0*]*],
[*[*2*]*]*],[*[*"B",2*],150,[*[*0,1*],[*0,3*]*],[*[*2,3,7*]*]*],[*[*
"A",1*],150,[*[*1*],[*2*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],
150,[*[*1*],[*2*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],150,[*[*1*],[*4*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],150,[*[*1*],[*74*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],150,[*[*2*],[*4
*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],150,[*[*2*],[*49*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],150,
[*[*4*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],
150,[*[*4*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],150,[*[*5*],[*
24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],150,[*[*9*],
[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],150,[*[*149*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131,137,139,149*]*]*],[*[*
"C",75*],150,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",75*],150,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",149*],150,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",149*],150,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"B",8*],151,[*[*0,0,0,0,0,0,0,2*]*],17*],[*[*
"C",9*],151,[*[*0,0,0,0,0,0,0,1,0*]*],3*],[*[*"A",1*],151,[*[*
150*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149*]*]*],[*[*"B",75*],151,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",150*],151,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",150
*],151,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",2*],152,[*[*1,1*],[*2,2*]*],5*],[*[*"B",8*],
152,[*[*0,0,0,0,0,0,0,2*]*],[*[*2,17*]*]*],[*[*"B",9*],152,[*[*
0,0,0,0,0,0,0,1,0*]*],2*],[*[*"C",9*],152,[*[*0,0,0,0,0,0,
0,1,0*]*],[*[*3*]*]*],[*[*"D",4*],152,[*[*0,0,1,1*]*],7*],[*[*"D",
4*],152,[*[*0,1,1,0*]*],7*],[*[*"D",4*],152,[*[*1,0,1,0*]*],7*],
[*[*"D",9*],152,[*[*0,0,0,0,0,0,0,1,0*]*],2*],[*[*"A",1*],152,
[*[*1*],[*3*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],152,
[*[*1*],[*75*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73*]*]*],[*[*"A",1*],152,[*[*1*],[*1*],[*37*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],152,[*[*1*],
[*1*],[*1*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],152,[*
[*3*],[*37*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*
"A",1*],152,[*[*7*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1
*],152,[*[*151*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151*]*]*],[*[*"C",76*],152,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",76*],
152,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",151*],152,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",151*],152,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
16*],153,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]
*],[*[*"A",16*],153,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[*2*]*]*],[*[*"A",17*],153,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1,0*]*],[*[**]*]*],[*[*"A",17*],153,[*[*0,1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],153,[*[*0,16*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",2*],153,[*[*16,0*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"D",9*],153,[*[*0,0,0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*
[*"A",1*],153,[*[*2*],[*50*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47*]*]*],[*[*"A",1*],153,[*[*2*],[*2*],[*16*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",1*],153,[*[*8*],[*16*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],153,[*[*152*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139,149,151*]*]*],[*[*"B",76*],153,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",152*],153,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",152*],153,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2
*],154,[*[*3,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],154,[*[*6,3*]*],[*[*
2,3,5*]*]*],[*[*"A",5*],154,[*[*0,1,0,1,0*]*],2*],[*[*"B",2*],
154,[*[*2,3*]*],[*[*2,3,5*]*]*],[*[*"A",1*],154,[*[*1*],[*6*],[*10
*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],154,[*[*1*],[*76*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]*]*],[*
[*"A",1*],154,[*[*6*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],154,[*[*10*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
154,[*[*153*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151*]*]*],[*[*"C",77*],154,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",77*],
154,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",153*],154,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",153*],154,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"B",3*],155,[*[*0,0,4*]*],11*],[*[*"G",2*],155,[*[*0,4*]
*],11*],[*[*"A",1*],155,[*[*4*],[*30*]*],[*[*2,3,5,7,11,13,17,19,
23,29*]*]*],[*[*"A",1*],155,[*[*154*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139,149,151*]*]*],[*[*"B",77*],155,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",154*],155,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",154*],155,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",3*],156,[*[*1,2,1*]*],3*],[*[*"A",8*],156,[*[*0,0,0,
0,0,0,1,1*]*],3*],[*[*"A",8*],156,[*[*1,1,0,0,0,0,0,0*]*],3
*],[*[*"B",2*],156,[*[*1,1*],[*0,2*]*],5*],[*[*"B",4*],156,[*[*0,
0,0,3*]*],[*[*2,3,11*]*]*],[*[*"A",1*],156,[*[*1*],[*2*],[*25*]*],
[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],156,[*[*1*],[*5*],[*
12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],156,[*[*1*],[*77*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]
*]*],[*[*"A",1*],156,[*[*1*],[*1*],[*2*],[*12*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],156,[*[*1*],[*1*],[*38*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37*]*]*],[*[*"A",1*],156,[*[*2*],[*3*],[*12*]*],[*[*2,
3,5,7,11*]*]*],[*[*"A",1*],156,[*[*2*],[*51*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],156,[*[*3*],[*38
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],156,
[*[*5*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
156,[*[*11*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],156,[*[*155
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151*]*]*],[*[*"C",78*],156,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",78*],156,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"A",155*],156,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",155*],156,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",1*],157,[*[*156*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151*]*]*],[*[*"B",78*],157,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",156*],157,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",156*],157,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"C",3*],158,[*[*1,0,2*]*],7*],[*[*"A",1*],158,[*[*1
*],[*78*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73*]*]*],[*[*"A",1*],158,[*[*157*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157*]*]*],[*[*"C",
79*],158,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",79*],158,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",157*],158,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",157*],158,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",2*],159,[*[*2,9*]*],11*],[*[*"A",2*],159,[*[*9,2*]*],
11*],[*[*"A",1*],159,[*[*2*],[*52*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47*]*]*],[*[*"A",1*],159,[*[*158*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157*]*]*],[*[*
"B",79*],159,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",158*],159,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",158*],159,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],160,[*
[*0,0,1*],[*1,0,0*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*
0,0,1*],[*1,0,0*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*0,
0,1*],[*0,0,1*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*0,0,
1*],[*0,0,1*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*1,0,0
*],[*1,0,0*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*1,0,0*],
[*1,0,0*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],160,[*[*0,0,2*],[*
0,1,1*]*],3*],[*[*"A",3*],160,[*[*0,0,2*],[*1,1,0*]*],3*],[*[*
"A",3*],160,[*[*2,0,0*],[*0,1,1*]*],3*],[*[*"A",3*],160,[*[*2,0,
0*],[*1,1,0*]*],3*],[*[*"A",3*],160,[*[*2,0,3*]*],[*[*2,3,7*]*]
*],[*[*"A",3*],160,[*[*3,0,2*]*],[*[*2,3,7*]*]*],[*[*"A",4*],160,[*
[*1,0,0,3*]*],[*[*2,3,7*]*]*],[*[*"A",4*],160,[*[*1,0,1,1*]*],2
*],[*[*"A",4*],160,[*[*1,1,0,1*]*],2*],[*[*"A",4*],160,[*[*3,0,0,
1*]*],[*[*2,3,7*]*]*],[*[*"B",2*],160,[*[*1,0*],[*1,2*]*],[*[*2,7
*]*]*],[*[*"B",2*],160,[*[*1,0*],[*1,0*],[*2,0*]*],[*[*2*]*]*],[*[*
"B",2*],160,[*[*2,0*],[*1,1*]*],[*[*2,5*]*]*],[*[*"B",2*],160,[*[*
5,1*]*],[*[*2,3,5,7*]*]*],[*[*"B",4*],160,[*[*0,0,1,1*]*],2*],[*
[*"C",4*],160,[*[*0,0,1,1*]*],[*[*3*]*]*],[*[*"D",4*],160,[*[*0,0,
1,1*]*],[*[*3,7*]*]*],[*[*"D",4*],160,[*[*0,1,1,0*]*],[*[*3,7*]*]
*],[*[*"D",4*],160,[*[*1,0,1,0*]*],[*[*3,7*]*]*],[*[*"D",5*],160,[*
[*0,0,0,0,1*],[*0,1,0,0,0*]*],[*[**]*]*],[*[*"D",5*],160,[*[*0,
0,0,0,1*],[*1,0,0,0,0*]*],[*[**]*]*],[*[*"A",1*],160,[*[*1*],[*3
*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],160,[*[*1*],
[*3*],[*3*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],160,[*[*1*],[*4*],[*15
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],160,[*[*1*],[*7*],[*9*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],160,[*[*1*],[*79*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79*]*]*],
[*[*"A",1*],160,[*[*1*],[*1*],[*3*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],160,[*[*1*],[*1*],[*4*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],160,[*[*1*],[*1*],[*39*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37*]*]*],[*[*"A",1*],160,[*[*1*],[*1*],[*1*],[*3*],[*4*]*],[*[*2,
3*]*]*],[*[*"A",1*],160,[*[*1*],[*1*],[*1*],[*19*]*],[*[*2,3,5,7,
11,13,17,19*]*]*],[*[*"A",1*],160,[*[*1*],[*1*],[*1*],[*1*],[*9*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],160,[*[*1*],[*1*],[*1*],[*1*],[*1
*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],160,[*[*3*],[*4*],[*7*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],160,[*[*3*],[*39*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37*]*]*],[*[*"A",1*],160,[*[*3*],[*3*],[*9*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],160,[*[*4*],[*31*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],160,[*[*7*],[*19*]*],[*[*2,3,
5,7,11,13,17,19*]*]*],[*[*"A",1*],160,[*[*9*],[*15*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],160,[*[*159*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157*]*]*],[*[*"C",80*],
160,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"D",80*],160,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",159*],160,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",159*],160,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",6*],161,[*[*0,0,0,0,2,0*]*],3*],[*[*"A",6
*],161,[*[*0,2,0,0,0,0*]*],3*],[*[*"A",1*],161,[*[*6*],[*22*]*],[*
[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],161,[*[*160*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131,137,139,149,151,157*]*]*],
[*[*"B",80*],161,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",160*],161,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",160*],161,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",2*],162,[*[*0,1*],[*1,0*],[*1,3*]*],5*],[*[*"A",2*],162,[*[*
0,1*],[*1,0*],[*3,1*]*],5*],[*[*"A",2*],162,[*[*0,1*],[*1,0*],[*
1,0*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*0,1*],[*1,0*],[*1,
0*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*0,1*],[*0,1*],[*1,0
*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*0,1*],[*0,1*],[*1,0*],
[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*0,1*],[*0,1*],[*1,3*]*],
5*],[*[*"A",2*],162,[*[*0,1*],[*0,1*],[*3,1*]*],5*],[*[*"A",2*],
162,[*[*0,1*],[*0,1*],[*0,1*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],
162,[*[*0,1*],[*0,1*],[*0,1*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],
162,[*[*1,0*],[*1,0*],[*1,3*]*],5*],[*[*"A",2*],162,[*[*1,0*],[*
1,0*],[*3,1*]*],5*],[*[*"A",2*],162,[*[*1,0*],[*1,0*],[*1,0*],[*
0,2*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*1,0*],[*1,0*],[*1,0*],[*2,
0*]*],[*[*2*]*]*],[*[*"A",2*],162,[*[*0,2*],[*2,2*]*],[*[*2,5*]*]*],
[*[*"A",2*],162,[*[*2,0*],[*2,2*]*],[*[*2,5*]*]*],[*[*"A",2*],162,
[*[*2,8*]*],[*[*2,5,7,11*]*]*],[*[*"A",2*],162,[*[*4,5*]*],7*],[*[*
"A",2*],162,[*[*5,4*]*],7*],[*[*"A",2*],162,[*[*8,2*]*],[*[*2,5,
7,11*]*]*],[*[*"A",1*],162,[*[*1*],[*2*],[*26*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],162,[*[*1*],[*2*],[*2*],[*8*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],162,[*[*1*],[*2*],[*2*],[*2*],[*2*]*],[*
[*2*]*]*],[*[*"A",1*],162,[*[*1*],[*8*],[*8*]*],[*[*2,3,5,7*]*]*],[*
[*"A",1*],162,[*[*1*],[*80*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79*]*]*],[*[*"A",1*],162,[*[*2
*],[*5*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],162,[*[*2*],[*53*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*
"A",1*],162,[*[*2*],[*2*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*
"A",1*],162,[*[*2*],[*2*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1
*],162,[*[*5*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",
1*],162,[*[*8*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],
162,[*[*161*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157*]*]*],[*[*"C",81*],162,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"D",81*],162,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",161*],162,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",161*],162,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",1*],163,[*[*162*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157*]*]*],[*[*"B",81*],163,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",162*],163,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",162*],163,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",2*],
164,[*[*3,4*]*],13*],[*[*"B",2*],164,[*[*5,2*]*],11*],[*[*"B",5*],
164,[*[*0,1,0,0,0*]*],2*],[*[*"C",5*],164,[*[*0,1,0,0,0*]*],2
*],[*[*"D",5*],164,[*[*1,1,0,0,0*]*],2*],[*[*"A",1*],164,[*[*1*],
[*81*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79*]*]*],[*[*"A",1*],164,[*[*1*],[*1*],[*40*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],164,[*[*3*],[*
40*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],
164,[*[*163*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163*]*]*],[*[*"C",82*],164,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",82*],164,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",163*],
164,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",163*],164,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],165,[*[*0,0,0,0,0,0,0,1,
0,0*]*],[*[**]*]*],[*[*"A",10*],165,[*[*0,0,1,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",2*],165,[*[*0,1*],[*0,9*]*],[*[*2,3,5,7*]*]*],[*
[*"A",2*],165,[*[*0,1*],[*9,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],
165,[*[*1,0*],[*0,9*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],165,[*[*1,0
*],[*9,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],165,[*[*4,5*]*],[*[*2,
3,5,7*]*]*],[*[*"A",2*],165,[*[*5,4*]*],[*[*2,3,5,7*]*]*],[*[*"A",
3*],165,[*[*0,0,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",3*],165,[*[*8,0,
0*]*],[*[*2,3,5,7*]*]*],[*[*"A",4*],165,[*[*0,1,1,1*]*],3*],[*[*
"A",4*],165,[*[*1,1,1,0*]*],3*],[*[*"A",8*],165,[*[*0,0,0,0,0,
0,0,3*]*],[*[*2,3*]*]*],[*[*"A",8*],165,[*[*3,0,0,0,0,0,0,0*]
*],[*[*2,3*]*]*],[*[*"B",2*],165,[*[*8,0*]*],[*[*2,3,5,7*]*]*],[*[*
"B",5*],165,[*[*0,0,1,0,0*]*],[*[*2*]*]*],[*[*"C",5*],165,[*[*0,
1,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",1*],165,[*[*2*],[*4*],[*10*]*],[*
[*2,3,5,7*]*]*],[*[*"A",1*],165,[*[*2*],[*54*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],165,[*[*4*],
[*32*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],165,
[*[*10*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],165,[*[*164*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163*]*]*],[*[*"B",82*],165,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",
164*],165,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",164*],165,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],166,[*[*1*],[*82*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79*]*]*],[*[*"A",1*],166,[*[*165*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163*]*]*],[*[*"C",83
*],166,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",83*],166,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",165*],166,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",165*],166,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",12*],167,[*[*1,0,0,0,0,0,0,0,0,0,0,1*]*],13*],[*[*"A",1
*],167,[*[*166*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163*]*]*],[*[*"B",83*],167,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[*2*]*]*],[*[*"A",166*],167,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",166*],167,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",12*],168,[*[*1,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*13*]*]*],
[*[*"A",2*],168,[*[*0,2*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
168,[*[*0,2*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],168,[*[*2,0*],
[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],168,[*[*2,0*],[*6,0*]*],[*[*
2,3,5*]*]*],[*[*"A",2*],168,[*[*1,1*],[*0,5*]*],[*[*2,3,5*]*]*],[*
[*"A",2*],168,[*[*1,1*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],168,
[*[*1,11*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",2*],168,[*[*1,15*]*],
17*],[*[*"A",2*],168,[*[*11,1*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",
2*],168,[*[*15,1*]*],17*],[*[*"A",7*],168,[*[*0,0,0,0,0,1,1*]*],
[*[*3*]*]*],[*[*"A",7*],168,[*[*1,1,0,0,0,0,0*]*],[*[*3*]*]*],[*[*
"B",3*],168,[*[*1,0,0*],[*0,1,0*]*],[*[*2*]*]*],[*[*"B",3*],168,[*
[*0,2,0*]*],[*[*2,3,5*]*]*],[*[*"B",3*],168,[*[*1,0,2*]*],[*[*2,3
*]*]*],[*[*"B",3*],168,[*[*2,0,1*]*],5*],[*[*"D",4*],168,[*[*0,1,
0,2*]*],5*],[*[*"D",4*],168,[*[*0,2,0,1*]*],5*],[*[*"D",4*],168,
[*[*1,0,0,2*]*],5*],[*[*"D",4*],168,[*[*1,2,0,0*]*],5*],[*[*"D",
4*],168,[*[*2,0,0,1*]*],5*],[*[*"D",4*],168,[*[*2,1,0,0*]*],5*],
[*[*"A",1*],168,[*[*1*],[*2*],[*3*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",
1*],168,[*[*1*],[*2*],[*27*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],
[*[*"A",1*],168,[*[*1*],[*3*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]
*]*],[*[*"A",1*],168,[*[*1*],[*5*],[*13*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],168,[*[*1*],[*6*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",1*],168,[*[*1*],[*83*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],168,[*[*1
*],[*1*],[*2*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],168,[*
[*1*],[*1*],[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],168,[*[*1*],[*
1*],[*41*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*
"A",1*],168,[*[*1*],[*1*],[*1*],[*2*],[*6*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],168,[*[*1*],[*1*],[*1*],[*20*]*],[*[*2,3,5,7,11,13,17,19
*]*]*],[*[*"A",1*],168,[*[*2*],[*3*],[*13*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],168,[*[*2*],[*6*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],168,[*[*2*],[*55*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53*]*]*],[*[*"A",1*],168,[*[*3*],[*5*],[*6*]*],[*[*2,3,5
*]*]*],[*[*"A",1*],168,[*[*3*],[*41*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41*]*]*],[*[*"A",1*],168,[*[*5*],[*27*]*],[*[*2,3,5,
7,11,13,17,19,23*]*]*],[*[*"A",1*],168,[*[*6*],[*23*]*],[*[*2,3,
5,7,11,13,17,19,23*]*]*],[*[*"A",1*],168,[*[*7*],[*20*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],168,[*[*11*],[*13*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",1*],168,[*[*167*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137,139,149,151,157,163,167*]*]*],[*
[*"C",84*],168,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",84*],168,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",167*],168,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",167*],168,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",12*],169,[*[*0,0,0,0,0,0,0,0,0,0,
0,1*],[*1,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",12*],
169,[*[*0,0,0,0,0,0,0,0,0,0,0,1*],[*0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",12*],169,[*[*1,0,0,0,0,0,0,0,0,0,0,
0*],[*1,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"B",2*],169,[*
[*0,2*],[*0,2*]*],5*],[*[*"B",6*],169,[*[*0,0,0,0,0,1*],[*0,0,
0,0,0,1*]*],[*[*2*]*]*],[*[*"C",3*],169,[*[*0,1,0*],[*0,1,0*]*],
3*],[*[*"D",9*],169,[*[*0,0,0,0,0,0,0,0,2*]*],3*],[*[*"A",1*],
169,[*[*12*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],169,[*[*168
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167*]*]*],[*[*"B",84*],169,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",168*],169,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",168*],169,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",4*],170,[*[*0,2,0,1*]*],5*],[*[*"A",4*],170,[*[*1,
0,2,0*]*],5*],[*[*"D",9*],170,[*[*0,0,0,0,0,0,0,0,2*]*],[*[*2,
3*]*]*],[*[*"A",1*],170,[*[*1*],[*4*],[*16*]*],[*[*2,3,5,7,11,13*]
*]*],[*[*"A",1*],170,[*[*1*],[*84*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],
170,[*[*4*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],170,[*[*9*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
170,[*[*169*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167*]*]*],[*[*"C",85*],170,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",85*],170,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",169*],170,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",169*],170,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",17*],171,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,2*]*],[*[*2*]*]*],[*[*"A",17*],171,[*[*2,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",18*],171,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",18*],171,[*[*
0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
2*],171,[*[*0,1*],[*1,0*],[*2,2*]*],5*],[*[*"A",2*],171,[*[*0,1
*],[*0,1*],[*2,2*]*],5*],[*[*"A",2*],171,[*[*1,0*],[*1,0*],[*2,2
*]*],5*],[*[*"A",2*],171,[*[*0,17*]*],[*[*2,3,5,7,11,13,17*]*]*],
[*[*"A",2*],171,[*[*3,8*]*],11*],[*[*"A",2*],171,[*[*8,3*]*],11*],
[*[*"A",2*],171,[*[*17,0*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"B",9
*],171,[*[*0,0,0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"C",3*],171,[*[*
2,1,0*]*],7*],[*[*"C",9*],171,[*[*0,0,0,0,0,0,0,0,2*]*],[*[*2
*]*]*],[*[*"A",1*],171,[*[*2*],[*56*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],171,[*[*2*],[*2*],[*18
*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],171,[*[*8*],[*18*]*],[*
[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],171,[*[*170*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167*]
*]*],[*[*"A",18*],171,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",18*],171,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"B",85*],171,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",170*],171,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",170*],171,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"C",3*],172,[*[*3,0,0*]*],7*],[*[*
"A",1*],172,[*[*1*],[*85*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],172,[*[*1
*],[*1*],[*42*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],
[*[*"A",1*],172,[*[*3*],[*42*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41*]*]*],[*[*"A",1*],172,[*[*171*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167*]*]*],[*[*
"C",86*],172,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",86*],
172,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",171*],172,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",171*],172,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],
173,[*[*1,1,3*]*],5*],[*[*"A",3*],173,[*[*3,1,1*]*],5*],[*[*"A",1
*],173,[*[*172*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167*]*]*],[*[*"B",86*],173,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[*2*]*]*],[*[*"A",172*],173,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",172*],173,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],174,[*[*0,
0,2,1*]*],3*],[*[*"A",4*],174,[*[*1,2,0,0*]*],3*],[*[*"B",2*],
174,[*[*0,7*]*],13*],[*[*"A",1*],174,[*[*1*],[*2*],[*28*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],174,[*[*1*],[*86*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83*]*]*],[*[*"A",1*],174,[*[*2*],[*57*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],174,[*[*5*],
[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],174,[*[*173
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167,173*]*]*],[*[*"C",87*],174,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"D",87*],174,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"A",173*],174,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",173*],174,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],175,[*[*1,2,1*]*],[*[*2,3
*]*]*],[*[*"A",4*],175,[*[*0,0,0,1*],[*0,0,0,3*]*],[*[*2,3*]*]*],
[*[*"A",4*],175,[*[*0,0,0,1*],[*3,0,0,0*]*],[*[*2,3*]*]*],[*[*
"A",4*],175,[*[*1,0,0,0*],[*0,0,0,3*]*],[*[*2,3*]*]*],[*[*"A",4
*],175,[*[*1,0,0,0*],[*3,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",4*],175,
[*[*0,0,2,1*]*],[*[*2,3*]*]*],[*[*"A",4*],175,[*[*0,0,3,0*]*],[*[*
2,3*]*]*],[*[*"A",4*],175,[*[*0,3,0,0*]*],[*[*2,3*]*]*],[*[*"A",4
*],175,[*[*1,0,1,1*]*],[*[*2,3*]*]*],[*[*"A",4*],175,[*[*1,1,0,1
*]*],[*[*2,3*]*]*],[*[*"A",4*],175,[*[*1,2,0,0*]*],[*[*2,3*]*]*],[*
[*"A",5*],175,[*[*0,0,2,0,0*]*],[*[*2,3*]*]*],[*[*"A",6*],175,[*[*
0,0,0,1,0,1*]*],2*],[*[*"A",6*],175,[*[*1,0,1,0,0,0*]*],2*],[*
[*"B",2*],175,[*[*0,1*],[*2,1*]*],[*[*2,3*]*]*],[*[*"B",2*],175,[*
[*0,1*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",1*],175,[*[*4*],[*34*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],175,[*[*4*],[*
4*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],175,[*[*6*],[*24*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],175,[*[*174*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,
167,173*]*]*],[*[*"B",87*],175,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]
*]*],[*[*"A",174*],175,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",174*],175,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],176,[*[*0,0,1*],[*1,
1,1*]*],3*],[*[*"A",3*],176,[*[*1,0,0*],[*1,1,1*]*],3*],[*[*"A",
4*],176,[*[*2,0,0,2*]*],7*],[*[*"B",2*],176,[*[*1,0*],[*3,1*]*],
7*],[*[*"A",1*],176,[*[*1*],[*3*],[*21*]*],[*[*2,3,5,7,11,13,17,
19*]*]*],[*[*"A",1*],176,[*[*1*],[*7*],[*10*]*],[*[*2,3,5,7*]*]*],[*
[*"A",1*],176,[*[*1*],[*87*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],176,[*
[*1*],[*1*],[*3*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],176,[*[*1
*],[*1*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]
*],[*[*"A",1*],176,[*[*1*],[*1*],[*1*],[*21*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",1*],176,[*[*1*],[*1*],[*1*],[*1*],[*10*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],176,[*[*3*],[*43*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],176,[*[*3*],[*3*],[*
10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],176,[*[*7*],[*21*]*],[*[*2,3,
5,7,11,13,17,19*]*]*],[*[*"A",1*],176,[*[*10*],[*15*]*],[*[*2,3,
5,7,11,13*]*]*],[*[*"A",1*],176,[*[*175*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137,139,149,151,157,163,167,173*]*]
*],[*[*"C",88*],176,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",88*],176,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",175
*],176,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",175*],176,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],177,[*[*2*],[*58*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],177,[*
[*176*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173*]*]*],[*[*"B",88*],177,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[*2*]*]*],[*[*"A",176*],177,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",176*],177,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",1*],178,[*[*1*],[*88*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],178,
[*[*177*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149,151,157,163,167,173*]*]*],[*[*"C",89*],178,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",89*],178,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",177*],178,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",177*],178,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",2*],179,[*[*4,7*]*],11*],[*[*"A",2*],179,[*[*7,4*]*],
11*],[*[*"B",2*],179,[*[*2,4*]*],11*],[*[*"A",1*],179,[*[*178*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173*]*]*],[*[*"B",89*],179,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",178*],179,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",178*],179,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",2*],180,[*[*0,1*],[*0,2*],[*0,3*]*],[*[*2,3*]*]*],[*[*
"A",2*],180,[*[*0,1*],[*0,2*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],
180,[*[*0,1*],[*2,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],180,[*[*
0,1*],[*2,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],180,[*[*0,1*],[*
2,4*]*],[*[*2,3,7*]*]*],[*[*"A",2*],180,[*[*0,1*],[*4,2*]*],[*[*2,
3,7*]*]*],[*[*"A",2*],180,[*[*1,0*],[*0,2*],[*0,3*]*],[*[*2,3*]*]
*],[*[*"A",2*],180,[*[*1,0*],[*0,2*],[*3,0*]*],[*[*2,3*]*]*],[*[*
"A",2*],180,[*[*1,0*],[*2,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],
180,[*[*1,0*],[*2,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],180,[*[*
1,0*],[*2,4*]*],[*[*2,3,7*]*]*],[*[*"A",2*],180,[*[*1,0*],[*4,2*]
*],[*[*2,3,7*]*]*],[*[*"A",2*],180,[*[*0,3*],[*1,3*]*],5*],[*[*"A",
2*],180,[*[*0,3*],[*3,1*]*],5*],[*[*"A",2*],180,[*[*3,0*],[*1,3*]
*],5*],[*[*"A",2*],180,[*[*3,0*],[*3,1*]*],5*],[*[*"A",3*],180,[*[*
0,0,1*],[*0,1,2*]*],[*[*2*]*]*],[*[*"A",3*],180,[*[*0,0,1*],[*2,
1,0*]*],[*[*2*]*]*],[*[*"A",3*],180,[*[*1,0,0*],[*0,1,2*]*],[*[*2
*]*]*],[*[*"A",3*],180,[*[*1,0,0*],[*2,1,0*]*],[*[*2*]*]*],[*[*"A",
3*],180,[*[*0,3,3*]*],7*],[*[*"A",3*],180,[*[*3,3,0*]*],7*],[*[*
"B",2*],180,[*[*1,8*]*],19*],[*[*"B",2*],180,[*[*3,3*]*],7*],[*[*
"A",1*],180,[*[*1*],[*2*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1
*],180,[*[*1*],[*2*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]
*],[*[*"A",1*],180,[*[*1*],[*2*],[*2*],[*9*]*],[*[*2,3,5,7*]*]*],[*
[*"A",1*],180,[*[*1*],[*4*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*
[*"A",1*],180,[*[*1*],[*5*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],180,[*[*1*],[*8*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
180,[*[*1*],[*89*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],180,[*[*1*],
[*1*],[*2*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],180,[*[*1
*],[*1*],[*2*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],180,[*[*1*],
[*1*],[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],180,[*[*1*],[*1
*],[*44*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*
"A",1*],180,[*[*2*],[*3*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],180,[*[*2*],[*4*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1
*],180,[*[*2*],[*5*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],180,[*[*
2*],[*59*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59*]*]*],[*[*"A",1*],180,[*[*2*],[*2*],[*3*],[*4*]*],[*[*2,3*]*]*],[*
[*"A",1*],180,[*[*2*],[*2*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",1*],180,[*[*3*],[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],180,[*[*3*],[*44*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43*]*]*],[*[*"A",1*],180,[*[*4*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*
[*"A",1*],180,[*[*4*],[*35*]*],[*[*2,3,5,7,11,13,17,19,23,29,31
*]*]*],[*[*"A",1*],180,[*[*5*],[*29*]*],[*[*2,3,5,7,11,13,17,19,
23,29*]*]*],[*[*"A",1*],180,[*[*8*],[*19*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",1*],180,[*[*9*],[*17*]*],[*[*2,3,5,7,11,13,17
*]*]*],[*[*"A",1*],180,[*[*11*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*
[*"A",1*],180,[*[*179*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179*]*]*],[*[*"C",90*],
180,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",90*],
180,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",179*],
180,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",179*],180,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],181,[*[*2,14
*]*],17*],[*[*"A",2*],181,[*[*14,2*]*],17*],[*[*"B",2*],181,[*[*0,9
*]*],19*],[*[*"A",1*],181,[*[*180*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179*]*]*],[*[*
"B",90*],181,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",180*],181,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",180*],181,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",3
*],182,[*[*0,0,1*],[*0,0,2*]*],7*],[*[*"B",3*],182,[*[*0,0,4*]*],
[*[*2,3,11*]*]*],[*[*"C",3*],182,[*[*0,1,0*],[*1,0,0*]*],3*],[*[*
"G",2*],182,[*[*0,1*],[*0,2*]*],7*],[*[*"G",2*],182,[*[*0,4*]*],[*
[*2,3,11*]*]*],[*[*"A",1*],182,[*[*1*],[*6*],[*12*]*],[*[*2,3,5,7,
11*]*]*],[*[*"A",1*],182,[*[*1*],[*90*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*
"A",1*],182,[*[*6*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],182,[*[*12*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],182,[*[*181*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181*]*]*],[*[*"C",91*],182,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",91*],182,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",181*],182,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",181*],182,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],183,[*[*
5,6*]*],11*],[*[*"A",2*],183,[*[*6,5*]*],11*],[*[*"A",1*],183,[*[*
2*],[*60*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59*]*]*],[*[*"A",1*],183,[*[*182*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181*]*]*],[*
[*"B",91*],183,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],
[*[*"A",182*],183,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",182*],183,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",3*],184,[*[*1,3,1*]*],5*],[*[*"B",2*],184,[*[*1,5*]*],
11*],[*[*"A",1*],184,[*[*1*],[*3*],[*22*]*],[*[*2,3,5,7,11,13,17,
19*]*]*],[*[*"A",1*],184,[*[*1*],[*91*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*
"A",1*],184,[*[*1*],[*1*],[*45*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43*]*]*],[*[*"A",1*],184,[*[*1*],[*1*],[*1*],[*22*]*],
[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],184,[*[*3*],[*45*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],184,
[*[*7*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],184,[*
[*183*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181*]*]*],[*[*"C",92*],184,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",92*],184,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",183*],184,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",183*],184,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],
185,[*[*0,0,2,2*]*],5*],[*[*"A",4*],185,[*[*2,2,0,0*]*],5*],[*[*
"A",1*],185,[*[*4*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]
*]*],[*[*"A",1*],185,[*[*184*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181*]*]*],[*
[*"B",92*],185,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]
*],[*[*"A",184*],185,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",184*],185,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",1*],186,[*[*1*],[*2*],[*30*]*],[*[*2,3,5,7,
11,13,17,19,23,29*]*]*],[*[*"A",1*],186,[*[*1*],[*92*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89*]*]*],[*[*"A",1*],186,[*[*2*],[*61*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],186,[*[*5
*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],186,
[*[*185*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149,151,157,163,167,173,179,181*]*]*],[*[*"C",93*],186,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",93*],186,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",185*],
186,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",185*],186,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],187,[*[*10*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*
[*"A",1*],187,[*[*186*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181*]*]*],[*[*"B",93
*],187,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",186*],187,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",186*],187,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",5*],188,[*[*0,1,0,1,0*]*],5*],[*[*
"B",10*],188,[*[*0,0,0,0,0,0,0,0,1,0*]*],2*],[*[*"B",9*],188,
[*[*0,0,0,0,0,0,0,0,2*]*],19*],[*[*"C",10*],188,[*[*0,0,0,0,
0,0,0,0,1,0*]*],2*],[*[*"C",10*],188,[*[*0,0,0,0,0,0,0,0,1,
0*]*],5*],[*[*"D",10*],188,[*[*0,0,0,0,0,0,0,0,1,0*]*],2*],[*[*
"A",1*],188,[*[*1*],[*93*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],188,[*
[*1*],[*1*],[*46*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43
*]*]*],[*[*"A",1*],188,[*[*3*],[*46*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43*]*]*],[*[*"A",1*],188,[*[*187*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181*]*]*],[*[*"C",94*],188,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"D",94*],188,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",187*],188,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",187*],188,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*
"A",2*],189,[*[*0,1*],[*1,0*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2
*],189,[*[*0,1*],[*1,0*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
189,[*[*0,1*],[*1,0*],[*1,0*],[*1,1*]*],3*],[*[*"A",2*],189,[*[*
0,1*],[*1,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*0,1*],[*3,3*]
*],5*],[*[*"A",2*],189,[*[*0,1*],[*6,1*]*],[*[*2,3,5*]*]*],[*[*"A",
2*],189,[*[*0,1*],[*0,1*],[*1,0*],[*1,1*]*],3*],[*[*"A",2*],189,
[*[*0,1*],[*0,1*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*0,
1*],[*0,1*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*0,1*],[*
0,1*],[*0,1*],[*1,1*]*],3*],[*[*"A",2*],189,[*[*1,0*],[*1,6*]*],
[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*1,0*],[*3,3*]*],5*],[*[*"A",2
*],189,[*[*1,0*],[*6,1*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*1,0
*],[*1,0*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*1,0*],[*
1,0*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],189,[*[*1,0*],[*1,0
*],[*1,0*],[*1,1*]*],3*],[*[*"A",2*],189,[*[*1,1*],[*2,2*]*],3*],
[*[*"A",3*],189,[*[*1,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",3*],189,[*[*
5,0,1*]*],[*[*2,3,5*]*]*],[*[*"A",5*],189,[*[*0,1,0,1,0*]*],[*[*
2,5*]*]*],[*[*"A",6*],189,[*[*1,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",
6*],189,[*[*2,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"B",3*],189,[*[*0,0,
1*],[*0,0,2*]*],[*[*2,7*]*]*],[*[*"B",3*],189,[*[*2,0,1*]*],[*[*2,
5*]*]*],[*[*"B",3*],189,[*[*2,1,0*]*],3*],[*[*"B",9*],189,[*[*0,0,
0,0,0,0,0,0,2*]*],[*[*2,19*]*]*],[*[*"C",10*],189,[*[*0,0,0,0,
0,0,0,0,1,0*]*],[*[*2,5*]*]*],[*[*"C",3*],189,[*[*0,1,2*]*],[*[*
2*]*]*],[*[*"G",2*],189,[*[*0,1*],[*0,2*]*],[*[*2,7*]*]*],[*[*"G",2
*],189,[*[*0,1*],[*2,0*]*],3*],[*[*"G",2*],189,[*[*1,0*],[*0,2*]
*],3*],[*[*"G",2*],189,[*[*1,0*],[*2,0*]*],3*],[*[*"G",2*],189,[*[*
1,2*]*],[*[*2,11*]*]*],[*[*"G",2*],189,[*[*2,1*]*],3*],[*[*"A",1*],
189,[*[*2*],[*6*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],189,[*[*2
*],[*62*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61*]*]*],[*[*"A",1*],189,[*[*2*],[*2*],[*20*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],189,[*[*2*],[*2*],[*2*],[*6*]*],[*[*2,3,
5*]*]*],[*[*"A",1*],189,[*[*6*],[*26*]*],[*[*2,3,5,7,11,13,17,19,
23*]*]*],[*[*"A",1*],189,[*[*8*],[*20*]*],[*[*2,3,5,7,11,13,17,19
*]*]*],[*[*"A",1*],189,[*[*188*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181*]*]*],[*
[*"B",94*],189,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",188*],189,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",188
*],189,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",18*],190,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",18
*],190,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]
*]*],[*[*"A",19*],190,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1,0*]*],[*[**]*]*],[*[*"A",19*],190,[*[*0,1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",2*],190,[*[*0,3*],[*2,
2*]*],5*],[*[*"A",2*],190,[*[*3,0*],[*2,2*]*],5*],[*[*"A",2*],190,
[*[*0,18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",2*],190,[*[*18,0
*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",3*],190,[*[*0,0,2*],[*0,
2,0*]*],3*],[*[*"A",3*],190,[*[*2,0,0*],[*0,2,0*]*],3*],[*[*"D",
10*],190,[*[*0,0,0,0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"D",5*],
190,[*[*0,0,0,1,1*]*],3*],[*[*"A",1*],190,[*[*1*],[*4*],[*18*]*],
[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],190,[*[*1*],[*94*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89*]*]*],[*[*"A",1*],190,[*[*4*],[*37*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37*]*]*],[*[*"A",1*],190,[*[*9*],[*18*]*],[*[*2,
3,5,7,11,13,17*]*]*],[*[*"A",1*],190,[*[*189*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181*]*]*],[*[*"A",18*],190,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",18*],190,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,2*]*],[*[**]*]*],[*[*"A",19*],190,[*[*0,1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",19*],
190,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]
*],[*[*"C",95*],190,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"D",95*],190,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",189*],190,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",189*],190,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*
"A",1*],191,[*[*190*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181*]*]*],[*[*"B",95
*],191,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]
*],[*[*"A",190*],191,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",190
*],191,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],192,[*[*
0,1*],[*1,1*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2*],192,[*[*0,1*],[*3,
3*]*],[*[*3,5,7*]*]*],[*[*"A",2*],192,[*[*1,0*],[*1,1*],[*1,1*]*],
[*[*3*]*]*],[*[*"A",2*],192,[*[*1,0*],[*3,3*]*],[*[*3,5,7*]*]*],[*[*
"A",2*],192,[*[*1,1*],[*1,3*]*],[*[*3,5*]*]*],[*[*"A",2*],192,[*[*
1,1*],[*3,1*]*],[*[*3,5*]*]*],[*[*"A",2*],192,[*[*3,7*]*],[*[*3,5,
7,11*]*]*],[*[*"A",2*],192,[*[*3,13*]*],17*],[*[*"A",2*],192,[*[*7,
3*]*],[*[*3,5,7,11*]*]*],[*[*"A",2*],192,[*[*13,3*]*],17*],[*[*"A",
3*],192,[*[*0,1,0*],[*1,0,2*]*],5*],[*[*"A",3*],192,[*[*0,1,0*],
[*2,0,1*]*],5*],[*[*"A",3*],192,[*[*0,2,3*]*],5*],[*[*"A",3*],
192,[*[*3,2,0*]*],5*],[*[*"B",2*],192,[*[*1,0*],[*1,0*],[*1,1*]
*],5*],[*[*"A",1*],192,[*[*1*],[*2*],[*3*],[*7*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],192,[*[*1*],[*2*],[*31*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31*]*]*],[*[*"A",1*],192,[*[*1*],[*3*],[*23*]*],[*[*2,3,
5,7,11,13,17,19,23*]*]*],[*[*"A",1*],192,[*[*1*],[*3*],[*3*],[*5
*]*],[*[*2,3,5*]*]*],[*[*"A",1*],192,[*[*1*],[*5*],[*15*]*],[*[*2,3,
5,7,11,13*]*]*],[*[*"A",1*],192,[*[*1*],[*7*],[*11*]*],[*[*2,3,5,
7,11*]*]*],[*[*"A",1*],192,[*[*1*],[*95*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*
"A",1*],192,[*[*1*],[*1*],[*2*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*
"A",1*],192,[*[*1*],[*1*],[*2*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],
[*[*"A",1*],192,[*[*1*],[*1*],[*3*],[*11*]*],[*[*2,3,5,7,11*]*]*],
[*[*"A",1*],192,[*[*1*],[*1*],[*5*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],192,[*[*1*],[*1*],[*47*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47*]*]*],[*[*"A",1*],192,[*[*1*],[*1*],[*1*],[*2*],
[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],192,[*[*1*],[*1*],[*1*],[*3
*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],192,[*[*1*],[*1*],[*1*],[*23
*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],192,[*[*1*],[*1
*],[*1*],[*1*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],192,[*[*1*],[*
1*],[*1*],[*1*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],192,[*[*
1*],[*1*],[*1*],[*1*],[*1*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],
192,[*[*1*],[*1*],[*1*],[*1*],[*1*],[*1*],[*2*]*],[*[*2*]*]*],[*[*
"A",1*],192,[*[*2*],[*3*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],192,[*[*2*],[*3*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],
192,[*[*2*],[*7*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],192,[*[*2
*],[*63*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61*]*]*],[*[*"A",1*],192,[*[*3*],[*5*],[*7*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],192,[*[*3*],[*47*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47*]*]*],[*[*"A",1*],192,[*[*3*],[*3*],[*11*]*],[*[*
2,3,5,7,11*]*]*],[*[*"A",1*],192,[*[*5*],[*31*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31*]*]*],[*[*"A",1*],192,[*[*7*],[*23*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],192,[*[*11*],[*15*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",1*],192,[*[*191*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191*]*]*],[*[*"C",96*],192,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",96*],192,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",191*],192,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",191*],192,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],193,[*[*192*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191*]*]*],[*[*"B",96*],193,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",192*],193,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",192*],193,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",13*],194,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,1*]*],2*],[*[*"A",13*],194,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,1*]*],7*],[*[*"A",1*],194,[*[*1*],[*96*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]
*],[*[*"A",1*],194,[*[*193*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193*]*]
*],[*[*"C",97*],194,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"D",97*],194,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",193*],194,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",193*],194,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",13*],195,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[*2,7*]*]*],[*[*"A",2*],195,[*[*1,12*]*],[*[*2,3,
5,7,11*]*]*],[*[*"A",2*],195,[*[*2,9*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",2*],195,[*[*9,2*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],195,[*[*
12,1*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",4*],195,[*[*0,2,0,1*]*],3
*],[*[*"A",4*],195,[*[*1,0,2,0*]*],3*],[*[*"D",4*],195,[*[*0,0,2,
0*]*],3*],[*[*"A",1*],195,[*[*2*],[*4*],[*12*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],195,[*[*2*],[*64*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],195,[*[*4*],[*38*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],195,[*
[*12*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],195,[*[*194*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193*]*]*],[*[*"B",97*],195,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",194
*],195,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",194*],
195,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",13*],
196,[*[*0,0,0,0,0,0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",13*],196,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,1*],[*0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
13*],196,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0*],[*1,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],196,[*[*1,0,1*],[*1,0,1*]
*],2*],[*[*"A",3*],196,[*[*0,5,0*]*],[*[*2,3,5*]*]*],[*[*"A",6*],
196,[*[*0,0,0,0,0,1*],[*0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",6
*],196,[*[*0,0,0,0,0,1*],[*2,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",
6*],196,[*[*1,0,0,0,0,0*],[*0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*
"A",6*],196,[*[*1,0,0,0,0,0*],[*2,0,0,0,0,0*]*],[*[*2*]*]*],[*
[*"A",6*],196,[*[*0,0,0,0,2,0*]*],[*[*2,3*]*]*],[*[*"A",6*],196,
[*[*0,2,0,0,0,0*]*],[*[*2,3*]*]*],[*[*"B",2*],196,[*[*0,2*],[*0,
2*]*],[*[*2,5*]*]*],[*[*"B",3*],196,[*[*0,1,0*],[*0,1,0*]*],2*],[*
[*"B",7*],196,[*[*0,0,0,0,0,0,1*],[*0,0,0,0,0,0,1*]*],2*],[*
[*"C",3*],196,[*[*0,1,0*],[*1,0,0*]*],[*[*2,3*]*]*],[*[*"C",3*],
196,[*[*0,1,0*],[*0,1,0*]*],[*[*3*]*]*],[*[*"C",3*],196,[*[*1,0,0
*],[*1,0,0*]*],[*[*2*]*]*],[*[*"C",6*],196,[*[*0,0,0,1,0,0*]*],5
*],[*[*"C",7*],196,[*[*0,0,0,0,0,0,1*],[*0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",7*],196,[*[*0,0,0,0,0,0,1*],[*0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"F",4*],196,[*[*0,0,1,0*]*],3*],[*[*"G",2*],196,[*
[*1,0*],[*1,0*]*],[*[*3*]*]*],[*[*"G",2*],196,[*[*3,0*]*],5*],[*[*
"A",1*],196,[*[*1*],[*6*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],196,[*[*1*],[*97*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97*]*]*],[*[*"A",1*],
196,[*[*1*],[*1*],[*6*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],196,[*[*
1*],[*1*],[*48*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47*]*]*],[*[*"A",1*],196,[*[*3*],[*6*],[*6*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],196,[*[*3*],[*48*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47*]*]*],[*[*"A",1*],196,[*[*6*],[*27*]*],[*[*2,3,5,7,
11,13,17,19,23*]*]*],[*[*"A",1*],196,[*[*13*],[*13*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],196,[*[*195*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193*]*]*],[*[*"C",98*],196,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",98*],196,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",195*],
196,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",195*],
196,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1
*],197,[*[*196*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193*]*]*],[*[*"B",98
*],197,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",196*],197,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",196*],197,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",2*],198,[*[*0,1*],[*0,10*]*],[*[*2,3,5,
7*]*]*],[*[*"A",2*],198,[*[*0,1*],[*10,0*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],198,[*[*1,0*],[*0,10*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],
198,[*[*1,0*],[*10,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],198,[*[*0,
2*],[*1,5*]*],7*],[*[*"A",2*],198,[*[*0,2*],[*5,1*]*],7*],[*[*"A",
2*],198,[*[*2,0*],[*1,5*]*],7*],[*[*"A",2*],198,[*[*2,0*],[*5,1*]
*],7*],[*[*"A",1*],198,[*[*1*],[*2*],[*32*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],198,[*[*1*],[*2*],[*2*],[*10*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],198,[*[*1*],[*8*],[*10*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],198,[*[*1*],[*98*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97*]*]*],
[*[*"A",1*],198,[*[*2*],[*5*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],198,[*[*2*],[*65*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61*]*]*],[*[*"A",1*],198,[*[*2*],[*2*],[*21*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],198,[*[*5*],[*32*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],198,[*[*8*],[*21*]*],
[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],198,[*[*10*],[*17*]*],[*
[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],198,[*[*197*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197*]*]*],[*[*"C",99*],198,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",99*],
198,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"A",197*],198,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",197*],198,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],199,[*[*2,0,0,2*]*],3*],[*[*"B",
2*],199,[*[*0,7*]*],11*],[*[*"B",2*],199,[*[*2,5*]*],7*],[*[*"A",1
*],199,[*[*198*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197*]*]*],[*[*
"B",99*],199,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",198*],199,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",198*],199,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],200,[*[*0,0,1
*],[*0,3,0*]*],[*[*2,3*]*]*],[*[*"A",3*],200,[*[*1,0,0*],[*0,3,0
*]*],[*[*2,3*]*]*],[*[*"A",3*],200,[*[*0,0,2*],[*0,0,3*]*],[*[*2,3
*]*]*],[*[*"A",3*],200,[*[*0,0,2*],[*0,1,1*]*],[*[*2,3*]*]*],[*[*
"A",3*],200,[*[*0,0,2*],[*0,2,0*]*],[*[*2,3*]*]*],[*[*"A",3*],
200,[*[*0,0,2*],[*1,1,0*]*],[*[*2,3*]*]*],[*[*"A",3*],200,[*[*0,
0,2*],[*3,0,0*]*],[*[*2,3*]*]*],[*[*"A",3*],200,[*[*2,0,0*],[*0,
0,3*]*],[*[*2,3*]*]*],[*[*"A",3*],200,[*[*2,0,0*],[*0,1,1*]*],[*[*
2,3*]*]*],[*[*"A",3*],200,[*[*2,0,0*],[*0,2,0*]*],[*[*2,3*]*]*],[*
[*"A",3*],200,[*[*2,0,0*],[*1,1,0*]*],[*[*2,3*]*]*],[*[*"A",3*],
200,[*[*2,0,0*],[*3,0,0*]*],[*[*2,3*]*]*],[*[*"A",4*],200,[*[*0,
0,0,1*],[*0,0,1,1*]*],[*[*3*]*]*],[*[*"A",4*],200,[*[*0,0,0,1*],
[*0,1,0,1*]*],2*],[*[*"A",4*],200,[*[*0,0,0,1*],[*1,0,1,0*]*],
2*],[*[*"A",4*],200,[*[*0,0,0,1*],[*1,1,0,0*]*],[*[*3*]*]*],[*[*
"A",4*],200,[*[*1,0,0,0*],[*0,0,1,1*]*],[*[*3*]*]*],[*[*"A",4*],
200,[*[*1,0,0,0*],[*0,1,0,1*]*],2*],[*[*"A",4*],200,[*[*1,0,0,
0*],[*1,0,1,0*]*],2*],[*[*"A",4*],200,[*[*1,0,0,0*],[*1,1,0,0
*]*],[*[*3*]*]*],[*[*"A",4*],200,[*[*2,0,0,2*]*],[*[*2,3,7*]*]*],[*
[*"B",2*],200,[*[*1,0*],[*0,1*],[*2,0*]*],[*[*2*]*]*],[*[*"B",2*],
200,[*[*0,1*],[*1,2*]*],[*[*2,7*]*]*],[*[*"B",2*],200,[*[*2,0*],[*
3,0*]*],[*[*2,3*]*]*],[*[*"B",2*],200,[*[*7,1*]*],11*],[*[*"A",1*],
200,[*[*1*],[*3*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],200,[*[*1
*],[*3*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
200,[*[*1*],[*4*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",
1*],200,[*[*1*],[*9*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],200,[*
[*1*],[*99*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97*]*]*],[*[*"A",1*],200,[*[*1*],[*1
*],[*4*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],200,[*[*1*],[*1*],[*
49*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*
"A",1*],200,[*[*1*],[*1*],[*1*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],200,[*[*1*],[*1*],[*1*],[*24*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],200,[*[*3*],[*4*],[*9*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],200,[*[*3*],[*49*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47*]*]*],[*[*"A",1*],200,[*[*4*],[*39*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],200,[*[*4*],[*4*],
[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],200,[*[*7*],[*24*]*],[*[*2,3,
5,7,11,13,17,19,23*]*]*],[*[*"A",1*],200,[*[*9*],[*19*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],200,[*[*199*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199*]*]*],[*[*"C",100*],200,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",
100*],200,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",199*],200,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",199*],200,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],201,[*[*4,12*]
*],17*],[*[*"A",2*],201,[*[*12,4*]*],17*],[*[*"A",1*],201,[*[*2*],[*
66*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61
*]*]*],[*[*"A",1*],201,[*[*200*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199*]*]*],[*[*"B",100*],201,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",200*],201,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",200
*],201,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"C",3*],202,[*[*1,0,2*]*],5*],[*[*"A",1*],202,[*[*1*],[*
100*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97*]*]*],[*[*"A",1*],202,[*[*201*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,
163,167,173,179,181,191,193,197,199*]*]*],[*[*"C",101*],202,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"D",101*],202,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",201*],202,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",201*],202,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*
"A",6*],203,[*[*0,0,1,0,0,1*]*],5*],[*[*"A",6*],203,[*[*1,0,0,
1,0,0*]*],5*],[*[*"A",1*],203,[*[*6*],[*28*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],203,[*[*202*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199*]*]*],[*[*"B",101*],203,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",
202*],203,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",202*],203,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",5*],204,[*[*0,0,0,0,1*],[*
1,0,0,0,1*]*],2*],[*[*"A",5*],204,[*[*0,0,0,0,1*],[*1,0,0,0,
1*]*],3*],[*[*"A",5*],204,[*[*1,0,0,0,0*],[*1,0,0,0,1*]*],2*],
[*[*"A",5*],204,[*[*1,0,0,0,0*],[*1,0,0,0,1*]*],3*],[*[*"A",5
*],204,[*[*0,0,1,1,0*]*],2*],[*[*"A",5*],204,[*[*0,1,1,0,0*]*],
2*],[*[*"B",2*],204,[*[*0,7*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"B",2
*],204,[*[*3,3*]*],5*],[*[*"A",1*],204,[*[*1*],[*2*],[*33*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],204,[*[*1*],[*5*],
[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],204,[*[*1*],[*101*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101*]*]*],[*[*"A",1*],204,[*[*1*],[*1*],[*2*],[*
16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],204,[*[*1*],[*1*],[*50
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*
"A",1*],204,[*[*2*],[*3*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],204,[*[*2*],[*67*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],204,[*[*3*],[*50*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],
204,[*[*5*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],204,[*[*11*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
204,[*[*203*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199*]*]*],[*
[*"C",102*],204,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",102*],204,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",203
*],204,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",203*],204,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],205,[*[*4*],[*40*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],205,[*[*
204*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199*]*]*],[*[*"B",102
*],205,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[*2*]*]*],[*[*"A",204*],205,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",204*],205,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"A",3*],206,[*[*0,4,2*]*],7*],[*[*"A",3*],206,[*[*2,4,0*]
*],7*],[*[*"B",2*],206,[*[*2,4*]*],7*],[*[*"A",1*],206,[*[*1*],[*
102*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101*]*]*],[*[*"A",1*],206,[*[*205*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199*]*]*],[*[*"C",103*],206,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"D",103*],206,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",205*],206,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",205*],206,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],207,[*[*1,17*]*],19*],[*
[*"A",2*],207,[*[*17,1*]*],19*],[*[*"A",1*],207,[*[*2*],[*68*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67*]*]
*],[*[*"A",1*],207,[*[*2*],[*2*],[*22*]*],[*[*2,3,5,7,11,13,17,19
*]*]*],[*[*"A",1*],207,[*[*8*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]
*]*],[*[*"A",1*],207,[*[*206*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199*]*]*],[*[*"B",103*],207,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",206*],
207,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",206*],207,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],208,[*[*
5,11*]*],17*],[*[*"A",2*],208,[*[*11,5*]*],17*],[*[*"A",3*],208,[*
[*0,0,1*],[*0,1,3*]*],5*],[*[*"A",3*],208,[*[*0,0,1*],[*3,1,0*]
*],5*],[*[*"A",3*],208,[*[*1,0,0*],[*0,1,3*]*],5*],[*[*"A",3*],
208,[*[*1,0,0*],[*3,1,0*]*],5*],[*[*"A",7*],208,[*[*0,1,0,0,0,
0,1*]*],7*],[*[*"A",7*],208,[*[*1,0,0,0,0,1,0*]*],7*],[*[*"B",2
*],208,[*[*1,0*],[*3,1*]*],5*],[*[*"B",2*],208,[*[*1,0*],[*1,0*],
[*0,2*]*],5*],[*[*"B",3*],208,[*[*1,0,0*],[*0,0,2*]*],7*],[*[*
"B",4*],208,[*[*0,0,0,1*],[*0,0,1,0*]*],2*],[*[*"B",6*],208,[*[*
0,0,0,1,0,0*]*],2*],[*[*"C",4*],208,[*[*0,0,0,1*],[*0,0,1,0*]
*],2*],[*[*"C",6*],208,[*[*0,0,0,1,0,0*]*],[*[*5*]*]*],[*[*"D",10
*],208,[*[*0,0,0,0,0,0,0,0,0,2*]*],5*],[*[*"D",4*],208,[*[*0,
0,0,1*],[*0,0,1,0*]*],2*],[*[*"D",4*],208,[*[*0,1,0,0*],[*0,0,
1,0*]*],2*],[*[*"D",4*],208,[*[*1,0,0,0*],[*0,0,1,0*]*],2*],[*[*
"D",6*],208,[*[*0,0,0,1,0,0*]*],2*],[*[*"A",1*],208,[*[*1*],[*3
*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],208,[*[*
1*],[*7*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],208,[*[*1*],[*
103*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103*]*]*],[*[*"A",1*],208,[*[*1*],[*
1*],[*3*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],208,[*[*1*],[*
1*],[*51*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]
*],[*[*"A",1*],208,[*[*1*],[*1*],[*1*],[*25*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"A",1*],208,[*[*1*],[*1*],[*1*],[*1*],[*12*]*],[*
[*2,3,5,7,11*]*]*],[*[*"A",1*],208,[*[*3*],[*51*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],208,[*[*3*],
[*3*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],208,[*[*7*],[*25*]
*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],208,[*[*12*],[*15
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],208,[*[*207*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,
167,173,179,181,191,193,197,199*]*]*],[*[*"C",104*],208,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",104*],208,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",207*],208,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",207*],208,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"D",10*],209,[*[*0,0,0,
0,0,0,0,0,0,2*]*],[*[*2,5*]*]*],[*[*"A",1*],209,[*[*10*],[*18*]
*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],209,[*[*208*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,
167,173,179,181,191,193,197,199*]*]*],[*[*"B",104*],209,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",208*],209,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",208*],209,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",19*],210,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",19*],210,[*[*2,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",20*],210,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*
[*"A",20*],210,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",2*],210,[*[*0,2*],[*1,4*]*],[*[*2,3*]*]*],[*
[*"A",2*],210,[*[*0,2*],[*4,1*]*],[*[*2,3*]*]*],[*[*"A",2*],210,[*
[*2,0*],[*1,4*]*],[*[*2,3*]*]*],[*[*"A",2*],210,[*[*2,0*],[*4,1*]
*],[*[*2,3*]*]*],[*[*"A",2*],210,[*[*0,3*],[*0,5*]*],[*[*2,3,5*]*]
*],[*[*"A",2*],210,[*[*0,3*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
210,[*[*3,0*],[*0,5*]*],[*[*2,3,5*]*]*],[*[*"A",2*],210,[*[*3,0*],
[*5,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],210,[*[*0,19*]*],[*[*2,3,5,
7,11,13,17,19*]*]*],[*[*"A",2*],210,[*[*4,6*]*],[*[*2,3,5,11*]*]
*],[*[*"A",2*],210,[*[*6,4*]*],[*[*2,3,5,11*]*]*],[*[*"A",2*],210,
[*[*19,0*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",3*],210,[*[*0,
1,0*],[*0,0,4*]*],[*[*2,3*]*]*],[*[*"A",3*],210,[*[*0,1,0*],[*4,
0,0*]*],[*[*2,3*]*]*],[*[*"A",4*],210,[*[*0,0,0,6*]*],[*[*2,3,5*]
*]*],[*[*"A",4*],210,[*[*0,2,0,1*]*],[*[*2,3,5*]*]*],[*[*"A",4*],
210,[*[*1,0,2,0*]*],[*[*2,3,5*]*]*],[*[*"A",4*],210,[*[*6,0,0,0
*]*],[*[*2,3,5*]*]*],[*[*"A",5*],210,[*[*0,0,0,0,1*],[*1,0,0,0,
1*]*],[*[*2,3*]*]*],[*[*"A",5*],210,[*[*1,0,0,0,0*],[*1,0,0,0,1
*]*],[*[*2,3*]*]*],[*[*"A",5*],210,[*[*0,0,0,1,2*]*],[*[*2*]*]*],[*
[*"A",5*],210,[*[*0,0,1,1,0*]*],[*[*2,3*]*]*],[*[*"A",5*],210,[*[*
0,1,1,0,0*]*],[*[*2,3*]*]*],[*[*"A",5*],210,[*[*2,1,0,0,0*]*],[*
[*2*]*]*],[*[*"A",6*],210,[*[*0,0,0,0,0,4*]*],[*[*2,3*]*]*],[*[*
"A",6*],210,[*[*0,0,0,1,0,1*]*],[*[*2*]*]*],[*[*"A",6*],210,[*[*
1,0,1,0,0,0*]*],[*[*2*]*]*],[*[*"A",6*],210,[*[*4,0,0,0,0,0*]
*],[*[*2,3*]*]*],[*[*"A",9*],210,[*[*0,0,0,0,0,0,0,1,1*]*],3*],
[*[*"A",9*],210,[*[*0,0,0,0,0,1,0,0,0*]*],[*[**]*]*],[*[*"A",9*],
210,[*[*0,0,0,1,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",9*],210,[*[*1,
1,0,0,0,0,0,0,0*]*],3*],[*[*"B",10*],210,[*[*0,0,0,0,0,0,0,
0,1,0*]*],[*[*2*]*]*],[*[*"C",10*],210,[*[*0,0,0,0,0,0,0,0,0,2
*]*],[*[*2*]*]*],[*[*"C",5*],210,[*[*0,0,0,1,1*]*],3*],[*[*"D",5*],
210,[*[*0,0,0,0,3*]*],[*[*2,3*]*]*],[*[*"D",5*],210,[*[*1,1,0,0,
0*]*],[*[*2*]*]*],[*[*"A",1*],210,[*[*1*],[*2*],[*4*],[*6*]*],[*[*2,
3,5*]*]*],[*[*"A",1*],210,[*[*1*],[*2*],[*34*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],210,[*[*1*],[*4*],[*20*]*],[*[*
2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],210,[*[*1*],[*6*],[*14*]*],
[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],210,[*[*1*],[*104*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103*]*]*],[*[*"A",1*],210,[*[*2*],[*4*],[*13*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",1*],210,[*[*2*],[*6*],[*9*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],210,[*[*2*],[*69*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],210,[*
[*4*],[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],210,[*[*4*],[*41*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],
210,[*[*5*],[*34*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],210,[*[*6*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]
*],[*[*"A",1*],210,[*[*9*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],
[*[*"A",1*],210,[*[*13*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",
1*],210,[*[*209*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199*]*]*],[*
[*"A",19*],210,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
*]*],[*[**]*]*],[*[*"A",19*],210,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,2*]*],[*[**]*]*],[*[*"A",20*],210,[*[*0,1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",20*],210,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*
"C",105*],210,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",105*],210,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"A",209*],210,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",209*],210,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",3*],211,[*[*1,1,3*]*],7*],[*[*"A",3*],211,[*[*
3,1,1*]*],7*],[*[*"A",1*],211,[*[*210*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193,197,199*]*]*],[*[*"B",105*],211,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",210*],211,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",210*],211,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",1*],212,[*[*1*],[*105*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103*]*]*],[*[*"A",1*],212,[*[*1*],[*1*],[*52*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],212,[*[*3*],
[*52*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*
[*"A",1*],212,[*[*211*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211*]*]*],[*[*"C",106*],212,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",106*],
212,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*],[*[*"A",211*],212,[*[*1,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",211*],212,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],213,[*[*0,
1*],[*2,5*]*],7*],[*[*"A",2*],213,[*[*0,1*],[*5,2*]*],7*],[*[*"A",
2*],213,[*[*1,0*],[*2,5*]*],7*],[*[*"A",2*],213,[*[*1,0*],[*5,2*]
*],7*],[*[*"A",2*],213,[*[*6,10*]*],17*],[*[*"A",2*],213,[*[*10,6*]
*],17*],[*[*"A",1*],213,[*[*2*],[*70*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],213,[*[*212*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211*]*]*],[*[*"B",106*],
213,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",212*],213,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",212*],213,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],
214,[*[*1*],[*106*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79,83,89,97,101,103*]*]*],[*[*"A",1*],
214,[*[*213*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211*]*]
*],[*[*"C",107*],214,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",107*],214,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"A",213*],214,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",213*],214,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],215,[*[*
2,11*]*],13*],[*[*"A",2*],215,[*[*5,5*]*],7*],[*[*"A",2*],215,[*[*
11,2*]*],13*],[*[*"A",1*],215,[*[*4*],[*42*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41*]*]*],[*[*"A",1*],215,[*[*214*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211*]*]*],[*[*"B",107*],215,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",214*],215,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",214*],215,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],
216,[*[*0,1*],[*1,0*],[*1,3*]*],[*[*3,5*]*]*],[*[*"A",2*],216,[*[*
0,1*],[*1,0*],[*3,1*]*],[*[*3,5*]*]*],[*[*"A",2*],216,[*[*0,1*],[*
1,0*],[*1,0*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2*],216,[*[*0,1*],[*0,
1*],[*1,0*],[*1,1*]*],[*[*3*]*]*],[*[*"A",2*],216,[*[*0,1*],[*0,1
*],[*1,3*]*],[*[*3,5*]*]*],[*[*"A",2*],216,[*[*0,1*],[*0,1*],[*3,1
*]*],[*[*3,5*]*]*],[*[*"A",2*],216,[*[*0,1*],[*0,1*],[*0,1*],[*1,1
*]*],[*[*3*]*]*],[*[*"A",2*],216,[*[*1,0*],[*1,0*],[*1,3*]*],[*[*3,
5*]*]*],[*[*"A",2*],216,[*[*1,0*],[*1,0*],[*3,1*]*],[*[*3,5*]*]*],
[*[*"A",2*],216,[*[*1,0*],[*1,0*],[*1,0*],[*1,1*]*],[*[*3*]*]*],[*
[*"A",2*],216,[*[*0,2*],[*2,0*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],
216,[*[*0,2*],[*0,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],216,[*[*0,2
*],[*2,4*]*],7*],[*[*"A",2*],216,[*[*0,2*],[*4,2*]*],7*],[*[*"A",2
*],216,[*[*0,2*],[*7,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],216,[*[*
0,2*],[*0,2*],[*2,0*]*],[*[*2*]*]*],[*[*"A",2*],216,[*[*0,2*],[*0,
2*],[*0,2*]*],[*[*2*]*]*],[*[*"A",2*],216,[*[*2,0*],[*0,7*]*],[*[*
2,3,5,7*]*]*],[*[*"A",2*],216,[*[*2,0*],[*2,4*]*],7*],[*[*"A",2
*],216,[*[*2,0*],[*4,2*]*],7*],[*[*"A",2*],216,[*[*2,0*],[*7,0*]
*],[*[*2,3,5,7*]*]*],[*[*"A",2*],216,[*[*2,0*],[*2,0*],[*2,0*]*],
[*[*2*]*]*],[*[*"A",2*],216,[*[*1,1*],[*2,2*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],216,[*[*5,5*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],216,[*[*
7,9*]*],17*],[*[*"A",2*],216,[*[*9,7*]*],17*],[*[*"A",3*],216,[*[*
0,1,0*],[*1,0,2*]*],[*[*2,5*]*]*],[*[*"A",3*],216,[*[*0,1,0*],[*
2,0,1*]*],[*[*2,5*]*]*],[*[*"A",3*],216,[*[*0,1,0*],[*0,1,0*],[*
0,1,0*]*],[*[**]*]*],[*[*"A",3*],216,[*[*0,1,5*]*],[*[*2,3,5,7*]
*]*],[*[*"A",3*],216,[*[*5,1,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",5*],
216,[*[*0,0,0,0,1*],[*1,0,0,0,0*],[*1,0,0,0,0*]*],[*[**]*]*],[*
[*"A",5*],216,[*[*0,0,0,0,1*],[*0,0,0,0,1*],[*1,0,0,0,0*]*],
[*[**]*]*],[*[*"A",5*],216,[*[*0,0,0,0,1*],[*0,0,0,0,1*],[*0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",5*],216,[*[*1,0,0,0,0*],[*1,0,0,0,
0*],[*1,0,0,0,0*]*],[*[**]*]*],[*[*"A",7*],216,[*[*0,1,0,0,0,0,
1*]*],[*[*7*]*]*],[*[*"A",7*],216,[*[*1,0,0,0,0,1,0*]*],[*[*7*]*]
*],[*[*"B",2*],216,[*[*1,0*],[*0,4*]*],7*],[*[*"B",3*],216,[*[*0,
0,1*],[*0,0,1*],[*0,0,1*]*],2*],[*[*"B",3*],216,[*[*1,0,0*],[*
0,0,2*]*],[*[*2,7*]*]*],[*[*"C",3*],216,[*[*0,0,1*],[*0,0,1*],[*
0,0,1*]*],[*[**]*]*],[*[*"C",3*],216,[*[*1,0,2*]*],[*[*2,5,7*]*]
*],[*[*"C",4*],216,[*[*0,0,0,1*],[*0,0,1,0*]*],[*[*2*]*]*],[*[*
"G",2*],216,[*[*0,1*],[*0,1*],[*0,1*]*],2*],[*[*"A",1*],216,[*[*1
*],[*2*],[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],216,[*[*1*],[*
2*],[*5*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],216,[*[*1*],[*2*],[*
35*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],216,[*
[*1*],[*2*],[*2*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],216,[*
[*1*],[*2*],[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],216,[*[*1
*],[*3*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
216,[*[*1*],[*5*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],
216,[*[*1*],[*8*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],216,[*
[*1*],[*107*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107*]*]*],[*[*"A",1*],
216,[*[*1*],[*1*],[*2*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*
"A",1*],216,[*[*1*],[*1*],[*2*],[*2*],[*5*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],216,[*[*1*],[*1*],[*5*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],216,[*[*1*],[*1*],[*53*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53*]*]*],[*[*"A",1*],216,[*[*1*],[*1*],[*1*],[*2
*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],216,[*[*1*],[*1*],[*1*],[*
2*],[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],216,[*[*1*],[*1*],[*1*],[*
26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],216,[*[*2*],
[*3*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],216,[*[*2*],
[*5*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],216,[*[*2*],[*7*],
[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],216,[*[*2*],[*71*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*
[*"A",1*],216,[*[*2*],[*2*],[*3*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1
*],216,[*[*2*],[*2*],[*23*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],216,[*[*2*],[*2*],[*2*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],216,[*[*3*],[*5*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
216,[*[*3*],[*53*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53*]*]*],[*[*"A",1*],216,[*[*5*],[*35*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],216,[*[*5*],[*5*],[*5*]*],[*[*
2,3,5*]*]*],[*[*"A",1*],216,[*[*7*],[*26*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"A",1*],216,[*[*8*],[*23*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],216,[*[*11*],[*17*]*],[*[*2,3,5,7,
11,13,17*]*]*],[*[*"A",1*],216,[*[*215*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193,197,199,211*]*]*],[*[*"C",108*],216,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",108*],216,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",215*],
216,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",215*],216,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],217,[*[*8,8*]*],17*],[*[*"A",
1*],217,[*[*6*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*
"A",1*],217,[*[*216*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211*]*]*],[*[*"B",108*],217,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",216*],217,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",216*],
217,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],218,[*[*1
*],[*108*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107*]*]*],[*[*"A",1*],218,[*
[*217*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211*]*]*],[*[*
"C",109*],218,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",109*],218,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",217*],218,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",217*],218,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],219,[*[*2*],[*72*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*[*"A",1*],219,[*[*
218*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211*]*]*],[*[*
"B",109*],219,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",218*],219,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",218*],219,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",11*],220,[*[*0,0,0,0,0,
0,0,0,1,0,0*]*],[*[**]*]*],[*[*"A",11*],220,[*[*0,0,1,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],220,[*[*0,0,9*]*],[*[*2,3,5,7
*]*]*],[*[*"A",3*],220,[*[*1,2,2*]*],5*],[*[*"A",3*],220,[*[*2,2,1
*]*],5*],[*[*"A",3*],220,[*[*9,0,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",9
*],220,[*[*0,0,0,0,0,0,0,0,3*]*],[*[*2,3*]*]*],[*[*"A",9*],220,
[*[*3,0,0,0,0,0,0,0,0*]*],[*[*2,3*]*]*],[*[*"B",2*],220,[*[*1,0
*],[*0,4*]*],[*[*2,3,7*]*]*],[*[*"B",2*],220,[*[*0,1*],[*3,1*]*],7
*],[*[*"B",2*],220,[*[*4,2*]*],[*[*2,3,5,7*]*]*],[*[*"B",2*],220,[*
[*9,0*]*],[*[*2,3,5,7*]*]*],[*[*"C",5*],220,[*[*0,0,0,0,3*]*],[*
[*2,3*]*]*],[*[*"D",6*],220,[*[*0,0,0,1,0,0*]*],[*[*2*]*]*],[*[*
"A",1*],220,[*[*1*],[*4*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],
[*[*"A",1*],220,[*[*1*],[*9*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1
*],220,[*[*1*],[*109*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109*]*]*],
[*[*"A",1*],220,[*[*1*],[*1*],[*4*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],220,[*[*1*],[*1*],[*54*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53*]*]*],[*[*"A",1*],220,[*[*3*],[*4*],[*10*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],220,[*[*3*],[*54*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],220,[*[*
4*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*
[*"A",1*],220,[*[*9*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],220,[*[*10*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],220,[*[*219*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211*]*]*],[*[*"C",110*],220,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",110*],220,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",219*],220,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",219*],220,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],221,[*[*12*],[*
16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],221,[*[*220*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,
163,167,173,179,181,191,193,197,199,211*]*]*],[*[*"B",110*],221,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",220*],221,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",220*],221,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],222,[*[*0,2*],[*3,3*]
*],7*],[*[*"A",2*],222,[*[*2,0*],[*3,3*]*],7*],[*[*"A",2*],222,[*[*
2,16*]*],19*],[*[*"A",2*],222,[*[*16,2*]*],19*],[*[*"A",1*],222,[*
[*1*],[*2*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],222,[*[*1*],[*110*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109*]*]
*],[*[*"A",1*],222,[*[*2*],[*73*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],222,[*[*5
*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],
222,[*[*221*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211*]*]
*],[*[*"C",111*],222,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",111*],
222,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",221*],222,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",221*],222,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",14*],223,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],3*],[*[*"A",14*],223,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],5*],[*[*"A",1*],223,[*[*222*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211*]*]*],[*[*"B",111*],
223,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",222*],223,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",222*],223,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",14*],224,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*3,5*]*]*],[*[*"A",2
*],224,[*[*1,1*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],224,[*[*1,1
*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],224,[*[*1,13*]*],[*[*2,3,
5,7,11,13*]*]*],[*[*"A",2*],224,[*[*13,1*]*],[*[*2,3,5,7,11,13*]
*]*],[*[*"A",3*],224,[*[*0,0,1*],[*1,0,0*],[*1,0,1*]*],2*],[*[*
"A",3*],224,[*[*0,0,1*],[*0,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",3*],
224,[*[*0,0,1*],[*5,0,0*]*],[*[*2,3,5*]*]*],[*[*"A",3*],224,[*[*
0,0,1*],[*0,0,1*],[*1,0,1*]*],2*],[*[*"A",3*],224,[*[*1,0,0*],
[*0,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",3*],224,[*[*1,0,0*],[*5,0,0
*]*],[*[*2,3,5*]*]*],[*[*"A",3*],224,[*[*1,0,0*],[*1,0,0*],[*1,0,
1*]*],2*],[*[*"A",3*],224,[*[*0,2,3*]*],[*[*2,3,5*]*]*],[*[*"A",3
*],224,[*[*0,5,1*]*],7*],[*[*"A",3*],224,[*[*1,5,0*]*],7*],[*[*
"A",3*],224,[*[*3,2,0*]*],[*[*2,3,5*]*]*],[*[*"A",4*],224,[*[*0,
0,1,3*]*],[*[*2,3,5*]*]*],[*[*"A",4*],224,[*[*3,1,0,0*]*],[*[*2,
3,5*]*]*],[*[*"A",6*],224,[*[*0,0,1,0,0,1*]*],[*[*5*]*]*],[*[*"A",
6*],224,[*[*1,0,0,1,0,0*]*],[*[*5*]*]*],[*[*"A",7*],224,[*[*0,0,
0,0,0,0,1*],[*0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",7*],224,[*[*
0,0,0,0,0,0,1*],[*0,1,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",7*],
224,[*[*1,0,0,0,0,0,0*],[*0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"A",
7*],224,[*[*1,0,0,0,0,0,0*],[*0,1,0,0,0,0,0*]*],[*[**]*]*],[*[*
"B",2*],224,[*[*1,0*],[*5,0*]*],[*[*2,3,5*]*]*],[*[*"B",2*],224,[*
[*1,0*],[*1,0*],[*0,2*]*],[*[*2,5*]*]*],[*[*"B",2*],224,[*[*0,2*],
[*1,1*]*],[*[*2,5*]*]*],[*[*"B",2*],224,[*[*1,5*]*],[*[*2,3,5,11,
13*]*]*],[*[*"D",4*],224,[*[*0,0,0,1*],[*0,0,1,0*]*],[*[*2*]*]*],
[*[*"D",4*],224,[*[*0,1,0,0*],[*0,0,1,0*]*],[*[*2*]*]*],[*[*"D",4
*],224,[*[*1,0,0,0*],[*0,0,1,0*]*],[*[*2*]*]*],[*[*"D",4*],224,[*
[*0,1,0,2*]*],[*[*2,5*]*]*],[*[*"D",4*],224,[*[*0,2,0,1*]*],[*[*
2,5*]*]*],[*[*"D",4*],224,[*[*1,0,0,2*]*],[*[*2,5*]*]*],[*[*"D",4
*],224,[*[*1,2,0,0*]*],[*[*2,5*]*]*],[*[*"D",4*],224,[*[*2,0,0,1
*]*],[*[*2,5*]*]*],[*[*"D",4*],224,[*[*2,1,0,0*]*],[*[*2,5*]*]*],[*
[*"A",1*],224,[*[*1*],[*3*],[*27*]*],[*[*2,3,5,7,11,13,17,19,23
*]*]*],[*[*"A",1*],224,[*[*1*],[*3*],[*3*],[*6*]*],[*[*2,3,5*]*]*],[*
[*"A",1*],224,[*[*1*],[*6*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],224,[*[*1*],[*7*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],224,[*[*1*],[*111*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109*]*]
*],[*[*"A",1*],224,[*[*1*],[*1*],[*3*],[*13*]*],[*[*2,3,5,7,11,13
*]*]*],[*[*"A",1*],224,[*[*1*],[*1*],[*6*],[*7*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],224,[*[*1*],[*1*],[*55*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],224,[*[*1*],[*1*],
[*1*],[*3*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],224,[*[*1*],[*1*],[*
1*],[*27*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],224,[*
[*1*],[*1*],[*1*],[*1*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],224,[*[*1*],[*1*],[*1*],[*1*],[*1*],[*6*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],224,[*[*3*],[*6*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
224,[*[*3*],[*55*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53*]*]*],[*[*"A",1*],224,[*[*3*],[*3*],[*13*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],224,[*[*6*],[*31*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],224,[*[*7*],[*27*]*],[*[*2,3,5,
7,11,13,17,19,23*]*]*],[*[*"A",1*],224,[*[*13*],[*15*]*],[*[*2,3,
5,7,11,13*]*]*],[*[*"A",1*],224,[*[*223*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223*]*]*],[*[*"C",112*],224,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",112*],224,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*],[*[*"A",223*],224,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",223*],224,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",14*],225,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",14*],225,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,1*],[*0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",14*],225,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0*],[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0*]*],[*[**]*]*],[*[*"A",2*],225,[*[*0,1*],[*1,9*]*],11*],[*[*"A",
2*],225,[*[*0,1*],[*3,4*]*],7*],[*[*"A",2*],225,[*[*0,1*],[*4,3*]
*],7*],[*[*"A",2*],225,[*[*0,1*],[*9,1*]*],11*],[*[*"A",2*],225,[*
[*1,0*],[*1,9*]*],11*],[*[*"A",2*],225,[*[*1,0*],[*3,4*]*],7*],[*
[*"A",2*],225,[*[*1,0*],[*4,3*]*],7*],[*[*"A",2*],225,[*[*1,0*],[*
9,1*]*],11*],[*[*"A",2*],225,[*[*0,4*],[*1,2*]*],[*[*2,3*]*]*],[*[*
"A",2*],225,[*[*0,4*],[*2,1*]*],[*[*2,3*]*]*],[*[*"A",2*],225,[*[*
0,4*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],225,[*[*0,4*],[*0,4*]*],
[*[*2,3*]*]*],[*[*"A",2*],225,[*[*1,2*],[*2,1*]*],[*[*2*]*]*],[*[*
"A",2*],225,[*[*1,2*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],225,[*[*
1,2*],[*1,2*]*],[*[*2*]*]*],[*[*"A",2*],225,[*[*2,1*],[*4,0*]*],[*
[*2,3*]*]*],[*[*"A",2*],225,[*[*2,1*],[*2,1*]*],[*[*2*]*]*],[*[*"A",
2*],225,[*[*4,0*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",3*],225,[*[*1,0,
1*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",4*],225,[*[*0,0,0,1*],[*0,0,
2,0*]*],3*],[*[*"A",4*],225,[*[*0,0,0,1*],[*0,1,0,1*]*],[*[*2*]
*]*],[*[*"A",4*],225,[*[*0,0,0,1*],[*0,2,0,0*]*],3*],[*[*"A",4*],
225,[*[*0,0,0,1*],[*1,0,1,0*]*],[*[*2*]*]*],[*[*"A",4*],225,[*[*
1,0,0,0*],[*0,0,2,0*]*],3*],[*[*"A",4*],225,[*[*1,0,0,0*],[*0,
1,0,1*]*],[*[*2*]*]*],[*[*"A",4*],225,[*[*1,0,0,0*],[*0,2,0,0*]
*],3*],[*[*"A",4*],225,[*[*1,0,0,0*],[*1,0,1,0*]*],[*[*2*]*]*],[*
[*"A",4*],225,[*[*0,0,0,2*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",4
*],225,[*[*0,0,0,2*],[*0,0,0,2*]*],[*[*2*]*]*],[*[*"A",4*],225,[*
[*2,0,0,0*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",5*],225,[*[*0,0,0,
1,0*],[*0,1,0,0,0*]*],[*[**]*]*],[*[*"A",5*],225,[*[*0,0,0,1,0
*],[*0,0,0,1,0*]*],[*[**]*]*],[*[*"A",5*],225,[*[*0,1,0,0,0*],[*
0,1,0,0,0*]*],[*[**]*]*],[*[*"B",7*],225,[*[*0,0,0,0,0,0,1*],[*
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",1*],225,[*[*2*],[*4*],[*14
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],225,[*[*2*],[*74*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73
*]*]*],[*[*"A",1*],225,[*[*2*],[*2*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],225,[*[*2*],[*2*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]
*]*],[*[*"A",1*],225,[*[*4*],[*44*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43*]*]*],[*[*"A",1*],225,[*[*4*],[*4*],[*8*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],225,[*[*8*],[*24*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"A",1*],225,[*[*14*],[*14*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"A",1*],225,[*[*224*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223*]*]*],[*[*"B",112*],225,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",224*],225,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",224*],225,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],226,[*[*1*],[*112*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109*]*]*],[*[*"A",1*],226,[*[*225*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223*]*]*],[*[*"C",113
*],226,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",113*],226,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",225*],226,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",225*],226,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
1*],227,[*[*226*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,223
*]*]*],[*[*"B",113*],227,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",226*],227,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]
*],[*[**]*]*],[*[*"A",226*],227,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],228,[*[*1*],[*2*],[*37*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],228,[*[*1
*],[*5*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],228,[*[*1
*],[*113*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1
*],228,[*[*1*],[*1*],[*2*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*
"A",1*],228,[*[*1*],[*1*],[*56*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53*]*]*],[*[*"A",1*],228,[*[*2*],[*3*],[*18*]*],
[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],228,[*[*2*],[*75*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]
*]*],[*[*"A",1*],228,[*[*3*],[*56*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53*]*]*],[*[*"A",1*],228,[*[*5*],[*37*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],228,[*[*11*],
[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],228,[*[*227*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227*]*]*],[*[*
"C",114*],228,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",114
*],228,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",227*],228,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",227*],228,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"B",10*],229,[*[*0,0,0,0,0,0,0,0,0,2*]*],3
*],[*[*"B",10*],229,[*[*0,0,0,0,0,0,0,0,0,2*]*],7*],[*[*"C",11
*],229,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],11*],[*[*"A",1*],229,[*[*
228*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211,223,227*]*]
*],[*[*"B",114*],229,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",228*],229,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",228*],229,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],230,[*[*0,0,1,0
*],[*1,0,0,1*]*],5*],[*[*"A",4*],230,[*[*0,1,0,0*],[*1,0,0,1*]
*],5*],[*[*"B",10*],230,[*[*0,0,0,0,0,0,0,0,0,2*]*],[*[*2,3,7
*]*]*],[*[*"B",11*],230,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],2*],[*[*
"C",11*],230,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],[*[*11*]*]*],[*[*
"D",11*],230,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],2*],[*[*"A",1*],
230,[*[*1*],[*4*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",
1*],230,[*[*1*],[*114*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]
*]*],[*[*"A",1*],230,[*[*4*],[*45*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43*]*]*],[*[*"A",1*],230,[*[*9*],[*22*]*],[*[*2,3,5,
7,11,13,17,19*]*]*],[*[*"A",1*],230,[*[*229*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229*]*]*],[*[*"C",115*],230,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",115*],230,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",229*],230,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
229*],230,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",20*],231,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",20*],231,[*[*2,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",2*],
231,[*[*0,20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",2*],231,[*
[*2,10*]*],[*[*2,3,5,7,13*]*]*],[*[*"A",2*],231,[*[*3,10*]*],13*],
[*[*"A",2*],231,[*[*10,2*]*],[*[*2,3,5,7,13*]*]*],[*[*"A",2*],231,
[*[*10,3*]*],13*],[*[*"A",2*],231,[*[*20,0*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",3*],231,[*[*0,6,0*]*],7*],[*[*"B",2*],231,[*[*
6,1*]*],[*[*2,3,5*]*]*],[*[*"B",4*],231,[*[*0,0,1,1*]*],[*[*2,3*]
*]*],[*[*"D",11*],231,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],[*[*2*]*]*]
,[*[*"A",1*],231,[*[*2*],[*6*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",
1*],231,[*[*2*],[*76*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],231,[*[*6*],[*32*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],231,[*[*10
*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],231,[*[*230
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167,173,179,181,191,193,197,199,211,223,227,229*]*]
*],[*[*"A",20*],231,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",20*],231,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,2*]*],[*[**]*]*],[*[*"A",21*],231,[*[*0,1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
21*],231,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]
*],[*[**]*]*],[*[*"B",115*],231,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[*2*]*]*],[*[*"A",230*],231,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",230*],231,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
3*],232,[*[*0,0,1*],[*1,1,1*]*],5*],[*[*"A",3*],232,[*[*1,0,0*],
[*1,1,1*]*],5*],[*[*"A",1*],232,[*[*1*],[*3*],[*28*]*],[*[*2,3,5,
7,11,13,17,19,23*]*]*],[*[*"A",1*],232,[*[*1*],[*115*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],232,[*[*1*],[*1*],
[*57*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53*]*]
*],[*[*"A",1*],232,[*[*1*],[*1*],[*1*],[*28*]*],[*[*2,3,5,7,11,13,
17,19,23*]*]*],[*[*"A",1*],232,[*[*3*],[*57*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],232,[*[*7*],
[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],232,[*[*231
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167,173,179,181,191,193,197,199,211,223,227,229*]*]
*],[*[*"C",116*],232,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"D",116*],232,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",231*],232,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",231*],232,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],
233,[*[*232*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229*]*]*],[*[*"B",116*],233,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",232*],233,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",232*],233,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",2*],234,[*[*0,1*],[*0,11*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",2*],234,[*[*0,1*],[*11,0*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",2*],234,[*[*1,0*],[*0,11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2
*],234,[*[*1,0*],[*11,0*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],234,
[*[*0,2*],[*2,3*]*],5*],[*[*"A",2*],234,[*[*0,2*],[*3,2*]*],5*],[*
[*"A",2*],234,[*[*2,0*],[*2,3*]*],5*],[*[*"A",2*],234,[*[*2,0*],[*
3,2*]*],5*],[*[*"A",2*],234,[*[*3,8*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",2*],234,[*[*8,3*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],234,[*[*
1*],[*2*],[*38*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*
[*"A",1*],234,[*[*1*],[*2*],[*2*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*
[*"A",1*],234,[*[*1*],[*8*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",
1*],234,[*[*1*],[*116*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]
*]*],[*[*"A",1*],234,[*[*2*],[*5*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*
[*"A",1*],234,[*[*2*],[*77*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],234,[*[*2*],
[*2*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],234,
[*[*5*],[*38*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*
"A",1*],234,[*[*8*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],234,[*[*12*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*
"A",1*],234,[*[*233*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233*]*]*],[*[*"C",117*],234,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",117*],234,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",233*],234,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",233*],234,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],235,[*[*3,15*]*],19*],[*
[*"A",2*],235,[*[*15,3*]*],19*],[*[*"A",3*],235,[*[*0,3,2*]*],5*],
[*[*"A",3*],235,[*[*2,3,0*]*],5*],[*[*"A",4*],235,[*[*0,1,1,1*]
*],5*],[*[*"A",4*],235,[*[*1,1,1,0*]*],5*],[*[*"A",1*],235,[*[*4
*],[*46*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*
"A",1*],235,[*[*234*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233*]*]*],[*[*"B",117*],235,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",234*],235,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",234*],235,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],236,[*[*2,1,2*]*],
7*],[*[*"B",2*],236,[*[*5,3*]*],13*],[*[*"A",1*],236,[*[*1*],[*117
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],236,
[*[*1*],[*1*],[*58*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53*]*]*],[*[*"A",1*],236,[*[*3*],[*58*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],236,[*[*235*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233*]*]
*],[*[*"C",118*],236,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",118*],236,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"A",235*],236,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",235*],236,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",1*],237,[*[*2*],[*78*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73*]*]*],[*[*"A",
1*],237,[*[*236*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233*]*]*],[*[*"B",118*],237,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",236*],237,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",236*],237,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],238,[*[*1*],[*
6*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],238,[*[*1*],[*118
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],238,
[*[*6*],[*33*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",
1*],238,[*[*13*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],238,
[*[*237*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149,151,157,163,167,173,179,181,191,193,197,199,211,223,
227,229,233*]*]*],[*[*"C",119*],238,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",119*],238,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",237*],238,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",237*],238,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],239,[*[*
238*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211,223,227,
229,233*]*]*],[*[*"B",119*],239,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",238*],239,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",238*],239,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],240,[*[*
0,1*],[*1,1*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],240,[*[*0,1*],[*
1,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],240,[*[*0,1*],[*1,7*]*],
[*[*2,3,5,7*]*]*],[*[*"A",2*],240,[*[*0,1*],[*7,1*]*],[*[*2,3,5,
7*]*]*],[*[*"A",2*],240,[*[*1,0*],[*1,1*],[*0,3*]*],[*[*2,3*]*]*],
[*[*"A",2*],240,[*[*1,0*],[*1,1*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",
2*],240,[*[*1,0*],[*1,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],240,[*[*
1,0*],[*7,1*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],240,[*[*0,3*],[*1,
3*]*],[*[*2,3,5*]*]*],[*[*"A",2*],240,[*[*0,3*],[*3,1*]*],[*[*2,3,
5*]*]*],[*[*"A",2*],240,[*[*3,0*],[*1,3*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],240,[*[*3,0*],[*3,1*]*],[*[*2,3,5*]*]*],[*[*"A",3*],240,[*
[*0,0,1*],[*1,0,0*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*
0,0,1*],[*0,1,0*],[*0,0,2*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*0,
0,1*],[*0,1,0*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*0,0,
1*],[*0,2,1*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*0,0,1*],[*1,2,0*]
*],[*[*2*]*]*],[*[*"A",3*],240,[*[*0,0,1*],[*0,0,1*],[*1,0,1*]*],
[*[*2*]*]*],[*[*"A",3*],240,[*[*1,0,0*],[*0,1,0*],[*0,0,2*]*],[*[*
2*]*]*],[*[*"A",3*],240,[*[*1,0,0*],[*0,1,0*],[*2,0,0*]*],[*[*2*]
*]*],[*[*"A",3*],240,[*[*1,0,0*],[*0,2,1*]*],[*[*2*]*]*],[*[*"A",3
*],240,[*[*1,0,0*],[*1,2,0*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*1,
0,0*],[*1,0,0*],[*1,0,1*]*],[*[*2*]*]*],[*[*"A",3*],240,[*[*1,0,
1*],[*0,1,1*]*],3*],[*[*"A",3*],240,[*[*1,0,1*],[*1,1,0*]*],3*],
[*[*"A",4*],240,[*[*0,0,1,0*],[*1,0,0,1*]*],[*[*5*]*]*],[*[*"A",4
*],240,[*[*0,1,0,0*],[*1,0,0,1*]*],[*[*5*]*]*],[*[*"A",8*],240,[*
[*0,0,0,0,0,0,1,1*]*],[*[*3*]*]*],[*[*"A",8*],240,[*[*1,1,0,0,
0,0,0,0*]*],[*[*3*]*]*],[*[*"B",2*],240,[*[*1,0*],[*0,1*],[*1,1*]
*],5*],[*[*"B",2*],240,[*[*1,0*],[*1,4*]*],11*],[*[*"B",2*],240,[*
[*2,0*],[*1,2*]*],7*],[*[*"B",2*],240,[*[*1,1*],[*3,0*]*],5*],[*[*
"C",4*],240,[*[*1,0,0,1*]*],7*],[*[*"A",1*],240,[*[*1*],[*2*],[*3
*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],240,[*[*1*],[*2*],[*4*],[*
7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],240,[*[*1*],[*2*],[*39*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],240,[*[*1*],
[*3*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],240,[*[*1*],[*3*],[*
29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],240,[*[*1
*],[*4*],[*23*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
240,[*[*1*],[*5*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",
1*],240,[*[*1*],[*7*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],240,[*[*1*],[*9*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],240,
[*[*1*],[*119*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*
"A",1*],240,[*[*1*],[*1*],[*2*],[*3*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],240,[*[*1*],[*1*],[*2*],[*19*]*],[*[*2,3,5,7,11,13,17,19
*]*]*],[*[*"A",1*],240,[*[*1*],[*1*],[*3*],[*14*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"A",1*],240,[*[*1*],[*1*],[*4*],[*11*]*],[*[*2,3,5,7,
11*]*]*],[*[*"A",1*],240,[*[*1*],[*1*],[*5*],[*9*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],240,[*[*1*],[*1*],[*59*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59*]*]*],[*[*"A",1*],240,[*[*1*],[*1
*],[*1*],[*2*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],240,[*[*1*],[*
1*],[*1*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],240,[*[*1*],[*1
*],[*1*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1
*],240,[*[*1*],[*1*],[*1*],[*1*],[*2*],[*4*]*],[*[*2,3*]*]*],[*[*"A",
1*],240,[*[*1*],[*1*],[*1*],[*1*],[*14*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],240,[*[*2*],[*3*],[*19*]*],[*[*2,3,5,7,11,13,17,19
*]*]*],[*[*"A",1*],240,[*[*2*],[*3*],[*3*],[*4*]*],[*[*2,3*]*]*],[*[*
"A",1*],240,[*[*2*],[*4*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*
"A",1*],240,[*[*2*],[*7*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
240,[*[*2*],[*79*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79*]*]*],[*[*"A",1*],240,[*[*3*],[*4*],[*
11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],240,[*[*3*],[*5*],[*9*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],240,[*[*3*],[*59*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59*]*]*],[*[*"A",1*],240,[*
[*3*],[*3*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],240,[*[*4
*],[*5*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],240,[*[*4*],[*47*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1
*],240,[*[*5*],[*39*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]
*],[*[*"A",1*],240,[*[*7*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,
29*]*]*],[*[*"A",1*],240,[*[*9*],[*23*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],240,[*[*11*],[*19*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",1*],240,[*[*14*],[*15*]*],[*[*2,3,5,7,11,13*]
*]*],[*[*"A",1*],240,[*[*239*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239*]*]*],[*[*"C",120*],240,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",120*],
240,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",239*],240,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",239*],240,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],241,[*[*240*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239*]*]*],[*[*
"B",120*],241,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",240*],241,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",240
*],241,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],242,[*[*1*],[*
10*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],242,[*[*1*],[*120*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],242,[*[*10
*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],242,[*[*241
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167,173,179,181,191,193,197,199,211,223,227,229,
233,239,241*]*]*],[*[*"C",121*],242,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",121*],242,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",241*],242,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",241*],242,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",2*],243,[*[*0,1*],[*1,0*],[*2,2*]*],[*[*2,5*]*]
*],[*[*"A",2*],243,[*[*0,1*],[*1,0*],[*1,0*],[*1,0*],[*1,0*]*],[*
[**]*]*],[*[*"A",2*],243,[*[*0,1*],[*2,5*]*],[*[*2,5,7*]*]*],[*[*
"A",2*],243,[*[*0,1*],[*5,2*]*],[*[*2,5,7*]*]*],[*[*"A",2*],243,[*
[*0,1*],[*0,1*],[*1,0*],[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],
243,[*[*0,1*],[*0,1*],[*2,2*]*],[*[*2,5*]*]*],[*[*"A",2*],243,[*[*
0,1*],[*0,1*],[*0,1*],[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],
243,[*[*0,1*],[*0,1*],[*0,1*],[*0,1*],[*1,0*]*],[*[**]*]*],[*[*"A",
2*],243,[*[*0,1*],[*0,1*],[*0,1*],[*0,1*],[*0,1*]*],[*[**]*]*],[*[*
"A",2*],243,[*[*1,0*],[*2,5*]*],[*[*2,5,7*]*]*],[*[*"A",2*],243,[*
[*1,0*],[*5,2*]*],[*[*2,5,7*]*]*],[*[*"A",2*],243,[*[*1,0*],[*1,0
*],[*2,2*]*],[*[*2,5*]*]*],[*[*"A",2*],243,[*[*1,0*],[*1,0*],[*1,0
*],[*1,0*],[*1,0*]*],[*[**]*]*],[*[*"A",2*],243,[*[*4,9*]*],13*],[*[*
"A",2*],243,[*[*9,4*]*],13*],[*[*"A",1*],243,[*[*2*],[*8*],[*8*]*],
[*[*2,3,5,7*]*]*],[*[*"A",1*],243,[*[*2*],[*80*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79*]*]*],
[*[*"A",1*],243,[*[*2*],[*2*],[*26*]*],[*[*2,3,5,7,11,13,17,19,
23*]*]*],[*[*"A",1*],243,[*[*2*],[*2*],[*2*],[*8*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],243,[*[*2*],[*2*],[*2*],[*2*],[*2*]*],[*[*2*]*]*],[*
[*"A",1*],243,[*[*8*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],
[*[*"A",1*],243,[*[*242*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,
197,199,211,223,227,229,233,239,241*]*]*],[*[*"B",121*],243,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",242
*],243,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",242*],243,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",2*],244,[*[*1,0*],[*0,5
*]*],11*],[*[*"G",2*],244,[*[*3,1*]*],13*],[*[*"A",1*],244,[*[*1*],
[*121*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],
244,[*[*1*],[*1*],[*60*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59*]*]*],[*[*"A",1*],244,[*[*3*],[*60*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59*]*]*],[*[*"A",1*],
244,[*[*243*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241*]*]*],[*[*"C",122*],244,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",122*],244,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",243
*],244,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",243*],244,[*[*
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",6*],245,[*[*0,0,0,
0,0,1*],[*0,0,0,1,0,0*]*],[*[**]*]*],[*[*"A",6*],245,[*[*0,0,0,
0,0,1*],[*0,0,1,0,0,0*]*],[*[**]*]*],[*[*"A",6*],245,[*[*1,0,0,
0,0,0*],[*0,0,0,1,0,0*]*],[*[**]*]*],[*[*"A",6*],245,[*[*1,0,0,
0,0,0*],[*0,0,1,0,0,0*]*],[*[**]*]*],[*[*"B",3*],245,[*[*0,0,1
*],[*2,0,0*]*],[*[*2*]*]*],[*[*"A",1*],245,[*[*4*],[*6*],[*6*]*],[*[*
2,3,5*]*]*],[*[*"A",1*],245,[*[*4*],[*48*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],245,[*[*6*],[*34*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],245,[*[*
244*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211,223,227,
229,233,239,241*]*]*],[*[*"B",122*],245,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",244*],245,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",244*],245,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],246,[*[*0,1*],[*2,8*]*],11
*],[*[*"A",2*],246,[*[*0,1*],[*8,2*]*],11*],[*[*"A",2*],246,[*[*1,
0*],[*2,8*]*],11*],[*[*"A",2*],246,[*[*1,0*],[*8,2*]*],11*],[*[*
"A",2*],246,[*[*4,14*]*],19*],[*[*"A",2*],246,[*[*14,4*]*],19*],[*
[*"A",5*],246,[*[*0,0,0,1,3*]*],5*],[*[*"A",5*],246,[*[*0,1,0,
0,2*]*],3*],[*[*"A",5*],246,[*[*2,0,0,1,0*]*],3*],[*[*"A",5*],
246,[*[*3,1,0,0,0*]*],5*],[*[*"B",4*],246,[*[*0,1,0,1*]*],2*],[*
[*"C",3*],246,[*[*0,1,3*]*],5*],[*[*"C",4*],246,[*[*0,1,0,1*]*],
2*],[*[*"D",4*],246,[*[*1,1,0,1*]*],2*],[*[*"F",4*],246,[*[*0,0,
1,0*]*],2*],[*[*"F",4*],246,[*[*0,1,0,0*]*],2*],[*[*"A",1*],246,
[*[*1*],[*2*],[*40*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]
*],[*[*"A",1*],246,[*[*1*],[*122*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113*]*]*],[*[*"A",1*],246,[*[*2*],[*81*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79*]*]*],[*[*
"A",1*],246,[*[*5*],[*40*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37*]*]*],[*[*"A",1*],246,[*[*245*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241*]*]*],[*[*"C",123*],246,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",123*],246,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",245*],246,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",245*],246,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],247,[*[*12*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]
*],[*[*"A",1*],247,[*[*246*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,
197,199,211,223,227,229,233,239,241*]*]*],[*[*"B",123*],247,[*[*0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*
"A",246*],247,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",246*],247,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"B",3*],248,[*[*1,2,0*]*],7*],[*[*"E",8*],248,[*[*0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"G",2*],248,[*[*2,1*]*],7*],[*[*"A",1*],
248,[*[*1*],[*3*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*
[*"A",1*],248,[*[*1*],[*123*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113*]*]*],[*[*"A",1*],248,[*[*1*],[*1*],[*61*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],
248,[*[*1*],[*1*],[*1*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]
*]*],[*[*"A",1*],248,[*[*3*],[*61*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],248,[*[*7*],[*30*]
*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],248,[*[*247*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241*]*]*],[*[*"C",124*],248,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",124*],248,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",247*],
248,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",247*],
248,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1
*],249,[*[*2*],[*82*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79*]*]*],[*[*"A",1*],249,[*[*248*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241*]*]*],[*[*"B",124*],249,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",248*],249,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",248*],249,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",4*],250,[*[*0,
0,0,1*],[*1,0,0,0*],[*0,0,1,0*]*],[*[**]*]*],[*[*"A",4*],250,[*[*
0,0,0,1*],[*1,0,0,0*],[*0,1,0,0*]*],[*[**]*]*],[*[*"A",4*],250,
[*[*0,0,0,1*],[*0,0,2,0*]*],[*[*2,3*]*]*],[*[*"A",4*],250,[*[*0,
0,0,1*],[*0,2,0,0*]*],[*[*2,3*]*]*],[*[*"A",4*],250,[*[*0,0,0,1
*],[*0,0,0,1*],[*0,0,1,0*]*],[*[**]*]*],[*[*"A",4*],250,[*[*0,0,
0,1*],[*0,0,0,1*],[*0,1,0,0*]*],[*[**]*]*],[*[*"A",4*],250,[*[*1,
0,0,0*],[*0,0,2,0*]*],[*[*2,3*]*]*],[*[*"A",4*],250,[*[*1,0,0,0
*],[*0,2,0,0*]*],[*[*2,3*]*]*],[*[*"A",4*],250,[*[*1,0,0,0*],[*1,
0,0,0*],[*0,0,1,0*]*],[*[**]*]*],[*[*"A",4*],250,[*[*1,0,0,0*],[*
1,0,0,0*],[*0,1,0,0*]*],[*[**]*]*],[*[*"B",2*],250,[*[*0,1*],[*0,
1*],[*2,0*]*],[*[*2*]*]*],[*[*"B",2*],250,[*[*2,0*],[*0,3*]*],7*],
[*[*"B",2*],250,[*[*2,0*],[*2,1*]*],3*],[*[*"A",1*],250,[*[*1*],[*
4*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],250,[*
[*1*],[*4*],[*4*],[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],250,[*[*1*],[*
124*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],
250,[*[*4*],[*49*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47*]*]*],[*[*"A",1*],250,[*[*4*],[*4*],[*9*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],250,[*[*9*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]
*]*],[*[*"A",1*],250,[*[*249*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241*]*]*],[*[*"C",125*],250,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*],[*[*"D",125*],250,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",249*],250,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",249*],250,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],251,[*[*5,8*]*],13
*],[*[*"A",2*],251,[*[*8,5*]*],13*],[*[*"B",2*],251,[*[*2,5*]*],13
*],[*[*"D",11*],251,[*[*0,0,0,0,0,0,0,0,0,0,2*]*],11*],[*[*"A",
1*],251,[*[*250*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241*]*]*],[*[*"B",125*],251,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",250*],
251,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",250*],251,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",2*],252,[*[*0,1*],[*1,0*],[*0,6*]*],[*[*2,3,5*]*]
*],[*[*"A",2*],252,[*[*0,1*],[*1,0*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],252,[*[*0,1*],[*0,1*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2
*],252,[*[*0,1*],[*0,1*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],
252,[*[*1,0*],[*1,0*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],252,[*
[*1,0*],[*1,0*],[*6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],252,[*[*0,2
*],[*2,0*],[*1,1*]*],3*],[*[*"A",2*],252,[*[*0,2*],[*2,3*]*],[*[*
2,3,5*]*]*],[*[*"A",2*],252,[*[*0,2*],[*3,2*]*],[*[*2,3,5*]*]*],[*
[*"A",2*],252,[*[*0,2*],[*0,2*],[*1,1*]*],3*],[*[*"A",2*],252,[*[*
2,0*],[*2,3*]*],[*[*2,3,5*]*]*],[*[*"A",2*],252,[*[*2,0*],[*3,2*]
*],[*[*2,3,5*]*]*],[*[*"A",2*],252,[*[*2,0*],[*2,0*],[*1,1*]*],3
*],[*[*"A",2*],252,[*[*3,9*]*],11*],[*[*"A",2*],252,[*[*9,3*]*],11
*],[*[*"A",5*],252,[*[*0,0,0,0,5*]*],[*[*2,3,5*]*]*],[*[*"A",5*],
252,[*[*5,0,0,0,0*]*],[*[*2,3,5*]*]*],[*[*"A",9*],252,[*[*0,0,0,
0,1,0,0,0,0*]*],[*[**]*]*],[*[*"C",3*],252,[*[*0,0,5*]*],[*[*2,3,
5*]*]*],[*[*"D",11*],252,[*[*0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2,
11*]*]*],[*[*"A",1*],252,[*[*1*],[*2*],[*5*],[*6*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],252,[*[*1*],[*2*],[*41*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41*]*]*],[*[*"A",1*],252,[*[*1*],[*2*],[*2*],[*13*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],252,[*[*1*],[*5*],[*20*]*],
[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],252,[*[*1*],[*6*],[*17
*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],252,[*[*1*],[*8*],[*13
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],252,[*[*1*],[*125*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113*]*]*],[*[*"A",1*],252,[*[*1*],
[*1*],[*2*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],
252,[*[*1*],[*1*],[*2*],[*2*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],
252,[*[*1*],[*1*],[*6*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],252,
[*[*1*],[*1*],[*62*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61*]*]*],[*[*"A",1*],252,[*[*2*],[*3*],[*20*]*],[*[*2,
3,5,7,11,13,17,19*]*]*],[*[*"A",1*],252,[*[*2*],[*5*],[*13*]*],[*
[*2,3,5,7,11,13*]*]*],[*[*"A",1*],252,[*[*2*],[*6*],[*11*]*],[*[*
2,3,5,7,11*]*]*],[*[*"A",1*],252,[*[*2*],[*83*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]
*],[*[*"A",1*],252,[*[*2*],[*2*],[*3*],[*6*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],252,[*[*2*],[*2*],[*27*]*],[*[*2,3,5,7,11,13,17,19,23*]
*]*],[*[*"A",1*],252,[*[*3*],[*6*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],252,[*[*3*],[*62*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61*]*]*],[*[*"A",1*],252,[*[*5*],[*41*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],252,[*[*5*],
[*5*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],252,[*[*6*],[*35*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],252,[*[*8*],[*27
*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],252,[*[*11*],[*
20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],252,[*[*13*],[*
17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],252,[*[*251*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241,251*]*]*],[*[*"C",126*],252,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",126*],252,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"A",251*],252,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",251*],252,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],253,[*[*0,21*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",2*],253,[*[*21,0*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"B",11*],253,[*[*0,0,0,0,0,0,0,0,0,1,0*]*],[*[*
2*]*]*],[*[*"C",11*],253,[*[*0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2*]
*]*],[*[*"A",1*],253,[*[*10*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]
*]*],[*[*"A",1*],253,[*[*252*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251*]*]*],[*[*"A",21*],
253,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",21*],253,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,2*]*],[*[**]*]*],[*[*"A",22*],253,[*[*0,1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",22*],
253,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],
[*[**]*]*],[*[*"B",126*],253,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",252*],253,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",252*],253,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*
[*"A",15*],254,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],2*],[*
[*"A",1*],254,[*[*1*],[*126*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113*]*]*],[*[*"A",1*],254,[*[*253*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193,197,199,211,223,227,229,233,239,241,251*]*]*],[*[*
"C",127*],254,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",127*],254,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",253*],254,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",253*],254,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",15*],255,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",2*],255,[*[*1,14*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",2*],255,[*[*5,13*]*],19*],[*[*"A",2*],255,[*[*
6,7*]*],13*],[*[*"A",2*],255,[*[*7,6*]*],13*],[*[*"A",2*],255,[*[*
13,5*]*],19*],[*[*"A",2*],255,[*[*14,1*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",4*],255,[*[*0,0,0,1*],[*0,1,1,0*]*],3*],[*[*"A",4*],
255,[*[*1,0,0,0*],[*0,1,1,0*]*],3*],[*[*"A",4*],255,[*[*0,0,3,
1*]*],5*],[*[*"A",4*],255,[*[*1,3,0,0*]*],5*],[*[*"A",1*],255,[*[*
2*],[*4*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],255,[*[*2
*],[*84*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83*]*]*],[*[*"A",1*],255,[*[*4*],[*50*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],255,
[*[*14*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],255,[*[*254*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241,251*]*]*],[*[*"B",127*],255,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",254*],255,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",254*],255,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",15*],256,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*],[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*
[*"A",15*],256,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*],[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",15*],256,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0*],[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],256,[*[*0,0,1*],[*1,0,0*],[*
0,1,1*]*],3*],[*[*"A",3*],256,[*[*0,0,1*],[*1,0,0*],[*1,1,0*]
*],3*],[*[*"A",3*],256,[*[*0,0,1*],[*1,0,0*],[*1,0,0*],[*1,0,0
*]*],[*[**]*]*],[*[*"A",3*],256,[*[*0,0,1*],[*1,1,1*]*],[*[*3,5*]*]
*],[*[*"A",3*],256,[*[*0,0,1*],[*0,0,1*],[*1,0,0*],[*1,0,0*]*],
[*[**]*]*],[*[*"A",3*],256,[*[*0,0,1*],[*0,0,1*],[*0,1,1*]*],3*],
[*[*"A",3*],256,[*[*0,0,1*],[*0,0,1*],[*1,1,0*]*],3*],[*[*"A",3
*],256,[*[*0,0,1*],[*0,0,1*],[*0,0,1*],[*1,0,0*]*],[*[**]*]*],[*[*
"A",3*],256,[*[*0,0,1*],[*0,0,1*],[*0,0,1*],[*0,0,1*]*],[*[**]*]
*],[*[*"A",3*],256,[*[*1,0,0*],[*1,1,1*]*],[*[*3,5*]*]*],[*[*"A",3
*],256,[*[*1,0,0*],[*1,0,0*],[*0,1,1*]*],3*],[*[*"A",3*],256,[*[*
1,0,0*],[*1,0,0*],[*1,1,0*]*],3*],[*[*"A",3*],256,[*[*1,0,0*],
[*1,0,0*],[*1,0,0*],[*1,0,0*]*],[*[**]*]*],[*[*"A",3*],256,[*[*0,
1,1*],[*1,1,0*]*],3*],[*[*"A",3*],256,[*[*0,1,1*],[*0,1,1*]*],3
*],[*[*"A",3*],256,[*[*1,1,0*],[*1,1,0*]*],3*],[*[*"A",3*],256,[*
[*1,1,3*]*],[*[*3,5,7*]*]*],[*[*"A",3*],256,[*[*3,1,1*]*],[*[*3,
5,7*]*]*],[*[*"B",2*],256,[*[*0,1*],[*1,0*],[*1,1*]*],2*],[*[*"B",
2*],256,[*[*0,1*],[*1,0*],[*1,0*],[*1,0*]*],2*],[*[*"B",2*],256,
[*[*0,1*],[*3,1*]*],2*],[*[*"B",2*],256,[*[*0,1*],[*0,1*],[*1,0
*],[*1,0*]*],2*],[*[*"B",2*],256,[*[*0,1*],[*0,1*],[*1,1*]*],2*],
[*[*"B",2*],256,[*[*0,1*],[*0,1*],[*0,1*],[*1,0*]*],2*],[*[*"B",2
*],256,[*[*0,1*],[*0,1*],[*0,1*],[*0,1*]*],2*],[*[*"B",2*],256,[*
[*1,0*],[*3,1*]*],[*[*3,5,7*]*]*],[*[*"B",2*],256,[*[*1,0*],[*1,0
*],[*1,1*]*],[*[*5*]*]*],[*[*"B",2*],256,[*[*1,0*],[*1,0*],[*1,0*],
[*1,0*]*],[*[**]*]*],[*[*"B",2*],256,[*[*1,1*],[*1,1*]*],[*[*5*]*]*],
[*[*"B",2*],256,[*[*1,6*]*],13*],[*[*"B",2*],256,[*[*3,3*]*],[*[*3,
5,7,11*]*]*],[*[*"B",4*],256,[*[*1,0,0,0*],[*1,0,0,0*]*],[*[**]*]
*],[*[*"B",8*],256,[*[*0,0,0,0,0,0,0,1*],[*0,0,0,0,0,0,0,1*]
*],2*],[*[*"B",8*],256,[*[*1,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"C",
4*],256,[*[*1,0,0,0*],[*1,0,0,0*]*],2*],[*[*"C",8*],256,[*[*0,
0,0,0,0,0,0,1*],[*0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"C",8*],
256,[*[*1,0,0,0,0,0,0,0*]*],2*],[*[*"D",5*],256,[*[*0,1,0,0,0
*],[*1,0,0,0,0*]*],[*[**]*]*],[*[*"D",5*],256,[*[*0,1,0,0,0*],[*
0,1,0,0,0*]*],[*[**]*]*],[*[*"D",5*],256,[*[*1,0,0,0,0*],[*1,0,
0,0,0*]*],[*[**]*]*],[*[*"D",8*],256,[*[*0,0,0,0,0,0,0,1*],[*0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",9*],256,[*[*0,1,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"D",9*],256,[*[*1,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",1*],256,[*[*1*],[*3*],[*31*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],256,[*[*1*],[*3*],[*3*],[*7*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],256,[*[*1*],[*7*],[*15*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",1*],256,[*[*1*],[*127*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],
[*3*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],256,[*[*1*],[*1
*],[*3*],[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],
[*7*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],[*63
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61*]
*]*],[*[*"A",1*],256,[*[*1*],[*1*],[*1*],[*3*],[*7*]*],[*[*2,3,5,7
*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],[*1*],[*31*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],[*1*],[*1*],
[*3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],[*1*],[*1
*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],256,[*[*1*],[*1*],
[*1*],[*1*],[*1*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],256,[*[*1
*],[*1*],[*1*],[*1*],[*1*],[*1*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],
256,[*[*1*],[*1*],[*1*],[*1*],[*1*],[*1*],[*1*],[*1*]*],[*[**]*]*],[*
[*"A",1*],256,[*[*3*],[*7*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
256,[*[*3*],[*63*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61*]*]*],[*[*"A",1*],256,[*[*3*],[*3*],[*15*]*],[*[*2,
3,5,7,11,13*]*]*],[*[*"A",1*],256,[*[*3*],[*3*],[*3*],[*3*]*],[*[*
2,3*]*]*],[*[*"A",1*],256,[*[*7*],[*31*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31*]*]*],[*[*"A",1*],256,[*[*15*],[*15*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],256,[*[*255*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139,149,151,157,163,167,173,179,181,
191,193,197,199,211,223,227,229,233,239,241,251*]*]*],[*[*"C",128
*],256,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*],[*[*"D",128*],256,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",255*],256,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",255*],256,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],257,[*[*256*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251*]*]
*],[*[*"B",128*],257,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",256*],257,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",256
*],257,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*]*],[*[*[*"A",5*],258,[*[*0,0,1,0,2*]*],5*],[*[*"A",5*],
258,[*[*2,0,1,0,0*]*],5*],[*[*"A",1*],258,[*[*1*],[*2*],[*42*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*"A",1*],258,[*
[*1*],[*128*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127*]*]*],[*
[*"A",1*],258,[*[*2*],[*85*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1*],258,[*
[*5*],[*42*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41*]*]*],[*[*
"A",1*],258,[*[*257*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233,239,241,251,257*]*]*],[*[*"C",129*],258,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"D",129*],258,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",257*],258,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",257*],258,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"G",2*],259,[*[*1,3*]*],13*],[*
[*"A",1*],259,[*[*6*],[*36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31
*]*]*],[*[*"A",1*],259,[*[*258*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251,257*]*]*],[*[*"B",129
*],259,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",258*],259,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",258*],
259,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",2*],260,[*[*4,7*]*],[*[*2,3,5,7,11*]*]
*],[*[*"A",2*],260,[*[*7,4*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",3*],
260,[*[*0,4,1*]*],5*],[*[*"A",3*],260,[*[*1,4,0*]*],5*],[*[*"A",3
*],260,[*[*2,0,4*]*],7*],[*[*"A",3*],260,[*[*4,0,2*]*],7*],[*[*
"B",2*],260,[*[*1,0*],[*0,1*],[*0,2*]*],5*],[*[*"B",2*],260,[*[*
0,1*],[*3,1*]*],5*],[*[*"B",2*],260,[*[*0,2*],[*3,0*]*],5*],[*[*
"B",2*],260,[*[*2,4*]*],[*[*2,3,7,11*]*]*],[*[*"B",2*],260,[*[*4,
4*]*],7*],[*[*"A",1*],260,[*[*1*],[*4*],[*25*]*],[*[*2,3,5,7,11,
13,17,19,23*]*]*],[*[*"A",1*],260,[*[*1*],[*9*],[*12*]*],[*[*2,3,
5,7,11*]*]*],[*[*"A",1*],260,[*[*1*],[*129*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127*]*]*],[*[*"A",1*],260,[*[*1*],[*1*],[*4*],
[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],260,[*[*1*],[*1*],[*64*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61*]*]
*],[*[*"A",1*],260,[*[*3*],[*4*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*
"A",1*],260,[*[*3*],[*64*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61*]*]*],[*[*"A",1*],260,[*[*4*],[*51*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],260,
[*[*9*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],
260,[*[*12*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],
260,[*[*259*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257*]*]*],[*[*"C",130*],260,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",130*],260,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",259*],260,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",259*],260,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],261,[*[*0,1*],[*
3,7*]*],11*],[*[*"A",2*],261,[*[*0,1*],[*7,3*]*],11*],[*[*"A",2*],
261,[*[*1,0*],[*3,7*]*],11*],[*[*"A",2*],261,[*[*1,0*],[*7,3*]*],
11*],[*[*"A",1*],261,[*[*2*],[*86*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]*],[*[*"A",1
*],261,[*[*2*],[*2*],[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],261,[*[*8*],[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],
[*[*"A",1*],261,[*[*260*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,
197,199,211,223,227,229,233,239,241,251,257*]*]*],[*[*"B",130*],
261,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",260*],261,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",260
*],261,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],262,[*[*6,12*]*],19*],[*[*
"A",2*],262,[*[*12,6*]*],19*],[*[*"A",1*],262,[*[*1*],[*130*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127*]*]*],[*[*"A",1*],262,[*[*
261*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211,223,227,
229,233,239,241,251,257*]*]*],[*[*"C",131*],262,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],
[*[*"D",131*],262,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",261*],262,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",261*],262,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],263,[*[*262
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,
151,157,163,167,173,179,181,191,193,197,199,211,223,227,229,
233,239,241,251,257*]*]*],[*[*"B",131*],263,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*
[*"A",262*],263,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",262*],263,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"A",2*],264,[*[*1,1*],[*1,5*]*],7*],[*[*"A",2*],
264,[*[*1,1*],[*5,1*]*],7*],[*[*"A",3*],264,[*[*0,1,0*],[*1,1,1
*]*],3*],[*[*"B",2*],264,[*[*1,10*]*],23*],[*[*"B",5*],264,[*[*0,0,
0,0,3*]*],13*],[*[*"B",5*],264,[*[*0,0,0,1,1*]*],3*],[*[*"A",1
*],264,[*[*1*],[*2*],[*3*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
264,[*[*1*],[*2*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43*]*]*],[*[*"A",1*],264,[*[*1*],[*3*],[*32*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31*]*]*],[*[*"A",1*],264,[*[*1*],[*5*],[*21*]*],[*[*
2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],264,[*[*1*],[*10*],[*11*]
*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],264,[*[*1*],[*131*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131*]*]*],[*[*"A",1*],264,[*[*1
*],[*1*],[*2*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],
264,[*[*1*],[*1*],[*5*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
264,[*[*1*],[*1*],[*65*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61*]*]*],[*[*"A",1*],264,[*[*1*],[*1*],[*1*],[*2*],
[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],264,[*[*1*],[*1*],[*1*],[*32
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],264,[*[*
2*],[*3*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],264,
[*[*2*],[*7*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],264,[*[*2*],[*
87*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83*]*]*],[*[*"A",1*],264,[*[*3*],[*5*],[*10*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],264,[*[*3*],[*65*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*"A",1*],264,[*
[*5*],[*43*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],
[*[*"A",1*],264,[*[*7*],[*32*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31*]*]*],[*[*"A",1*],264,[*[*10*],[*23*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],264,[*[*11*],[*21*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",1*],264,[*[*263*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139,149,151,157,163,167,173,179,181,
191,193,197,199,211,223,227,229,233,239,241,251,257,263*]*]*],[*
[*"C",132*],264,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",132*],264,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"A",263*],264,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",263*],
264,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",2*],265,[*[*0,11*]*],23*],[*
[*"A",1*],265,[*[*4*],[*52*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47*]*]*],[*[*"A",1*],265,[*[*264*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,263
*]*]*],[*[*"B",132*],265,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",264*],265,[*
[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",264*],265,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",6*],266,[*[*0,0,0,1,1,0*]*],3*],[*[*"A",6*],266,[*
[*0,1,1,0,0,0*]*],3*],[*[*"A",7*],266,[*[*0,0,0,0,0,2,0*]*],3
*],[*[*"A",7*],266,[*[*0,2,0,0,0,0,0*]*],3*],[*[*"C",4*],266,[*[*
0,0,2,0*]*],3*],[*[*"G",2*],266,[*[*0,1*],[*1,1*]*],7*],[*[*"A",
1*],266,[*[*1*],[*6*],[*18*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",
1*],266,[*[*1*],[*132*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131*]*]*],[*[*"A",1*],266,[*[*6*],[*37*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37*]*]*],[*[*"A",1*],266,[*[*13*],[*18*]*],[*[*2,3,
5,7,11,13,17*]*]*],[*[*"A",1*],266,[*[*265*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,263
*]*]*],[*[*"C",133*],266,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",133*],266,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",265*],266,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",265*],266,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],
267,[*[*4,8*]*],11*],[*[*"A",2*],267,[*[*7,11*]*],19*],[*[*"A",2*],
267,[*[*8,4*]*],11*],[*[*"A",2*],267,[*[*11,7*]*],19*],[*[*"G",2*],
267,[*[*4,0*]*],7*],[*[*"A",1*],267,[*[*2*],[*88*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83*]*]
*],[*[*"A",1*],267,[*[*266*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,
197,199,211,223,227,229,233,239,241,251,257,263*]*]*],[*[*"B",133
*],267,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",266*],267,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",266*],267,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*
[*[*"A",1*],268,[*[*1*],[*133*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131*]*]*],[*[*"A",1*],268,[*[*1*],[*1*],[*66*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61*]*]*],[*[*
"A",1*],268,[*[*3*],[*66*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61*]*]*],[*[*"A",1*],268,[*[*267*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263*]*]*],[*[*"C",134*],268,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",
134*],268,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",267*],268,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0*]*],[*[**]*]*],[*[*"A",267*],268,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",1*],269,[*[*268*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251,257,263*]*]*],[*[*
"B",134*],269,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",268*],269,[*[*1,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",268*],269,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],270,[*[*0,1*],[*1,0*],[*1,0*],[*
0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],270,[*[*0,1*],[*1,0*],[*1,0*],[*
3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],270,[*[*0,1*],[*0,2*],[*0,4*]*],
[*[*2,3*]*]*],[*[*"A",2*],270,[*[*0,1*],[*0,2*],[*1,2*]*],[*[*2*]*]
*],[*[*"A",2*],270,[*[*0,1*],[*0,2*],[*2,1*]*],[*[*2*]*]*],[*[*"A",
2*],270,[*[*0,1*],[*0,2*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],270,
[*[*0,1*],[*2,0*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],270,[*[*0,1
*],[*2,0*],[*1,2*]*],[*[*2*]*]*],[*[*"A",2*],270,[*[*0,1*],[*2,0*],
[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],270,[*[*0,1*],[*2,0*],[*4,0*]*],
[*[*2,3*]*]*],[*[*"A",2*],270,[*[*0,1*],[*3,4*]*],[*[*2,3,7*]*]*],
[*[*"A",2*],270,[*[*0,1*],[*4,3*]*],[*[*2,3,7*]*]*],[*[*"A",2*],
270,[*[*0,1*],[*4,6*]*],11*],[*[*"A",2*],270,[*[*0,1*],[*6,4*]*],
11*],[*[*"A",2*],270,[*[*0,1*],[*0,1*],[*1,0*],[*0,3*]*],[*[*2,3
*]*]*],[*[*"A",2*],270,[*[*0,1*],[*0,1*],[*1,0*],[*3,0*]*],[*[*2,3
*]*]*],[*[*"A",2*],270,[*[*0,1*],[*0,1*],[*0,1*],[*0,3*]*],[*[*2,3
*]*]*],[*[*"A",2*],270,[*[*0,1*],[*0,1*],[*0,1*],[*3,0*]*],[*[*2,3
*]*]*],[*[*"A",2*],270,[*[*1,0*],[*0,2*],[*0,4*]*],[*[*2,3*]*]*],[*
[*"A",2*],270,[*[*1,0*],[*0,2*],[*1,2*]*],[*[*2*]*]*],[*[*"A",2*],
270,[*[*1,0*],[*0,2*],[*2,1*]*],[*[*2*]*]*],[*[*"A",2*],270,[*[*1,
0*],[*0,2*],[*4,0*]*],[*[*2,3*]*]*],[*[*"A",2*],270,[*[*1,0*],[*2,
0*],[*0,4*]*],[*[*2,3*]*]*],[*[*"A",2*],270,[*[*1,0*],[*2,0*],[*1,
2*]*],[*[*2*]*]*],[*[*"A",2*],270,[*[*1,0*],[*2,0*],[*2,1*]*],[*[*2
*]*]*],[*[*"A",2*],270,[*[*1,0*],[*2,0*],[*4,0*]*],[*[*2,3*]*]*],[*
[*"A",2*],270,[*[*1,0*],[*3,4*]*],[*[*2,3,7*]*]*],[*[*"A",2*],270,
[*[*1,0*],[*4,3*]*],[*[*2,3,7*]*]*],[*[*"A",2*],270,[*[*1,0*],[*4,
6*]*],11*],[*[*"A",2*],270,[*[*1,0*],[*6,4*]*],11*],[*[*"A",2*],
270,[*[*1,0*],[*1,0*],[*1,0*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],
270,[*[*1,0*],[*1,0*],[*1,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],
270,[*[*0,2*],[*0,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],270,[*[*0,2
*],[*8,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],270,[*[*2,0*],[*0,8*]
*],[*[*2,3,5,7*]*]*],[*[*"A",2*],270,[*[*2,0*],[*8,0*]*],[*[*2,3,
5,7*]*]*],[*[*"A",2*],270,[*[*0,3*],[*2,2*]*],[*[*2,3,5*]*]*],[*[*
"A",2*],270,[*[*3,0*],[*2,2*]*],[*[*2,3,5*]*]*],[*[*"A",2*],270,[*
[*0,4*],[*1,3*]*],5*],[*[*"A",2*],270,[*[*0,4*],[*3,1*]*],5*],[*[*
"A",2*],270,[*[*1,2*],[*1,3*]*],5*],[*[*"A",2*],270,[*[*1,2*],[*
3,1*]*],5*],[*[*"A",2*],270,[*[*2,1*],[*1,3*]*],5*],[*[*"A",2*],
270,[*[*2,1*],[*3,1*]*],5*],[*[*"A",2*],270,[*[*4,0*],[*1,3*]*],5
*],[*[*"A",2*],270,[*[*4,0*],[*3,1*]*],5*],[*[*"A",2*],270,[*[*2,
11*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",2*],270,[*[*8,10*]*],19*],
[*[*"A",2*],270,[*[*10,8*]*],19*],[*[*"A",2*],270,[*[*11,2*]*],[*[*
2,3,5,7,11,13*]*]*],[*[*"A",3*],270,[*[*0,1,0*],[*0,1,2*]*],[*[*
2*]*]*],[*[*"A",3*],270,[*[*0,1,0*],[*2,1,0*]*],[*[*2*]*]*],[*[*
"A",3*],270,[*[*2,0,4*]*],[*[*2,3,7*]*]*],[*[*"A",3*],270,[*[*4,
0,2*]*],[*[*2,3,7*]*]*],[*[*"B",2*],270,[*[*0,1*],[*0,4*]*],7*],[*
[*"A",1*],270,[*[*1*],[*2*],[*4*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*
"A",1*],270,[*[*1*],[*2*],[*44*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43*]*]*],[*[*"A",1*],270,[*[*1*],[*2*],[*2*],[*14*]*],
[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],270,[*[*1*],[*2*],[*2*],[*2*],
[*4*]*],[*[*2,3*]*]*],[*[*"A",1*],270,[*[*1*],[*4*],[*26*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],270,[*[*1*],[*8*],[*14*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],270,[*[*1*],[*134*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131*]*]*],[*[*"A",1*],270,[*
[*2*],[*4*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],270,[*
[*2*],[*5*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],270,[*[*2
*],[*8*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],270,[*[*2*],[*89*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89*]*]*],[*[*"A",1*],270,[*[*2*],[*2*],[*4*],[*5*]*],[*
[*2,3,5*]*]*],[*[*"A",1*],270,[*[*2*],[*2*],[*29*]*],[*[*2,3,5,7,
11,13,17,19,23,29*]*]*],[*[*"A",1*],270,[*[*2*],[*2*],[*2*],[*9*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],270,[*[*4*],[*5*],[*8*]*],[*[*2,3,
5,7*]*]*],[*[*"A",1*],270,[*[*4*],[*53*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],270,[*[*5*],[*44*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],
270,[*[*8*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",
1*],270,[*[*9*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],270,[*[*14*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",
1*],270,[*[*269*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257,263,269*]*]*],[*[*"C",135*],270,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",135*],270,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"A",269*],270,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",269*],270,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2
*],271,[*[*9,9*]*],19*],[*[*"B",2*],271,[*[*0,8*]*],13*],[*[*"A",1
*],271,[*[*270*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257,263,269*]*]*],[*[*"B",135*],271,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",270*],271,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",270*],271,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",3*],272,[*[*0,0,1*],[*0,2,2*]*],5*],[*
[*"A",3*],272,[*[*0,0,1*],[*2,2,0*]*],5*],[*[*"A",3*],272,[*[*1,
0,0*],[*0,2,2*]*],5*],[*[*"A",3*],272,[*[*1,0,0*],[*2,2,0*]*],5
*],[*[*"A",7*],272,[*[*1,0,0,0,0,0,2*]*],3*],[*[*"A",7*],272,[*[*
2,0,0,0,0,0,1*]*],3*],[*[*"B",2*],272,[*[*1,0*],[*2,2*]*],5*],
[*[*"A",1*],272,[*[*1*],[*3*],[*33*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31*]*]*],[*[*"A",1*],272,[*[*1*],[*7*],[*16*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],272,[*[*1*],[*135*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131*]*]*],[*[*"A",1*],272,[*[*1*],[*1*],[*
3*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],272,[*[*1*],[*1
*],[*67*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67*]*]*],[*[*"A",1*],272,[*[*1*],[*1*],[*1*],[*33*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],272,[*[*1*],[*1*],
[*1*],[*1*],[*16*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],272,[*[*3
*],[*67*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67*]*]*],[*[*"A",1*],272,[*[*3*],[*3*],[*16*]*],[*[*2,3,5,7,
11,13*]*]*],[*[*"A",1*],272,[*[*7*],[*33*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],272,[*[*15*],[*16*]*],[*[*2,3,5,
7,11,13*]*]*],[*[*"A",1*],272,[*[*271*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193,197,199,211,223,227,229,233,239,241,251,257,263,
269,271*]*]*],[*[*"C",136*],272,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",136*],272,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",271*],272,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",271*],272,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],273,[*[*0,1*],[*
0,12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],273,[*[*0,1*],[*5,5*]
*],11*],[*[*"A",2*],273,[*[*0,1*],[*12,0*]*],[*[*2,3,5,7,11*]*]*],
[*[*"A",2*],273,[*[*1,0*],[*0,12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",
2*],273,[*[*1,0*],[*5,5*]*],11*],[*[*"A",2*],273,[*[*1,0*],[*12,0
*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",2*],273,[*[*5,6*]*],[*[*2,3,5,
11*]*]*],[*[*"A",2*],273,[*[*6,5*]*],[*[*2,3,5,11*]*]*],[*[*"C",3
*],273,[*[*0,1,0*],[*0,0,2*]*],3*],[*[*"F",4*],273,[*[*0,0,1,0*]
*],[*[*2,3*]*]*],[*[*"G",2*],273,[*[*3,0*]*],[*[*2,3,5,11*]*]*],[*[*
"A",1*],273,[*[*2*],[*6*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1
*],273,[*[*2*],[*90*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],273,[*[*6*],
[*38*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37*]*]*],[*[*"A",1*],
273,[*[*12*],[*20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],
273,[*[*272*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257,263,269,271*]*]*],[*[*"B",136*],
273,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",272*],273,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",272*],273,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"B",11*],274,[*[*0,0,0,0,
0,0,0,0,0,0,2*]*],23*],[*[*"A",1*],274,[*[*1*],[*136*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131*]*]*],[*[*"A",1*],274,[*
[*273*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,
149,151,157,163,167,173,179,181,191,193,197,199,211,223,227,
229,233,239,241,251,257,263,269,271*]*]*],[*[*"C",12*],274,[*[*0,
0,0,0,0,0,0,0,0,0,1,0*]*],2*],[*[*"C",12*],274,[*[*0,0,0,0,
0,0,0,0,0,0,1,0*]*],3*],[*[*"D",12*],274,[*[*0,0,0,0,0,0,0,
0,0,0,1,0*]*],2*],[*[*"C",137*],274,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",137*],274,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",273
*],274,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",273
*],274,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
10*],275,[*[*0,0,0,0,0,0,0,0,1,1*]*],3*],[*[*"A",10*],275,[*[*
1,1,0,0,0,0,0,0,0,0*]*],3*],[*[*"B",11*],275,[*[*0,0,0,0,0,
0,0,0,0,0,2*]*],[*[*2,23*]*]*],[*[*"B",2*],275,[*[*0,1*],[*0,4*]
*],[*[*2,3,7*]*]*],[*[*"B",5*],275,[*[*0,0,0,0,3*]*],[*[*2,3,13*]
*]*],[*[*"A",1*],275,[*[*4*],[*54*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53*]*]*],[*[*"A",1*],275,[*[*4*],[*4*],[*10*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],275,[*[*10*],[*24*]*],[*[*2,3,5,
7,11,13,17,19,23*]*]*],[*[*"A",1*],275,[*[*274*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263,269,271*]*]*],[*[*"C",12*],275,[*[*0,0,0,0,0,0,0,0,0,
0,1,0*]*],[*[*2,3*]*]*],[*[*"B",137*],275,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[*2*]*]*],[*[*"A",274*],275,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0*]*],[*[**]*]*],[*[*"A",274*],275,[*[*0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],276,[*[*0,22*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",2*],276,[*[*5,7*]*],11*],[*[*"A",2*],276,[*
[*7,5*]*],11*],[*[*"A",2*],276,[*[*22,0*]*],[*[*2,3,5,7,11,13,
17,19*]*]*],[*[*"A",3*],276,[*[*0,0,1*],[*2,0,2*]*],3*],[*[*"A",3
*],276,[*[*1,0,0*],[*2,0,2*]*],3*],[*[*"A",1*],276,[*[*1*],[*2*],
[*45*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*
"A",1*],276,[*[*1*],[*5*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],
[*[*"A",1*],276,[*[*1*],[*137*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137*]*]*],[*[*"A",1*],276,[*[*1*],[*1*],[*2*],[*22
*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],276,[*[*1*],[*1*],
[*68*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67*]*]*],[*[*"A",1*],276,[*[*2*],[*3*],[*22*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],276,[*[*2*],[*91*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]
*],[*[*"A",1*],276,[*[*3*],[*68*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],276,[*[*5*],[*45
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1
*],276,[*[*11*],[*22*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1
*],276,[*[*275*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257,263,269,271*]*]*],[*[*"D",12*],
276,[*[*0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[*2*]*]*],[*[*"A",22*],
276,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",22*],276,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,2*]*],[*[**]*]*],[*[*"A",23*],276,[*[*0,1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
23*],276,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1,0*]*],[*[**]*]*],[*[*"C",138*],276,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",138*],276,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
275*],276,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",275*],276,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",1*],277,[*[*276*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251,257,263,269,271*]*]
*],[*[*"B",138*],277,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",
276*],277,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",276*],277,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",1*],278,[*[*1*],[*138*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137*]*]*],[*[*"A",1*],278,[*[*277*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241,251,257,263,269,271,277*]*]*],[*[*"C",139*],278,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",139*],278,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",277*],278,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",277*],278,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],279,[*[*6,
6*]*],11*],[*[*"A",5*],279,[*[*1,0,0,1,1*]*],3*],[*[*"A",5*],279,
[*[*1,1,0,0,1*]*],3*],[*[*"C",4*],279,[*[*0,1,0,1*]*],3*],[*[*
"A",1*],279,[*[*2*],[*92*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],279,[*
[*2*],[*2*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",
1*],279,[*[*8*],[*30*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*
"A",1*],279,[*[*278*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233,239,241,251,257,263,269,271,277*]*]*],[*
[*"B",139*],279,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",
278*],279,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",278*],279,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,1*]*],[*[**]*]*]*],[*[*[*"A",2*],280,[*[*1,1*],[*1,4*]*],[*[*2,3*]
*]*],[*[*"A",2*],280,[*[*1,1*],[*4,1*]*],[*[*2,3*]*]*],[*[*"A",2*],
280,[*[*0,3*],[*0,6*]*],[*[*2,3,5*]*]*],[*[*"A",2*],280,[*[*0,3*],
[*6,0*]*],[*[*2,3,5*]*]*],[*[*"A",2*],280,[*[*3,0*],[*0,6*]*],[*[*
2,3,5*]*]*],[*[*"A",2*],280,[*[*3,0*],[*6,0*]*],[*[*2,3,5*]*]*],[*
[*"A",2*],280,[*[*3,9*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",2*],
280,[*[*9,3*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",3*],280,[*[*0,0,
1*],[*1,0,3*]*],[*[*2,3*]*]*],[*[*"A",3*],280,[*[*0,0,1*],[*3,0,
1*]*],[*[*2,3*]*]*],[*[*"A",3*],280,[*[*1,0,0*],[*1,0,3*]*],[*[*2,
3*]*]*],[*[*"A",3*],280,[*[*1,0,0*],[*3,0,1*]*],[*[*2,3*]*]*],[*[*
"A",3*],280,[*[*1,0,1*],[*0,1,1*]*],2*],[*[*"A",3*],280,[*[*1,0,
1*],[*1,1,0*]*],2*],[*[*"A",3*],280,[*[*0,3,2*]*],[*[*2,3,5*]*]
*],[*[*"A",3*],280,[*[*0,4,1*]*],[*[*2,3,5*]*]*],[*[*"A",3*],280,[*
[*1,0,6*]*],[*[*2,3,5*]*]*],[*[*"A",3*],280,[*[*1,4,0*]*],[*[*2,
3,5*]*]*],[*[*"A",3*],280,[*[*2,3,0*]*],[*[*2,3,5*]*]*],[*[*"A",3
*],280,[*[*6,0,1*]*],[*[*2,3,5*]*]*],[*[*"A",4*],280,[*[*0,1,0,3
*]*],[*[*2,3*]*]*],[*[*"A",4*],280,[*[*0,1,1,1*]*],[*[*3,5*]*]*],[*
[*"A",4*],280,[*[*1,1,1,0*]*],[*[*3,5*]*]*],[*[*"A",4*],280,[*[*3,
0,1,0*]*],[*[*2,3*]*]*],[*[*"A",5*],280,[*[*0,1,0,0,2*]*],[*[*2,
3*]*]*],[*[*"A",5*],280,[*[*2,0,0,1,0*]*],[*[*2,3*]*]*],[*[*"A",7
*],280,[*[*1,0,0,0,0,0,2*]*],[*[*2,3*]*]*],[*[*"A",7*],280,[*[*2,
0,0,0,0,0,1*]*],[*[*2,3*]*]*],[*[*"B",2*],280,[*[*1,0*],[*0,1*],
[*0,2*]*],[*[*2,5*]*]*],[*[*"B",2*],280,[*[*0,1*],[*5,0*]*],[*[*2,
3,5*]*]*],[*[*"B",2*],280,[*[*0,2*],[*3,0*]*],[*[*2,3,5*]*]*],[*[*
"B",3*],280,[*[*0,0,1*],[*1,0,1*]*],7*],[*[*"B",3*],280,[*[*1,0,
0*],[*2,0,0*]*],[*[*2*]*]*],[*[*"B",3*],280,[*[*1,0,3*]*],11*],[*[*
"D",4*],280,[*[*0,0,0,1*],[*0,0,0,2*]*],[*[*2*]*]*],[*[*"D",4*],
280,[*[*0,0,0,1*],[*0,2,0,0*]*],[*[*2*]*]*],[*[*"D",4*],280,[*[*
0,0,0,1*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"D",4*],280,[*[*0,1,0,0
*],[*0,0,0,2*]*],[*[*2*]*]*],[*[*"D",4*],280,[*[*0,1,0,0*],[*0,2,
0,0*]*],[*[*2*]*]*],[*[*"D",4*],280,[*[*0,1,0,0*],[*2,0,0,0*]*],
[*[*2*]*]*],[*[*"D",4*],280,[*[*1,0,0,0*],[*0,0,0,2*]*],[*[*2*]*]
*],[*[*"D",4*],280,[*[*1,0,0,0*],[*0,2,0,0*]*],[*[*2*]*]*],[*[*
"D",4*],280,[*[*1,0,0,0*],[*2,0,0,0*]*],[*[*2*]*]*],[*[*"A",1*],
280,[*[*1*],[*3*],[*4*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],280,[*[*
1*],[*3*],[*34*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*
"A",1*],280,[*[*1*],[*4*],[*27*]*],[*[*2,3,5,7,11,13,17,19,23*]
*]*],[*[*"A",1*],280,[*[*1*],[*6*],[*19*]*],[*[*2,3,5,7,11,13,17,
19*]*]*],[*[*"A",1*],280,[*[*1*],[*9*],[*13*]*],[*[*2,3,5,7,11,13
*]*]*],[*[*"A",1*],280,[*[*1*],[*139*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,
107,109,113,127,131,137,139*]*]*],[*[*"A",1*],280,[*[*1*],[*1*],[*
4*],[*13*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],280,[*[*1*],[*1
*],[*6*],[*9*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],280,[*[*1*],[*1*],[*
69*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67*]*]*],[*[*"A",1*],280,[*[*1*],[*1*],[*1*],[*4*],[*6*]*],[*[*2,
3,5*]*]*],[*[*"A",1*],280,[*[*1*],[*1*],[*1*],[*34*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],280,[*[*3*],[*4*],[*13*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],280,[*[*3*],[*6*],[*9*]*],[*
[*2,3,5,7*]*]*],[*[*"A",1*],280,[*[*3*],[*69*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",1*],
280,[*[*4*],[*6*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],280,[*[*4
*],[*55*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53*]
*]*],[*[*"A",1*],280,[*[*6*],[*39*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37*]*]*],[*[*"A",1*],280,[*[*7*],[*34*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31*]*]*],[*[*"A",1*],280,[*[*9*],[*27*]*],[*[*2,
3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],280,[*[*13*],[*19*]*],[*[*
2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],280,[*[*279*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263,269,271,277*]*]*],[*[*"C",140*],280,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"D",140*],280,[*[*0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[**]*]*],[*[*"A",279*],280,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",279*],280,[*[*0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"C",4*],281,[*[*
0,0,2,0*]*],5*],[*[*"A",1*],281,[*[*280*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,
263,269,271,277*]*]*],[*[*"B",140*],281,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",280*],281,[*[*1,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",280*],281,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1
*],282,[*[*1*],[*2*],[*46*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43*]*]*],[*[*"A",1*],282,[*[*1*],[*140*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139*]*]*],[*[*"A",1*],282,[*[*
2*],[*93*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],282,[*[*5*],[*46*]*],[*
[*2,3,5,7,11,13,17,19,23,29,31,37,41,43*]*]*],[*[*"A",1*],282,
[*[*281*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149,151,157,163,167,173,179,181,191,193,197,199,211,223,
227,229,233,239,241,251,257,263,269,271,277,281*]*]*],[*[*"C",141
*],282,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",141*],
282,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",281*],282,
[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",281*],282,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],283,[*[*282*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263,269,271,277,281*]*]*],[*[*"B",141*],283,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",282*],283,[*[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",282*],
283,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*]*],[*[*[*"B",2*],284,[*[*1,0*],[*2,2*]*],7*],[*[*"B",2*],
284,[*[*0,8*]*],11*],[*[*"B",2*],284,[*[*3,6*]*],17*],[*[*"A",1*],
284,[*[*1*],[*141*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,
131,137,139*]*]*],[*[*"A",1*],284,[*[*1*],[*1*],[*70*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67*]*]*],[*[*"A",
1*],284,[*[*3*],[*70*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67*]*]*],[*[*"A",1*],284,[*[*283*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263,269,271,277,281,283*]*]*],[*[*"C",142*],284,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",142*],284,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",283*],284,[*[*1,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",283
*],284,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
1*]*],[*[**]*]*]*],[*[*[*"A",2*],285,[*[*0,4*],[*2,2*]*],5*],[*[*"A",
2*],285,[*[*1,2*],[*2,2*]*],5*],[*[*"A",2*],285,[*[*2,1*],[*2,2*]
*],5*],[*[*"A",2*],285,[*[*4,0*],[*2,2*]*],5*],[*[*"A",3*],285,[*[*
1,0,1*],[*0,2,0*]*],3*],[*[*"A",3*],285,[*[*3,0,3*]*],7*],[*[*
"B",2*],285,[*[*0,8*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"B",2*],
285,[*[*4,3*]*],11*],[*[*"A",1*],285,[*[*2*],[*4*],[*18*]*],[*[*2,
3,5,7,11,13,17*]*]*],[*[*"A",1*],285,[*[*2*],[*94*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89*]*]*],[*[*"A",1*],285,[*[*4*],[*56*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],285,[*[*14*],[*18*]
*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],285,[*[*284*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,
83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,
167,173,179,181,191,193,197,199,211,223,227,229,233,239,241,
251,257,263,269,271,277,281,283*]*]*],[*[*"B",142*],285,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",284*],285,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",284*],285,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",10*],286,[*[*0,0,0,0,0,
0,0,0,0,3*]*],[*[*2,3*]*]*],[*[*"A",10*],286,[*[*3,0,0,0,0,0,
0,0,0,0*]*],[*[*2,3*]*]*],[*[*"A",12*],286,[*[*0,0,0,0,0,0,0,
0,0,1,0,0*]*],[*[**]*]*],[*[*"A",12*],286,[*[*0,0,1,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",3*],286,[*[*0,0,10*]*],[*[*2,3,5,7
*]*]*],[*[*"A",3*],286,[*[*10,0,0*]*],[*[*2,3,5,7*]*]*],[*[*"B",2
*],286,[*[*10,0*]*],[*[*2,3,5,7*]*]*],[*[*"B",6*],286,[*[*0,0,0,
1,0,0*]*],[*[*2*]*]*],[*[*"C",3*],286,[*[*0,2,1*]*],3*],[*[*"G",2
*],286,[*[*2,1*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],286,[*[*1*],[*10
*],[*12*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],286,[*[*1*],[*142*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*
"A",1*],286,[*[*10*],[*25*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*
[*"A",1*],286,[*[*12*],[*21*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*
"A",1*],286,[*[*285*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233,239,241,251,257,263,269,271,277,281,283
*]*]*],[*[*"C",143*],286,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"D",143*],286,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*],[*[*"A",285*],286,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",285*],286,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",16*],287,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1
*]*],17*],[*[*"A",1*],287,[*[*6*],[*40*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37*]*]*],[*[*"A",1*],287,[*[*286*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,
263,269,271,277,281,283*]*]*],[*[*"B",143*],287,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",286*],287,[*[*1,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",286*],287,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",16*],288,[*[*1,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*17*]*]*],[*[*"A",2*],288,[*[*
0,2*],[*2,0*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],288,[*[*0,2*],[*
1,5*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],288,[*[*0,2*],[*5,1*]*],[*[*
2,3,5,7*]*]*],[*[*"A",2*],288,[*[*0,2*],[*0,2*],[*1,1*]*],[*[*2,
3*]*]*],[*[*"A",2*],288,[*[*2,0*],[*1,5*]*],[*[*2,3,5,7*]*]*],[*[*
"A",2*],288,[*[*2,0*],[*5,1*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],
288,[*[*2,0*],[*2,0*],[*1,1*]*],[*[*2,3*]*]*],[*[*"A",2*],288,[*[*
1,1*],[*0,7*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],288,[*[*1,1*],[*2,
4*]*],7*],[*[*"A",2*],288,[*[*1,1*],[*4,2*]*],7*],[*[*"A",2*],288,
[*[*1,1*],[*7,0*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],288,[*[*1,15*]
*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",2*],288,[*[*15,1*]*],[*[*2,
3,5,7,11,13,17*]*]*],[*[*"A",7*],288,[*[*0,0,0,0,0,0,1*],[*0,
0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",7*],288,[*[*0,0,0,0,0,0,1
*],[*2,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"A",7*],288,[*[*1,0,0,0,
0,0,0*],[*0,0,0,0,0,0,2*]*],[*[*2*]*]*],[*[*"A",7*],288,[*[*1,
0,0,0,0,0,0*],[*2,0,0,0,0,0,0*]*],[*[*2*]*]*],[*[*"B",3*],288,
[*[*0,0,1*],[*1,0,1*]*],2*],[*[*"B",3*],288,[*[*0,0,1*],[*0,0,1
*],[*1,0,0*]*],2*],[*[*"B",5*],288,[*[*1,0,0,0,1*]*],11*],[*[*
"C",3*],288,[*[*0,0,1*],[*1,0,1*]*],2*],[*[*"C",3*],288,[*[*0,0,
1*],[*0,0,1*],[*1,0,0*]*],2*],[*[*"C",4*],288,[*[*0,0,0,1*],[*
0,0,0,2*]*],[*[*2*]*]*],[*[*"C",4*],288,[*[*1,0,0,1*]*],[*[*2,7*]
*]*],[*[*"A",1*],288,[*[*1*],[*2*],[*3*],[*11*]*],[*[*2,3,5,7,11*]
*]*],[*[*"A",1*],288,[*[*1*],[*2*],[*5*],[*7*]*],[*[*2,3,5,7*]*]*],
[*[*"A",1*],288,[*[*1*],[*2*],[*47*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31,37,41,43,47*]*]*],[*[*"A",1*],288,[*[*1*],[*2*],[*2*],[*
3*],[*3*]*],[*[*2,3*]*]*],[*[*"A",1*],288,[*[*1*],[*2*],[*2*],[*15
*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],288,[*[*1*],[*3*],[*5*],[*
5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],288,[*[*1*],[*3*],[*35*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],288,[*[*1*],[*3*],
[*3*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],288,[*[*1*],[*5*],[*23
*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],288,[*[*1*],[*7
*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],288,[*[*1*],[*8
*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],288,[*[*1*],[*11*],
[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],288,[*[*1*],[*143*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139*]*]*],[*[*
"A",1*],288,[*[*1*],[*1*],[*2*],[*3*],[*5*]*],[*[*2,3,5*]*]*],[*[*
"A",1*],288,[*[*1*],[*1*],[*2*],[*23*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*2*],[*2*],[*7*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*3*],[*17*]*],[*[*2,3,
5,7,11,13,17*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*5*],[*11*]*],[*
[*2,3,5,7,11*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*7*],[*8*]*],[*[*
2,3,5,7*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*71*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*[*"A",
1*],288,[*[*1*],[*1*],[*1*],[*2*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*
[*"A",1*],288,[*[*1*],[*1*],[*1*],[*2*],[*2*],[*3*]*],[*[*2,3*]*]*],
[*[*"A",1*],288,[*[*1*],[*1*],[*1*],[*3*],[*8*]*],[*[*2,3,5,7*]*]
*],[*[*"A",1*],288,[*[*1*],[*1*],[*1*],[*5*],[*5*]*],[*[*2,3,5*]*]
*],[*[*"A",1*],288,[*[*1*],[*1*],[*1*],[*35*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*1*],[*1*],[*2
*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],288,[*[*1*],[*1*],[*1*],[*1
*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],288,[*[*1*],[*1
*],[*1*],[*1*],[*1*],[*2*],[*2*]*],[*[*2*]*]*],[*[*"A",1*],288,[*[*1
*],[*1*],[*1*],[*1*],[*1*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],
288,[*[*2*],[*3*],[*23*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*
"A",1*],288,[*[*2*],[*3*],[*3*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1
*],288,[*[*2*],[*5*],[*15*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],
288,[*[*2*],[*7*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],288,[*
[*2*],[*95*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89*]*]*],[*[*"A",1*],288,[*[*2*],[*2*],[*
3*],[*7*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],288,[*[*2*],[*2*],[*31*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],288,[*[*3
*],[*5*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],288,[*[*3*],[*7
*],[*8*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],288,[*[*3*],[*71*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],
[*[*"A",1*],288,[*[*3*],[*3*],[*17*]*],[*[*2,3,5,7,11,13,17*]*]*],
[*[*"A",1*],288,[*[*5*],[*47*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47*]*]*],[*[*"A",1*],288,[*[*5*],[*5*],[*7*]*],[*[*2,
3,5,7*]*]*],[*[*"A",1*],288,[*[*7*],[*35*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31*]*]*],[*[*"A",1*],288,[*[*8*],[*31*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],288,[*[*11*],[*23*]*],[*
[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],288,[*[*15*],[*17*]*],
[*[*2,3,5,7,11,13,17*]*]*],[*[*"A",1*],288,[*[*287*]*],[*[*2,3,5,
7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,
89,97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,
173,179,181,191,193,197,199,211,223,227,229,233,239,241,251,
257,263,269,271,277,281,283*]*]*],[*[*"C",144*],288,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",144*],288,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",287*],288,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],
[*[**]*]*],[*[*"A",287*],288,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",16*],289,
[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*],[*1,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",16*],289,[*[*0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,1*],[*0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[**]*]*],[*[*"A",16*],289,[*[*1,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0*],[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[*
*]*]*],[*[*"B",8*],289,[*[*0,0,0,0,0,0,0,1*],[*0,0,0,0,0,0,0,
1*]*],[*[*2*]*]*],[*[*"A",1*],289,[*[*16*],[*16*]*],[*[*2,3,5,7,11,
13*]*]*],[*[*"A",1*],289,[*[*288*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251,257,263,269,271,
277,281,283*]*]*],[*[*"B",144*],289,[*[*0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,1*]*],[*[*2*]*]*],[*[*"A",288*],289,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*
"A",288*],289,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],290,[*[*1*],[*4
*],[*28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],290,[*[*
1*],[*144*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139*]*]*],[*[*"A",1*],290,[*[*4*],[*57*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1*],290,[*[*9*],[*
28*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],290,[*[*289*]
*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241,251,257,263,269,271,277,281,283*]*]*],[*[*"C",145*],290,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",145*],
290,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",
289*],290,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",289*],290,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*
[**]*]*]*],[*[*[*"A",1*],291,[*[*2*],[*96*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89*]*]*],[*[*
"A",1*],291,[*[*290*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,
41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,
127,131,137,139,149,151,157,163,167,173,179,181,191,193,197,
199,211,223,227,229,233,239,241,251,257,263,269,271,277,281,283
*]*]*],[*[*"B",145*],291,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[*2*]*]*],[*[*"A",290*],291,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",
290*],291,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",1*],292,[*[*1*],[*
145*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]
*]*],[*[*"A",1*],292,[*[*1*],[*1*],[*72*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71*]*]*],[*[*"A",1*],292,[*
[*3*],[*72*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71*]*]*],[*[*"A",1*],292,[*[*291*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,
263,269,271,277,281,283*]*]*],[*[*"C",146*],292,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",146*],292,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",291*],292,[*[*
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",291*],292,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"B",3*],293,[*[*4,0,0*]*],5*],[*[*"D",4*],293,[*[*0,0,
0,4*]*],5*],[*[*"D",4*],293,[*[*0,4,0,0*]*],5*],[*[*"D",4*],293,
[*[*4,0,0,0*]*],5*],[*[*"A",1*],293,[*[*292*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,
179,181,191,193,197,199,211,223,227,229,233,239,241,251,257,
263,269,271,277,281,283*]*]*],[*[*"B",146*],293,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[*2*]*]*],[*[*"A",292*],293,[*[*1,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",292*],293,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*]*],[*[*[*"A",2*],294,[*[*0,2*],[*1,1*],[*1,1*]*],3*],[*[*"A",2
*],294,[*[*2,0*],[*1,1*],[*1,1*]*],3*],[*[*"A",3*],294,[*[*2,1,2
*]*],3*],[*[*"B",2*],294,[*[*0,9*]*],17*],[*[*"B",3*],294,[*[*4,0,
0*]*],[*[*2,3,5*]*]*],[*[*"C",3*],294,[*[*0,1,0*],[*0,0,2*]*],[*[*
2,3*]*]*],[*[*"C",3*],294,[*[*1,0,0*],[*0,0,2*]*],[*[*2*]*]*],[*[*
"D",4*],294,[*[*0,0,0,4*]*],[*[*2,3,5*]*]*],[*[*"D",4*],294,[*[*
0,4,0,0*]*],[*[*2,3,5*]*]*],[*[*"D",4*],294,[*[*4,0,0,0*]*],[*[*
2,3,5*]*]*],[*[*"A",1*],294,[*[*1*],[*2*],[*6*],[*6*]*],[*[*2,3,5
*]*]*],[*[*"A",1*],294,[*[*1*],[*2*],[*48*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],294,[*[*1*],[*6*],[*
20*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],294,[*[*1*],[*
146*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]
*]*],[*[*"A",1*],294,[*[*2*],[*6*],[*13*]*],[*[*2,3,5,7,11,13*]*]
*],[*[*"A",1*],294,[*[*2*],[*97*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97*]*]*],[*[*"A",
1*],294,[*[*5*],[*6*],[*6*]*],[*[*2,3,5*]*]*],[*[*"A",1*],294,[*[*5
*],[*48*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],
[*[*"A",1*],294,[*[*6*],[*41*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41*]*]*],[*[*"A",1*],294,[*[*13*],[*20*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],294,[*[*293*]*],[*[*2,3,5,7,11,13,17,
19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,
103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,
181,191,193,197,199,211,223,227,229,233,239,241,251,257,263,
269,271,277,281,283,293*]*]*],[*[*"C",147*],294,[*[*0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"D",147*],294,[*[*0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",293*],
294,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",293*],294,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*],[*[*[*"C",3*],295,[*[*0,3,0*]*],5*],[*[*
"G",2*],295,[*[*2,2*]*],11*],[*[*"A",1*],295,[*[*4*],[*58*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53*]*]*],[*[*"A",1
*],295,[*[*294*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,211,
223,227,229,233,239,241,251,257,263,269,271,277,281,283,293*]*]
*],[*[*"B",147*],295,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[*2*]*]*],[*[*"A",294*],295,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]
*],[*[*"A",294*],295,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
2*],296,[*[*1,1*],[*3,3*]*],7*],[*[*"B",2*],296,[*[*7,2*]*],13*],
[*[*"A",1*],296,[*[*1*],[*3*],[*36*]*],[*[*2,3,5,7,11,13,17,19,
23,29,31*]*]*],[*[*"A",1*],296,[*[*1*],[*147*]*],[*[*2,3,5,7,11,
13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,
97,101,103,107,109,113,127,131,137,139*]*]*],[*[*"A",1*],296,[*[*
1*],[*1*],[*73*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,
47,53,59,61,67,71,73*]*]*],[*[*"A",1*],296,[*[*1*],[*1*],[*1*],[*
36*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],296,[*
[*3*],[*73*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,
53,59,61,67,71,73*]*]*],[*[*"A",1*],296,[*[*7*],[*36*]*],[*[*2,3,
5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],296,[*[*295*]*],[*[*2,
3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,
79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,157,
163,167,173,179,181,191,193,197,199,211,223,227,229,233,239,
241,251,257,263,269,271,277,281,283,293*]*]*],[*[*"C",148*],296,[*
[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*
"D",148*],296,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],
[*[**]*]*],[*[*"A",295*],296,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],
[*[*"A",295*],296,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",
2*],297,[*[*0,1*],[*1,0*],[*1,5*]*],7*],[*[*"A",2*],297,[*[*0,1
*],[*1,0*],[*5,1*]*],7*],[*[*"A",2*],297,[*[*0,1*],[*1,8*]*],[*[*
2,3,5,7*]*]*],[*[*"A",2*],297,[*[*0,1*],[*8,1*]*],[*[*2,3,5,7*]
*]*],[*[*"A",2*],297,[*[*0,1*],[*0,1*],[*1,5*]*],7*],[*[*"A",2*],
297,[*[*0,1*],[*0,1*],[*5,1*]*],7*],[*[*"A",2*],297,[*[*1,0*],[*
1,8*]*],[*[*2,3,5,7*]*]*],[*[*"A",2*],297,[*[*1,0*],[*8,1*]*],[*[*
2,3,5,7*]*]*],[*[*"A",2*],297,[*[*1,0*],[*1,0*],[*1,5*]*],7*],[*
[*"A",2*],297,[*[*1,0*],[*1,0*],[*5,1*]*],7*],[*[*"A",2*],297,[*[*
1,21*]*],23*],[*[*"A",2*],297,[*[*21,1*]*],23*],[*[*"A",1*],297,[*
[*2*],[*8*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],297,[*[*2*],[*98
*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,
67,71,73,79,83,89,97*]*]*],[*[*"A",1*],297,[*[*2*],[*2*],[*32*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],297,[*[*2*],
[*2*],[*2*],[*10*]*],[*[*2,3,5,7*]*]*],[*[*"A",1*],297,[*[*8*],[*
32*]*],[*[*2,3,5,7,11,13,17,19,23,29,31*]*]*],[*[*"A",1*],297,[*
[*10*],[*26*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],297,
[*[*296*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149,151,157,163,167,173,179,181,191,193,197,199,211,223,
227,229,233,239,241,251,257,263,269,271,277,281,283,293*]*]*],[*
[*"B",148*],297,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]
*],[*[*2*]*]*],[*[*"A",296*],297,[*[*1,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*
[**]*]*],[*[*"A",296*],297,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*]
*],[*[*[*"F",4*],298,[*[*0,0,0,2*]*],7*],[*[*"A",1*],298,[*[*1*],[*
148*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,
61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,139*]
*]*],[*[*"A",1*],298,[*[*297*]*],[*[*2,3,5,7,11,13,17,19,23,29,
31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,
109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
193,197,199,211,223,227,229,233,239,241,251,257,263,269,271,
277,281,283,293*]*]*],[*[*"D",12*],298,[*[*0,0,0,0,0,0,0,0,0,
0,0,2*]*],2*],[*[*"D",12*],298,[*[*0,0,0,0,0,0,0,0,0,0,0,2*]
*],3*],[*[*"C",149*],298,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"D",149*],298,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]*],[*[*"A",297*],298,[*[*1,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",297*],298,[*[*0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,1*]*],[*[**]*]*]*],[*[*[*"A",3*],299,[*[*2,2,2*]*],5*],
[*[*"D",4*],299,[*[*0,0,2,0*]*],7*],[*[*"A",1*],299,[*[*12*],[*22
*]*],[*[*2,3,5,7,11,13,17,19*]*]*],[*[*"A",1*],299,[*[*298*]*],[*[*
2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,
73,79,83,89,97,101,103,107,109,113,127,131,137,139,149,151,
157,163,167,173,179,181,191,193,197,199,211,223,227,229,233,
239,241,251,257,263,269,271,277,281,283,293*]*]*],[*[*"D",12*],
299,[*[*0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[*2,3*]*]*],[*[*"B",149
*],299,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[*2
*]*]*],[*[*"A",298*],299,[*[*1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*]*],[*[**]
*]*],[*[*"A",298*],299,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]*]
*]*],[*[*[*"A",2*],300,[*[*0,1*],[*0,3*],[*3,0*]*],[*[*2,3*]*]*],[*
[*"A",2*],300,[*[*0,1*],[*0,3*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2
*],300,[*[*0,1*],[*3,0*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],300,[*
[*1,0*],[*0,3*],[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],300,[*[*1,0*],
[*0,3*],[*0,3*]*],[*[*2,3*]*]*],[*[*"A",2*],300,[*[*1,0*],[*3,0*],
[*3,0*]*],[*[*2,3*]*]*],[*[*"A",2*],300,[*[*0,23*]*],[*[*2,3,5,7,
11,13,17,19,23*]*]*],[*[*"A",2*],300,[*[*23,0*]*],[*[*2,3,5,7,
11,13,17,19,23*]*]*],[*[*"A",3*],300,[*[*0,1,0*],[*0,3,0*]*],[*[*
2,3*]*]*],[*[*"A",3*],300,[*[*1,0,1*],[*0,0,3*]*],[*[*2,3*]*]*],[*
[*"A",3*],300,[*[*1,0,1*],[*0,1,1*]*],[*[*2,3*]*]*],[*[*"A",3*],
300,[*[*1,0,1*],[*0,2,0*]*],[*[*2,3*]*]*],[*[*"A",3*],300,[*[*1,
0,1*],[*1,1,0*]*],[*[*2,3*]*]*],[*[*"A",3*],300,[*[*1,0,1*],[*3,
0,0*]*],[*[*2,3*]*]*],[*[*"A",3*],300,[*[*1,2,2*]*],7*],[*[*"A",3
*],300,[*[*2,1,2*]*],[*[*2,3,7*]*]*],[*[*"A",3*],300,[*[*2,2,1*]
*],7*],[*[*"A",3*],300,[*[*3,0,3*]*],[*[*2,3,7*]*]*],[*[*"A",4*],
300,[*[*0,0,1,0*],[*0,0,1,1*]*],3*],[*[*"A",4*],300,[*[*0,0,1,
0*],[*1,1,0,0*]*],3*],[*[*"A",4*],300,[*[*0,1,0,0*],[*0,0,1,1
*]*],3*],[*[*"A",4*],300,[*[*0,1,0,0*],[*1,1,0,0*]*],3*],[*[*"A",
5*],300,[*[*0,0,0,0,1*],[*0,0,0,1,1*]*],3*],[*[*"A",5*],300,[*
[*0,0,0,0,1*],[*1,1,0,0,0*]*],3*],[*[*"A",5*],300,[*[*1,0,0,
0,0*],[*0,0,0,1,1*]*],3*],[*[*"A",5*],300,[*[*1,0,0,0,0*],[*1,
1,0,0,0*]*],3*],[*[*"A",5*],300,[*[*0,0,0,1,0*],[*0,0,1,0,0*]
*],[*[**]*]*],[*[*"A",5*],300,[*[*0,1,0,0,0*],[*0,0,1,0,0*]*],[*[*
*]*]*],[*[*"B",2*],300,[*[*0,1*],[*1,4*]*],11*],[*[*"B",2*],300,[*[*
0,1*],[*0,1*],[*1,1*]*],5*],[*[*"B",2*],300,[*[*2,0*],[*0,3*]*],
[*[*2,3,7*]*]*],[*[*"B",2*],300,[*[*5,2*]*],7*],[*[*"C",3*],300,[*
[*0,0,1*],[*0,1,1*]*],3*],[*[*"D",4*],300,[*[*0,0,2,0*]*],[*[*2,
3,7*]*]*],[*[*"A",1*],300,[*[*1*],[*2*],[*4*],[*9*]*],[*[*2,3,5,7
*]*]*],[*[*"A",1*],300,[*[*1*],[*2*],[*49*]*],[*[*2,3,5,7,11,13,
17,19,23,29,31,37,41,43,47*]*]*],[*[*"A",1*],300,[*[*1*],[*4*],[*
29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]*],[*[*"A",1*],300,[*[*1
*],[*4*],[*4*],[*5*]*],[*[*2,3,5*]*]*],[*[*"A",1*],300,[*[*1*],[*5
*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]*]*],[*[*"A",1*],300,[*[*
1*],[*9*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],300,[*[*1
*],[*149*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,
59,61,67,71,73,79,83,89,97,101,103,107,109,113,127,131,137,
139,149*]*]*],[*[*"A",1*],300,[*[*1*],[*1*],[*2*],[*4*],[*4*]*],[*[*
2,3*]*]*],[*[*"A",1*],300,[*[*1*],[*1*],[*2*],[*24*]*],[*[*2,3,5,
7,11,13,17,19,23*]*]*],[*[*"A",1*],300,[*[*1*],[*1*],[*4*],[*14*]
*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1*],300,[*[*1*],[*1*],[*74*]*],
[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,
71,73*]*]*],[*[*"A",1*],300,[*[*2*],[*3*],[*4*],[*4*]*],[*[*2,3*]*]
*],[*[*"A",1*],300,[*[*2*],[*3*],[*24*]*],[*[*2,3,5,7,11,13,17,
19,23*]*]*],[*[*"A",1*],300,[*[*2*],[*4*],[*19*]*],[*[*2,3,5,7,11,
13,17,19*]*]*],[*[*"A",1*],300,[*[*2*],[*9*],[*9*]*],[*[*2,3,5,7*]
*]*],[*[*"A",1*],300,[*[*2*],[*99*]*],[*[*2,3,5,7,11,13,17,19,23,
29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97*]*]*],[*[*"A",
1*],300,[*[*3*],[*4*],[*14*]*],[*[*2,3,5,7,11,13*]*]*],[*[*"A",1
*],300,[*[*3*],[*74*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,
43,47,53,59,61,67,71,73*]*]*],[*[*"A",1*],300,[*[*4*],[*5*],[*9*]
*],[*[*2,3,5,7*]*]*],[*[*"A",1*],300,[*[*4*],[*59*]*],[*[*2,3,5,7,
11,13,17,19,23,29,31,37,41,43,47,53,59*]*]*],[*[*"A",1*],300,[*
[*4*],[*4*],[*11*]*],[*[*2,3,5,7,11*]*]*],[*[*"A",1*],300,[*[*5*],
[*49*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,37,41,43,47*]*]*],[*
[*"A",1*],300,[*[*9*],[*29*]*],[*[*2,3,5,7,11,13,17,19,23,29*]*]
*],[*[*"A",1*],300,[*[*11*],[*24*]*],[*[*2,3,5,7,11,13,17,19,23*]
*]*],[*[*"A",1*],300,[*[*14*],[*19*]*],[*[*2,3,5,7,11,13,17,19*]*]
*],[*[*"A",1*],300,[*[*299*]*],[*[*2,3,5,7,11,13,17,19,23,29,31,
37,41,43,47,53,59,61,67,71,73,79,83,89,97,101,103,107,109,
113,127,131,137,139,149,151,157,163,167,173,179,181,191,193,
197,199,211,223,227,229,233,239,241,251,257,263,269,271,277,
281,283,293*]*]*],[*[*"B",12*],300,[*[*0,0,0,0,0,0,0,0,0,0,1,
0*]*],[*[*2*]*]*],[*[*"C",12*],300,[*[*0,0,0,0,0,0,0,0,0,0,0,2
*]*],[*[**]*]*],[*[*"A",23*],300,[*[*2,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0*]*],[*[**]*]*],[*[*"A",23*],300,[*[*0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2*]*],[*[**]*]*],[*[*
"A",24*],300,[*[*0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0*]*],[*[**]*]*],[*[*"A",24*],300,[*[*0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0*]*],[*[**]*]*],[*[*"C",150*],
300,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1*]*],[*[**]
*]*],[*[*"D",150*],300,[*[*0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*],[*[*"A",299*],300,[*[*1,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0*]*],[*[**]*]*],[*[*"A",299*],300,[*[*0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,1*]*],[*[**]*]*]*]*];

/*
function AllCases(n, F)
    p := Characteristic(F);

    if n gt #DEFREPDATA then
        return false;
    end if;
    
    // all cases for this degree
    l := DEFREPDATA[n];
    
    // only those for characteristic p
    lp := [a : a in l | a[4] cmpeq p or
        (Category(a[4]) eq List and p notin [x : x in a[4][1]])];
    
    // only those for q = #F
    lq := [a : a in lp | #a[3] le Degree(F)];
    return [[* a[1][1], a[1][2], a[3] *] : a in lq];
end function;

function AllHighestWeightsOfCase(c, F)
    p := Characteristic(F);
    f := Degree(F);

    print c;
    comb := [l : a in CartesianProduct([[1 .. f] : i in [1 .. #c[3]]]) |
        #SequenceToSet(l) eq #c[3] where l is [a[i] : i in [1 .. #a]]];
    print comb;
    
    pps := [p^(i - 1) : i in [1 .. f]];
    v := Matrix([[i : i in k] : k in c[3]]);
    
    return {@ ElementToSequence(Vector(pps[a]) * v) : a in comb @};
end function;
*/

// Convert raw data to associative array

DefRepData := AssociativeArray();
for n in [1 .. #DEFREPDATA] do
    for c in DEFREPDATA[n] do
	// Hash on name, rank and dim
	key := <c[1][1], c[1][2], c[2]>;

	assert Category(c[3]) eq List;
	
	if Category(c[4]) eq List then
	    assert #c[4] eq 1;
	    assert Category(c[4][1]) eq List;
	    data := [* {p : p in c[4][1]}, [[j : j in i] : i in c[3]] *];
        else
            data := [* c[4], [[j : j in i] : i in c[3]] *];
	end if;

	flag, oldData := IsDefined(DefRepData, key);
	if flag then
	    DefRepData[key] := Append(oldData, data);
	else
	    DefRepData[key] := [data];
	end if;
    end for;
end for;

// Calculates maximum possible sum of the highest weights for a given group,
// dimension and field.
function MaximumHighestWeightSum(name, rank, dim, q)

    // Pick out representations for this group and dimension
    key := <name, rank, dim>;
    flag, reps := IsDefined(DefRepData, key);
    if not flag then
	return false, _;
    end if;

    flag, p, e := IsPrimePower(q);
    if not flag then
	return false, _;
    end if;
    
    // only those for characteristic p
    reps_p := [rep : rep in reps | rep[1] cmpeq p or
	(Category(rep[1]) eq SetEnum and p notin rep[1])];
    if #reps_p eq 0 then
	return false, _;
    end if;
        
    // only those for F
    repsF := [rep : rep in reps_p | #rep[2] le e];
    if #repsF eq 0 then
	return false, _;
    end if;

    pps := [p^(i - 1) : i in [1 .. e]];
    maxSum := 0;
    
    // Obtain highest weight sums
    for c in repsF do
	comb := [l : a in CartesianProduct([[1 .. e] : i in [1 .. #c[2]]]) |
	    #SequenceToSet(l) eq #c[2] where l is [a[i] : i in [1 .. #a]]];
	
	v := Matrix(c[2]);
	
	weights := {@ ElementToSequence(Vector(pps[a]) * v) : a in comb @};
	maxSum := Max(maxSum, Max([&+l : l in weights]));
    end for;
    
    return true, maxSum;
end function;

// G has as its stable derivative a quasisimple group H, 
// a central extension of a simple group of Lie type. 
// Use statement from Frank Luebeck about upper bound
// to index of H in G;
// namely it's d * g * log_p q where d and g are listed in
// Table 5 of Atlas p xvi

MaxIndex := function (name, n, q, p, e) 
   if name eq "A" then
      if n eq 1 then 
         max := Gcd (2, q - 1) * 1;
      else 
         max := Gcd (n + 1, q - 1) * 2;
      end if;
   elif name eq "2A" and n ge 2 then 
      max := Gcd (n + 1, q + 1) * 1;
   elif name eq "B" then 
      if n eq 2 and p eq 2 then g := 2; else g := 1; end if;
      max := Gcd (2, q - 1) * g;
   elif name eq "2B" then 
      max := 1;
   elif name eq "C" and n ge 3 then 
      max := Gcd (2, q - 1) * 1;
   elif name eq "D" and n ge 4 then 
      if n eq 4 then max := Gcd (2, (q - 1))^2 * 6;
      elif IsEven (q) then max := Gcd (2, (q - 1))^2 * 2;
      else max := Gcd (4, (q^n - 1)) * 2;
      end if;
   elif name eq "2D" and n ge 4 then 
      max := Gcd (4, q^n + 1) * 1;
   elif name eq "3D" and n eq 4 then 
      max := 1;
   elif name eq "E" then
      if n eq 6 then max := Gcd (3, q - 1) * 2;
      elif n eq 7 then max := Gcd (2, q - 1) * 1;
      else max := 1; end if;
   elif name eq "F" then 
      max := p eq 2 select 2 else 1;
   elif name eq "2F" then 
      max := 1;
   elif name eq "2E" then 
      max := Gcd (3, q + 1) * 1;
   elif name eq "G" then
      max := p eq 3 select 2 else 1;
   elif name eq "2G" then
      max := 1;
   else
      return 0;
   end if;
   return max * e; 
end function;

// Use statement from Frank Luebeck about index of stable derivative

function MaximumStableDerivativeIndex(name, rank, dim, q, F)
    // Result only applies for defining characteristic
    flag, p, e := IsPrimePower(q);
    if not flag or p ne Characteristic(F) then
	return 0;
    end if;
    return MaxIndex (name, rank, q, p, e); 
    
    // Obtain sums of highest weights from Frank's data
    // flag, MaxWeightSum := MaximumHighestWeightSum(name, rank, dim, q);
end function;
